// ===========================================================================
// CONTENT  : CLASS BasicUtil
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 15/05/2020
// HISTORY  :
//  15/05/2020  mdu  CREATED
//
// Copyright (c) 2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.bif;

import java.util.Collection;
import java.util.Map;

import org.pfsw.bif.execution.IExecutable;
import org.pfsw.bif.text.IStringRepresentation;

/**
 * Basic utility methods that should actually be part of the JDK. 
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class BasicUtil
{
  public static final BasicUtil BU = new BasicUtil();

  public static BasicUtil current()
  {
    return BU;
  }

  /**
   * This method is intentionally doing nothing at all.
   * It can be used in situations where a runnable callback is required that should do nothing.
   */
  public static void noop()
  {
    // intentionally nothing
  }

  protected BasicUtil()
  {
    super();
  }

  /**
   * Returns true if the given array contains no null value.
   * 
   * @param objects The array to check (must not be null).
   */
  public boolean isNoneNull(Object... objects)
  {
    return !isAnyNull(objects);
  }

  /**
   * Returns true if the given array contains at least one null value.
   * 
   * @param objects The array to check (must not be null).
   */
  public boolean isAnyNull(Object... objects)
  {
    for (Object object : objects)
    {
      if (object == null)
      {
        return true;
      }
    }
    return false;
  }

  /**
   * Returns true if the given collection is null or has no elements.
   * 
   * @param collection The collection to check
   */
  public boolean isNullOrEmpty(Collection<?> collection)
  {
    if (collection == null)
    {
      return true;
    }
    return collection.isEmpty();
  }

  /**
   * Returns true if the given array is null or has no elements.
   * 
   * @param array The array to check
   */
  public boolean isNullOrEmpty(Object[] array)
  {
    if (array == null)
    {
      return true;
    }
    return array.length == 0;
  }

  /**
   * Returns true if the given map is null or has no elements.
   * 
   * @param map The map to check
   */
  public boolean isNullOrEmpty(Map<?, ?> map)
  {
    if (map == null)
    {
      return true;
    }
    return map.isEmpty();
  }

  /**
   * Returns true if any one of the given strings is null or blank.
   * Blanks means "" as well as strings filled with white spaces only (e.g. "\t  ").
   */
  public boolean isAnyNullOrBlank(String... strings)
  {
    if (strings == null)
    {
      return true;
    }
    for (String string : strings)
    {
      if (isNullOrBlank(string))
      {
        return true;
      }
    }
    return false;
  }

  /**
   * Returns true if none of the given strings is null or blank.
   * Blanks means "" as well as strings filled with white spaces only (e.g. "\t  ").
   */
  public boolean isNoneNullOrBlank(String... strings)
  {
    return !isAnyNullOrBlank(strings);
  }

  /**
   * Returns true if the given string array is null or empty
   * 
   * @param strings The string array to check
   */
  public boolean isNullOrEmpty(String[] strings)
  {
    return (strings == null) || (strings.length == 0);
  }

  /**
   * Returns true if the given string is null or empty
   * 
   * @param str The string to check
   */
  public boolean isNullOrEmpty(String str)
  {
    return ((str == null) || (str.length() == 0));
  }

  /**
   * Returns true if the given string is null or empty or consists of 
   * whitespace characters only.
   * @param str The string to check
   */
  public boolean isNullOrBlank(String str)
  {
    return (str == null) || (str.trim().length() == 0);
  }

  /**
   * Returns true if the given string is not null and not empty.
   *
   * @param str The string to check.
   */
  public boolean notNullOrEmpty(String str)
  {
    return !isNullOrEmpty(str);
  }

  /**
   * Returns true if the given string is not null and not empty and does not 
   * consist of whitespace characters only.
   * @param str The string to check.
   */
  public boolean notNullOrBlank(String str)
  {
    return !isNullOrBlank(str);
  }

  /**
   * Returns whether or not the given string contains any white space character.
   * Returns false for null or empty string.
   * 
   * @param str The string to check.
   * @return true if the given string contains any white space character.
   */
  public boolean containsWhitespaceChars(String str)
  {
    if (notNullOrEmpty(str))
    {
      for (char ch : str.toCharArray())
      {
        if (Character.isWhitespace(ch))
        {
          return true;
        }
      }
    }
    return false;
  }

  /**
   * Returns whether or not the first given object is less than the second.
   * 
   * @param object1 The object to be compared if it is less than the other object (must not be null).
   * @param object2 The object to be compared with (must not be null).
   * @return true if the first given object is less than the second.
   */
  public <T> boolean isLess(Comparable<T> object1, T object2)
  {
    return compare(object1, object2).isLess();
  }

  /**
   * Returns whether or not the first given object is equal to the second.
   * 
   * @param object1 The object to be compared if it is equal to the other object (must not be null).
   * @param object2 The object to be compared with (must not be null).
   * @return true if the first given object is equal to the second.
   */
  public <T> boolean isEqual(Comparable<T> object1, T object2)
  {
    return compare(object1, object2).isEqual();
  }

  /**
   * Returns whether or not the first given object is greater than the second.
   * 
   * @param object1 The object to be compared if it is greater than the other object (must not be null).
   * @param object2 The object to be compared with (must not be null).
   * @return true if the first given object is greater than the second.
   */
  public <T> boolean isGreater(Comparable<T> object1, T object2)
  {
    return compare(object1, object2).isGreater();
  }

  /**
   * Compares the first object to the second and returns an enum value for the result.
   * 
   * @param object1 The object to be compared to the other object (must not be null).
   * @param object2 The object to be compared with (must not be null).
   * @return The outcome of the comparison.
   */
  public <T> CompareResult compare(Comparable<T> object1, T object2)
  {
    return CompareResult.of(object1.compareTo(object2));
  }

  /**
   * Returns the string representation of the given object utilizing either asString() 
   * if it implements {@link IStringRepresentation}, name() if it is an enum value
   * or toString() for all other objects.
   * 
   * @param object The object to get the string representation for (may be null).
   * @return Returns the string representation of the given object or null if it is null.
   */
  @SuppressWarnings("rawtypes")
  public String asString(Object object)
  {
    if (object == null)
    {
      return null;
    }
    if (object instanceof IStringRepresentation)
    {
      return ((IStringRepresentation)object).asString();
    }
    if (object instanceof Enum)
    {
      return ((Enum)object).name();
    }
    return object.toString();
  }

  /**
   * Returns whether or not the given object is exactly of the specified type.
   * This is more restrictive than <code>instanceof</code> checking because it
   * ignores inheritance. 
   * Returns null if any of the arguments is null.
   */
  public boolean isTypeOf(Object object, Class<?> type)
  {
    if (isNoneNull(object, type))
    {
      return object.getClass() == type;
    }
    return false;
  }

  /**
   * Returns whether or not the given object is the same type as the specified type
   * of a sub-type of this type.
   * Returns null if any of the arguments is null.
   */
  public boolean isTypeOrSubTypeOf(Object object, Class<?> type)
  {
    if (isNoneNull(object, type))
    {
      return type.isAssignableFrom(object.getClass());
    }
    return false;
  }

  /**
   * Returns the first value of the given value array that is not NaN 
   * or null if none found.
   */
  public Double firstValidValue(Double... values)
  {
    for (Double value : values)
    {
      if (!value.isNaN())
      {
        return value;
      }
    }
    return null;
  }

  /**
   * Returns the first value of the given value array that is not NaN 
   * or null if none found.
   */
  public Float firstValidValue(Float... values)
  {
    for (Float value : values)
    {
      if (!value.isNaN())
      {
        return value;
      }
    }
    return null;
  }

  /**
   * Converts the given runnable into an {@link IExecutable};
   */
  public IExecutable asExecutable(final Runnable runnable)
  {
    if (runnable == null)
    {
      return null;
    }
    return new IExecutable()
    {
      @Override
      public void execute()
      {
        runnable.run();
      }
    };
  }

  /**
   * Converts the given executable into an {@link Runnable};
   */
  public Runnable asRunnable(final IExecutable executable)
  {
    if (executable == null)
    {
      return null;
    }
    return new Runnable()
    {
      @Override
      public void run()
      {
        executable.execute();
      }
    };
  }
}
