// ===========================================================================
// CONTENT  : CLASS RemoteCallException
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 19/04/2021
// HISTORY  :
//  19/04/2021  mdu  CREATED
//
// Copyright (c) 2021, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.bif.exceptions;

import java.util.HashMap;
import java.util.Map;

import org.pfsw.bif.BasicUtil;

/**
 * Generic runtime exception that signals a problem with a remote call.
 */
public class RemoteCallException extends RuntimeException
{
  private static final long serialVersionUID = 8970626293378773236L;

  public static final int UNDEFINED_STATUS_CODE = 0;

  private int statusCode = UNDEFINED_STATUS_CODE;
  private final transient Map<String, Object> data = new HashMap<String, Object>();

  public static RemoteCallException create(String message, Object... args)
  {
    return new RemoteCallException(message, args);
  }

  public static RemoteCallException create(Exception cause, String message, Object... args)
  {
    return new RemoteCallException(cause, message, args);
  }

  public static RemoteCallException create(int statusCode, String message, Object... args)
  {
    return create(message, args).setStatusCode(statusCode);
  }
  
  public RemoteCallException(String message, Object... args)
  {
    super(String.format(message, args));
  }

  public RemoteCallException(Exception cause, String message, Object... args)
  {
    super(String.format(message, args), cause);
  }

  /**
   * Returns the additional object that might have been registered under the given key
   * or null if not found.
   */
  @SuppressWarnings("unchecked")
  public <T> T get(String key)
  {
    if (key != null)
    {
      return (T)getData().get(key);
    }
    return null;
  }

  /**
   * Register an arbitrary object under the given key. 
   */
  public RemoteCallException put(String key, Object value)
  {
    if (BasicUtil.BU.isNoneNull(key, value))
    {
      getData().put(key, value);
    }
    return this;
  }

  public Map<String, Object> getData()
  {
    return this.data;
  }

  public int getStatusCode()
  {
    return this.statusCode;
  }

  public RemoteCallException setStatusCode(int statusCode)
  {
    this.statusCode = statusCode;
    return this;
  }

  @Override
  public String toString()
  {
    return String.format("%s(%d, '%s')", getClass().getSimpleName(), getStatusCode(), getMessage());
  }
}
