// ===========================================================================
// CONTENT  : CLASS HttpMethod
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.0 - 14/05/2020
// HISTORY  :
//  19/08/2014  mdu  CREATED
//  14/05/2020  mdu   changed -> from class to enum
//
// Copyright (c) 2014-2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.bif.http;

/**
 * Represents an HTTP method like GET, PUT and so on.
 * https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods
 *
 * @author Manfred Duchrow
 * @version 2.0
 */
public enum HttpMethod implements HttpVerb
{
  //@formatter:off
  GET("GET", false),
  POST("POST", true),
  PUT("PUT", true),
  DELETE("DELETE", false),
  HEAD("HEAD", false),
  OPTIONS("OPTIONS", false),
  CONNECT("CONNECT", false),
  TRACE("TRACE", false),
  PATCH("PATCH", true); // See RFC-5789
  //@formatter:on

  private static final HttpMethod[] HTTP11_METHODS = { GET, POST, PUT, DELETE, HEAD, OPTIONS, CONNECT, TRACE, PATCH };

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private final String name;
  private final boolean hasBody;

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Returns the corresponding HTTP/1.1 method that corresponds to the given name
   * (case-insensitive) or null if the name does not match any HTTP/1.1 method.
   */
  public static HttpMethod getHttp11Method(String methodName)
  {
    for (HttpMethod httpMethod : HTTP11_METHODS)
    {
      if (httpMethod.isMethod(methodName))
      {
        return httpMethod;
      }
    }
    return null;
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  private HttpMethod(String name, boolean hasBody)
  {
    if ((name == null) || (name.trim().length() == 0))
    {
      throw new IllegalArgumentException("Required argument 'name' is not valid!");
    }
    this.name = name.trim();
    this.hasBody = hasBody;
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * The (upper-case) name of the HTTP method (e.g. "GET").
   */
  public String getName()
  {
    return name;
  }

  /**
   * Returns true if this method typically is used with a request body.
   */
  public boolean hasBody()
  {
    return hasBody;
  }

  /**
   * Returns true if the given string equals (case-insensitive) this
   * HttpMethod's name.
   */
  public boolean isMethod(String method)
  {
    return getName().equalsIgnoreCase(method);
  }

  /**
   * Returns the string representation of the method (e.g. "DELETE").
   */
  @Override
  public String asString()
  {
    return getName();
  }

  @Override
  public String toString()
  {
    return String.format("%s('%s')", getClass().getSimpleName(), asString());
  }
}
