// ===========================================================================
// CONTENT  : CLASS ObjectReference
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 16/01/2021
// HISTORY  :
//  16/01/2021  mdu  CREATED
//
// Copyright (c) 2021, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.bif.text;

import org.pfsw.bif.filter.IObjectFilter;

/**
 * A unique string reference to identify a specific object (in-memory).
 * 
 * @author M.Duchrow
 */
public class ObjectReference implements IStringRepresentation, IObjectFilter<Object>
{
  /**
   * The object reference for null.
   */
  public static final ObjectReference NULL_REFERENCE = new ObjectReference(makeReference(null));

  private final String referenceString;

  /**
   * Returns an object reference for the given object.
   * 
   * @param object The object to create the reference for (may be null).
   */
  public static ObjectReference create(Object object)
  {
    if (object == null)
    {
      return NULL_REFERENCE;
    }
    return new ObjectReference(makeReference(object));
  }

  private static String makeReference(String typeName, int hashValue)
  {
    return String.format("%s#%d", typeName, hashValue);
  }

  private static String makeReference(Object object)
  {
    if (object == null)
    {
      return makeReference(Void.class.getCanonicalName(), 0);
    }
    return makeReference(object.getClass().getCanonicalName(), object.hashCode());
  }

  private ObjectReference(String referenceString)
  {
    super();
    this.referenceString = referenceString;
  }

  @Override
  public String asString()
  {
    return getReferenceString();
  }

  /**
   * Returns true if the given object has the same reference representation as this object reference. 
   */
  @Override
  public boolean matches(Object object)
  {
    return asString().equals(makeReference(object));
  }

  @Override
  public int hashCode()
  {
    return getReferenceString().hashCode();
  }

  @Override
  public boolean equals(Object object)
  {
    if ((object != null) && (object.getClass() == ObjectReference.class))
    {
      ObjectReference other = (ObjectReference)object;
      return asString().equals(other.asString());
    }
    return false;
  }

  @Override
  public String toString()
  {
    return String.format("%s('%s')", getClass().getSimpleName(), asString());
  }

  protected String getReferenceString()
  {
    return this.referenceString;
  }
}
