// ===========================================================================
// CONTENT  : CLASS ArchiveTOC
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.0 - 26/01/2017
// HISTORY  :
//  22/07/2002  mdu  CREATED
//	26/08/2012	mdu	changed	--> Generic types		
//  26/01/2017  mdu changed --> hold internally List<FileInfo> rather than FileInfo[]
//
// Copyright (c) 2002-2017, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.file;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import org.pfsw.util.CollectionUtil;
import org.pfsw.util.NamedValue;
import org.pfsw.util.NamedValueList;

/**
 * This is a wrapper class around a zip archive's table of contents.
 * It reads in the table of contents of an archive and provides it as
 * a NamedValueList where the names are the directories and the values
 * are List<FileInfo>.
 *
 * @author Manfred Duchrow
 * @version 2.0
 */
public class ArchiveTOC
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final CollectionUtil CU = CollectionUtil.current();

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private NamedValueList<List<FileInfo>> directories = null;
  private String zipFilename = null;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with the archive named by the given filename.
   * Reads in automatically the archive's table of contents.
   */
  public ArchiveTOC(String filename) throws IOException
  {
    super();
    directories(new NamedValueList<List<FileInfo>>());
    init(filename);
  }

  /**
   * Initialize the new instance with the archive identified by the given file.
   * Reads in automatically the archive's table of contents.
   */
  public ArchiveTOC(File archive) throws IOException
  {
    this(archive.getAbsolutePath());
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns a named value list, where the names are directory names and
   * the values are FileInfo[] objects containing all files inside the
   * directories.
   * 
   */
  public NamedValueList<FileInfo[]> toc()
  {
    NamedValueList<FileInfo[]> toc;

    toc = new NamedValueList<FileInfo[]>();
    for (NamedValue<List<FileInfo>> entry : directories().namedValueArray())
    {
      toc.add(entry.name(), CU.toArray(entry.value()));
    }
    return toc;
  }

  /**
   * Returns the information about the files of this archive as a TableOfContents
   * object. 
   */
  public TableOfContents asTableOfContents()
  {
    TableOfContents toc;
    NamedValueList<List<FileInfo>> namedValueList;
    DirectoryContents dirContents;

    toc = new TableOfContents(getZipFilename());
    namedValueList = directories();
    for (int i = 0; i < namedValueList.size(); i++)
    {
      dirContents = new DirectoryContents(namedValueList.nameAt(i), namedValueList.valueAt(i));
      toc.add(dirContents);
    }
    return toc;
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected void init(String filename) throws IOException
  {
    ZipFile zipFile = null;

    zipFile = new ZipFile(filename);
    try
    {
      readTOC(zipFile);
    }
    finally
    {
      FileUtil.current().close(zipFile);
    }
    setZipFilename(filename);
  }

  protected void readTOC(ZipFile zipFile)
  {
    Enumeration<? extends ZipEntry> entries = null;
    ZipEntry entry = null;
    FileInfo fileInfo = null;

    entries = zipFile.entries();
    while (entries.hasMoreElements())
    {
      entry = entries.nextElement();
      if (isDirectory(entry))
      {
        // IGNORE all directory entries !
      }
      else
      {
        fileInfo = new FileInfo(zipFile.getName(), entry);
        add(fileInfo);
      }
    }
  }

  protected void add(FileInfo info)
  {
    NamedValue<List<FileInfo>> pathEntry;
    List<FileInfo> classList;

    pathEntry = directories().findNamedValue(info.getPath());
    if (pathEntry == null)
    {
      pathEntry = new NamedValue<List<FileInfo>>(info.getPath(), new ArrayList<FileInfo>());
      directories().add(pathEntry);
    }
    classList = pathEntry.value();
    classList.add(info);
  }

  protected boolean isDirectory(ZipEntry entry)
  {
    return entry.isDirectory();
  }

  protected NamedValueList<List<FileInfo>> directories()
  {
    return directories;
  }

  protected void directories(NamedValueList<List<FileInfo>> newValue)
  {
    directories = newValue;
  }

  protected String getZipFilename()
  {
    return zipFilename;
  }

  protected void setZipFilename(String newValue)
  {
    zipFilename = newValue;
  }
}
