// ===========================================================================
// CONTENT  : CLASS Classpath
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.3 - 30/08/2012
// HISTORY  :
//  02/02/2003  mdu		CREATED
//	07/11/2003	mdu		changed	-->	Use FileUtil.getClasspath()
//	25/12/2004	mdu		added		-->	Support of wildcards in element definition
//	26/12/2004	mdu		added		-->	removeDuplicates()
//  30/08/2012	mdu		changed	-->	appendElement()
//
// Copyright (c) 2003-2012, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.file;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import org.pfsw.text.StringPattern;
import org.pfsw.text.StringUtil;

/**
 * Represents a classpath definition whereas the elements are already parsed
 * an validated.
 * <br>
 * Since version 1.2 elements are also collected from patterns with wildcards
 * (e.g. "base.jar:lib/*.jar:xalan.jar")
 *
 * @author Manfred Duchrow
 * @version 1.3
 */
public class Classpath
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  protected final static String DEFAULT_SEPARATOR = File.pathSeparator;

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private List<ClasspathElement> elements = null;

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Returns a new Classpath object initialized with the given path elemments.
   * If the method is called with no parameter (i.e. an empty array), 
   * the default (system) classpath will be returned.
   * 
   * @param classpathElements The arbitrary number of elements to be put into the classpath.
   * @return A newly created Classpath instance.
   */
  public static Classpath create(String... classpathElements)
  {
    if (classpathElements.length == 0)
    {
      return new Classpath();
    }
    return new Classpath(classpathElements);
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with the classpath from System property 
   * "java.class.path" or from the Class-Path value in the manifest file of
   * the sole JAR file if the program was started with option -jar. 
   */
  public Classpath()
  {
    this((String)null);
  }

  /**
   * Initialize the new instance with a classpath definition.
   * 
   * @param classpath A classpath definition with the current platform's path separator
   */
  public Classpath(String classpath)
  {
    this(classpath, DEFAULT_SEPARATOR);
  }

  /**
   * Initialize the new instance with a classpath definition where the separator
   * is the one given as parameter.
   * 
   * @param classpath A classpath definition with the specified path separator
   * @param separator The separator thst is used to separate the elements
   */
  public Classpath(final String classpath, final String separator)
  {
    super();

    String cp = classpath;
    if (cp == null)
    {
      cp = fileUtil().getClasspath();
    }
    init(cp, separator == null ? DEFAULT_SEPARATOR : separator);
  }

  /**
   * Initialize the new instance with the given elements.
   * 
   * @param classpathElements An array with the elements of the classpath
   */
  public Classpath(String[] classpathElements)
  {
    super();
    initFromParts(classpathElements);
  }

  /**
   * Initialize the new instance with the given elements.
   * 
   * @param classpathElements A collection of String elements that build the classpath.
   */
  public Classpath(Collection<String> classpathElements)
  {
    super();
    initFromParts(str().asStrings(classpathElements));
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns all classpath elements.
   */
  public ClasspathElement[] allElements()
  {
    return elementArray(getElements());
  }

  /**
   * Returns the classpath elements that are valid.
   */
  public ClasspathElement[] validElements()
  {
    return collectElements(true);
  }

  /**
   * Returns the classpath elements that are invalid.
   */
  public ClasspathElement[] invalidElements()
  {
    return collectElements(false);
  }

  /**
   * Returns all elements that contain a file with the specified name.
   */
  public ClasspathElement[] elementsContaining(String filename)
  {
    return elementsWithFile(filename);
  }

  /**
   * Returns the first element that contains a file with the specified name or
   * null if the file cannot be found.
   */
  public ClasspathElement firstElementContaining(String filename)
  {
    ClasspathElement[] cpElements;

    cpElements = elementsWithFile(filename, true);
    if ((cpElements == null) || (cpElements.length == 0))
      return null;

    return cpElements[0];
  }

  /**
   * Removes all duplicates of classpath elements that are already in the 
   * classpath. 
   */
  public void removeDuplicates()
  {
    List<ClasspathElement> newList;
    Iterator<ClasspathElement> iter;
    ClasspathElement element;

    newList = new ArrayList<ClasspathElement>(getElements().size());
    iter = getElements().iterator();
    while (iter.hasNext())
    {
      element = iter.next();
      if (!newList.contains(element))
      {
        newList.add(element);
      }
    }
    setElements(newList);
  }

  /**
   * Returns the string representation of this classpath using the current 
   * platform's path separator.
   */
  @Override
  public String toString()
  {
    return toString(DEFAULT_SEPARATOR);
  }

  /**
   * Returns the string representation of this classpath using the given 
   * path separator.
   * 
   * @param separator The separator to separate the elements 
   */
  public String toString(String separator)
  {
    StringBuffer buffer;
    ClasspathElement element;

    buffer = new StringBuffer(getElements().size() * 20);

    for (int i = 0; i < getElements().size(); i++)
    {
      element = getElements().get(i);
      if (i > 0)
      {
        buffer.append(separator);
      }
      buffer.append(element.getName());
    }
    return buffer.toString();
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected void init(String classpath, String separator)
  {
    String[] parts;

    if (classpath == null)
    {
      setElements(new ArrayList<ClasspathElement>());
      return;
    }

    parts = str().parts(classpath, separator);
    initFromParts(parts);
  }

  protected void initFromParts(String[] parts)
  {
    String elementName;

    if (str().isNullOrEmpty(parts))
    {
      setElements(new ArrayList<ClasspathElement>());
      return;
    }

    setElements(new ArrayList<ClasspathElement>(parts.length));
    for (int i = 0; i < parts.length; i++)
    {
      elementName = parts[i].trim();
      appendElement(elementName);
    }
  }

  protected void appendElement(final String elementName)
  {
    File[] files;
    String pattern;

    if (StringPattern.containsWildcard(elementName))
    {
      pattern = elementName;
      if (pattern.endsWith("*"))
      {
        pattern += ".jar";
      }
      files = FileFinder.findFiles(pattern);
      for (int i = 0; i < files.length; i++)
      {
        basicAppendElement(files[i].getPath());
      }
    }
    else
    {
      basicAppendElement(elementName);
    }
  }

  protected void basicAppendElement(String elementName)
  {
    getElements().add(new ClasspathElement(elementName));
  }

  /**
   * Returns all elements that contain a file with the given name.
   */
  protected ClasspathElement[] elementsWithFile(String filename)
  {
    return elementsWithFile(filename, false);
  }

  /**
   * Returns all elements or the first element that contain(s) a file with the 
   * given name.
   */
  protected ClasspathElement[] elementsWithFile(String filename, boolean onlyFirst)
  {
    List<ClasspathElement> collection;
    ClasspathElement element;

    collection = new ArrayList<ClasspathElement>(getElements().size());
    for (int i = 0; i < getElements().size(); i++)
    {
      element = getElements().get(i);
      if (element.contains(filename))
      {
        collection.add(element);
        if (onlyFirst)
        {
          break;
        }
      }
    }

    return elementArray(collection);
  }

  protected ClasspathElement[] collectElements(boolean valid)
  {
    Iterator<ClasspathElement> iter;
    List<ClasspathElement> collection;
    ClasspathElement element;

    collection = new ArrayList<ClasspathElement>(getElements().size());

    iter = getElements().iterator();
    while (iter.hasNext())
    {
      element = iter.next();
      if (valid)
      {
        if (element.isValid())
        {
          collection.add(element);
        }
      }
      else
      {
        if (!element.isValid())
        {
          collection.add(element);
        }
      }
    }
    return elementArray(collection);
  }

  protected ClasspathElement[] elementArray(List<ClasspathElement> elementList)
  {
    ClasspathElement[] array;

    array = new ClasspathElement[elementList.size()];
    return elementList.toArray(array);
  }

  protected List<ClasspathElement> getElements()
  {
    return elements;
  }

  protected void setElements(List<ClasspathElement> newValue)
  {
    elements = newValue;
  }

  protected FileUtil fileUtil()
  {
    return FileUtil.current();
  }

  protected StringUtil str()
  {
    return StringUtil.current();
  }
}
