// ===========================================================================
// CONTENT  : CLASS ClasspathElement
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.6.1 - 10/09/2006
// HISTORY  :
//  02/02/2003  mdu		CREATED
//	07/11/2003	mdu		added		-->	javadoc for open(), new method createURL(), DEBUG
//	20/12/2003	mdu		changed	-->	Using LoggerProvider
//	08/05/2004	mdu		changed	-->	createURL() now returns jar:file:/ as protocol for URLs that point into JAR files
//	25/12/2004	mdu		added		-->	toString(), equals(), hashCode()
//	26/05/2006	mdu		changed	-->	cache zipFile for re-use and performance optimization
//	01/09/2006	mdu		bugfix	--> Added isOpen() and fixed close()
//	10/09/2006	mdu		bugfix	--> Use "file:/" rather than "file://" in createURL()
//
// Copyright (c) 2003-2006, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.file;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;

import org.pfsw.logging.Logger;
import org.pfsw.util.SysUtil;

/**
 * Represents one entry in a classpath. It can be asked if it s a directory
 * or an archive. It also can be asked if it exists and it is valid (really
 * a zip archive).
 *
 * @author Manfred Duchrow
 * @version 1.6.1
 */
public class ClasspathElement
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final String FILE_PROTOCOL_INDICATOR = "file:/";
  private static final String JAR_FILE_PROTOCOL_INDICATOR = "jar:file:/";
  private static final String ARCHIVE_INDICATOR = "!/";

  private static final boolean DEBUG = "true".equals(System.getProperty("org.pfsw.file.debug"));
  private static final String DEBUG_PREFIX = "org.pfsw.file.ClasspathElement: ";

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private String name = null;
  private File file = null;
  private boolean valid = false;
  private String id = null;
  private boolean leaveArchiveOpen = false;
  private ZipFile zipFile = null;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with a file representing either a directory
   * or an archive file.
   * 
   * @param aFile A file representing an element in a classpath
   */
  public ClasspathElement(File aFile)
  {
    super();
    if (aFile == null)
    {
      throw new IllegalArgumentException("ClasspathElement: File parameter must not be null");
    }
    init(aFile);
  }

  /**
   * Initialize the new instance with default values.
   */
  public ClasspathElement(String elementName)
  {
    super();
    if (elementName == null)
    {
      throw new IllegalArgumentException("ClasspathElement: elementName parameter must not be null");
    }

    init(new File(elementName));
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /** 
   * Returns the name of this classpath element 
   */
  public String getName()
  {
    return name;
  }

  /**
   * Returns true, if an opened archive stays open until method close()
   * gets called. Returns true if the archive gets closed after each access.
   * The default is false.
   * @see #close()
   */
  public boolean leaveArchiveOpen()
  {
    return leaveArchiveOpen;
  }

  /**
   * Sets whether an opened archive stays open until method close()
   * gets called. If set to false the archive gets closed after each access.
   * @see #close()
   */
  public void leaveArchiveOpen(boolean newValue)
  {
    leaveArchiveOpen = newValue;
  }

  /**
   * Returns whether or not this classpath element exists.
   */
  public boolean exists()
  {
    return getFile().exists();
  }

  /**
   * Returns true, if this element is a file.
   * Actually it only determines, if this element is a file. 
   * This doesn't ensure that it is really an archive file.
   */
  public boolean isFile()
  {
    return getFile().isFile();
  }

  /**
   * Returns true, if this element is an archive file.
   */
  public boolean isArchive()
  {
    return (isFile() && isValid());
  }

  /**
   * Returns true, if this element is a directory
   */
  public boolean isDirectory()
  {
    return getFile().isDirectory();
  }

  /**
   * Returns true, if this element is either a valid directory or really an
   * archive (zip file).
   */
  public boolean isValid()
  {
    return getValid();
  }

  /**
   * Returns true, if this element represents an archive (zip file) and that
   * archive is currently open.
   * If it represents a directory this method returns false
   */
  public boolean isOpen()
  {
    return zipFile != null;
  }

  /**
   * Returns whether or not this element contains a file with the given name.
   * 
   * @param filename The name of the file to be looked for
   */
  public boolean contains(String filename)
  {
    if (isValid())
    {
      if (isDirectory())
      {
        return dirContainsFile(filename);
      }
      return archiveContainsFile(filename);
    }
    return false;
  }

  /**
   * Returns a file info object for the given filename which is expected
   * to be inside the directory or archive represented by this classpath element.
   * 
   * @param filename The name of the file to be looked for
   * @return A file info object or null if the specified file does not exist
   */
  public FileInfo getFileInfo(String filename)
  {
    File aFile;
    ZipEntry zipEntry;

    if (isValid())
    {
      if (isDirectory())
      {
        aFile = getFileFromDirectory(filename);
        if ((aFile != null) && (aFile.exists()))
        {
          return new FileInfo(aFile);
        }
      }
      else
      {
        zipEntry = getFileFromArchive(filename);
        if (zipEntry != null)
        {
          return new FileInfo(getName(), zipEntry);
        }
      }
    }
    return null;
  }

  /**
   * Opens the file with the given name in this classpath element and returns 
   * the input stream. If the file doesn't exist or this classpath element is
   * not valid, null will be returned.
   * 
   * @param filename The name of the file in this container to be opened
   */
  public InputStream open(String filename) throws IOException
  {
    InputStream stream = null;
    File aFile;
    ZipFile aZipFile;
    ZipEntry zipEntry;

    if (isValid())
    {
      if (isArchive())
      {
        aZipFile = getZipFile();
        zipEntry = aZipFile.getEntry(filename);
        if (zipEntry != null)
        {
          stream = aZipFile.getInputStream(zipEntry);
        }
        else
        {
          // System.out.println( filename ) ;
        }
      }
      else
      {
        aFile = new File(getFile(), filename);
        stream = new FileInputStream(aFile);
      }
    }
    else
    {
      throw new IOException("Invalid classpath element can't contain '" + filename + "'");
    }
    return stream;
  }

  /**
   * Close any underlying (maybe) opened archive 
   */
  public synchronized void close()
  {
    if (zipFile != null)
    {
      try
      {
        zipFile.close();
      }
      catch (IOException e)
      {
        handleException(e);
      }
      setZipFile(null);
    }
  }

  /**
   * Returns a URL that points to this classpath element.
   */
  public URL getURL()
  {
    return createURL(null);
  }

  /**
   * Creates a URL for the file with the given name that is in this file
   * container. It is not checked, if the file really exists.
   * If this classpath element is invalid null will be returned. 
   */
  public URL createURL(String filename)
  {
    StringBuffer urlStr;
    URL url;

    if (!isValid())
      return null;

    urlStr = new StringBuffer(200);
    if (isArchive() && (filename != null))
    {
      urlStr.append(JAR_FILE_PROTOCOL_INDICATOR);
    }
    else
    {
      urlStr.append(FILE_PROTOCOL_INDICATOR);
    }
    urlStr.append(futil().standardize(getFile().getAbsolutePath()));

    if (filename != null)
    {
      urlStr.append(isArchive() ? ARCHIVE_INDICATOR : "/");
      urlStr.append(filename);
    }
    try
    {
      url = new URL(urlStr.toString());
    }
    catch (MalformedURLException e)
    {
      if (DEBUG)
      {
        logger().logException(e);
      }
      return null;
    }
    return url;
  }

  /**
   * Returns the element's string representation.
   * The string always contains slashes, never backslashes.
   */
  @Override
  public String toString()
  {
    return getName();
  }

  /**
   * Returns the hash value of this object
   * @see java.lang.Object#hashCode()
   */
  @Override
  public int hashCode()
  {
    return getId().hashCode();
  }

  /**
   * Returns true if the given object is equal to this object
   */
  @Override
  public boolean equals(Object object)
  {
    if (object instanceof ClasspathElement)
    {
      ClasspathElement otherElement = (ClasspathElement)object;
      return getId().equals(otherElement.getId());
    }
    return false;
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected void init(File aFile)
  {
    setFile(aFile);
    setName(futil().javaFilename(aFile.getPath()));
    setId(futil().standardize(aFile.getAbsolutePath()));
    if (sys().isWindows())
    {
      setId(getId().toLowerCase());
    }
    if (exists())
    {
      if (isDirectory())
      {
        setValid(true);
      }
      else
      {
        if (isFile())
        {
          setValid(checkArchive());
        }
      }
    }
  }

  protected boolean checkArchive()
  {
    try
    {
      getZipFile();
      if (!leaveArchiveOpen())
      {
        close();
      }
      return true;
    }
    catch (Exception e)
    {
      handleException(e);
      return false;
    }
  }

  protected boolean dirContainsFile(String filename)
  {
    File aFile;

    aFile = getFileFromDirectory(filename);
    if (aFile == null)
    {
      return false;
    }
    return aFile.exists();
  }

  protected File getFileFromDirectory(String filename)
  {
    File aFile;

    aFile = new File(getFile(), filename);
    return aFile;
  }

  protected boolean archiveContainsFile(String filename)
  {
    return getFileFromArchive(filename) != null;
  }

  protected ZipEntry getFileFromArchive(String filename)
  {
    ZipFile aZipFile;
    ZipEntry entry;

    try
    {
      aZipFile = getZipFile();
    }
    catch (Exception e)
    {
      handleException(e);
      return null;
    }
    try
    {
      entry = aZipFile.getEntry(filename);
      return entry;
    }
    catch (Exception e)
    {
      handleException(e);
      return null;
    }
    finally
    {
      if (!leaveArchiveOpen())
      {
        close();
      }
    }
  }

  /**
   * Opens and returns the zip file. The zip file stays open until close()
   * is called explicitly.
   * @see #close()
   */
  protected synchronized ZipFile getZipFile() throws ZipException, IOException
  {
    if (zipFile == null)
    {
      zipFile = new ZipFile(getFile());
    }
    return zipFile;
  }
  
  protected void handleException(Exception ex) 
  {
    if (DEBUG)
    {
      ex.printStackTrace();
    }
  }

  protected void setName(String newValue)
  {
    name = newValue;
  }

  protected File getFile()
  {
    return file;
  }

  protected void setFile(File newValue)
  {
    file = newValue;
  }

  protected boolean getValid()
  {
    return valid;
  }

  protected void setValid(boolean newValue)
  {
    valid = newValue;
  }

  protected String getId()
  {
    return id;
  }

  protected void setId(String newValue)
  {
    id = newValue;
  }

  protected void setZipFile(ZipFile newValue)
  {
    zipFile = newValue;
  }

  protected void debug(String text)
  {
    logger().logDebug(DEBUG_PREFIX + text);
  }

  protected Logger logger()
  {
    return LoggerProvider.getLogger();
  }

  protected FileUtil futil()
  {
    return FileUtil.current();
  }

  protected SysUtil sys()
  {
    return SysUtil.current();
  }
}
