// ===========================================================================
// CONTENT  : CLASS DirectoryContents
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 16/08/2003
// HISTORY  :
//  16/08/2003  mdu  CREATED
//
// Copyright (c) 2003, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.file;

import java.util.ArrayList;
import java.util.List;

import org.pfsw.util.CollectionUtil;

/**
 * Represents one directory path with all its contained files.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class DirectoryContents
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final CollectionUtil CU = CollectionUtil.current();

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private String dirName = null;
  private List<FileInfo> files = null;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with the directory's name.
   */
  public DirectoryContents(String name)
  {
    this(name, new ArrayList<FileInfo>());
  }

  /**
   * Initialize the new instance with the directory's name and its content
   */
  public DirectoryContents(String name, FileInfo[] fileInfos)
  {
    this(name, CU.asList(fileInfos));
  }

  /**
   * Initialize the new instance with the directory's name and its content
   */
  public DirectoryContents(String name, List<FileInfo> fileInfos)
  {
    super();

    if (name == null)
    {
      throw new IllegalArgumentException("name must not be null!");
    }

    setDirName(name);
    setFiles(new ArrayList<FileInfo>());
    getFiles().addAll(fileInfos);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the name of the directory this object represents
   */
  public String getDirectoryName()
  {
    return getDirName();
  }

  /**
   * Returns the number of files that are directly exist in this directory
   */
  public int numberOfContainedFiles()
  {
    return getFiles().size();
  }

  /**
   * Returns an array of information objects on all files contained in this 
   * directory
   */
  public FileInfo[] getContainedFiles()
  {
    FileInfo[] array;

    array = new FileInfo[numberOfContainedFiles()];
    getFiles().toArray(array);
    return array;
  }

  /**
   * Adds the given file info to the list of contained file info objects
   */
  public void addFileInfo(FileInfo info)
  {
    if (info != null)
    {
      getFiles().add(info);
    }
  }

  @Override
  public String toString()
  {
    return String.format("%s('%s')", getClass().getSimpleName(), getDirectoryName());
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================

  protected String getDirName()
  {
    return dirName;
  }

  protected void setDirName(String newValue)
  {
    dirName = newValue;
  }

  protected List<FileInfo> getFiles()
  {
    return files;
  }

  protected void setFiles(List<FileInfo> newValue)
  {
    files = newValue;
  }

  /**
   * Returns a display string for an inspector.
   * Active JOI support !
   */
  protected String inspectString()
  {
    return "DirectoryContents(\"" + getDirectoryName() + "\")";
  }
}
