// ===========================================================================
// CONTENT  : CLASS ExtendedFileFilter
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.0 - 06/12/2014
// HISTORY  :
//  24/01/2000  duma  CREATED
//	14/02/2003	duma	added		->	Support for digit wildcard characher
//	21/12/2005	duma	added		->	addPatterns()
//	25/02/2006	mdu		added		->	implements IObjectFilter
//  06/12/2014  mdu   changed ->  to hold StringFilter rather than StringPattern
//
// Copyright (c) 2000-2014, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.file;

import java.io.File;
import java.io.FilenameFilter;
import java.util.ArrayList;
import java.util.List;

import org.pfsw.bif.filter.IObjectFilter;
import org.pfsw.bif.text.IStringFilter;
import org.pfsw.text.StringPattern;
import org.pfsw.text.StringUtil;

/**
 * This filter implements the standard pattern matching on UNIX and Windows
 * platforms. It supports the wildcards '*' and '?' on file names.  <br>
 * It allows to set more than one pattern.
 * Apart from that it allows control over inclusion/exclusion of directories 
 * independently from name patterns.
 *
 * @author Manfred Duchrow
 * @version 2.0
 */
public class ExtendedFileFilter implements FilenameFilter, IObjectFilter<Object>
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  /**
   * The character to be used to separate filename patterns (';') as String.
   */
  public static final String PATTERN_SEPARATOR = ";";

  protected final static int DIR_CHECK_NAME = 1;
  protected final static int DIR_INCLUDE = 2;
  protected final static int DIR_EXCLUDE = 3;

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private List<IStringFilter> stringFilterList = new ArrayList<IStringFilter>();
  private int dirHandling = DIR_CHECK_NAME;
  private boolean restrictiveMode = false;
  private Character digitWildcard = null;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================

  /**
   * Initialize the new instance with default values.
   * <br>
   * The default for "restrictive" is false! See other constructor for 
   * explanation.
   */
  public ExtendedFileFilter()
  {
    super();
  }

  /**
   * Initialize the new instance with patterns.
   * 
   * @param patternList A list of patterns (may be just one)
   * @see #addPatterns(String) 
   */
  public ExtendedFileFilter(String patternList)
  {
    super();
    addPatterns(patternList);
  }

  /**
   * Initialize the new instance with filters.
   * 
   * @param nameFilters An optional array of name filters (must not be null).
   */
  public ExtendedFileFilter(IStringFilter... nameFilters)
  {
    super();
    addNameFilters(nameFilters);
  }

  /**
   * Initialize the new instance with a wildcard character for digits.
   * <br>
   * The default for "restrictive" is false! See other constructor for 
   * explanation.
   * 
   * @param wildcardForDigits The character that stands for any digit (0-9) in a filename
   */
  public ExtendedFileFilter(char wildcardForDigits)
  {
    super();
    setDigitWildcard(new Character(wildcardForDigits));
  }

  /**
   * Initialize the new instance with the option of a more restrictive handling
   * of wildcard matching. If the argument restrictive is true, empty strings
   * won't match the '*' wildcard character.
   * <p>
   * Example:
   * restrictive=true  => "text.doc" does NOT match the pattern "text*.doc"
   * <br>
   * restrictive=false => "text.doc" does match the pattern "text*.doc"
   * 
   * @param restrictive To restrict the match of '*' to at least one character
   */
  public ExtendedFileFilter(boolean restrictive)
  {
    super();
    setRestrictiveMode(restrictive);
  }

  /**
   * Initialize the new instance with a wildcard character for digits in 
   * filenames and the option of a more restrictive handling
   * of wildcard matching. If the argument restrictive is true, empty strings
   * won't match the '*' wildcard character.
   * <p>
   * See other constructor for examples.
   * 
   * @param wildcardForDigits The character that stands for any digit (0-9) in a filename
   * @param restrictive To restrict the match of '*' to at least one character
   */
  public ExtendedFileFilter(char wildcardForDigits, boolean restrictive)
  {
    this(wildcardForDigits);
    setRestrictiveMode(restrictive);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================

  /**
   * Adds a pattern. All filenames match this pattern are acceptable.   <br>
   * Case sensitivity is switched on !
   * 
   * @param pattern The pattern string containing  optional wildcards ( '*', '?' ) 
   */
  public void addPattern(String pattern)
  {
    StringPattern stringPattern = null;

    stringPattern = createStringPattern(pattern, false);
    getStringFilterList().add(stringPattern);
  }

  /**
   * Adds a pattern. All filenames match this pattern are acceptable.   <br>
   * Case sensitivity is switched on !
   * The second parameter specifies a character that will be recognized in the
   * pattern as a placeholder for a single digit character.  <p>
   * A patterb "XX-####.log" with a digitWildcard set to '#' wil match to
   * "XX-2000.log" and "XX-7376.log" but not to "XX-C363.log" and "XX-dddd.log".
   * 
   * @param pattern The pattern string containing  optional wildcards ( '*', '?' ) 
   * @param digitWildcardChar The character that will be treated as wildcard for digits ('0'-'9')
   */
  public void addPattern(String pattern, char digitWildcardChar)
  {
    StringPattern stringPattern = null;

    stringPattern = createStringPattern(pattern, false);
    stringPattern.setDigitWildcardChar(digitWildcardChar);
    getStringFilterList().add(stringPattern);
  }

  /**
   * Adds a pattern. All filenames match this pattern are acceptable.
   * 
   * @param pattern The pattern string containing  optional wildcards ( '*', '?' ) 
   * @param ignoreCase If true, all character comparisons are ignoring uppercase/lowercase
   */
  public void addPattern(String pattern, boolean ignoreCase)
  {
    StringPattern stringPattern = null;

    stringPattern = createStringPattern(pattern, ignoreCase);
    getStringFilterList().add(stringPattern);
  }

  /**
   * Adds a pattern. All filenames that match this pattern are acceptable.
   * Additionally to the standard wildcards '*' and '?' a wildcard for single
   * digit characters ('0' - '9') can be specified here.
   * 
   * @param pattern The pattern string containing  optional wildcards ( '*', '?' ) 
   * @param ignoreCase If true, all character comparisons are ignoring uppercase/lowercase
   * @param digitWildcardChar The character that will be treated as wildcard for digits ('0'-'9')
   */
  public void addPattern(String pattern, boolean ignoreCase, char digitWildcardChar)
  {
    StringPattern stringPattern = null;

    stringPattern = createStringPattern(pattern, ignoreCase);
    stringPattern.setDigitWildcardChar(digitWildcardChar);
    getStringFilterList().add(stringPattern);
  }

  /**
   * Adds one or more patterns separated by semi-colon (';').
   * 
   * @param patternList The list of filename patterns
   */
  public void addPatterns(String patternList)
  {
    String[] parts;

    if (str().isNullOrEmpty(patternList))
    {
      return;
    }

    parts = extractPatterns(patternList);
    for (int i = 0; i < parts.length; i++)
    {
      if (hasDigitWildcard())
      {
        addPattern(parts[i], true, getDigitWildcardChar());
      }
      else
      {
        addPattern(parts[i], true);
      }
    }
  }

  public void addNameFilters(IStringFilter... stringFilters)
  {
    for (IStringFilter stringFilter : stringFilters)
    {
      getStringFilterList().add(stringFilter);
    }
  }

  /**
   * Sets the filter to only accept directories that match a defined pattern.
   */
  public void checkNameOfDirectories()
  {
    setDirHandling(DIR_CHECK_NAME);
  }

  /**
   * Sets the filter to always accept directories, even if they don't match
   * a given pattern.
   */
  public void alwaysIncludeDirectories()
  {
    setDirHandling(DIR_INCLUDE);
  }

  /**
   * Sets the filter to never accept directories.
   */
  public void alwaysExcludeDirectories()
  {
    setDirHandling(DIR_EXCLUDE);
  }

  /**
   * Tests if a specified file should be included in a file list.
   *
   * @param dir the directory in which the file was found.
   * @param name the name of the file.
   * @return true if and only if the name should be included in the file list, false otherwise.
   */
  @Override
  public boolean accept(File dir, String name)
  {
    File fileOrDir;
    boolean accepted;

    fileOrDir = new File(dir, name);
    if (fileOrDir.isDirectory())
    {
      if (mustIncludeDirectories())
      {
        postAcceptCheck(dir, name, fileOrDir, true);
        return true;
      }
      if (mustExcludeDirectories())
      {
        postAcceptCheck(dir, name, fileOrDir, false);
        return false;
      }
    }
    accepted = checkAgainstStringFilters(name);
    if (LoggerProvider.TRACING)
    {
      System.out.println("ExtendedFileFilter.accept('" + ((dir == null) ? "null" : dir.getAbsolutePath()) + "', '" + name + "') : " + accepted);
    }
    postAcceptCheck(dir, name, fileOrDir, accepted);
    return accepted;
  }

  /**
   * Returns if this filter matches the given object.
   * Currently the supported object types are java.lang.String (i.e. a filename),
   * java.io.File, org.pfsw.file.FileLocator (not for remote files), org.pfsw.file.FileInfo.
   * 
   *  @param object Any object
   *  @return true if the object matches this filter
   */
  @Override
  public boolean matches(Object object)
  {
    if (object instanceof String)
    {
      return matchesFile(new File((String)object));
    }
    if (object instanceof File)
    {
      return matchesFile((File)object);
    }
    if (object instanceof FileInfo)
    {
      return matchesFile(new File(((FileInfo)object).getFullName()));
    }
    if (object instanceof FileLocator)
    {
      FileLocator locator = (FileLocator)object;
      if (!locator.isRemote())
      {
        return matchesFile(new File(locator.getOriginalFileName()));
      }
    }
    return false;
  }

  /**
   * Returns true if the filter always accepts directories, even if they don't match
   * a given pattern.
   */
  public boolean mustIncludeDirectories()
  {
    return (getDirHandling() == DIR_INCLUDE);
  }

  /**
   * Returns true if the filter never accepts directories.
   */
  public boolean mustExcludeDirectories()
  {
    return (getDirHandling() == DIR_EXCLUDE);
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  /**
   * This method gets called after the accept check had happened.
   * Here it does nothing, but subclasses may override to do things like
   * logging or monitoring.
   * 
   * @param dir The directory in which the file is located
   * @param filename The name of the file without path 
   * @param file The combined directory and filename 
   * @param accepted The result of the acceptance check
   */
  protected void postAcceptCheck(File dir, String filename, File file, boolean accepted)
  {
    // This is a template method which optionally can be overridden by subclasses
  }

  protected boolean checkAgainstStringFilters(String name)
  {
    for (IStringFilter filter : getStringFilterList())
    {
      if (filter.matches(name))
      {
        return true;
      }
    }
    return false; // No filter matched
  }

  protected StringPattern createStringPattern(String pattern, boolean ignoreCase)
  {
    StringPattern strPattern;

    strPattern = new StringPattern(pattern, ignoreCase);
    strPattern.multiCharWildcardMatchesEmptyString(!getRestrictiveMode());
    return strPattern;
  }

  protected String[] extractPatterns(String pattern)
  {
    return str().parts(pattern, PATTERN_SEPARATOR);
  }

  protected char getDigitWildcardChar()
  {
    if (hasDigitWildcard())
    {
      return getDigitWildcard().charValue();
    }
    return '\0';
  }

  protected boolean hasDigitWildcard()
  {
    return getDigitWildcard() != null;
  }

  protected boolean matchesFile(File file)
  {
    File dir;
    String filename;

    dir = file.getParentFile();
    filename = file.getName();
    return accept(dir, filename);
  }

  protected List<IStringFilter> getStringFilterList()
  {
    return stringFilterList;
  }

  protected void setStringFilterList(List<IStringFilter> newValue)
  {
    stringFilterList = newValue;
  }

  protected int getDirHandling()
  {
    return dirHandling;
  }

  protected void setDirHandling(int newValue)
  {
    dirHandling = newValue;
  }

  protected boolean getRestrictiveMode()
  {
    return restrictiveMode;
  }

  protected void setRestrictiveMode(boolean newValue)
  {
    restrictiveMode = newValue;
  }

  protected Character getDigitWildcard()
  {
    return digitWildcard;
  }

  protected void setDigitWildcard(Character newValue)
  {
    digitWildcard = newValue;
  }

  protected StringUtil str()
  {
    return StringUtil.current();
  }
}
