// ===========================================================================
// CONTENT  : CLASS FileCopyProcessor
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 12/01/2014
// HISTORY  :
//  12/01/2014  mdu  CREATED
//
// Copyright (c) 2014, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.file;

import java.io.File;
import java.io.IOException;

/**
 * Copies each file to a specific target folder. 
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class FileCopyProcessor extends AFileProcessor
{
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private File targetFolder;
  private Exception exception = null;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initializes the new instance with a target folder.
   * 
   * @param targetFolder The folder to which the files must be copied.
   * @throws IllegalArgumentException if the given target folder is no directory
   * or does not exist. 
   */
  public FileCopyProcessor(File targetFolder)
  {
    super();
    if (targetFolder == null)
    {
      throw new IllegalArgumentException("The target folder must not be null!");
    }
    if (!targetFolder.exists())
    {
      throw new IllegalArgumentException("The target folder " + targetFolder.getAbsolutePath() + " does not exist!");
    }
    if (!targetFolder.isDirectory())
    {
      throw new IllegalArgumentException("The specified target " + targetFolder.getAbsolutePath() + " is no directory!");
    }
    setTargetFolder(targetFolder);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  public File getTargetFolder()
  {
    return targetFolder;
  }

  /**
   * Returns the exception that occurred during file copying or null if 
   * all was ok.
   */
  public Exception getException()
  {
    return exception;
  }

  /**
   * Copies the given file to the target folder. If any exception occurs, the
   * method returns false (i.e. terminate processing) and provides the exception
   * via {@link #getException()} method.
   */
  @Override
  public boolean processFile(final File file)
  {
    File targetFile;

    targetFile = construcTargetFile(file);
    try
    {
      fileUtil().copyFile(file, targetFile, true);
    }
    catch (IOException ex)
    {
      handleException(ex);
      return false;
    }
    return true;
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected File construcTargetFile(File file)
  {
    return new File(getTargetFolder(), file.getName());
  }

  protected void setTargetFolder(File folder)
  {
    if (LoggerProvider.TRACING)
    {
      System.out.println("FileCopyProcessor.setTargetFolder('" + folder + "')");
    }
    targetFolder = new File(fileUtil().standardize(folder.getAbsolutePath()));
    if (LoggerProvider.TRACING)
    {
      System.out.println("FileCopyProcessor.setTargetFolder() using targetFolder: " + getTargetFolder().getAbsolutePath());
    }
  }

  protected void handleException(Exception ex)
  {
    if (LoggerProvider.TRACING)
    {
      System.out.println("FileCopyProcessor.handleException() : targetFolder: " + getTargetFolder().getAbsolutePath());
      ex.printStackTrace(System.out);
    }
    LoggerProvider.getLogger().logWarning("FileCopyProcessor problem: ", ex);
    exception = ex;
  }

  protected FileUtil fileUtil()
  {
    return FileUtil.current();
  }
}
