// ===========================================================================
// CONTENT  : CLASS FileDirectoryScannerFileHandler
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.1 - 20/12/2003
// HISTORY  :
//  16/08/2003  mdu  	CREATED
//	20/12/2003	mdu		changed	-->	Use LoggerProvider in handleException()
//
// Copyright (c) 2003, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.file;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Stack;

/**
 * This is a helper class for FileDirectoryScanner which implements the
 * FileHandler interface and collects the data provided by a FileWalker.
 *
 * @author Manfred Duchrow
 * @version 1.1
 */
class FileDirectoryScannerFileHandler implements FileHandler
{
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private Stack<DirectoryContents> dirStack = null;
  protected Stack<DirectoryContents> getDirStack()
  {
    return dirStack;
  }
  protected void setDirStack(Stack<DirectoryContents> newValue)
  {
    dirStack = newValue;
  }

  private TableOfContents toc = null;
  protected TableOfContents toc()
  {
    return toc;
  }
  protected void toc(TableOfContents newValue)
  {
    toc = newValue;
  }

  private int baseDirLength = 0;
  protected int baseDirLength()
  {
    return baseDirLength;
  }
  protected void baseDirLength(int newValue)
  {
    baseDirLength = newValue;
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   * @throws IOException If the start directory doesn't exist or is no directory
   */
  protected FileDirectoryScannerFileHandler(String baseDir) throws IOException
  {
    super();
    initialize(baseDir);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * This method is called for each directory, that a FileWalker starts to walk through.
   * It must return true, if the FileWalker should continue. To stop the
   * calling FileWalker it can return false.
   *
   * @param dir The directory, the FileWalker is starting to walk through
   * @param count The number of files and directories the FileWalker found in the directory
   * @return true to continue, false to terminate processing of files
   */
  @Override
  public boolean directoryStart(File dir, int count)
  {
    DirectoryContents dirContents;
    String dirName;

    dirName = dir.getAbsolutePath();
    dirName = normalizeDirName(dirName);
    if (dirName.length() > baseDirLength())
    {
      dirName = dirName.substring(baseDirLength());
    }
    else
    {
      dirName = "";
    }

    dirContents = new DirectoryContents(dirName);
    toc().add(dirContents);
    setCurrentDirContents(dirContents);

    return true;
  }

  /**
   * Remove the completed directory contents object from the stack
   */
  @Override
  public boolean directoryEnd(File dir)
  {
    removeCurrentDirContents();
    return true;
  }

  /**
   * This method is called for whenever an exception occurs in walking through
   * the directories.   <br>
   * The method must return true, if the FileWalker should continue. To stop the
   * calling FileWalker it can return false.
   * By default this method forwards the given exception to the 
   * LoggerProvider.logError( msg, exception ).
   *
   * @param ex The exception to handle
   * @param The file, currently found by the FileWalker instance
   * @return true to continue, false to terminate processing of files
   */
  @Override
  public boolean handleException(Exception ex, File file)
  {
    LoggerProvider.getLogger().logError("Error on walking over file: " + file.getPath(), ex);
    return true;
  }

  /**
   * This method is called for each file, that a FileWalker instance finds.
   * It must return true, if the FileWalker should continue. To stop the
   * calling FileWalker it can return false.
   *
   * @param The file, currently found by the FileWalker instance
   * @return true to continue, false to terminate processing of files
   */
  @Override
  public boolean handleFile(File file)
  {
    registerFileInfo(new FileInfo(file));
    return true;
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected void registerFileInfo(FileInfo fileInfo)
  {
    getCurrentDirContents().addFileInfo(fileInfo);
  }

  protected DirectoryContents getCurrentDirContents()
  {
    return getDirStack().peek();
  }

  protected void setCurrentDirContents(DirectoryContents newValue)
  {
    getDirStack().push(newValue);
  }

  protected DirectoryContents removeCurrentDirContents()
  {
    return getDirStack().pop();
  }

  protected String normalizeDirName(String dirName)
  {
    String pathName = null;

    pathName = dirName.replace('\\', '/');
    if (pathName.endsWith("/"))
    {
      pathName = pathName.substring(0, pathName.length() - 1);
    }
    return pathName;
  }

  protected void initialize(String baseDirName) throws IOException
  {
    int len;
    String dirName;
    File startDir;

    startDir = new File(baseDirName);
    if (!startDir.exists())
    {
      throw new FileNotFoundException("Directory '" + baseDirName + "' not found.");
    }

    if (!startDir.isDirectory())
    {
      throw new IOException("'" + baseDirName + "' is not a directory.");
    }

    dirName = startDir.getAbsolutePath();
    dirName = normalizeDirName(dirName);
    len = dirName.length();
    if (!dirName.endsWith("/"))
    {
      len++;
    }
    setDirStack(new Stack<DirectoryContents>());
    toc(new TableOfContents(baseDirName));
    baseDirLength(len);
  }

}
