// ===========================================================================
// CONTENT  : CLASS FileInfo
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.3 - 27/04/2008
// HISTORY  :
//  22/07/2002  duma  CREATED
//	14/03/2003	duma	changed	->	Use FileUtil now
//	20/12/2005	duma	added		->	asFileLocator()
//	27/04/2008	mdu		added		->	asFile()
//
// Copyright (c) 2002-2008, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.file;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.zip.ZipEntry;

/**
 * This data object holds the name, size and modification date of a file.
 * The file can either be a normal file or one inside a zip archive.
 *
 * @author Manfred Duchrow
 * @version 1.3
 */
public class FileInfo
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final String PATH_SEPARATOR = "/";
  private static final String ARCHIVE_INDICATOR = "!";

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private String archiveName = null;
  private String path = null;
  private String name = null;
  private long size = 0L;
  private long lastModified = 0L;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance from the given file.
   */
  public FileInfo(File file)
  {
    super();
    setPathAndName(file);
    setSize(file.length());
    lastModified(file.lastModified());
  }

  /**
   * Initialize the new instance from the given zip file entry.
   */
  public FileInfo(String archiveName, ZipEntry zipEntry)
  {
    super();
    setArchiveName(archiveName);
    setPathAndName(new File(zipEntry.getName()));
    setSize(zipEntry.getSize());
    lastModified(zipEntry.getTime());
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the name of the archive, the file is included in or null if the
   * file is located in the normal directory structure.
   */
  public boolean isInArchive()
  {
    return getArchiveName() != null;
  }

  /**
   * Returns the name of the archive, the file is included in or null if the
   * file is located in the normal directory structure.
   */
  public String getArchiveName()
  {
    return archiveName;
  }

  /**
   * Returns the full location name, that is in case of a normal file the
   * path and filename combined.
   * In case of an archived file it is the archive name + path + filename.
   * <br>
   * The path separator is a forward slash ('/') on any platform !
   */
  public String getLocation()
  {
    if (isInArchive())
    {
      return getArchiveName() + PATH_SEPARATOR + getFullName();
    }
    return getFullName();
  }

  /**
   * Returns the path + name of the file.
   * <br>
   * The path separator is a forward slash ('/') on any platform !
   */
  public String getFullName()
  {
    return getPath() + PATH_SEPARATOR + getName();
  }

  /**
   * Returns the path (without filename) to the file.
   * <br>
   * The path separator is a forward slash ('/') on any platform !
   */
  public String getPath()
  {
    return path;
  }

  /**
   * Returns the name of the file (without any path).
   */
  public String getName()
  {
    return name;
  }

  /**
   * Returns the size of the file in bytes
   */
  public long getSize()
  {
    return size;
  }

  /**
   * Returns the timestamp of the file's last modification
   */
  public long lastModified()
  {
    return lastModified;
  }

  /**
   * Returns a FileLocator that corresponds to the file specified by this
   * FileInfo object.
   */
  public FileLocator asFileLocator()
  {
    return FileLocator.create(getLocation());
  }

  /**
   * Returns the file this info objects describes or null if it is inside
   * an archive.
   */
  public File asFile()
  {
    if (isInArchive())
    {
      return null;
    }
    return new File(getPath(), getName());
  }

  /**
   * Returns the URL of the file this info objects describes.
   * 
   * @throws MalformedURLException if no appropriate URL can be created. 
   */
  public URL asURL() throws MalformedURLException
  {
    File file;
    String fileStr;
    String urlStr;

    if (isInArchive())
    {
      file = new File(getArchiveName());
      fileStr = file.getAbsolutePath() + ARCHIVE_INDICATOR + PATH_SEPARATOR + getFullName();
    }
    else
    {
      fileStr = asFile().getAbsolutePath();
    }
    urlStr = "jar:file:/" + FileUtil.current().javaFilename(fileStr);
    return new URL(urlStr);
  }

  @Override
  public String toString()
  {
    return "FileInfo(" + getName() + ")";
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected void setArchiveName(String newValue)
  {
    archiveName = newValue;
  }

  protected void setPath(String newValue)
  {
    path = newValue;
  }

  protected void setName(String newValue)
  {
    name = newValue;
  }

  protected void setSize(long newValue)
  {
    size = newValue;
  }

  protected void lastModified(long newValue)
  {
    lastModified = newValue;
  }

  protected void setPathAndName(File file)
  {
    setPath(util().javaFilename(file.getParent()));
    setName(file.getName());
  }

  protected FileUtil util()
  {
    return FileUtil.current();
  }

}
