// ===========================================================================
// CONTENT  : CLASS FileLocator
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.0 - 12/02/2017
// HISTORY  :
//  17/05/2002  duma  CREATED
//	24/05/2002	duma	added		->	toURL(), isFile(), isDirectory(), getAbsolutePath()
//	21/06/2002	duma	added		->	realFile()
//	14/03/2003	duma	added		->	getStandardizedPath(), getStandardizedAbsolutePath()
//	20/12/2003	duma	added		->	logger() / changed all debug statements to use logger
//	03/09/2004	duma	added		->	support for remote files via http://
//	02/10/2004	duma	added		->	support for autho authentication with remote files
//	03/03/2006	mdu		bugfix	->	convertFromURLSyntax() has cut off leading slash on unix
//  12/02/2017  mdu   added   ->  better handling of temporary files
//
// Copyright (c) 2002-2017, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.file ;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import org.pfsw.bif.lifecycle.IReleasable;
import org.pfsw.bif.text.IStringFilter;
import org.pfsw.logging.Logger;
import org.pfsw.security.authentication.AutoAuthenticationManager;
import org.pfsw.text.StringPatternCollection;
import org.pfsw.text.StringUtil;
import org.pfsw.util.SysUtil;

/**
 * This class mainly supports access to files which can be in the normal
 * file directory structure or inside zip archives.
 * The main purpose is to provide methods that transparently treat files 
 * the same way whether they are in the normal directory structure or
 * inside archives.
 * The syntax is simply to allow archive names in a path name at any place
 * where a sub-directory name can be. <br>
 * Examples: <br>
 * <ul>
 *   <li>d:\temp\archive.zip\config\nls.properties</li>
 *	 <li>/usr/java/jdk1.3/src.jar/java/io/File.java</li>
 *	 <li>jar:http://myserver.com/code/src.jar!/settings.properties</li>
 * </ul>
 * @author Manfred Duchrow
 * @version 2.0
 */
public class FileLocator implements IReleasable
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
	private static final boolean DEBUG = "true".equals( System.getProperty( "org.pfsw.file.debug" ) ) ;
	private static final String ARCHIVE_INDICATOR				= "!" + File.separator ;
	private static final String JAR_INDICATOR						= "jar:" ;

	private static final String HTTP_INDICATOR 		= "http://" ;
	private static final String HTTPS_INDICATOR 	= "https://" ;
	private static final String JAR_HTTP_INDICATOR 		= "jar:http://" ;
	private static final String JAR_HTTPS_INDICATOR 	= "jar:https://" ;

	private static final String[] SUPPORTED_REMOTE_PROTOCOLS =
		{ HTTP_INDICATOR, HTTPS_INDICATOR, JAR_HTTP_INDICATOR, JAR_HTTPS_INDICATOR } ;
	
	private static final IStringFilter JAR_FILE_INDICATORS = 
		new StringPatternCollection( new String[] { "jar:file:/*", "jar:file:\\*" } ) ;
	
	public static final String TEMP_FILE_PREFIX = "FLOC_" ; 
	public static final String TEMP_FILE_SUFFIX = ".xtr" ; 
	
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private FileLocator parentLocator = null ;
  private File file = null ;
  private File tempFile = null ;
  private ZipFile zipFile = null ;
  private Boolean exists = Boolean.TRUE ;
  private Exception exception = null ;
  private String originalFileName = null ;
  private boolean remote = false ;
  private boolean isResolved = false ;
  
  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Create a file locator that corresponds to the given file name.
   */
  public static FileLocator create(File file)
  {
    FileLocator locator = new FileLocator(file.getPath(), false);

    return locator.createFrom(file);
  }

  /**
   * Create a file locator that corresponds to the given URL.
   * Allowed remote URLs start with either "http://" or "https://"
   * or "jar:http://" or "jar:https://". 
   * 
   * @param url The url that specifies the file
   */
  public static FileLocator create(URL url)
  {
    return create(url.toString());
  }

  /**
   * Create a file locator that corresponds to the given file name.
   * Remote files are also supported for filenames starting with 
   * "http://" or "https://". 
   * 
   * @param filename The name (or URL) that specifies the file
   */
  public static FileLocator create(String filename)
  {
    if (usesSupportedRemoteProtocol(filename))
    {
      return new FileLocator(filename, true);
    }
    if (JAR_FILE_INDICATORS.matches(filename))
    {
      return createFromJarFileUrl(filename);
    }
    return create(new File(filename));
  } 

  /**
   * Create a file locator that represents an already extracted file
   * from inside an archive.
   * 
   * @param parent The file locator that represents the parent archive (must not be null).
   * @param zipEntryName The full name of the entry in the zip archive (must not be null.
   * @param extractedFile The file that has been extracted from the parent archive (must not be null).
   * @return The new locator. 
   */
  public static FileLocator createExtractedFile(FileLocator parent, String zipEntryName, File extractedFile)
  {
    FileLocator locator;
    
    locator = new FileLocator();
    locator.setParentLocator(parent);
    locator.setResolvedAsTempFile(extractedFile);
    locator.setOriginalFileName(zipEntryName);
    return locator;
  } 
  
  private static FileLocator createFromJarFileUrl(String filename)
  {
    String simpleFilename;
    FileLocator locator;

    simpleFilename = filename.substring(JAR_INDICATOR.length());
    simpleFilename = str().replaceAll(simpleFilename, ARCHIVE_INDICATOR, "/");
    locator = create(new File(simpleFilename));
    locator.setOriginalFileName(filename);
    return locator;
  }

  private static FileLocator newWith(FileLocator aParent, String[] pathElements) throws Exception
  {
    FileLocator locator = new FileLocator();

    return locator.createFrom(aParent, pathElements);
  } 

  protected static boolean usesSupportedRemoteProtocol(String filename)
  {
    if (filename == null)
      return false;

    for (int i = 0; i < SUPPORTED_REMOTE_PROTOCOLS.length; i++)
    {
      if (filename.startsWith(SUPPORTED_REMOTE_PROTOCOLS[i]))
        return true;
    }
    return false;
  } 

  protected static StringUtil str()
  {
    return StringUtil.current();
  } 

  protected static FileUtil fileUtil()
  {
    return FileUtil.current();
  } 

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  FileLocator()
  {
    super();
  } 

  /**
   * Initialize the new instance with default values.
   */
  private FileLocator(String filename, boolean remoteFile)
  {
    this();
    setRemote(remoteFile);
    if (remoteFile)
    {
      setExists(null);
    }
    setOriginalFileName(filename);
  } 

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the file that contains the data the locator points to.
   * If the locator points to a normal file in a directory, than
   * this file will be returned.
   * If the locator points to a file inside an archive, the file
   * will be unzipped into the <i><b>tmp</i></b> directory and this
   * temp file will be returned.
   * If the locator points to a remote file it will be downloaded to the
   * <i><b>temp</i></b> directory and this temp file will be returned. 
   * If the locator points to a none existing file, this method 
   * returns null.
   */
  public File realFile()
  {
    File aFile = null;

    if (resolveIfNecessary())
    {
      try
      {
        aFile = fileRef();
      }
      catch (Exception e)
      {
        handleException(e);
        aFile = null;
      }
    }
    return aFile;
  }

  /**
   * Checks if the referred file has already been resolved.
   * If not it tries to resolve it. That includes unzipping from
   * archive as well as download of remote file.
   * 
   * @return true if file is already or just has been resolved, otherwise false.
   */
  protected boolean resolveIfNecessary()
  {
    if (isResolved())
    {
      return true;
    }
    if (isUnlocatedRemoteFile()) // Not yet checked
    {
      copyFromRemote();
      if (getExists() == null)
      {
        return false;
      }
    }
    try
    {
      // This ensures unzipping to a temporary file.
      fileRef();
      return true;
    }
    catch (Exception e)
    {
      handleException(e);
      return false;
    }
  }

  /**
   * Returns whether or not the file specified by this locator exists.
   * If the locator points to a remote file, it will be downloaded when
   * calling this method.
   */
  public boolean exists()
  {
    File aFile;
    
    resolveIfNecessary();
    aFile = getUnderlyingRealFile();
    return aFile == null ? false : aFile.exists();
  }

  /**
   * Returns whether or not the name specified by this locator 
   * points to a file.
   */
  public boolean isFile()
  {
    try
    {
      if (exists())
      {
        if (isTempFile())
        {
          return getTempFile().isFile();
        }
        return isFileElement(getFile());
      }
      return false;
    }
    catch (Exception e)
    {
      handleException(e);
      return false;
    }
  }

  /**
   * Returns whether or not the name specified by this locator 
   * points to a directory.
   */
  public boolean isDirectory()
  {
    try
    {
      return isDirectoryElement(getFile());
    }
    catch (Exception e)
    {
      handleException(e);
      return false;
    }
  } 

  /**
   * Returns true if the locator points to a remote file or directory.
   */
  public boolean isRemote()
  {
    return getRemote();
  }

  /**
   * Returns true if the underlying file has been resolved and copied
   * to a temporary file in the local file system.
   */
  public boolean isTempFile()
  {
    return getTempFile() != null;
  }
  
  /**
   * Returns the size of the file or 0 if it does not exist.
   */
  public long size()
  {
    ZipEntry entry;

    exists();

    try
    {
      if (isInArchive())
      {
        entry = archiveEntry();
        // if ( DEBUG ) com.mdcs.joi.Inspector.inspectWait( entry ) ;
        return entry.getSize();
      }
      return getFile().length();
    }
    catch (Exception ex)
    {
      if (DEBUG)
      {
        logger().logException(ex);
      }
      return 0L;
    }
  }

  /**
   * Returns the timestamp of when the file was last modified 
   * or 0 in any case of error.
   */
  public long lastModified()
  {
    ZipEntry entry;

    exists();

    try
    {
      if (isInArchive())
      {
        entry = archiveEntry();
        return entry.getTime();
      }
      return getFile().lastModified();
    }
    catch (Exception ex)
    {
      if (DEBUG)
        logger().logException(ex);
      return 0L;
    }
  }

  /**
   * Returns an opened input stream on the file defined by this locator.
   */
  public InputStream getInputStream() throws IOException
  {
    ZipEntry entry;

    if (!exists())
    {
      throw new FileNotFoundException(getAbsolutePath());
    }

    if (isInArchive())
    {
      if (isTempFile())
      {        
        return new FileInputStream(getTempFile());
      }
      entry = archiveEntry();
      return container().getInputStream(entry);
    }
    return new FileInputStream(getFile());
  } 

  /**
   * Returns the parent of the file represented by this locator
   */
  public FileLocator getParent()
  {
    String path;

    path = str().cutTail(getStandardizedAbsolutePath(), "/");
    return FileLocator.create(path);
  }

  /**
   * Returns whether or not the file specified by this locator 
   * is inside an archive.
   */
  public boolean isInArchive()
  {
    return getParentLocator() != null;
  }

  /**
   * Returns the full pathname.
   */
  public String getPath()
  {
    return fullFilePath(false).getPath();
  }

  /**
   * Returns the full absolute pathname.
   */
  public String getAbsolutePath()
  {
    if (isRemote())
    {
      return getOriginalFileName();
    }
    return fullFilePath(true).getPath();
  } 

  /**
   * Returns the full pathname in a standardized for.
   * That is all ".." and "." elements are removed and forward slashes are 
   * used as separators of the remaining elements.
   */
  public String getStandardizedPath()
  {
    return fileUtil().standardize(getPath());
  } 

  /**
   * Returns the full absolute pathname in a standardized form.
   * That is all ".." and "." elements are removed and forward slashes are 
   * used as separators of the remaining elements.
   */
  public String getStandardizedAbsolutePath()
  {
    return fileUtil().standardize(getAbsolutePath());
  }  

  /**
   * Returns the last exception that occured while using this locator
   * or null, if no exception was thrown at all.
   */
  public Exception exception()
  {
    return getException();
  } 

  // -------------------------------------------------------------------------

  /**
   * Returns the name of the file as an URL.
   */
  public URL toURL() throws MalformedURLException
  {
    StringBuffer buffer = new StringBuffer(128);

    urlPath(buffer);
    return new URL(buffer.toString());
  }

  @Override
  public String toString()
  {
    return getStandardizedAbsolutePath();
  }

  /**
   * Releases any allocated objects (e.g. deleting temporary file).
   */
  @Override
  public void release()
  {
    if (getTempFile() != null)
    {
      getTempFile().delete();
      setTempFile(null);
      setIsResolved(false);
    }
    closeZipFile();
    if (getParentLocator() != null)
    {
      getParentLocator().release();
    }
  }
  
  /**
   * Returns the original (unchanged) filename as it was given to the create 
   * method of this locator.
   */
  public String getOriginalFileName()
  {
    return originalFileName;
  }
  
  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected FileLocator createFrom(File filePath)
  {
    FileLocator locator = null;
    String[] parts = null;
    File path = filePath;

    if (fileUtil().isLocalFileURL(path.getPath()))
    {
      path = convertFromURLSyntax(path);
    }

    parts = str().parts(path.getPath(), File.separator);
    try
    {
      locator = initFromPath(parts, path.getPath().startsWith(File.separator));
    }
    catch (Exception ex)
    {
      setException(ex);
      doesNotExist(path);
      locator = this;
    }
    return locator;
  } 

  protected FileLocator createFrom(FileLocator aParent, String[] pathElements) throws Exception
  {
    setParentLocator(aParent);
    return initFromPath(pathElements, false);
  }  

  protected FileLocator initFromPath(String[] parts, boolean startsFromRoot) throws Exception
  {
    FileLocator locator = this;
    File pathElement = null;
    String[] rest = null;
    boolean elementExists = false;
    String originalName;

    if (startsFromRoot)
    {
      pathElement = new File(File.separator);
    }

    for (int i = 0; i < parts.length; i++)
    {
      if (pathElement == null)
      {
        pathElement = new File(asRootPart(parts[i]));
      }
      else
      {
        pathElement = new File(pathElement, parts[i]);
      }

      elementExists = doesElementExist(pathElement);

      if (elementExists)
      {
        setFile(pathElement);
        if (isFileElement(pathElement))
        {
          if (DEBUG)
          {
            logger().logDebug("Locator(" + pathElement + ")");
          }
          if (i < (parts.length - 1)) // Is not last element ? 
          {
            originalName = getOriginalFileName();
            setOriginalFileName(pathElement.getPath());
            rest = str().copyFrom(parts, i + 1);
            locator = FileLocator.newWith(this, rest);
            locator.setOriginalFileName(originalName);
          }
          break;
        }
      }
      else
      {
        if (isInArchive())
        {
          if (i < (parts.length - 1)) // Is not last element ? 
          {
            // Directories are not always identifiable individually in zip archives.
            // Therefore it must be accepted that they are not found.
            // So in such case no exception will be thrown.
          }
          else
          {
            throw new Exception("\"" + pathElement.getPath() + "\" does not exist");
          }
        }
        else
        {
          throw new Exception("\"" + pathElement.getPath() + "\" does not exist");
        }
      }
    }
    closeZipFile(); // It might have been opened during check of elements existing
    return locator;
  }

  protected boolean doesElementExist(File element) throws Exception
  {
    if (isInArchive())
    {
      return doesElementExistInArchive(element.getPath());
    }
    return element.exists();
  } 

  protected boolean isFileElement(File element) throws Exception
  {
    if (isInArchive())
    {
      return isFileInArchive(getNormalizedFilePath());
    }
    return element.isFile();
  } 

  protected boolean isDirectoryElement(File element) throws Exception
  {
    if (isInArchive())
    {
      return isDirectoryInArchive(getNormalizedFilePath());
    }
    return element.isDirectory();
  } 
  
  protected boolean doesElementExistInArchive(String elementName) throws Exception
  {
    ZipEntry entry;

    entry = entryFromArchive(elementName);

    return (entry != null);
  } 

  protected boolean isFileInArchive(String elementName) throws Exception
  {
    ZipEntry entry;
    
    if (isDirectoryInArchive(elementName))
    {
      return false;
    }
    entry = entryFromArchive(elementName);

    // Unfortunately entry.isDirectory() returns false even for
    // pure directory entries inside a zip archive, so it can't be used here.
    // The trick below is problematic, because apart from 
    // directories it will also not recognize files with size 0.
    return (entry != null) && (entry.getSize() > 0);
  }  

  protected boolean isDirectoryInArchive(final String elementName)
  {
    ZipEntry entry;
    String path;
    
    if (elementName == null)
    {
      return false;
    }
    path = elementName.endsWith("/") ? elementName : elementName + "/";
    try
    {
      entry = entryFromArchive(path);
    }
    catch (IOException e)
    {
      handleException(e);
      return false;
    }
    return (entry != null) && entry.isDirectory();
  } 
  
  protected ZipEntry entryFromArchive(String elementName) throws IOException
  {
    ZipEntry entry;
    ZipFile archive;
    String name;

    name = str().replaceAll(elementName, "\\", "/");
    archive = container();
    entry = archive.getEntry(name);

    if (DEBUG)
    {
      StringBuffer buffer = new StringBuffer(100);
      // if ( entry == null ) com.mdcs.joi.Inspector.inspect( name ) ;
      buffer.append(archive.getName() + "::" + name + " --- " + (entry != null));
      if (entry == null)
      {
        buffer.append("\n");
      }
      else
      {
        buffer.append(" (" + entry.getSize() + ")");
        buffer.append(" (T:" + entry.getTime() + ")");
        buffer.append(" (" + (entry.isDirectory() ? "Dir" : "File") + ")\n");
      }
      logger().logDebug(buffer.toString());
    }

    return entry;
  } 

  protected ZipEntry archiveEntry() throws IOException
  {
    return entryFromArchive(getFile().getPath());
  } 

  protected void doesNotExist(File aFile)
  {
    setExists(Boolean.FALSE);
    setFile(aFile);
  } 

  protected File fullFilePath(boolean absolute)
  {
    File full;
    String subPath;

    if (!exists() && isRemote())
    {
      return null;
    }

    if (isInArchive())
    {
      if (getFile() == null)
      {
        subPath = getOriginalFileName() == null ? getTempFile().getPath() : getOriginalFileName();
      }
      else
      {
        subPath = getFile().getPath();
      }
      full = new File(getParentLocator().fullFilePath(absolute), subPath);
    }
    else
    {
      full = absolute ? getFile().getAbsoluteFile() : getFile();
    }
    return full;
  }

  protected void urlPath(StringBuffer buffer)
  {
    if (isRemote())
    {
      buffer.append(getOriginalFileName());
    }
    else
    {
      if (isInArchive())
      {
        getParentLocator().urlPath(buffer);
        buffer.append(ARCHIVE_INDICATOR);
        buffer.append(getFile().getPath());
      }
      else
      {
        buffer.append(fileUtil().convertToURLSyntax(getFile().getPath()));
      }
    }
  }

  /**
   * This either returns the underlying file from the local file system or
   * unzipps the file from the underlying archive into a temporary file and returns 
   * In any case of error null wil be returned.
   */
  protected File getUnderlyingRealFile()
  {
    try
    {
      return fileRef();
    }
    catch (IOException e)
    {
      handleException(e);
      return null;
    }
  }
  
  /**
   * This either returns the underlying file from the local file system or
   * unzipps the file from the underlying archive into a temporary file and returns 
   */
  protected File fileRef() throws IOException
  {
    InputStream archiveStream;
    ZipEntry entry;
    File temporaryFile;

    if (!isResolved())
    {      
      if (isInArchive())
      {
        if (getTempFile() == null)
        {
          if (isDirectoryInArchive(getNormalizedFilePath()))
          {            
            return null;
          }
          entry = archiveEntry();
          if (entry.isDirectory())
          {
            return null;
          }
          archiveStream = container().getInputStream(entry);
          if (archiveStream == null)
          {
            return null;
          }
          temporaryFile = fileUtil().copyToTempFile(archiveStream, TEMP_FILE_PREFIX, TEMP_FILE_SUFFIX);
          setResolvedAsTempFile(temporaryFile);
        }
      }
    }
    if (isTempFile())
    {
      return getTempFile();      
    }
    return getFile();
  }

  /**
   * Returns the file this locator presents as opened zip file or
   * null in any case of error.
   */
  protected ZipFile archive() throws IOException
  {
    if (getZipFile() == null)
    {
      setZipFile(new ZipFile(fileRef()));
    }
    return getZipFile();
  } 

  /**
   * Returns the zip file which is presented by the parent container
   * or null in any case of error.
   */
  protected ZipFile container() throws IOException
  {
    if (isInArchive())
    {
      return getParentLocator().archive();
    }
    return null;
  }
  
  protected void closeZipFile() 
  {
    if (getZipFile() != null)
    {
      fileUtil().close(getZipFile());
      setZipFile(null);
    }
  }

  protected File convertFromURLSyntax(File aFile)
  {
    String newStr;

    newStr = fileUtil().convertFromURLSyntax(aFile.getPath());
    newStr = str().replaceAll(newStr, ARCHIVE_INDICATOR, File.separator);

    return new File(newStr);
  }  

  protected String getNormalizedFilePath()
  {
    if (getFile() == null)
    {      
      return null;
    }
    return fileUtil().standardizeFilename(getFile().getPath());
  }

  protected boolean isUnlocatedRemoteFile()
  {
    return (getExists() == null);
  } 

  /**
   * This method must only be called if the original filename contains a
   * URL starting with one of the supported remote protocols!
   */
  protected void copyFromRemote()
  {
    File temporaryFile;

    if (getOriginalFileName() == null) // No filename available
    {
      return;
    }
    try
    {
      AutoAuthenticationManager.instance().aboutToAccess(getOriginalFileName());
      temporaryFile = fileUtil().copyToTempFile(getOriginalFileName(), TEMP_FILE_PREFIX, TEMP_FILE_SUFFIX, false);
    }
    catch (IOException e)
    {
      setException(e);
      if (DEBUG)
      {
        logger().logException(e);
      }
      return;
    }
    if (temporaryFile != null)
    {
      setExists(temporaryFile.exists() ? Boolean.TRUE : Boolean.FALSE);
      setResolvedAsTempFile(temporaryFile);
    }
  }  

  protected void setResolvedAsTempFile(File temporaryFile)
  {
    setTempFile(temporaryFile);
    setIsResolved(true);
  }

  protected String asRootPart(final String part) 
  {
    String rootPart;
    
    rootPart = part;
    if (SysUtil.current().isWindows())
    {
      if (rootPart.endsWith(":"))
      {
        rootPart += File.separator;
      }
    }
    return rootPart;
  }
  
  protected void setOriginalFileName(String newValue)
  {
    originalFileName = fileUtil().standardize(newValue);
  } 

  protected FileLocator getParentLocator()
  {
    return parentLocator;
  }
  
  protected void setParentLocator(FileLocator newValue)
  {
    parentLocator = newValue;
  }

  protected File getFile()
  {
    return file;
  }
  
  protected void setFile(File newValue)
  {
    file = newValue;
  }

  protected ZipFile getZipFile()
  {
    return zipFile;
  }
  
  protected void setZipFile(ZipFile newValue)
  {
    zipFile = newValue;
  }

  protected Boolean getExists()
  {
    return exists;
  }
  protected void setExists(Boolean newValue)
  {
    exists = newValue;
  }

  protected Exception getException()
  {
    return exception;
  }
  protected void setException(Exception newValue)
  {
    exception = newValue;
  }

  protected boolean getRemote()
  {
    return remote;
  }
  protected void setRemote(boolean newValue)
  {
    remote = newValue;
  }

  protected File getTempFile()
  {
    return tempFile;
  }
  
  protected void setTempFile(File aFile)
  {
    tempFile = aFile;
  }

  protected boolean isResolved()
  {
    return isResolved;
  }
  
  protected void setIsResolved(boolean resolved)
  {
    this.isResolved = resolved;
  }

  protected void handleException(Exception ex) 
  {
    if (DEBUG)
    {
      ex.printStackTrace();
    }
  }

  protected Logger logger()
  {
    return LoggerProvider.getLogger();
  } 
} 
