// ===========================================================================
// CONTENT  : CLASS FileUtil
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.8 - 17/12/2020
// HISTORY  :
//  17/05/2002  duma  CREATED
//	14/03/2003	duma	added		->	standardize(), javaFilename()
//	20/12/2003	duma	added		->	close(zipFile)
//	05/02/2004	duma	added		->	copyFile() methods
//	06/03/2004	duma	added		->	close(reader), close(writer)
//	03/09/2004	duma	added		->	4 copyToTempFile() methods
//	24/02/2006	mdu		added		->	readTextLines(...)
//	03/03/2006	mdu		added		->	convertFromURLSyntax()
//	28/07/2006	mdu		changed	->	performance optimization
//	27/04/2008	mdu		changed	->	calculateClasspath()
//	28/08/2012	mdu		added		->	readTextFrom() and readTextLinesFrom() methods supporting encoding
//	11/01/2014	mdu		added		->	methods with Charset parameter
//  01/10/2014  mdu   added   ->  openFile()
//  30/11/2014  mdu   added   ->  createFile()
//  14/05/2015  mdu   added   ->  openFileForRead()
//  12/09/2015  mdu   added   ->  ensureEmptyFolder()
//  13/09/2015  mdu   bugfix  ->  copyAll() use "*" rather than "*."
//  02/11/2017  mdu   added   ->  isExistingFolder()
//  28/11/2020  mdu   added   ->  getRelativePath()
//  17/12/2020  mdu   added   ->  ensureFolderExists()
//
// Copyright (c) 2002-2020, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.file;

import java.io.ByteArrayInputStream;
import java.io.Closeable;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.Reader;
import java.io.StringWriter;
import java.net.URL;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Properties;
import java.util.jar.Attributes;
import java.util.jar.JarFile;
import java.util.jar.Manifest;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

import org.pfsw.bif.configuration.IConfigSettings;
import org.pfsw.bif.text.ICheckedCharset;
import org.pfsw.bif.text.ILineProcessor;
import org.pfsw.bif.text.IStringFilter;
import org.pfsw.reflect.Dynamic;
import org.pfsw.reflect.ReflectUtil;
import org.pfsw.text.CheckedCharsetName;
import org.pfsw.text.StringPattern;
import org.pfsw.text.StringUtil;
import org.pfsw.textx.BasicVariableContainer;
import org.pfsw.textx.TextEngine;
import org.pfsw.textx.TextReplacementException;
import org.pfsw.textx.VariableResolver;
import org.pfsw.util.IOUtil;
import org.pfsw.util.OrderedSet;
import org.pfsw.util.StreamCopy;
import org.pfsw.util.SysUtil;

/**
 * This class provides helper methods for file and stream handling.
 * It's an add-on to the java.io package.
 * The service is implemented as a singleton, so use the
 * <b>FileUtil.current()</b> method to get the sole instance.
 *
 * @author Manfred Duchrow
 * @version 2.8
 */
public class FileUtil
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  public static final FileUtil FU = new FileUtil();

  private static final boolean DEBUG = "true".equals(System.getProperty("org.pfsw.file.debug"));
  private static final String SYS_PROP_CLASSPATH = "java.class.path";
  private static final String SYS_PROP_EXT_DIRS = "java.ext.dirs";
  private static final String JAR_FILE_EXTENSION = "jar";
  /** The lines.separator from the system properties as a constant */
  public static final String LINE_SEPARATOR = System.getProperty("line.separator");
  public static final int DEFAULT_BUFFER_SIZE = 4096;

  /**
   * The constant that contains the default prefix for temporary file.
   * It is "PFFU" (<b>P</b>rogrammer's <b>F</b>riend <b>F</b>ile<b>U</b>til).
   */
  public static final String DEFAULT_TEMP_FILE_PREFIX = "PFFU";

  /**
   * Contains the default URL string encoding used in this class: "UTF-8"
   */
  public static final String DEFAULT_URL_STR_ENCODING = "UTF-8";

  /**
   * Defines the prefix that specifies that the remaining part of the
   * string with that prefix is a file name that must be looked-up
   * using the current classloader: "classpath:".
   */
  public static final String CLASSPATH_PREFIX = "classpath:";

  protected static final String FILE_PROTOCOL_INDICATOR = "file:";
  protected static final StringPattern LOCAL_PATTERN = StringPattern.create("*file:?*");
  protected static final StringPattern REMOTE_PATTERN = StringPattern.create("*?://?*");
  protected static final StringPattern JAR_PATTERN = StringPattern.create("*jar:*:*!/*");
  protected static final StringPattern DRIVE_LETTER_PATTERN_1 = StringPattern.create("?:");
  protected static final StringPattern DRIVE_LETTER_PATTERN_2 = StringPattern.create("?:?*");
  protected static final StringPattern WINDOWS_DRIVE_PATTERN_1 = StringPattern.create("\\?:*");
  protected static final StringPattern WINDOWS_DRIVE_PATTERN_2 = StringPattern.create("/?:*");

  protected static final Long NO_LIMIT = Long.valueOf(-1L);

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private Classpath calculatedClasspath = null;
  private String classpathString = null;
  private Classpath systemClasspath = null;
  private final ClassLoader currentClassLoader;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Create a new instance with with an external classloader.
   */
  public FileUtil(ClassLoader classLoader)
  {
    super();
    if (classLoader == null)
    {
      throw new IllegalArgumentException("classLoader must not be null");
    }
    currentClassLoader = classLoader;
  }

  /**
   * Initialize the new instance with default values.
   */
  protected FileUtil()
  {
    this(FileUtil.class.getClassLoader());
  }

  // =========================================================================
  // PUBLIC CLASS METHODS
  // =========================================================================
  /**
   * Returns the default instance of this class.
   * This instance is internally using the same classloader that was used
   * to load FileUtil class.
   */
  public static FileUtil current()
  {
    return FileUtil.FU;
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Creates the file and writes the given content to it.
   * 
   * @param file The file to be created (must not be null).
   * @param content The (optional) content to be written to the file.
   */
  public void createFile(File file, byte... content) throws IOException
  {
    FileOutputStream outStream;
    ByteArrayInputStream inStream;

    outStream = new FileOutputStream(file);
    try
    {
      if ((content != null) && (content.length > 0))
      {
        inStream = new ByteArrayInputStream(content);
        copyStream(inStream, outStream);
      }
    }
    finally
    {
      close(outStream);
    }
  }

  /**
   * Creates the file and writes the given stream to it.
   * Afterwards the inStream will be closed.
   * 
   * @param inStream The stream to be written to the file.
   * @param file The file to be created (must not be null).
   */
  public void createFile(InputStream inStream, File file) throws IOException
  {
    FileOutputStream outStream;

    outStream = new FileOutputStream(file);
    copyStream(inStream, outStream);
  }

  /**
   * Creates the file and writes the given stream to it.
   * Afterwards the inStream (and the file stream) will be closed.
   * 
   * @param inStream The stream to be written to the file.
   * @param file The file to be created (must not be null).
   * @param bufferSize The size of the buffer that is used for the transfer of byte chunks (must be a positive value).
   * @param limit The maximum number of bytes that will be copied. If exceeded an IOException will be thrown.
   *   Setting this parameter to null, zero or a negative value will specify no limit.
   */
  public void createFile(InputStream inStream, File file, int bufferSize, Long limit) throws IOException
  {
    FileOutputStream outStream;

    outStream = new FileOutputStream(file);
    copyStream(inStream, outStream, bufferSize, limit);
  }

  /**
   * Copies all data from the input stream to the output stream using
   * a buffer with the default size (4096 bytes).
   * After all data is copied both streams will be closed !
   */
  public void copyStream(InputStream inStream, OutputStream outStream) throws IOException
  {
    ioUtil().copyStream(inStream, outStream);
  }

  /**
   * Copies all data from the input stream to the output stream using
   * a buffer of the given size in bytes.
   * After all data has been copied both streams will be closed!
   */
  public void copyStream(InputStream inStream, OutputStream outStream, int bufSize) throws IOException
  {
    ioUtil().copyStream(inStream, outStream, bufSize);
  }

  /**
   * Copies all data from the input stream to the output stream using
   * a buffer of the given size in bytes.
   * After all data has been copied both streams will be closed!
   * 
   * @param inStream The source stream.
   * @param outStream The target stream.
   * @param bufSize The size of the buffer that is used for the transfer of byte chunks (must be positive value).
   * @param limit The maximum number of bytes that will be copied. If exceeded an IOException will be thrown.
   *   Setting this parameter to null, zero or a negative value will specify no limit.
   */
  public void copyStream(InputStream inStream, OutputStream outStream, int bufSize, Long limit) throws IOException
  {
    ioUtil().copyStream(inStream, outStream, bufSize, limit);
  }

  /**
   * Copies the whole contents of the file specified by sourceFile to the file
   * specified by destFile.
   * If destFile already exists, it will be overwritten without warning!
   * If preservetimestamp is true, the last modified time of the destFile will
   * be set to that of the sourceFile. Otherwise it will be the current date.
   *
   * @param sourceFile The file to copy
   * @param destFile The destination where to copy the file content
   * @param preserveTimestamp A flag that indicates if destFile must have the same timestamp as sourceFile
   * @throws IOException
   */
  public void copyFile(File sourceFile, File destFile, boolean preserveTimestamp) throws IOException
  {
    FileInputStream inStream = null;
    FileOutputStream outStream = null;

    try
    {
      inStream = new FileInputStream(sourceFile);
      outStream = new FileOutputStream(destFile);
      copyStream(inStream, outStream);
    }
    finally
    {
      close(inStream);
      close(outStream);
    }
    if (preserveTimestamp)
    {
      destFile.setLastModified(sourceFile.lastModified());
    }
  }

  /**
   * Copies the whole contents of the file specified by sourceFile to the file
   * specified by destFile.
   * If destFile already exists, it will be overwritten without warning!
   * If preservetimestamp is true, the last modified time of the destFile will
   * be set to that of the sourceFile. Otherwise it will be the current date.
   *
   * @param sourceFile The file to copy
   * @param destFile The destination where to copy the file content
   * @param preserveTimestamp A flag that indicates if destFile must have the same timestamp as sourceFile
   * @throws IOException
   */
  public void copyFile(FileLocator sourceFile, File destFile, boolean preserveTimestamp) throws IOException
  {
    InputStream inStream = null;
    FileOutputStream outStream = null;

    try
    {
      inStream = sourceFile.getInputStream();
      outStream = new FileOutputStream(destFile);
      copyStream(inStream, outStream);
    }
    finally
    {
      close(inStream);
      close(outStream);
    }
    if (preserveTimestamp)
    {
      destFile.setLastModified(sourceFile.lastModified());
    }
  }

  /**
   * Copies the whole contents of the file specified by source URL to the file
   * specified by destFile.
   * If destFile already exists, it will be overwritten without warning!
   *
   * @param source An URL that specifies the source file that must be copied
   * @param destFile The destination where to copy the file content
   * @throws IOException
   */
  public void copyFile(URL source, File destFile) throws IOException
  {
    InputStream inStream = null;
    FileOutputStream outStream = null;

    try
    {
      inStream = source.openStream();
      outStream = new FileOutputStream(destFile);
      copyStream(inStream, outStream);
    }
    finally
    {
      close(inStream);
      close(outStream);
    }
  }

  /**
   * Copies the whole contents of the file specified by sourceFile to the file
   * specified by destFile.
   * If destFile already exists, it will be overwritten without warning!
   * The modification timestamp of destFile will be the current date.
   *
   * @param sourceFile The file to copy
   * @param destFile The destination where to copy the file content
   * @throws IOException
   */
  public void copyFile(File sourceFile, File destFile) throws IOException
  {
    copyFile(sourceFile, destFile, false);
  }

  /**
   * Copies the whole contents of the file specified by sourceFilename to the
   * file specified by destFilename.
   * If destFilename already exists, it will be overwritten without warning!
   * If preservetimestamp is true, the last modified time of the destFilename will
   * be set to that of the sourceFilename. Otherwise it will be the current date.
   *
   * @param sourceFilename The file to copy
   * @param destFilename The destinition where to copy the file content
   * @param preserveTimestamp A flag that indicates if destFile must have the same timestamp as sourceFile
   * @throws IOException
   */
  public void copyFile(String sourceFilename, String destFilename, boolean preserveTimestamp) throws IOException
  {
    copyFile(new File(sourceFilename), new File(destFilename), preserveTimestamp);
  }

  /**
   * Copies the whole contents of the file specified by sourceFilename to the
   * file specified by destFilename.
   * If destFilename already exists, it will be overwritten without warning!
   * The modification timestamp of destFilename will be the current date.
   *
   * @param sourceFilename The file to copy
   * @param destFilename The destinition where to copy the file content
   * @throws IOException
   */
  public void copyFile(String sourceFilename, String destFilename) throws IOException
  {
    copyFile(sourceFilename, destFilename, false);
  }

  /**
   * Copies all files and folders (recursively) from the sourceFolder
   * to the targetFolder. The target folder will be created if necessary.
   * 
   * @return The number of files copied.
   */
  public long copyAll(final File sourceFolder, final File targetFolder)
  {
    RecursiveFileCopyProcessor copyProcessor;
    FileWalker fileWalker;
    long copiedFilesCounter;

    if (!targetFolder.exists())
    {
      targetFolder.mkdirs();
    }

    if (LoggerProvider.TRACING)
    {
      System.out.println("FileUtil.copyAll('" + sourceFolder.getAbsolutePath() + "', '" + targetFolder.getAbsolutePath() + "')");
    }

    copyProcessor = new RecursiveFileCopyProcessor(sourceFolder, targetFolder);
    fileWalker = new FileWalker(copyProcessor);
    copiedFilesCounter = fileWalker.walkThrough(sourceFolder.getAbsolutePath(), "*", true);
    if (LoggerProvider.TRACING)
    {
      System.out.println(
          "FileUtil.copyAll('" + sourceFolder.getAbsolutePath() + "', '" + targetFolder.getAbsolutePath() + "') copied " + copiedFilesCounter + " files.");
    }
    return copiedFilesCounter;
  }

  /**
   * Returns all (full-path) names of file entries in the given zipFile that match the specified filter.
   *  
   * @param zipFile The zip file to look for the matching entries (must not be null).
   * @param filter The filter that defines which names to return (must not be null).
   * @return A collection of the found names (never null).
   * @throws IOException In case of errors with the zip file handling.
   */
  public Collection<String> findZippedFiles(final File zipFile, final IZipEntryFilter filter) throws IOException
  {
    ZipEntry zipEntry;
    Collection<String> result;
    ZipInputStream zipInStream;

    zipInStream = new ZipInputStream(new FileInputStream(zipFile));
    result = new ArrayList<String>();
    try
    {
      zipEntry = zipInStream.getNextEntry();
      while (zipEntry != null)
      {
        try
        {
          if (!zipEntry.isDirectory())
          {
            if (filter.matches(zipEntry))
            {
              result.add(zipEntry.getName());
            }
          }
        }
        finally
        {
          closeEntry(zipInStream);
        }
        zipEntry = zipInStream.getNextEntry();
      }
    }
    finally
    {
      close(zipInStream);
    }
    return result;
  }

  /**
   * Copies all files from the given zip archive to the specified
   * destination directory.
   * 
   * @param zipFile The file to be extracted (must not be null).
   * @param destDir The file directory to which to extract the files (must not be null).
   */
  public void extractZipArchive(final File zipFile, final File destDir) throws IOException
  {
    extractZipArchive(zipFile, destDir, null);
  }

  /**
   * Copies all files from the given zip archive that match the given filter to the specified
   * destination directory.
   * 
   * @param zipFile The file to be extracted (must not be null).
   * @param destDir The file directory to which to extract the files (must not be null).
   * @param filter An optional filter - only matching entries will be extracted (may be null for all matching).
   */
  public void extractZipArchive(final File zipFile, final File destDir, final IZipEntryFilter filter) throws IOException
  {
    ZipInputStream inStream;

    inStream = new ZipInputStream(new FileInputStream(zipFile));
    try
    {
      extractZipArchive(inStream, destDir, filter);
    }
    finally
    {
      close(inStream);
    }
  }

  /**
   * Copies all files from the given zip archive that match the given filter to the specified
   * destination directory.
   * The zipInStream will NOT be closed after this method has finished!
   * 
   * @param zipInStream The opened stream on the zip file to be extracted (must not be null).
   * @param destDir The file directory to which to extract the files (must not be null).
   * @param filter An optional filter - only matching entries will be extracted (may be null for all matching).
   */
  public void extractZipArchive(final ZipInputStream zipInStream, final File destDir, final IZipEntryFilter filter) throws IOException
  {
    StreamCopy streamCopy;
    ZipEntry zipEntry;
    File targetFile;
    FileOutputStream targetStream;

    streamCopy = new StreamCopy(false);

    destDir.mkdirs();
    zipEntry = zipInStream.getNextEntry();
    while (zipEntry != null)
    {
      if (!zipEntry.isDirectory())
      {
        if ((filter == null) || filter.matches(zipEntry))
        {
          targetFile = new File(destDir, zipEntry.getName());
          targetFile.getParentFile().mkdirs();
          targetStream = new FileOutputStream(targetFile);
          try
          {
            streamCopy.copy(zipInStream, targetStream);
          }
          finally
          {
            close(targetStream);
            closeEntry(zipInStream);
          }
        }
      }
      zipEntry = zipInStream.getNextEntry();
    }
  }

  /**
   * Reads the whole content of the given input stream and returns
   * it as a string.
   * The stream will be closed after calling this method. Even if an exception
   * occurred!
   * The stream character encoding is the JVM default encoding.
   *
   * @param inStream The input stream to read.
   * @return The text content of the given stream.
   */
  public String readTextFrom(InputStream inStream) throws IOException
  {
    return ioUtil().readTextFrom(inStream);
  }

  /**
   * Reads the whole content of the given input stream and returns
   * it as a string.
   * The stream will be closed after calling this method. Even if an exception
   * occurred!
   *
   * @param inStream The input stream to read (must not be null).
   * @param charsetName A charset name already checked to be supported (must not be null).
   * @return The text content of the given stream
   */
  public String readTextFrom(InputStream inStream, ICheckedCharset charsetName) throws IOException
  {
    return ioUtil().readTextFrom(inStream, charsetName);
  }

  /**
   * Reads the whole content of the given input stream and returns
   * it as a string.
   * The stream will be closed after calling this method. Even if an exception
   * occurred!
   *
   * @param inStream The input stream to read (must not be null).
   * @param charset A character set defining the encoding of the stream (must not be null).
   * @return The text content of the given stream.
   */
  public String readTextFrom(InputStream inStream, Charset charset) throws IOException
  {
    return ioUtil().readTextFrom(inStream, charset);
  }

  /**
   * Reads the whole content of the file with the given name and returns
   * it as a string.
   * The file character encoding is the JVM default encoding.
   *
   * @param filename The name of the text containing file
   */
  public String readTextFrom(String filename) throws IOException
  {
    return readTextFrom(filename, Charset.defaultCharset());
  }

  /**
   * Reads the whole content of the file with the given name and returns
   * it as a string.
   *
   * @param filename The name of the text containing file (must not be null).
   * @param charset A character set defining the file encoding (must not be null).
   */
  public String readTextFrom(String filename, Charset charset) throws IOException
  {
    File file;

    file = new File(filename);
    return readTextFrom(file, charset);
  }

  /**
   * Reads the whole content of the file with the given name and returns
   * it as a string.
   *
   * @param filename The name of the text containing file
   * @param charsetName A charset name already checked to be supported (must not be null).
   */
  public String readTextFrom(String filename, ICheckedCharset charsetName) throws IOException
  {
    return readTextFrom(filename, charsetName.getCharset());
  }

  /**
   * Reads the whole content of the specified file and returns
   * it as a string.
   * The file character encoding is the JVM default encoding.
   */
  public String readTextFrom(File file) throws IOException
  {
    return readTextFrom(file, Charset.defaultCharset());
  }

  /**
   * Reads the whole content of the specified file and returns
   * it as a string.
   * 
   * @param file The text file from which to read (must not be null).
   * @param charsetName A charset name already checked to be supported (must not be null).
   */
  public String readTextFrom(File file, ICheckedCharset charsetName) throws IOException
  {
    return readTextFrom(file, charsetName.getCharset());
  }

  /**
   * Reads the whole content of the specified file and returns
   * it as a string.
   * 
   * @param file The text file from which to read (must not be null).
   * @param charset A character set defining the file encoding (must not be null).
   */
  public String readTextFrom(File file, Charset charset) throws IOException
  {
    FileInputStream inStream;

    inStream = new FileInputStream(file);
    return readTextFrom(inStream, charset);
  }

  /**
   * Reads the whole content of the specified file and returns
   * it as a string.
   * The file character encoding is the JVM default encoding.
   */
  public String readTextFrom(FileLocator fileLocator) throws IOException
  {
    return readTextFrom(fileLocator.getInputStream());
  }

  /**
   * Reads the whole content of the specified file and returns
   * it as a string.
   * 
   * @param fileLocator The locator of the text file to read (must not be null).
   * @param charsetName A charset name already checked to be supported (must not be null).
   */
  public String readTextFrom(FileLocator fileLocator, ICheckedCharset charsetName) throws IOException
  {
    return readTextFrom(fileLocator, charsetName.getCharset());
  }

  /**
   * Reads the whole content of the specified file and returns
   * it as a string.
   * 
   * @param fileLocator The locator of the text file to read (must not be null).
   * @param charset A character set defining the encoding of the file (must not be null).
   */
  public String readTextFrom(FileLocator fileLocator, Charset charset) throws IOException
  {
    return readTextFrom(fileLocator.getInputStream(), charset);
  }

  /**
   * Reads all text lines from the named file into a String array.
   * The file character encoding is the JVM default encoding.
   *
   * @param filename The name of the file from which to read the text lines
   */
  public String[] readTextLinesFrom(String filename) throws IOException
  {
    return readTextLinesFrom(filename, Charset.defaultCharset());
  }

  /**
   * Reads all text lines from the named file into a String array.
   *
   * @param filename The name of the file from which to read the text lines
   * @param charsetName A charset name already checked to be supported (must not be null).
   */
  public String[] readTextLinesFrom(String filename, ICheckedCharset charsetName) throws IOException
  {
    return readTextLinesFrom(filename, charsetName.getCharset());
  }

  /**
   * Reads all text lines from the named file into a String array.
   *
   * @param filename The name of the file from which to read the text lines (must not be null).
   * @param charset A character set defining the encoding of the file (must not be null).
   */
  public String[] readTextLinesFrom(String filename, Charset charset) throws IOException
  {
    File file;

    file = new File(filename);
    return readTextLinesFrom(file, charset);
  }

  /**
   * Reads all text lines from the specified file into a String array.
   * The file character encoding is the JVM default encoding.
   *
   * @param file The file from which to read the text lines.
   */
  public String[] readTextLinesFrom(File file) throws IOException
  {
    return readTextLinesFrom(file, Charset.defaultCharset());
  }

  /**
   * Reads all text lines from the specified file into a String array.
   *
   * @param file The file from which to read the text lines
   * @param charsetName A charset name already checked to be supported (must not be null).
   */
  public String[] readTextLinesFrom(File file, ICheckedCharset charsetName) throws IOException
  {
    return readTextLinesFrom(file, charsetName.getCharset());
  }

  /**
   * Reads all text lines from the specified file into a String array.
   *
   * @param file The file from which to read the text lines (must not be null).
   * @param charset A character set defining the encoding of the file (must not be null).
   */
  public String[] readTextLinesFrom(File file, Charset charset) throws IOException
  {
    FileInputStream inStream;

    inStream = new FileInputStream(file);
    return readTextLinesFrom(inStream, charset);
  }

  /**
   * Reads all text lines from the specified file location into a String array.
   * The file character encoding is the JVM default encoding.
   *
   * @param fileLocator The file from which to read the text lines
   */
  public String[] readTextLinesFrom(FileLocator fileLocator) throws IOException
  {
    return readTextLinesFrom(fileLocator.getInputStream());
  }

  /**
   * Reads all text lines from the specified file location into a String array.
   *
   * @param fileLocator The file from which to read the text lines
   * @param charsetName A charset name already checked to be supported (must not be null).
   */
  public String[] readTextLinesFrom(FileLocator fileLocator, ICheckedCharset charsetName) throws IOException
  {
    return readTextLinesFrom(fileLocator.getInputStream(), charsetName);
  }

  /**
   * Reads all text lines from the specified stream into a String array.
   * After that the stream will be closed. Even if an exception occurs.
   * The character encoding is the JVM default encoding.
   *
   * @param inStream The stream from which to read the text lines
   */
  public String[] readTextLinesFrom(InputStream inStream) throws IOException
  {
    return ioUtil().readTextLinesFrom(inStream);
  }

  // ------------------------------------------------------------------------

  /**
   * Reads all text lines from the specified stream into a String array.
   * After that the stream will be closed. Even if an exception occurs.
   *
   * @param inStream The stream from which to read the text lines
   * @param charsetName A charset name already checked to be supported (must not be null).
   */
  public String[] readTextLinesFrom(InputStream inStream, ICheckedCharset charsetName) throws IOException
  {
    return ioUtil().readTextLinesFrom(inStream, charsetName.getCharset());
  }

  // ------------------------------------------------------------------------

  /**
   * Reads all text lines from the specified stream into a String array.
   * After that the stream will be closed. Even if an exception occurs.
   *
   * @param inStream The stream from which to read the text lines (must not be null).
   * @param charset A character set defining the encoding of the stream (must not be null).
   */
  public String[] readTextLinesFrom(InputStream inStream, Charset charset) throws IOException
  {
    return ioUtil().readTextLinesFrom(inStream, charset);
  }

  // ------------------------------------------------------------------------

  /**
   * Reads all text lines from the specified reader into a String array.
   * After that the reader will be closed. Even if an exception occurs.
   *
   * @param reader The reader which provides the text
   */
  public String[] readTextLinesFrom(Reader reader) throws IOException
  {
    return ioUtil().readTextLinesFrom(reader);
  }

  // ------------------------------------------------------------------------

  /**
   * Reads all text lines from the named file into a String array.
   * The character encoding is the JVM default encoding.
   *
   * @param filename The name of the file from which to read the text lines
   * @param filter A filter that defines which lines to return (may be null -> all lines)
   */
  public String[] readTextLinesFrom(String filename, IStringFilter filter) throws IOException
  {
    return readTextLinesFrom(filename, Charset.defaultCharset(), filter);
  }

  /**
   * Reads all text lines from the named file into a String array.
   *
   * @param filename The name of the file from which to read the text lines
   * @param charsetName A charset name already checked to be supported (must not be null).
   * @param filter A filter that defines which lines to return (may be null -> all lines)
   */
  public String[] readTextLinesFrom(String filename, ICheckedCharset charsetName, IStringFilter filter) throws IOException
  {
    return readTextLinesFrom(filename, charsetName.getCharset(), filter);
  }

  /**
   * Reads all text lines from the named file into a String array.
   *
   * @param filename The name of the file from which to read the text lines
   * @param charset A character set defining the file encoding (must not be null).
   * @param filter A filter that defines which lines to return (may be null -> all lines)
   */
  public String[] readTextLinesFrom(String filename, Charset charset, IStringFilter filter) throws IOException
  {
    File file;

    file = new File(filename);
    return readTextLinesFrom(file, charset, filter);
  }

  /**
   * Reads all text lines matching the filter from the specified file into a
   * String array.
   * The character encoding is the JVM default encoding.
   *
   * @param file The file from which to read the text lines
   * @param filter A filter that defines which lines to return (may be null -> all lines)
   */
  public String[] readTextLinesFrom(File file, IStringFilter filter) throws IOException
  {
    return readTextLinesFrom(file, Charset.defaultCharset(), filter);
  }

  /**
   * Reads all text lines matching the filter from the specified file into a
   * String array.
   *
   * @param file The file from which to read the text lines
   * @param charsetName A charset name already checked to be supported (must not be null).
   * @param filter A filter that defines which lines to return (may be null -> all lines)
   */
  public String[] readTextLinesFrom(File file, ICheckedCharset charsetName, IStringFilter filter) throws IOException
  {
    return readTextLinesFrom(file, charsetName.getCharset(), filter);
  }

  /**
   * Reads all text lines matching the filter from the specified file into a
   * String array.
   *
   * @param file The file from which to read the text lines (must not be null).
   * @param charset A character set defining the encoding of the file (must not be null).
   * @param filter A filter that defines which lines to return (may be null -> all lines)
   */
  public String[] readTextLinesFrom(File file, Charset charset, IStringFilter filter) throws IOException
  {
    FileInputStream inStream;

    inStream = new FileInputStream(file);
    return readTextLinesFrom(inStream, charset, filter);
  }

  /**
   * Reads all text lines matching the filter from the specified file location
   * into a String array.
   * The character encoding is the JVM default encoding.
   *
   * @param fileLocator The file from which to read the text lines
   * @param filter A filter that defines which lines to return (may be null -> all lines)
   */
  public String[] readTextLinesFrom(FileLocator fileLocator, IStringFilter filter) throws IOException
  {
    return readTextLinesFrom(fileLocator, Charset.defaultCharset(), filter);
  }

  /**
   * Reads all text lines matching the filter from the specified file location
   * into a String array.
   *
   * @param fileLocator The file from which to read the text lines
   * @param charsetName A charset name already checked to be supported (must not be null).
   * @param filter A filter that defines which lines to return (may be null -> all lines)
   */
  public String[] readTextLinesFrom(FileLocator fileLocator, ICheckedCharset charsetName, IStringFilter filter) throws IOException
  {
    return readTextLinesFrom(fileLocator, charsetName.getCharset(), filter);
  }

  /**
   * Reads all text lines matching the filter from the specified file location
   * into a String array.
   *
   * @param fileLocator The file from which to read the text lines  (must not be null).
   * @param charset A character set defining the file encoding (must not be null).
   * @param filter A filter that defines which lines to return (may be null -> all lines)
   */
  public String[] readTextLinesFrom(FileLocator fileLocator, Charset charset, IStringFilter filter) throws IOException
  {
    return readTextLinesFrom(fileLocator.getInputStream(), charset, filter);
  }

  /**
   * Reads all text lines matching the filter from the specified stream into a
   * String array.
   * After that the stream will be closed. Even if an exception occurs.
   * The character encoding is the JVM default encoding.
   *
   * @param inStream The stream from which to read the text lines
   * @param filter A filter that defines which lines to return (may be null -> all lines)
   */
  public String[] readTextLinesFrom(InputStream inStream, IStringFilter filter) throws IOException
  {
    return readTextLinesFrom(inStream, Charset.defaultCharset(), filter);
  }

  // ------------------------------------------------------------------------

  /**
   * Reads all text lines matching the filter from the specified stream into a
   * String array using the specified character encoding.
   * After that the stream will be closed. Even if an exception occurs.
   *
   * @param inStream The stream from which to read the text lines.
   * @param charsetName A charset name already checked to be supported (must not be null).
   * @param filter A filter that defines which lines to return (may be null -> all lines).
   */
  public String[] readTextLinesFrom(InputStream inStream, ICheckedCharset charsetName, IStringFilter filter) throws IOException
  {
    return ioUtil().readTextLinesFrom(inStream, charsetName.getCharset(), filter);
  }

  // ------------------------------------------------------------------------

  /**
   * Reads all text lines matching the filter from the specified stream into a
   * String array using the specified character encoding.
   * After that the stream will be closed. Even if an exception occurs.
   *
   * @param inStream The stream from which to read the text lines (must not be null).
   * @param charset A character set defining the encoding of the input stream (must not be null).
   * @param filter A filter that defines which lines to return (may be null -> all lines).
   */
  public String[] readTextLinesFrom(InputStream inStream, Charset charset, IStringFilter filter) throws IOException
  {
    return ioUtil().readTextLinesFrom(inStream, charset, filter);
  }

  // ------------------------------------------------------------------------

  /**
   * Reads all text lines that match the filter from the specified reader into
   * a String array. If the filter is null all lines will be returned.
   * After that the reader will be closed. Even if an exception occurs.
   *
   * @param reader The reader which provides the text
   * @param filter A filter that defines which lines to return (may be null -> all lines)
   */
  public String[] readTextLinesFrom(Reader reader, IStringFilter filter) throws IOException
  {
    return ioUtil().readTextLinesFrom(reader, filter);
  }

  // ------------------------------------------------------------------------

  /**
   * Copies all text lines from the specified reader to the given writer.
   * After that the reader will be closed. Even if an exception occurs.
   *
   * @param reader The reader which provides the text to copy
   * @param writer The writer to which the text will be copied
   */
  public void copyText(Reader reader, final StringWriter writer) throws IOException
  {
    ioUtil().copyText(reader, writer);
  }

  // ------------------------------------------------------------------------

  /**
   * Reads all text lines from the file with the specified name and passes them
   * one by one to the given line processor.
   * The processing will be terminated, if the end of the text is reached or
   * if the processor returns <b>false</b>.<br>
   *
   * @param filename The name of the text file to read
   * @param processor The processor that receives the lines from the text
   */
  public void processTextLines(String filename, ILineProcessor processor) throws IOException
  {
    FileInputStream inStream;

    assertArgumentNotNull("filename", filename);
    inStream = new FileInputStream(filename);
    processTextLines(inStream, processor);
  }

  /**
   * Reads all text lines from the file with the specified name and passes them
   * one by one to the given line processor.
   * The processing will be terminated, if the end of the text is reached or
   * if the processor returns <b>false</b>.<br>
   *
   * @param filename The name of the text file to read
   * @param charsetName A charset name already checked to be supported (must not be null).
   * @param processor The processor that receives the lines from the text
   */
  public void processTextLines(String filename, ICheckedCharset charsetName, ILineProcessor processor) throws IOException
  {
    FileInputStream inStream;

    assertArgumentNotNull("filename", filename);
    inStream = new FileInputStream(filename);
    processTextLines(inStream, charsetName, processor);
  }

  /**
   * Reads all text lines from the specified input stream and passes them
   * one by one to the given line processor.
   * The processing will be terminated, if the end of the text is reached or
   * if the processor returns <b>false</b>.<br>
   * The given input stream will be closed after the execution of this method.
   * Even if an exception occurred.
   * The character encoding is the JVM default encoding.
   *
   * @param inStream The input stream that contains the text
   * @param processor The processor that receives the lines from the text
   */
  public void processTextLines(InputStream inStream, ILineProcessor processor) throws IOException
  {
    processTextLines(inStream, Charset.defaultCharset(), processor);
  }

  /**
   * Reads all text lines from the specified input stream and passes them
   * one by one to the given line processor.
   * The processing will be terminated, if the end of the text is reached or
   * if the processor returns <b>false</b>.<br>
   * The given input stream will be closed after the execution of this method.
   * Even if an exception occurred.
   *
   * @param inStream The input stream that contains the text (must not be null).
   * @param charsetName A charset name already checked to be supported (must not be null).
   * @param processor The processor that receives the lines from the text (must not be null).
   */
  public void processTextLines(InputStream inStream, ICheckedCharset charsetName, ILineProcessor processor) throws IOException
  {
    processTextLines(inStream, charsetName.getCharset(), processor);
  }

  /**
   * Reads all text lines from the specified input stream and passes them
   * one by one to the given line processor.
   * The processing will be terminated, if the end of the text is reached or
   * if the processor returns <b>false</b>.<br>
   * The given input stream will be closed after the execution of this method.
   * Even if an exception occurred.
   *
   * @param inStream The input stream that contains the text (must not be null).
   * @param charset A character set defining the encoding of the stream (must not be null).
   * @param processor The processor that receives the lines from the text (must not be null).
   */
  public void processTextLines(InputStream inStream, Charset charset, ILineProcessor processor) throws IOException
  {
    ioUtil().processTextLines(inStream, charset, processor);
  }

  /**
   * Reads all text lines from the specified reader and passes them one by one
   * to the given line processor.
   * The processing will be terminated, if the end of the text is reached or
   * if the processor returns <b>false</b>.
   *
   * @param reader The reader that contains a text stream
   * @param processor The processor that receives the lines from the text
   */
  public void processTextLines(Reader reader, ILineProcessor processor) throws IOException
  {
    ioUtil().processTextLines(reader, processor);
  }

  /**
   * Close the given ZIP file ignoring any exception.
   * Returns true, if the ZIP file was closed successfully, false otherwise
   */
  public boolean close(ZipFile zipFile)
  {
    if (zipFile == null)
    {
      return false;
    }
    try
    {
      zipFile.close();
      return true;
    }
    catch (IOException e)
    {
      handleException(e);
      return false;
    }
  }

  /**
   * Close the current entry in the given ZIP input stream ignoring any exception.
   * Returns true, if the entry was closed successfully, false otherwise
   */
  public boolean closeEntry(ZipInputStream zipInStream)
  {
    if (zipInStream == null)
    {
      return false;
    }
    try
    {
      zipInStream.closeEntry();
      return true;
    }
    catch (IOException e)
    {
      handleException(e);
      return false;
    }
  }

  /**
   * Close the current entry in the given ZIP output stream ignoring any exception.
   * Returns true, if the entry was closed successfully, false otherwise
   */
  public boolean closeEntry(ZipOutputStream zipOutStream)
  {
    if (zipOutStream == null)
    {
      return false;
    }
    try
    {
      zipOutStream.closeEntry();
      return true;
    }
    catch (IOException e)
    {
      handleException(e);
      return false;
    }
  }

  /**
   * Close the given closeable ignoring any exception and checking for null before closing.
   * Returns true, if the object was closed successfully, false otherwise.
   */
  public boolean close(Closeable closeable)
  {
    return ioUtil().close(closeable);
  }

  /**
   * Convert the filename to a canonical (see java.io.File.getCanonicalPath())
   * format and replace any backslashes '\' by slashes ('/').
   * If possible all "." and ".." elements in the path are eliminated.
   *
   * @param filename The filename which has to be standardized
   * @return An absolute filename that uses slashes to separate its elements
   */
  public String standardize(String filename)
  {
    if (filename == null)
    {
      return null;
    }

    if (isRemote(filename))
    {
      return filename;
    }

    return standardizeFilename(filename);
  }

  /**
   * Convert all the given filenames to a canonical (see java.io.File.getCanonicalPath())
   * format and replace any backslashes '\' by slashes ('/').
   * If possible all "." and ".." elements in the path are eliminated.
   *
   * @param filenames The file names which have to be standardized
   * @return An new array containing absolute filenames that uses slashes to separate its elements
   */
  public String[] standardize(String[] filenames)
  {
    String[] names;

    if (filenames == null)
      return null;

    names = new String[filenames.length];
    for (int i = 0; i < filenames.length; i++)
    {
      names[i] = standardize(filenames[i]);
    }
    return names;
  }

  /**
   * Returns the given filename in the platform independent way that Java
   * understands. That is all elements are separated by a forward slash rather
   * than back slashes.
   *
   * @param filename The name to be modified
   */
  public String javaFilename(String filename)
  {
    if (filename == null)
    {
      return null;
    }

    return filename.replace('\\', '/');
  }

  /**
   * Returns a new array of filenames where all backslashes are
   * replaced by forward slashes.
   *
   * @param filenames The file names to be modified
   * @return The same file names but with forward slashes only
   */
  public String[] javaFilenames(String[] filenames)
  {
    String[] newNames;

    if (filenames == null)
    {
      return null;
    }

    newNames = new String[filenames.length];
    for (int i = 0; i < filenames.length; i++)
    {
      newNames[i] = javaFilename(filenames[i]);
    }
    return newNames;
  }

  /**
   * Returns the current classpath as String.
   * If this is called from within an Eclipse plug-in the classpath is
   * created from the plug-in dependencies.
   * Otherwise the classpath is taken from the system property "java.class.path"
   * or the manifest file if the application was started with option -jar.
   */
  public String getClasspath()
  {
    Classpath classpath;

    if (classpathString == null)
    {
      classpath = calculateClasspath();
      if (classpath != null)
      {
        classpathString = classpath.toString();
      }
    }
    return classpathString;
  }

  /**
   * Returns the current classpath.
   * If this is called from within an Eclipse plug-in the classpath is
   * created from the plug-in dependencies.
   * Otherwise the classpath is taken from the system property "java.class.path"
   * or the manifest file if the application was started with option -jar.
   *
   * @see #getClasspath()
   */
  public Classpath calculateClasspath()
  {
    if (calculatedClasspath == null)
    {
      if (SysUtil.current().isEclipse())
      {
        calculatedClasspath = calculateEclipseClasspath();
      }
      if (calculatedClasspath == null)
      {
        calculatedClasspath = new Classpath(getClasspathFromSystemProperty());
      }
    }
    return calculatedClasspath;
  }

  /**
   * Returns a class path that is either the value of the specified system
   * property or the classpath calculated from the current Java environment.
   * <p>
   * If the given system property has a lookup path set it will be scanned
   * for ${xxx} placeholders which will be replaced by values of the corresponding
   * system properties.
   *
   * @param sysPropName A system property name of null
   */
  public Classpath getLookupPath(String sysPropName)
  {
    String propValue;

    if (!str().isNullOrEmpty(sysPropName))
    {
      propValue = System.getProperty(sysPropName);
      if (!str().isNullOrEmpty(propValue))
      {
        return determineLookupClasspath(sysPropName);
      }
    }
    if (SysUtil.current().isEclipse())
    {
      return calculateClasspath();
    }
    return determineLookupClasspath(sysPropName);
  }

  /**
   * Copies the contents of the given inStream to a temporary file and returns
   * this temporary file. When the VM terminates the temporary file will
   * automatically be deleted if it still exists.
   * The name of the temporary file starts with the default prefix "PFFU"
   * and ends with the default suffix ".tmp".
   *
   * @param inStream The stream to copy into the temporary file
   * @see FileUtil#copyToTempFile(InputStream, String)
   * @see FileUtil#copyToTempFile(InputStream, String, String)
   * @see FileUtil#copyToTempFile(InputStream, String, String, boolean)
   */
  public File copyToTempFile(InputStream inStream) throws IOException
  {
    return copyToTempFile(inStream, DEFAULT_TEMP_FILE_PREFIX);
  }

  /**
   * Copies the contents of the given inStream to a temporary file and returns
   * this temporary file. When the VM terminates the temporary file will
   * automatically be deleted if it still exists.
   * The name of the temporary file starts with the given prefix and ends with
   * the default ".tmp".
   *
   * @param inStream The stream to copy into the temporary file
   * @param filePrefix The prefix to be used for the temp file name (must not be null)
   * @see FileUtil#copyToTempFile(InputStream)
   * @see FileUtil#copyToTempFile(InputStream, String, String)
   * @see FileUtil#copyToTempFile(InputStream, String, String, boolean)
   */
  public File copyToTempFile(InputStream inStream, String filePrefix) throws IOException
  {
    return copyToTempFile(inStream, filePrefix, null);
  }

  /**
   * Copies the contents of the given inStream to a temporary file and returns
   * this temporary file. When the VM terminates the temporary file will
   * automatically be deleted if it still exists.
   * The name of the temporary file starts with the given prefix and ends with
   * the given suffix.
   *
   * @param inStream The stream to copy into the temporary file
   * @param filePrefix The prefix to be used for the temp file name (must not be null)
   * @param fileSuffix The prefix to be used for the temp file name (may be null)
   * @see File#createTempFile(java.lang.String, java.lang.String)
   * @see FileUtil#copyToTempFile(InputStream)
   * @see FileUtil#copyToTempFile(InputStream, String)
   * @see FileUtil#copyToTempFile(InputStream, String, String, boolean)
   */
  public File copyToTempFile(InputStream inStream, String filePrefix, String fileSuffix) throws IOException
  {
    return copyToTempFile(inStream, filePrefix, fileSuffix, true);
  }

  /**
   * Copies the contents of the given inStream to a temporary file and returns
   * this temporary file.
   * The name of the temporary file starts with the given prefix and ends with
   * the given suffix.
   *
   * @param inStream The stream to copy into the temporary file
   * @param filePrefix The prefix to be used for the temp file name (must not be null)
   * @param fileSuffix The prefix to be used for the temp file name (may be null)
   * @param deleteOnExit If true the temporary file will be deleted when the VM terminates
   * @see File#createTempFile(java.lang.String, java.lang.String)
   * @see FileUtil#copyToTempFile(InputStream)
   * @see FileUtil#copyToTempFile(InputStream, String)
   * @see FileUtil#copyToTempFile(InputStream, String, String)
   */
  public File copyToTempFile(InputStream inStream, String filePrefix, String fileSuffix, boolean deleteOnExit) throws IOException
  {
    File tempFile;
    OutputStream outStream;

    tempFile = File.createTempFile(filePrefix, fileSuffix);
    outStream = new FileOutputStream(tempFile);
    copyStream(inStream, outStream);
    if (deleteOnExit)
    {
      tempFile.deleteOnExit();
    }
    return tempFile;
  }

  /**
   * Copies the contents of the file specified by filename to a temporary file
   * and returns this temporary file.
   * The name of the temporary file starts with the given prefix and ends with
   * the given suffix.
   *
   * @param filename The name of the file which to copy into the temporary file
   * @param filePrefix The prefix to be used for the temp file name (must not be null)
   * @param fileSuffix The prefix to be used for the temp file name (may be null)
   * @param deleteOnExit If true the temporary file will be deleted when the VM terminates
   * @return The newly created temporary file or null if the input file (filename) cannot be found
   * @see File#createTempFile(java.lang.String, java.lang.String)
   * @throws IOException
   * @throws IllegalArgumentException If any input parameter is not ok
   */
  public File copyToTempFile(String filename, String filePrefix, String fileSuffix, boolean deleteOnExit) throws IOException
  {
    InputStream inStream;
    URL url;

    if (filename == null)
      throw new IllegalArgumentException("<filename> must not be null!");
    if (filePrefix == null)
      throw new IllegalArgumentException("<filePrefix> must not be null!");

    if (isLocal(filename))
    {
      inStream = new FileInputStream(filename);
    }
    else
    {
      url = new URL(filename);
      inStream = url.openStream();
    }

    if (inStream == null)
    {
      return null;
    }
    return copyToTempFile(inStream, filePrefix, fileSuffix, deleteOnExit);
  }

  /**
   * Returns true if the given filename specifies a file (or directory) on
   * the local file system using the URL syntax such as "file:/usr/bin/run"
   * or "file:\C:\temp\readme.txt.
   */
  public boolean isLocalFileURL(String filename)
  {
    if (filename == null)
      return false;

    return LOCAL_PATTERN.matches(filename);
  }

  /**
   * Returns true if the given filename specifies a file (or directory) on
   * the local file system. That is, it doesn't contain a protocol prefix
   * such as "http:" or "ftp:".
   */
  public boolean isLocal(String filename)
  {
    if (filename == null)
    {
      return false;
    }
    if (isLocalFileURL(filename))
    {
      return true;
    }
    return !isRemote(filename);
  }

  /**
   * Returns true if the given filename specifies a file (or directory) on
   * a remote system. That is, it contains a protocol prefix such as "http:"
   * or "ftp:".
   */
  public boolean isRemote(String filename)
  {
    if (filename == null)
    {
      return false;
    }
    return REMOTE_PATTERN.matches(filename);
  }

  /**
   * Returns true if the given filename specifies a file (or directory) that
   * is located inside a JAR file.
   */
  public boolean isInsideJavaArchive(String filename)
  {
    if (filename == null)
    {
      return false;
    }
    return JAR_PATTERN.matches(filename);
  }

  /**
   * Converts the given filename to URL syntax. That is, it prepends the
   * scheme "file:" and replaces all illegal URL characters by %xx values.
   * Be aware that this method uses %20 for replacing spaces and not the
   * plus ('+') as the URLEncoder class would do.
   * <p>
   * Examples:<br>
   *  "/path/name" -> "file:/path/name"<br>
   *  "C:\path\name" -> "file:/C:/path/name"<br>
   *  "path/name" -> "file:path/name"
   *  "C:\\folder name\\filename{v1}" -> "file:/C:/folder%20name/filename%7bv1%7d"<br>
   *
   * @param filename The filename to convert
   */
  public String convertToURLSyntax(String filename)
  {
    StringBuffer buffer;
    String[] parts;
    String part;

    if ((filename == null) || isRemote(filename) || isLocalFileURL(filename))
    {
      return filename;
    }
    buffer = new StringBuffer(filename.length() + FILE_PROTOCOL_INDICATOR.length() + 3);
    buffer.append(FILE_PROTOCOL_INDICATOR);
    parts = str().allParts(filename, "/\\");
    for (int i = 0; i < parts.length; i++)
    {
      if (i == 0)
      {
        if (DRIVE_LETTER_PATTERN_2.matches(filename))
        {
          buffer.append("/");
          part = parts[i];
        }
        else
        {
          if ((filename.startsWith("/")) || (filename.startsWith("\\")))
          {
            buffer.append("/");
          }
          part = urlEncode(parts[i]);
        }
      }
      else
      {
        buffer.append("/");
        part = urlEncode(parts[i]);
      }
      buffer.append(part);
    }
    return buffer.toString();
  }

  /**
   * Converts the given filename from URL syntax to normal file system syntax.
   * It removes the scheme and decodes all escaped characters.
   * <p>
   * Examples:<br>
   *  "file:/path/name" -> "/path/name"<br>
   *  "file:\C:\path\name" -> "C:\path\name"<br>
   *  "path/name" -> "path/name"
   *  "file:/C:/Program+Files/settings.ini" -> "C:\Program Files\settings.ini"<br>
   *
   *   @param filename The filename to convert
   */
  public String convertFromURLSyntax(String filename)
  {
    String newStr;

    if (filename == null)
    {
      return null;
    }
    if (LOCAL_PATTERN.matches(filename))
    {
      newStr = filename.substring(FILE_PROTOCOL_INDICATOR.length());
      newStr = urlDecode(newStr);
      if (WINDOWS_DRIVE_PATTERN_1.matches(newStr) || WINDOWS_DRIVE_PATTERN_2.matches(newStr))
      {
        newStr = newStr.substring(1); // Remove leading backslash
      }
      return newStr;
    }
    return filename;
  }

  /**
   * Converts the given file from URL syntax to normal file system syntax.
   * <p>
   * Examples:<br>
   *  "file:/path/name" -> "/path/name"<br>
   *  "file:\C:\path\name" -> "C:\path\name"<br>
   *  "path/name" -> "path/name"
   *
   *   @param file The file to convert
   */
  public File convertFromURLSyntax(File file)
  {
    return new File(convertFromURLSyntax(file.getPath()));
  }

  /**
   * Returns the given string with all %xx placeholders decoded using the
   * specified charset encoding.
   *
   * @param str The string to decode.
   * @param charsetName The name of the charset to be used for decoding. 
   * @return A string with all %xx placeholders decoded according to the specified charset.
   */
  public String urlDecode(String str, ICheckedCharset charsetName)
  {
    return ioUtil().urlDecode(str, charsetName.getCharset());
  }

  /**
   * Returns the given string with all %xx placeholders decoded.
   *
   * @param str The string to decode
   * @return A string with all %xx placeholders decoded (using UTF-8)
   */
  public String urlDecode(String str)
  {
    return ioUtil().urlDecode(str);
  }

  /**
   * Returns the given string with all characters that are not allowed directly
   * in URLs encoded as %xx placeholders. In particular it replaces spaces by "
   * %20" rather than '+'! That is reasonable in cases where files in the
   * local file system contain spaces. Such URLs (staring with "file:") cannot
   * open a stream if the URL contains a '+' because the '+' would not be
   * decoded back to a space. That is a bug (or at least an inconsistency) in
   * the URL class implementation.
   *
   * @param str The string to encode.
   * @param charsetName The name of the charset to be used for encoding. 
   * @return A string with %xx encodings for all invalid URL characters.
   */
  public String urlEncode(String str, ICheckedCharset charsetName)
  {
    return ioUtil().urlEncode(str, charsetName.getCharset());
  }

  /**
   * Returns the given string with all characters that are not allowed directly
   * in URLs encoded as %xx placeholders. In particular it replaces spaces by "
   * %20" rather than '+'! That is reasonable in cases where files in the
   * local file system contain spaces. Such URLs (staring with "file:") cannot
   * open a stream if the URL contains a '+' because the '+' would not be
   * decoded back to a space. That is a bug (or at least an inconsistency) in
   * the URL class implementation.
   *
   * @param str The string to encode.
   * @return A string with %xx (UTF-8) encodings for all invalid URL characters.
   */
  public String urlEncode(String str)
  {
    return ioUtil().urlEncode(str);
  }

  /**
   * Deletes all files and sub-directories in the given directory and the directory itself.
   * !!! BE CAREFUL WITH THIS METHOD!!!
   * <br/>
   * If called with the wrong directory name it can cause data loss.
   * 
   * @param dir The directory under which all files must be deleted (must not be null).
   * @return The number of deleted artifacts.
   */
  public long removeDirectory(final File dir)
  {
    long counter;

    if (!dir.exists() || !dir.isDirectory())
    {
      return 0;
    }
    counter = cleanDirectory(dir);
    if (dir.delete())
    {
      counter++;
    }
    return counter;
  }

  /**
   * Deletes all files and sub-directories in the given directory.
   * !!! BE CAREFUL WITH THIS METHOD!!!
   * <br/>
   * If called with the wrong directory name it can cause data loss.
   * 
   * @param dir The directory under which all files must be deleted (must not be null).
   * @return The number of deleted artifacts.
   */
  public long cleanDirectory(final File dir)
  {
    if (!dir.exists() || !dir.isDirectory())
    {
      return 0;
    }
    return cleanDirectory(dir, false);
  }

  /**
   * Deletes all files (and optionally sub-directories) in the given directory.
   * !!! BE CAREFUL WITH THIS METHOD!!!
   * <br/>
   * If called with the wrong directory name it can cause data loss.
   * 
   * @param dir The directory under which all files must be deleted (must not be null).
   * @param filesOnly If true only files will be deleted, if false sub-directories will be deleted as well.
   * @return The number of deleted artifacts.
   */
  public long cleanDirectory(final File dir, final boolean filesOnly)
  {
    FileWalker fileWalker;
    AFileProcessor fileProcessor;
    long deleteCounter;
    File startDir;

    startDir = new File(standardize(dir.getAbsolutePath()));
    if (!startDir.exists() || !startDir.isDirectory())
    {
      return 0;
    }
    fileProcessor = new FileDeleteProcessor();
    fileWalker = new FileWalker(fileProcessor);
    // Delete all files
    deleteCounter = fileWalker.walkThrough(startDir.getAbsolutePath(), DefaultFilenameFilter.ALL, true);
    if (filesOnly)
    {
      return deleteCounter;
    }
    // Delete all sub-directories
    deleteCounter += fileWalker.walkThroughDirectoriesSubDirsFirst(dir.getAbsolutePath(), DefaultFilenameFilter.ALL);
    return deleteCounter;
  }

  /**
   * Opens the file with the given name and returns it as an input stream.<br>
   * If the filename contains the prefix "classpath:" then it will be looked-up
   * using the underlying current classloader.
   * <p>
   * Without that prefix it will be treated as normal file on the file system
   * (relative or absolute).
   * 
   * @param filename The file to open (must not be null). It may contain a path (with slashes).
   * @return The input stream for the file.
   */
  public InputStream openFile(final String filename) throws FileNotFoundException
  {
    String name;
    InputStream inStream;
    File file;

    if (filename.startsWith(CLASSPATH_PREFIX))
    {
      name = str().suffix(filename, CLASSPATH_PREFIX);
      inStream = getCurrentClassLoader().getResourceAsStream(name);
      if (inStream == null)
      {
        throw new FileNotFoundException("File " + name + " not found on classpath.");
      }
    }
    else
    {
      file = new File(filename);
      inStream = new FileInputStream(file.getAbsolutePath()); // That ensures that any exception contains the absolute filename
    }
    return inStream;
  }

  /**
  * Opens the file with the given name and returns it as a named input stream.<br>
  * If the filename contains the prefix "classpath:" then it will be looked-up
  * using the underlying current classloader.
  * <p>
  * Without that prefix it will be treated as normal file on the file system
  * (relative or absolute).
  * 
  * @param filename The file to open (must not be null). It may contain a path (with slashes).
  * @return The input stream for the file along with the normalized absolute name of the file.
  */
  public NamedInputStream openFileForRead(final String filename) throws FileNotFoundException
  {
    InputStream inputStream;
    File file;
    String name;

    inputStream = openFile(filename);
    if (filename.startsWith(CLASSPATH_PREFIX))
    {
      name = filename;
    }
    else
    {
      file = new File(filename);
      name = standardize(file.getAbsolutePath());
    }
    return new NamedInputStream(name, inputStream);
  }

  /**
   * Loads all properties form the file with the given name.<br>
   * If the filename contains the prefix "classpath:" the it will be looked-up
   * using the underlying current classloader.
   * <p>
   * Without that prefix it will be treated as normal file on the file system
   * (relative or absolute).
   * 
   * @param filename The properties file (must not be null). It may contain a path (with slashes).
   * @return The loaded properties.
   */
  public Properties loadPropertiesFrom(final String filename) throws IOException
  {
    InputStream inStream;
    Properties properties;

    properties = new Properties();
    inStream = openFile(filename);
    try
    {
      properties.load(inStream);
    }
    finally
    {
      close(inStream);
    }
    return properties;
  }

  /**
   * Loads all properties form the file with the given name.<br>
   * If the filename contains the prefix "classpath:" then it will be looked-up
   * using the underlying current classloader.
   * <p>
   * Without that prefix it will be treated as normal file on the file system
   * (relative or absolute).
   * 
   * @param filename The properties file (must not be null). It may contain a path (with slashes).
   * @return The loaded properties as immutable name/value pairs.
   */
  public IConfigSettings loadNamedValuesFrom(final String filename) throws IOException
  {
    return str().asConfigSettings(loadPropertiesFrom(filename));
  }

  /**
   * Ensures that the specified folder exists and is empty.
   * So if it exists already, it will be cleaned, otherwise it will be created.
   * If the given folder is actually a file it will not be touched and the
   * method returns false.
   * 
   * @return true if the folder exists now, otherwise false.
   */
  public boolean ensureEmptyFolder(final File folder)
  {
    if (folder.exists())
    {
      if (!folder.isDirectory())
      {
        return false;
      }
      removeDirectory(folder);
    }
    folder.mkdirs();
    return folder.exists();
  }

  /**
   * Ensures that the specified folder exists.
   * If it does not exist yet, it will be created.
   * If the given folder is actually a file it will not be touched and the
   * method returns false.
   * 
   * @return true if the folder exists, otherwise false.
   */
  public boolean ensureFolderExists(final File folder)
  {
    if (folder == null)
    {
      return false;
    }
    if (!isExistingFolder(folder))
    {
      folder.mkdirs();
    }
    return isExistingFolder(folder);
  }
    
  /**
   * Returns true if the given path is an existing folder on the file system.
   */
  public boolean isExistingFolder(String path)
  {
    File folder;

    if (str().isNullOrBlank(path))
    {
      return false;
    }
    folder = new File(path.trim());
    return isExistingFolder(folder);
  }

  /**
   * Returns true if the given folder is an existing folder on the file system.
   */
  public boolean isExistingFolder(final File folder)
  {
    if (folder == null)
    {
      return false;
    }
    if (!folder.exists())
    {
      return false;
    }
    if (!folder.isDirectory())
    {
      return false;
    }
    return true;
  }

  /**
   * Returns the relative path of the given fileBelowDir in relation to the specified directory.
   * If the file is not below the specified dir, its absolute path will be returned.
   * 
   * @param dir The directory under which the file is supposed to be located (must not be null).
   * @param fileBelowDir The file under the given dir somewhere down the hierarchy of sub-directories (must not be null).
   * @return The relative path (with forward slashes), with no starting and trailing slashes. 
   */
  public String getRelativePath(File dir, File fileBelowDir)
  {
    String sourcePath;
    String filePath;
    String relativeFilePath;

    sourcePath = standardize(dir.getAbsolutePath());
    filePath = standardize(fileBelowDir.getAbsolutePath());
    relativeFilePath = str().cutHead(filePath, sourcePath);
    return str().trim(relativeFilePath, "/");
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected void copyText(InputStream inStream, StringWriter writer) throws IOException
  {
    copyText(new InputStreamReader(inStream), writer);
  }

  protected void copyText(InputStream inStream, CheckedCharsetName charsetName, StringWriter writer) throws IOException
  {
    assertArgumentNotNull("inStream", inStream);
    assertArgumentNotNull("charsetName", charsetName);
    copyText(inStream, charsetName.getCharset(), writer);
  }

  protected void copyText(InputStream inStream, Charset charset, StringWriter writer) throws IOException
  {
    InputStreamReader reader = null;

    assertArgumentNotNull("inStream", inStream);
    assertArgumentNotNull("charset", charset);

    try
    {
      reader = new InputStreamReader(inStream, charset);
      copyText(reader, writer);
    }
    finally
    {
      close(reader);
      close(inStream);
    }
  }

  protected String standardizeFilename(final String filename)
  {
    String[] nameElements;
    boolean hasDriveLetter;
    boolean startedFromRoot;
    boolean isAbsolute;
    String nameOfFile;
    int index;

    nameOfFile = javaFilename(filename);
    startedFromRoot = nameOfFile.startsWith("/");
    nameElements = str().parts(nameOfFile, "/");
    if (nameElements.length > 0)
    {
      hasDriveLetter = false;
      for (int i = 0; i < nameElements.length; i++)
      {
        if (DRIVE_LETTER_PATTERN_1.matches(nameElements[0]))
        {
          nameElements[0] = nameElements[0].toUpperCase();
          hasDriveLetter = true;
        }
      }
      if (!hasDriveLetter)
      {
        if (startedFromRoot)
        {
          nameElements = str().append(new String[] { "" }, nameElements);
        }
      }
      isAbsolute = hasDriveLetter || startedFromRoot;
      for (int i = 0; i < nameElements.length; i++)
      {
        if (".".equals(nameElements[i]))
        {
          nameElements[i] = null;
        }
        else
        {
          if ("..".equals(nameElements[i]))
          {
            index = indexOfPreceedingNotNullElement(nameElements, i - 1);
            if (index >= 0)
            {
              if ((index > 0) || (!isAbsolute))
              {
                nameElements[i] = null;
                nameElements[index] = null;
              }
            }
          }
        }
      }
      nameElements = str().removeNull(nameElements);
      return str().asString(nameElements, "/");
    }
    return "";
  }

  protected int indexOfPreceedingNotNullElement(String[] elements, int start)
  {
    for (int i = start; i >= 0; i--)
    {
      if (elements[i] != null)
      {
        if ("..".equals(elements[i])) // This is not a valid not null element
        {
          return -1;
        }
        return i;
      }
    }
    return -1;
  }

  /**
   * Returns the classpath definition from the manifest file in the specified
   * JAR or null if not found.
   */
  protected String classpathFromManifest(String jarFilename)
  {
    JarFile jarFile;
    Manifest manifest;
    String classpath;
    Attributes attrs;

    try
    {
      jarFile = new JarFile(jarFilename);
    }
    catch (IOException e)
    {
      handleException(e);
      return null;
    }
    try
    {
      manifest = jarFile.getManifest();
      if (manifest == null)
        return null;

      attrs = manifest.getMainAttributes();
      if (attrs == null)
        return null;

      classpath = attrs.getValue("Class-Path");
    }
    catch (IOException e1)
    {
      handleException(e1);
      // I hate code polution by unwanted exception handling !!!!!!
      return null;
    }
    finally
    {
      try
      {
        jarFile.close();
      }
      catch (Exception e)
      {
        handleException(e);
        // I hate code polution by unwanted exception handling !!!!!!
      }
    }
    return classpath;
  }

  protected String constructClasspathFrom(String jarFilename)
  {
    String classpath;
    String[] elements;

    classpath = classpathFromManifest(jarFilename);
    if (classpath != null)
    {
      classpath = jarFilename + " " + classpath;
      elements = str().parts(classpath, " ");
      classpath = str().asString(elements, File.pathSeparator);
    }
    return classpath;
  }

  protected Classpath calculateEclipseClasspath()
  {
    OrderedSet<String> elements;
    Object classLoader;
    Object bundleLoader;
    try
    {
      elements = new OrderedSet<String>(50);
      classLoader = getCurrentClassLoader();
      bundleLoader = reflect().getValueOf(classLoader, "delegate");
      appendClasspathEntriesFrom(elements, bundleLoader);
      return new Classpath(elements);
    }
    catch (Exception e)
    {
      handleException(e);
      // Don't care
    }
    return null;
  }

  protected void appendClasspathEntriesFrom(Collection<String> elements, Object bundleLoader)
  {
    Object classloader;
    Object bundlefile;
    Object[] classpathEntries;
    Object[] requiredBundles;
    Object bundleLoader2;
    File file;

    try
    {
      classloader = reflect().getValueOf(bundleLoader, "classloader");
      if (classloader == null)
      {
        return;
      }
      classpathEntries = (Object[])reflect().getValueOf(classloader, "classpathEntries");
      if (classpathEntries != null)
      {
        for (int i = 0; i < classpathEntries.length; i++)
        {
          bundlefile = reflect().getValueOf(classpathEntries[i], "bundlefile");
          file = (File)Dynamic.invoke(bundlefile, "getBaseFile");
          elements.add(file.getPath());
        }
      }
      requiredBundles = (Object[])reflect().getValueOf(bundleLoader, "requiredBundles");
      if (requiredBundles != null)
      {
        for (int i = 0; i < requiredBundles.length; i++)
        {
          bundleLoader2 = Dynamic.invoke(requiredBundles[i], "getBundleLoader");
          appendClasspathEntriesFrom(elements, bundleLoader2);
        }
      }
    }
    catch (Exception e)
    {
      handleException(e);
      // Don't care
    }
  }

  /**
   * Returns the classpath this JVM is using. This is either the current
   * value of the system property "java.class.path" or in case of JVM started
   * with option -jar the classpath from the manifest file in the startup JAR.
   */
  protected String getClasspathFromSystemProperty()
  {
    String classpath;
    String[] elements;
    String suffix;

    classpath = System.getProperty(SYS_PROP_CLASSPATH);
    elements = str().parts(classpath, File.pathSeparator);
    if (elements.length == 1)
    {
      suffix = str().cutHead(elements[0], ".");
      if (JAR_FILE_EXTENSION.equalsIgnoreCase(suffix))
      {
        classpath = constructClasspathFrom(elements[0]);
        if (classpath == null)
        {
          classpath = elements[0];
        }
      }
    }

    return classpath;
  }

  protected Classpath determineLookupClasspath(String sysPropName)
  {
    String classpath = null;

    if (!str().isNullOrEmpty(sysPropName))
    {
      classpath = System.getProperty(sysPropName);
    }
    if (!str().isNullOrBlank(classpath))
    {
      classpath = createClasspathFrom(classpath);
      if (classpath != null)
      {
        return new Classpath(classpath);
      }
    }
    return createSystemClasspath();
  }

  protected Classpath createSystemClasspath()
  {
    String classpath;

    if (systemClasspath == null)
    {
      classpath = System.getProperty(SYS_PROP_EXT_DIRS, "");
      if (!str().isNullOrEmpty(classpath))
      {
        classpath = classpath + File.pathSeparator;
      }
      classpath = classpath + getClasspath();
      systemClasspath = new Classpath(classpath);
    }
    return systemClasspath;
  }

  protected String createClasspathFrom(String path)
  {
    TextEngine textEngine;
    VariableResolver varResolver;

    varResolver = new BasicVariableContainer(System.getProperties());
    textEngine = new TextEngine(varResolver);
    textEngine.setVarStartPrefix(new Character('$'));
    textEngine.setVarStartDelimiter('{');
    textEngine.setVarEndDelimiter('}');
    textEngine.setSpecialNameCharacters(".");
    try
    {
      return textEngine.completeText(path);
    }
    catch (TextReplacementException e)
    {
      handleException(e);
      return path;
    }
  }

  protected void assertArgumentNotNull(String argName, Object arg)
  {
    if (arg == null)
    {
      throw new IllegalArgumentException("Parameter " + argName + " must not be null!");
    }
  }

  protected ClassLoader getCurrentClassLoader()
  {
    return currentClassLoader;
  }

  protected IOUtil ioUtil()
  {
    return IOUtil.current();
  }

  protected StringUtil str()
  {
    return StringUtil.current();
  }

  protected ReflectUtil reflect()
  {
    return ReflectUtil.current();
  }

  protected void handleException(Exception ex)
  {
    if (DEBUG)
    {
      ex.printStackTrace();
    }
  }
}
