// ===========================================================================
// CONTENT  : CLASS FileWalker
// AUTHOR   : Manfred Duchrow
// VERSION  : 3.0 - 28/11/2020
// HISTORY  :
//  21/01/2000  duma  CREATED
//	14/02/2003	duma	added		->	Support for patterns with wildcards for digits
//	04/07/2003	duma	bugfix	->	NullPointerException in walkThrough for protected directories
//	25/10/2003	duma	changed	->	To match '*' to empty strings by default
//	24/12/2004	duma	added		->	walkThrough( String )
//	09/06/2006	mdu		added		->	fileProcessor support
//	09/03/2007	mdu		bugfix	->	avoid NPE in walkThrough( String searchPattern )
//  12/01/2014  mdu   changed ->  from AFileProcessor to IFileProcessor
//  28/11/2020  mdu   added   ->  create(...), processAll(...), processAll(...)
//
// Copyright (c) 2000-2020, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.file;

import java.io.File;
import java.io.FilenameFilter;

import org.pfsw.util.CollectionUtil;

/**
 * This class provides services to navigate through a file directory and
 * handle files that match a name filter.
 * <p/>
 * It can be used either with a FileHandler or an IFileProcessor.
 *
 * @author Manfred Duchrow
 * @version 3.0
 */
public class FileWalker
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final CollectionUtil CU = CollectionUtil.current();

  /**
   * The character to be used to separate filename patterns (';').
   */
  public static final char PATTERN_SEPARATOR_CHAR = ';';

  /**
   * The character to be used to separate filename patterns (';') as String.
   */
  public static final String PATTERN_SEPARATOR = ExtendedFileFilter.PATTERN_SEPARATOR;

  /**
   * The wildcard pattern that indicates a recursive walk-through in a
   * single string definition (i.e. "**").
   */
  public static final String RECURSIVE_DIR_WILDCARD = "**";

  private static final FilenameFilter DIRECTORIES_ONLY_FILTER = new FilenameFilter()
  {
    @Override
    public boolean accept(File dir, String name)
    {
      File file;
      file = new File(dir, name);
      return file.isDirectory();
    }
  };

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private FileHandler fileHandler = null;
  private IFileProcessor fileProcessor = null;
  private boolean goOn = true;
  private Character digitWildcard = null;
  private boolean handleContentBeforeDirectory = false;

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Creates a new instance with a file processor.
   */
  public static FileWalker create(IFileProcessor processor)
  {
    return new FileWalker(processor);
  }

  /**
   * Creates a new instance with a file handler.
   */
  public static FileWalker create(FileHandler handler)
  {
    return new FileWalker(handler);
  }

  /**
   * Creates a new instance and processes with the given file processor all files 
   * under the given start directory and its sub-directories.
   */
  public static long processAll(String startDir, IFileProcessor processor)
  {
    return create(processor).walkThroughAllRecursive(startDir);
  }
  
  /**
   * Creates a new instance and handles with the given file handler all files 
   * under the given start directory and its sub-directories.
   */
  public static long handleAll(String startDir, FileHandler handler)
  {
    return create(handler).walkThroughAllRecursive(startDir);
  }
  
  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with a file processor.
   */
  public FileWalker(IFileProcessor processor)
  {
    setFileProcessor(processor);
  }

  /**
   * Initialize the new instance with a file handler.
   */
  public FileWalker(FileHandler handler)
  {
    setFileHandler(handler);
  }

  /**
   * Initialize the new instance with a file handler and a wildcard character
   * for digits.
   * 
   * @param handler The file handler that gets all found files
   * @param digitWildcard A character that is used as wildcard for digits in filname patterns
   */
  public FileWalker(FileHandler handler, char digitWildcard)
  {
    this(handler);
    setDigitWildcardChar(digitWildcard);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================

  /**
   * This method starts in the given directory to search for all files
   * matching the given pattern(s).   <br>
   * There can be more than one pattern in the pattern parameter. They have 
   * to be separated by the PATTERN_SEPARATOR (';').   <p>
   * If recursive is <b>true</b> it goes down to each subdirectory and doing
   * the same there.<br>
   * For each matching file (non-directory) the defined <i>FileHandler.handle()</i>
   * is called.
   *
   * @param dir The directory where to start
   * @param pattern The file name pattern(s) for filtering out the correct files ( wildcards '*' and '?' )
   * @param recursive If set to true, the file selection is going down to all subdirectories
   * @return the number of found files, that have matched the given pattern.
   */
  public long walkThrough(String dir, String pattern, boolean recursive)
  {
    ExtendedFileFilter filter = null;

    setGoOn(true);
    filter = createFileFilter();

    if (LoggerProvider.TRACING)
    {
      System.out.println("FileWalker.walkThrough('" + dir + "', '" + pattern + "', " + recursive + ")");
    }

    filter.addPatterns(pattern);

    if (recursive)
    {
      filter.alwaysIncludeDirectories();
    }
    else
    {
      filter.alwaysExcludeDirectories();
    }

    return walkThrough(dir, filter, recursive);
  }

  /**
   * Walks through all directories specified by the given pattern and calls
   * the file handler or file processor for each file that matches the file part 
   * in the given pattern.
   * <br>
   * The pattern may contain '**' in the path to define recursive walk through.
   * The file part of the pattern may contain '*' and '?' and optionally
   * the digit wildcard character for the files to be 
   * passed to the file handler or processor.
   * <p>
   * Examples:
   * <br>"prog/lib\/**\/a*.jar"  ==> Starts in directory prog\lib and walks
   * recursive through all sub-directories to pass all files that match pattern 
   * "a*.jar" to the file handler. 
   * <br>"d:/temp/tmp???.doc"  ==> Passes all files that match pattern 
   * "tmp???.doc" to the file handler.
   * 
   * @param searchPattern The pattern that specifies the path and files to walk through 
   */
  public long walkThrough(String searchPattern)
  {
    boolean recursive = false;
    String dir;
    String fileMatchPattern;
    File file;

    if (searchPattern == null)
    {
      return 0L;
    }

    file = new File(searchPattern);
    fileMatchPattern = file.getName();
    dir = file.getParent();
    if (dir == null)
    {
      dir = ".";
    }
    if (dir.endsWith(RECURSIVE_DIR_WILDCARD))
    {
      recursive = true;
      file = new File(dir);
      dir = file.getParent();
    }
    return walkThrough(dir, fileMatchPattern, recursive);
  }
  
  /**
   * Walks through the given directory and all its sub-directories as well.
   * Calls the file handler's handleFile() method or the file processor's processFile() 
   * for each file (not directory) found.
   * 
   * @param dir The directory to start from.
   * @return The number of files that have been processed. 
   */
  public long walkThroughAllRecursive(final String dir)
  {
    return walkThrough(dir, DefaultFilenameFilter.ALL, true);
  }

  /**
   * Walks through the given directory and optionally through all its 
   * sub-directories as well.
   * Calls the file handler's handleFile() method or the file processor's processFile() 
   * for each file (not directory) that matches the given filter.
   * 
   * @param dir The directory to start from.
   * @param filter The filter that decides whether or not a file is passed to the FileHandler.
   * @param recursive If true all sub-directories are scanned as well.
   * @return The number of files that have matched the filter and have been processed. 
   */
  public long walkThrough(final String dir, final FilenameFilter filter, final boolean recursive)
  {
    long counter = 0;
    File directory = null;
    File file = null;
    File[] files = null;
    int index = 0;

    directory = new File(dir);
    if (LoggerProvider.TRACING)
    {
      System.out.println("FileWalker.walkThrough() - directory: " + directory.getAbsolutePath());
    }
    files = directory.listFiles(filter);
    if (files == null) // BUGFIX suggested by Kyle Gossman
    {
      if (LoggerProvider.TRACING)
      {
        System.out.println("FileWalker.walkThrough() - directory.list() returned null");
      }
      return counter;
    }
    setGoOn(directoryStart(directory, files.length));
    if (!getGoOn())
    {
      if (LoggerProvider.TRACING)
      {
        System.out.println("FileWalker.walkThrough() - cancelled after directoryStart()");
      }
      return counter;
    }

    for (index = 0; index < files.length; index++)
    {
      if (LoggerProvider.TRACING)
      {
        System.out.println("FileWalker.walkThrough() - '" + files[index].getName() + "' index=" + index);
      }

      file = files[index];
      if (file.isDirectory())
      {
        if (LoggerProvider.TRACING)
        {
          System.out.println("FileWalker.walkThrough() - '" + file.getName() + "' is directory; recursive=" + recursive);
        }
        if (recursive)
        {
          if (LoggerProvider.TRACING)
          {
            System.out.println("FileWalker.walkThrough() - walk through sub-folder: " + file.getName());
          }
          counter += walkThrough(file.getPath(), filter, recursive);
        }
      }
      else
      {
        if (LoggerProvider.TRACING)
        {
          System.out.println("FileWalker.walkThrough() - before handleFile('" + file.getName() + "')");
        }
        setGoOn(handleFile(file));
        counter++;
      }
      if (!getGoOn())
      {
        if (LoggerProvider.TRACING)
        {
          System.out.println("FileWalker.walkThrough() - cancelled after handleFile() - counter: " + counter);
        }
        return counter;
      }
    }
    setGoOn(directoryEnd(directory));

    return counter;
  }

  /**
   * Walks through the given start directory and optionally through all its 
   * sub-directories as well.
   * Calls the file handler or file processor for each directory (not files) that 
   * match the given filter. 
   * However, it scans all directories, even those that do not match the filter.
   * 
   * @param startDir The directory to start from.
   * @param filter The filter that decides whether or not a directory is passed to the FileHandler.
   * @param recursive If true all sub-directories are scanned as well.
   * @see #walkThroughDirectoriesSubDirsFirst(String, FilenameFilter)
   */
  public long walkThroughDirectories(String startDir, FilenameFilter filter, boolean recursive)
  {
    File directory;

    directory = new File(startDir);
    return scanDirectories(directory, filter, recursive, false);
  }

  /**
   * Walks through the given start directory and recursively through all its sub-directories.
   * Calls the file handler or file processor for each directory (not files) that 
   * matches the given filter. It walks down the hierarchy of sub-directories before
   * it calls the handler/processor. That can be used for directory deletion. See {@link FileDeleteProcessor}. 
   * However, it scans all directories, even those that do not match the filter.
   * 
   * @param startDir The directory to start from.
   * @param filter The filter that decides whether or not a directory is passed to the FileHandler.
   * @see #walkThroughDirectories(String, FilenameFilter, boolean)
   */
  public long walkThroughDirectoriesSubDirsFirst(String startDir, FilenameFilter filter)
  {
    File directory;

    directory = new File(startDir);
    return scanDirectories(directory, filter, true, true);
  }

  /**
   * Sets the given character as a wildcard character to match
   * digits ('0'-'9') only.   <br>
   * 
   * @param digitWildcard The placeholder character for digits
   */
  public FileWalker setDigitWildcardChar(char digitWildcard)
  {
    if (digitWildcard <= 0)
    {
      setDigitWildcard(null);
    }
    else
    {
      setDigitWildcard(new Character(digitWildcard));
    }
    return this;
  }

  /**
   * Returns whether or not the contents of a directory should be walked through
   * before the directory is passed to the file processor.
   * <br/>
   * The default is false.
   */
  public boolean getHandleContentBeforeDirectory()
  {
    return handleContentBeforeDirectory;
  }

  /**
   * Defines whether or not the contents of a directory should be walked through
   * before the directory is passed to the file handler.
   */
  public FileWalker setHandleContentBeforeDirectory(boolean newValue)
  {
    handleContentBeforeDirectory = newValue;
    return this;
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected long scanDirectories(final File startDir, final FilenameFilter filter, final boolean recursive, final boolean subDirsFirst)
  {
    long counter = 0;
    File dir;
    File[] directories;

    directories = startDir.listFiles(DIRECTORIES_ONLY_FILTER);
    if (CU.isNullOrEmpty(directories))
    {
      return counter;
    }

    setGoOn(directoryStart(startDir, directories.length));
    if (!getGoOn())
    {
      return counter;
    }

    for (int i = 0; i < directories.length; i++)
    {
      dir = directories[i];

      if (subDirsFirst)
      {
        if (getGoOn() && recursive)
        {
          counter += scanDirectories(dir, filter, recursive, subDirsFirst);
        }
      }
      if (filter.accept(startDir, dir.getName()))
      {
        setGoOn(handleFile(dir));
        counter++;
      }
      if (!subDirsFirst)
      {
        if (getGoOn() && recursive)
        {
          counter += scanDirectories(dir, filter, recursive, subDirsFirst);
        }
      }
      if (!getGoOn())
      {
        return counter;
      }
    }
    setGoOn(directoryEnd(startDir));

    return counter;
  }

  protected char getDigitWildcardChar()
  {
    if (hasDigitWildcard())
    {
      return getDigitWildcard().charValue();
    }
    return '\0';
  }

  protected boolean hasDigitWildcard()
  {
    return getDigitWildcard() != null;
  }

  protected ExtendedFileFilter createFileFilter()
  {
    if (hasDigitWildcard())
    {
      return new ExtendedFileFilter(getDigitWildcardChar(), restrictedWilcardMatch());
    }
    return new ExtendedFileFilter(restrictedWilcardMatch());
  }

  /**
   * Returns true to define that '*' wildcards don't match empty strings.
   * The default result is false.
   * Subclasses may override this method to return true.
   */
  protected boolean restrictedWilcardMatch()
  {
    return false;
  }

  protected boolean handleFile(File file)
  {
    if (getFileProcessor() != null)
    {
      if (LoggerProvider.TRACING)
      {
        System.out.println("FileWalker.handleFile() - " + getFileProcessor().getClass().getName() + ".processFile('" + file.getName() + "')");
      }
      return getFileProcessor().processFile(file);
    }
    if (LoggerProvider.TRACING)
    {
      System.out.println("FileWalker.handleFile() - " + getFileHandler().getClass().getName() + ".handleFile('" + file.getName() + "')");
    }
    return getFileHandler().handleFile(file);
  }

  protected boolean directoryEnd(File dir)
  {
    if (LoggerProvider.TRACING)
    {
      System.out.println("FileWalker.directoryEnd(" + dir.getAbsolutePath() + ")");
    }

    if (getFileProcessor() != null)
    {
      return true;
    }
    return getFileHandler().directoryEnd(dir);
  }

  protected boolean directoryStart(File dir, int count)
  {
    if (LoggerProvider.TRACING)
    {
      System.out.println("FileWalker.directoryStart('" + dir.getAbsolutePath() + "', " + count + ")");
    }
    if (getFileProcessor() != null)
    {
      return true;
    }
    return getFileHandler().directoryStart(dir, count);
  }

  protected FileHandler getFileHandler()
  {
    return fileHandler;
  }

  protected void setFileHandler(FileHandler newValue)
  {
    fileHandler = newValue;
  }

  protected IFileProcessor getFileProcessor()
  {
    return fileProcessor;
  }

  protected void setFileProcessor(IFileProcessor newValue)
  {
    fileProcessor = newValue;
  }

  protected boolean getGoOn()
  {
    return goOn;
  }

  protected void setGoOn(boolean newValue)
  {
    goOn = newValue;
  }

  protected Character getDigitWildcard()
  {
    return digitWildcard;
  }

  protected void setDigitWildcard(Character newValue)
  {
    digitWildcard = newValue;
  }
}
