// ===========================================================================
// CONTENT  : CLASS NamedInputStream
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 14/05/2015
// HISTORY  :
//  14/05/2015  mdu  CREATED
//
// Copyright (c) 2015, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.file;

import java.io.Closeable;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;

/**
 * A container for an InputStream with an associated name.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class NamedInputStream extends NamedCloseable
{
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private final InputStream inputStream;

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Opens the provided file and returns the input stream along with the file name.
   * 
   * @param file The file to be opened (must not be null).
   * @return An object that contains the (normalized) absolute file name and the
   * FileInputStream on the just opened file.
   * @throws FileNotFoundException If the provided file cannot be found.
   */
  public static NamedInputStream open(File file) throws FileNotFoundException
  {
    FileInputStream inStream;
    String name;

    inStream = new FileInputStream(file);
    name = FileUtil.current().standardize(file.getAbsolutePath());
    return new NamedInputStream(name, inStream);
  }

  /**
   * Opens the file specified by the given filename and returns the input stream along with the file name.
   * 
   * @param filename The file to be opened (must not be null).
   * @return An object that contains the (normalized) absolute file name and the
   * FileInputStream on the just opened file.
   * @throws FileNotFoundException If a file with the given name cannot be found.
   */
  public static NamedInputStream open(String filename) throws FileNotFoundException
  {
    File file;

    file = new File(filename);
    return open(file);
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Create a new instance with the given name associated to the given input stream.
   * 
   * @param name The name of the input stream (must not be null).
   * @param stream The input stream to be associated with the name (must not be null).
   * @throws IllegalArgumentException If either name or stream is null.
   */
  public NamedInputStream(String name, InputStream stream)
  {
    super(name);
    if (stream == null)
    {
      throw new IllegalArgumentException("The provided argument for stream is null.");
    }
    this.inputStream = stream;
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the underlying InputStream (will never be null).
   */
  public InputStream getInputStream()
  {
    return inputStream;
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  @Override
  protected Closeable getCloseable()
  {
    return getInputStream();
  }
}
