// ===========================================================================
// CONTENT  : CLASS NamedWriter
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 14/05/2015
// HISTORY  :
//  14/05/2015  mdu  CREATED
//
// Copyright (c) 2015, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.file;

import java.io.Closeable;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.nio.charset.Charset;

/**
 * A container for an Writer with an associated name.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class NamedWriter extends NamedCloseable
{
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private final Writer writer;

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Opens the provided file and returns the writer along with the file name.
   * 
   * @param file The file to be opened (must not be null).
   * @param charset The character encoding of the file.
   * @return An object that contains the (normalized) absolute file name and the
   * OutputStreamWriter on the just opened file.
   * @throws FileNotFoundException If the provided file cannot be opened/created.
   */
  public static NamedWriter open(File file, Charset charset) throws FileNotFoundException
  {
    OutputStreamWriter streamWriter;
    String name;

    streamWriter = new OutputStreamWriter(new FileOutputStream(file), charset);

    name = FileUtil.current().standardize(file.getAbsolutePath());
    return new NamedWriter(name, streamWriter);
  }

  /**
   * Opens the file specified by the given filename and returns the writer along with the file name.
   * 
   * @param filename The file to be opened (must not be null).
   * @param charset The character encoding of the file.
   * @return An object that contains the (normalized) absolute file name and the
   * OutputStreamWriter on the just opened file.
   * @throws FileNotFoundException If a file with the given name cannot be opened/created.
   */
  public static NamedWriter open(String filename, Charset charset) throws FileNotFoundException
  {
    File file;

    file = new File(filename);
    return open(file, charset);
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Create a new instance with the given name associated to the given writer.
   * 
   * @param name The name of the writer (must not be null).
   * @param writer The writer to be associated with the name (must not be null).
   * @throws IllegalArgumentException If either name or writer is null.
   */
  public NamedWriter(String name, Writer writer)
  {
    super(name);
    if (writer == null)
    {
      throw new IllegalArgumentException("The provided argument for writer is null.");
    }
    this.writer = writer;
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the underlying Writer (will never be null).
   */
  public Writer getWriter()
  {
    return this.writer;
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  @Override
  protected Closeable getCloseable()
  {
    return getWriter();
  }
}
