// ===========================================================================
// CONTENT  : CLASS PropertiesFileContent
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.2 - 20/12/2021
// HISTORY  :
//  23/07/2004  mdu  CREATED
//	17/09/2004	mdu		changed	-->	superclass from Object to OrderedProperties
//  26/01/2017  mdu   changed --> to be typed with inner type PropertiesFileElement
//  29/12/2020  mdu   added   --> methods with IName
//  20/12/2021  mdu   added   --> renameProperty()
//
// Copyright (c) 2004-2021, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.file;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import org.pfsw.bif.BasicUtil;
import org.pfsw.bif.text.IName;
import org.pfsw.text.StringPattern;
import org.pfsw.util.OrderedProperties;

/**
 * Represents a properties collection with empty lines and comments.
 * The order will be preserved.
 *
 * @author Manfred Duchrow
 * @version 2.2
 */
public class PropertiesFileContent extends OrderedProperties<PropertiesFileElement>
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final long serialVersionUID = -3428801298203760647L;

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private List<Integer> commentIndex = null;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  public PropertiesFileContent()
  {
    this(INITIAL_CAPACITY);
  }

  /**
   * Initialize the new instance with an initial capacity.
   */
  public PropertiesFileContent(int initialCapacity)
  {
    super(initialCapacity);
    setPropertyIndex(new HashMap<String, Integer>(initialCapacity));
    setCommentIndex(new ArrayList<Integer>(initialCapacity));
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Adds the given line at the end of all contained elements
   */
  public void addLine(String line)
  {
    PropertiesFileElement element;

    if (line != null)
    {
      element = new PropertiesFileElement(line);
      addElement(element);
    }
  }

  /**
   * Modifies the value or adds the property specified by the given name.
   * Returns true if the property was found and modified or added, otherwise false.
   * 
   * @param name The name of the property to set
   * @param value The new value to be set
   */
  @Override
  public synchronized Object setProperty(String name, String value)
  {
    PropertiesFileElement property;
    String oldValue;

    property = findPropertyElement(name);
    if (property == null)
    {
      appendProperty(name, value);
      return null;
    }
    oldValue = property.text();
    property.changeValue(value);
    return oldValue;
  }

  /**
   * Modifies the value of the property specified by the given name.
   * Returns true if the property was found and modified, otherwise false.
   * 
   * @param name The name of the property to modify
   * @param newValue The new value to be set
   */
  public boolean modifyProperty(String name, String newValue)
  {
    PropertiesFileElement property;

    property = findPropertyElement(name);
    if (isNullOrDeleted(property))
    {
      return false;
    }
    property.changeValue(newValue);
    return true;
  }

  /**
   * Modifies the value of the property specified by the given name.
   * Returns true if the property was found and modified, otherwise false.
   * 
   * @param name The name of the property to modify
   * @param newValue The new value to be set
   */
  public boolean modifyProperty(IName name, String newValue)
  {
    return modifyProperty(name.asString(), newValue);
  }

  /**
   * Renames the the property specified with the given name to the specified new name.
   * Returns true if the property was found and renamed, otherwise false.
   * 
   * @param name The name of the property to rename.
   * @param newName The new name to be set.
   */
  public boolean renameProperty(String name, String newName)
  {
    PropertiesFileElement property;
    boolean renamed;
    int index;

    property = findPropertyElement(name);
    if (isNullOrDeleted(property))
    {
      return false;
    }
    index = indexOfProperty(name);
    renamed = property.renameProperty(newName);
    if (renamed)
    {
      getPropertyIndex().remove(name);
      registerProperty(property, index);
    }
    return renamed;
  }

  /**
   * Renames the the property specified with the given name to the specified new name.
   * Returns true if the property was found and renamed, otherwise false.
   * 
   * @param name The name of the property to rename.
   * @param newName The new name to be set.
   */
  public boolean renameProperty(IName name, IName newName)
  {
    if (BasicUtil.BU.isAnyNull(name, newName))
    {
      return false;
    }
    return renameProperty(name.asString(), newName.asString());
  }

  /**
   * Adds the property with the specified name and value.
   * Returns true if the property was added, otherwise false.
   * In particular it will not be added if a property with the same name
   * already exists (returns false then). 
   * 
   * @param name The name of the property to add.
   * @param value The value to be set.
   */
  public boolean addProperty(String name, String value)
  {
    PropertiesFileElement property;

    property = findPropertyElement(name);
    if (!isNullOrDeleted(property))
    {
      return false;
    }
    return appendProperty(name, value);
  }

  /**
   * Adds the property with the specified name and value.
   * Returns true if the property was added, otherwise false.
   * In particular it will not be added if a property with the same name
   * already exists (returns false then). 
   * 
   * @param name The name of the property to add.
   * @param value The value to be set.
   */
  public boolean addProperty(IName name, String value)
  {
    return addProperty(name.asString(), value);
  }

  /**
   * Turns the property specified by the given name to a comment.
   * Returns true if the property was found and changed to a comment, 
   * otherwise false.
   * 
   * @param name The name of the property to change to be a comment.
   */
  public boolean commentProperty(String name)
  {
    PropertiesFileElement property;
    Integer index;
    boolean success;

    property = findPropertyElement(name);
    if (isNullOrDeleted(property))
    {
      return false;
    }
    index = indexOfProperty(name);
    getPropertyIndex().remove(name);
    success = property.becomeComment();
    registerComment(index);
    return success;
  }

  /**
   * Turns the property specified by the given name to a comment.
   * Returns true if the property was found and changed to a comment, 
   * otherwise false.
   * 
   * @param name The name of the property to change to be a comment.
   */
  public boolean commentProperty(IName name)
  {
    return commentProperty(name.asString());
  }

  /**
   * Changes a comment containing the property specified by the given name 
   * to a real property setting, by removing the leading comment indicator '#'.
   * Returns true if the property was found in a comment and changed to 
   * a non-comment, otherwise false.
   * 
   * @param name The name of the property to change from a comment to a property.
   */
  public boolean uncommentProperty(String name)
  {
    Integer[] index;
    StringPattern pattern;
    PropertiesFileElement comment;
    boolean success;

    // Sample pattern: "#*property.name*=*"
    pattern = StringPattern.create(PropertiesFileElement.COMMENT_INDICATOR + "*" + name + "*" + PropertiesFileElement.ASSIGNMENT_CHAR + "*");
    index = getCommentIndex().toArray(new Integer[0]);
    for (int i = 0; i < index.length; i++)
    {
      comment = elementAt(index[i]);
      if (pattern.matches(comment.getLine()))
      {
        getCommentIndex().remove(index[i]);
        success = comment.becomeProperty();
        registerProperty(comment, index[i]);
        return success;
      }
    }
    return false;
  }

  /**
   * Changes a comment containing the property specified by the given name 
   * to a real property setting, by removing the leading comment indicator '#'.
   * Returns true if the property was found in a comment and changed to 
   * a non-comment, otherwise false.
   * 
   * @param name The name of the property to change from a comment to a property.
   */
  public boolean uncommentProperty(IName name)
  {
    return uncommentProperty(name.asString());
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  /**
   * Create a new property object with the given name and value. Subclasses
   * may override this to implement their own property class.
   */
  @Override
  protected PropertiesFileElement newProperty(String name, String value)
  {
    return new PropertiesFileElement(name, value);
  }

  @Override
  protected void addElement(PropertiesFileElement element)
  {
    Integer index;
    PropertiesFileElement propertyElement;

    propertyElement = element;
    index = new Integer(getElements().size());
    getElements().add(propertyElement);
    if (propertyElement.isProperty())
    {
      registerProperty(propertyElement, index);
    }
    else if (propertyElement.isComment())
    {
      registerComment(index);
    }
  }

  protected PropertiesFileElement findPropertyElement(String key)
  {
    return findProperty(key);
  }

  protected PropertiesFileElement elementAt(Integer index)
  {
    return elementAt(index.intValue());
  }

  protected PropertiesFileElement elementAt(int index)
  {
    return propertyAt(index);
  }

  protected void registerComment(Integer index)
  {
    getCommentIndex().add(index);
  }

  @Override
  protected boolean isValidProperty(PropertiesFileElement property)
  {
    if (property != null)
    {
      return property.isProperty();
    }
    return false;
  }

  protected List<Integer> getCommentIndex()
  {
    return commentIndex;
  }

  protected void setCommentIndex(List<Integer> newValue)
  {
    commentIndex = newValue;
  }
}
