// ===========================================================================
// CONTENT  : CLASS PropertiesFileElement
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.1 - 17/09/2004
// HISTORY  :
//  23/07/2004  mdu  CREATED
//	17/09/2004	mdu		changed	-->	superclass from Object to NamedText
//
// Copyright (c) 2004, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.file;

import org.pfsw.text.StringUtil;
import org.pfsw.util.NamedText;

/**
 * Represents one element in a properties file. That can be
 * <ul>
 * <li>An empty line
 * <li>A comment line starting with #
 * <li>A property setting x=y
 * </ul>
 *
 * @author Manfred Duchrow
 * @version 1.1
 */
class PropertiesFileElement extends NamedText
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final long serialVersionUID = -2049012158227903085L;
  protected static final String COMMENT_INDICATOR = "#";
  protected static final String ASSIGNMENT_CHAR = "=";

  protected static final int TYPE_UNKNOWN = -1;
  protected static final int TYPE_EMPTY = 0;
  protected static final int TYPE_COMMENT = 1;
  protected static final int TYPE_PROPERTY = 2;

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private String line = null;
  private int type = TYPE_UNKNOWN;
  private boolean deleted = false;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance from a single line.
   */
  protected PropertiesFileElement(String line)
  {
    super(null, null);
    setLine(line);
    init();
  }

  /**
   * Initialize the new instance from a name/value pair.
   */
  protected PropertiesFileElement(String name, String value)
  {
    this(name.trim() + ASSIGNMENT_CHAR + value.trim());
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Replaces the old value by the given new one.
   */
  public void changeValue(String newValue)
  {
    String newLine;

    if (isProperty() && (newValue != null))
    {
      newLine = str().replaceAll(getLine(), text(), newValue);
      setLine(newLine);
      text(newValue);
    }
  }

  /**
   * Replaces the old property name by the given new one.
   * Returns when the renaming was executed.
   */
  public boolean renameProperty(String newName)
  {
    String newLine;
    
    if (isProperty() && str().notNullOrBlank(newName))
    {
      newLine = str().replaceAll(getLine(), name(), newName);
      setLine(newLine);
      name(newName);
      return true;
    }
    return false;
  }
  
  /**
   * Changes this property to a comment
   */
  public boolean becomeComment()
  {
    String newLine;

    if (isProperty())
    {
      newLine = COMMENT_INDICATOR + getLine();
      setLine(newLine);
      init();
      return isComment();
    }
    return false;
  }

  /**
   * Changes this comment to a property
   */
  public boolean becomeProperty()
  {
    String newLine;

    if (isComment())
    {
      newLine = str().suffix(getLine(), COMMENT_INDICATOR);
      setLine(newLine);
      init();
      return isProperty();
    }
    return false;
  }

  /**
   * Returns true if this is a property element
   */
  public boolean isProperty()
  {
    return getType() == TYPE_PROPERTY;
  }

  /**
   * Returns true if this is an empty element
   */
  public boolean isEmptyLine()
  {
    return getType() == TYPE_EMPTY;
  }

  /**
   * Returns true if this is a comment element
   */
  public boolean isComment()
  {
    return getType() == TYPE_COMMENT;
  }

  /**
   * Marks this object as undeleted
   */
  public void undelete()
  {
    deleted(false);
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected void init()
  {
    String trimmedLine;
    String[] keyValue;

    trimmedLine = line.trim();
    if (trimmedLine.length() == 0)
    {
      setType(TYPE_EMPTY);
      return;
    }
    if (trimmedLine.startsWith(COMMENT_INDICATOR))
    {
      setType(TYPE_COMMENT);
      return;
    }
    if (trimmedLine.indexOf(ASSIGNMENT_CHAR) < 0)
    {
      setType(TYPE_UNKNOWN);
      return;
    }
    keyValue = str().splitNameValue(trimmedLine, ASSIGNMENT_CHAR);
    name(keyValue[0].trim());
    text(keyValue[1].trim());
    setType(TYPE_PROPERTY);
  }

  protected String inspectString()
  {
    return getLine();
  }

  protected String getLine()
  {
    return line;
  }

  protected void setLine(String newValue)
  {
    line = newValue;
  }

  protected int getType()
  {
    return type;
  }

  protected void setType(int newValue)
  {
    type = newValue;
  }

  protected boolean deleted()
  {
    return deleted;
  }

  protected void deleted(boolean newValue)
  {
    deleted = newValue;
  }

  protected StringUtil str()
  {
    return StringUtil.current();
  }
}
