// ===========================================================================
// CONTENT  : CLASS PropertiesFileWriter
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.2 - 12/09/2015
// HISTORY  :
//  27/07/2004  mdu  CREATED
//	17/09/2004	mdu		added		-->	writeTo( Writer, ..)
//	12/09/2015	mdu		added		-->	writeTo( File, ..), bugfix PF-56 in writeTo(Writer writer, ...)
//
// Copyright (c) 2004-2015, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.file;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;

/**
 * Writes ordered properties with preserved comments and blank lines to a file.
 *
 * @author Manfred Duchrow
 * @version 1.2
 */
public class PropertiesFileWriter
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  protected static final String NEWLINE = System.getProperty("line.separator");

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  public PropertiesFileWriter()
  {
    super();
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Writes the given properties including all comments and blank lines
   * to a file with the specified name.
   * 
   * @param filename The name of the file to write to (must not be null).
   * @param properties The properties to write (must not be null).
   */
  public void writeTo(String filename, PropertiesFileContent properties) throws IOException
  {
    writeTo(new File(filename), properties);
  }

  /**
   * Writes the given properties including all comments and blank lines
   * to the specified file.
   * 
   * @param file The file to write to (must not be null).
   * @param properties The properties to write (must not be null).
   */
  public void writeTo(File file, PropertiesFileContent properties) throws IOException
  {
    FileWriter writer;

    writer = new FileWriter(file);
    writeTo(writer, properties);
  }

  /**
   * Writes the given properties including all comments and blank lines
   * to the given writer.
   * When the method is finished the writer is closed. That is also true
   * if an exception was thrown.
   * 
   * @param writer The writer object to write the properties to (must not be null).
   * @param properties The properties to write (must not be null).
   */
  public void writeTo(Writer writer, PropertiesFileContent properties) throws IOException
  {
    PropertiesFileElement element;

    try
    {
      for (int i = 0; i < properties.numberOfElements(); i++)
      {
        element = properties.elementAt(i);
        writer.write(element.getLine());
        writer.write(NEWLINE);
      }
    }
    finally
    {
      fileUtil().close(writer);
    }
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected FileUtil fileUtil()
  {
    return FileUtil.current();
  }
}
