// ===========================================================================
// CONTENT  : CLASS TableOfContents
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.0 - 06/11/2017
// HISTORY  :
//  16/08/2003  mdu  CREATED
//  06/11/2017  mdu   added   --> findSubDirs()
//
// Copyright (c) 2003-2017, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.file;

import java.util.ArrayList;
import java.util.List;

import org.pfsw.text.StringUtil;

/**
 * An instance of this class contains a list of DirectoryContents objects.
 * The purpose is to hold in one object all directory and file information 
 * from either an archive (zip/jar) or a base directory (e.g. "/usr/apps/conf").
 * <br>
 * To represent both types in a common information structure allows to present
 * or analyze the content information in a common way. 
 *
 * @author Manfred Duchrow
 * @version 2.0
 */
public class TableOfContents
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final StringUtil SU = StringUtil.current();
  
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private String containerName = null;
  private List<DirectoryContents> dirInfos = null;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  public TableOfContents(String name)
  {
    super();

    if (name == null)
    {
      throw new IllegalArgumentException("name must not be null!");
    }
    setContainerName(name);
    setDirInfos(new ArrayList<DirectoryContents>());
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the name of the container this table of contents belongs to
   */
  public String getName()
  {
    return getContainerName();
  } 

  /**
   * Adds the given directory contents holder to the table of contents
   */
  public void add(DirectoryContents dir)
  {
    if (dir != null)
    {
      getDirInfos().add(dir);
    }
  }

  /**
   * Returns all contained directories
   */
  public DirectoryContents[] getDirectories()
  {
    DirectoryContents[] array;

    array = new DirectoryContents[getDirInfos().size()];
    getDirInfos().toArray(array);
    return array;
  }

  /**
   * Returns the number of directory entries in this table of contents
   */
  public int size()
  {
    return getDirInfos().size();
  }

  /**
   * Returns the DirectoryContents entry at the specified index.
   * 
   * @throws ArrayIndexOutOfBoundsException if the index is not valid
   */
  public DirectoryContents dirContentAt(int index)
  {
    return getDirInfos().get(index);
  } 

  /**
   * Returns the name of the directory entry at the specified index.
   * 
   * @throws ArrayIndexOutOfBoundsException if the index is not valid
   */
  public String dirNameAt(int index)
  {
    return dirContentAt(index).getDirName();
  } 

  /**
   * Returns the file info objects of the directory entry at the specified index.
   * 
   * @throws ArrayIndexOutOfBoundsException if the index is not valid
   */
  public FileInfo[] filesAt(int index)
  {
    return dirContentAt(index).getContainedFiles();
  }
  
  /**
   * Returns all sub-directories under the given base directory.
   * The returned directories do not contain the baseDir path and no
   * leading slash.
   * 
   * @param baseDir The directory under which to look for the sub directories.
   * @return A list of the found directories (maybe empty, but never null).
   */
  public List<DirectoryContents> findSubDirsOf(String baseDir)
  {
    List<DirectoryContents> dirs;
    DirectoryContents dir;
    String subDirName;
    
    dirs = new ArrayList<DirectoryContents>(200);
    for (DirectoryContents directory : getDirectories())
    {
      if (directory.getDirectoryName().startsWith(baseDir))
      {
        subDirName = SU.cutHead(directory.getDirectoryName(), baseDir);
        dir = new DirectoryContents(SU.trim(subDirName, "/"), directory.getContainedFiles());
        dirs.add(dir);
      }
    }
    return dirs;
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  /**
   * Returns a display string for an inspector.
   * Active JOI support !
   */
  protected String inspectString()
  {
    return "TableOfContents(\"" + getName() + "\")";
  }

  protected String getContainerName()
  {
    return containerName;
  }

  protected void setContainerName(String newValue)
  {
    containerName = newValue;
  }

  protected List<DirectoryContents> getDirInfos()
  {
    return dirInfos;
  }

  protected void setDirInfos(List<DirectoryContents> newValue)
  {
    dirInfos = newValue;
  }
}
