// ===========================================================================
// CONTENT  : CLASS ZipFileCreator
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 28/11/2020
// HISTORY  :
//  28/11/2020  mdu  CREATED
//
// Copyright (c) 2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.file;

import static org.pfsw.file.FileUtil.*;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import org.pfsw.bif.exceptions.IORuntimeException;
import org.pfsw.util.StreamCopy;

/**
 * Simplifies creation of Zip files.
 * 
 * @author M.Duchrow
 */
public class ZipFileCreator
{
  private boolean preserveBaseFolder = false;
  
  public static ZipFileCreator create()
  {
    return new ZipFileCreator();
  }
  
  public ZipFileCreator()
  {
    super();
  }
  
  /**
   * Create (or overwrite) the zip file and put all file under the specified start folder
   * and its sub-folders into it.
   * 
   * @param startFolder The folder from where to start recursive file processing (must not be null). 
   * @param zipFile The target zip file (must not be null).
   * @return The number of processed (zipped) files.
   * @throws IORuntimeException in any case of error.
   */
  public long zipAllInto(String startFolder, File zipFile)
  {
    ZipOutputStream zipStream;
    ZipFileCopyProcessor fileZipper;
    
    try
    {
      zipStream = new ZipOutputStream(new FileOutputStream(zipFile));
    }
    catch (FileNotFoundException e)
    {
      throw new IORuntimeException(e, "Failed to create zip file %s", zipFile.getAbsolutePath());
    }
    fileZipper = new ZipFileCopyProcessor(startFolder, zipStream);
    try
    {
      return FileWalker.processAll(startFolder, fileZipper);
    }
    finally
    {
      FU.close(zipStream);
    }
  }
  
  /**
   * Returns true if all files should be put under the base folder name in the zip file.
   * The default value is false.
   */
  public boolean preserveBaseFolder()
  {
    return this.preserveBaseFolder;
  }

  /**
   * Sets whether or not all files should be put under the base folder name in the zip file.
   * The default value is false.
   */
  public ZipFileCreator setPreserveBaseFolder(boolean preserveBaseFolder)
  {
    this.preserveBaseFolder = preserveBaseFolder;
    return this;
  }

  // =========================================================================
  // INNER CLASSES
  // =========================================================================
  private class ZipFileCopyProcessor extends AFileProcessor
  {
    private final File baseFolder;
    private final ZipOutputStream zipStream;
    private final StreamCopy streamCopy = StreamCopy.create().setBufferSize(1024).setCloseStreamsAfterCopy(false);

    ZipFileCopyProcessor(String baseFolder, ZipOutputStream zipStream)
    {
      super();
      this.baseFolder = new File(baseFolder);
      this.zipStream = zipStream;
    }
    
    @Override
    public boolean processFile(File file)
    {
      FileInputStream inStream;
      String relativeFilePath;
      ZipEntry zipEntry;

      try
      {
        inStream = new FileInputStream(file);
      }
      catch (IOException e)
      {
        throw new IORuntimeException(e, "Failed to open file '%s' to be zipped", file.getAbsolutePath());
      }
      try
      {
        relativeFilePath = getRootFolderInZip() + FU.getRelativePath(baseFolder, file);
        zipEntry = new ZipEntry(relativeFilePath);
        zipEntry.setTime(file.lastModified());
        
        zipStream.putNextEntry(zipEntry);
        streamCopy.copy(inStream, zipStream);
        zipStream.closeEntry();
      }
      catch (IOException e)
      {
        throw new IORuntimeException(e, "Failed to copy file '%s' into zip file", file.getAbsolutePath());
      }
      finally
      {
        FU.close(inStream);
      }
      return true;
    }
    
    private String getRootFolderInZip() 
    {
      if (preserveBaseFolder())
      {
        return this.baseFolder.getName() + "/";
      }
      return "";
    }
  }
}
