// ===========================================================================
// CONTENT  : CLASS MediaTypeConstant
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.0 - 14/05/2020
// HISTORY  :
//  19/08/2014  mdu  CREATED
//  14/05/2020  mdu   changed -> from class to enum and added new values
//
// Copyright (c) 2014-2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.net;

/**
 * A definition of internet media types that are comprised of a type and a sub-type.
 *
 * @author Manfred Duchrow
 * @version 2.0
 */
public enum MediaTypeConstant implements MediaType
{
  //@formatter:off
  APPLICATION_ATOM_XML("application", "atom+xml"),
  APPLICATION_GRAPHQL("application", "graphql"),
  APPLICATION_FORM_URLENCODED("application", "x-www-form-urlencoded"),
  APPLICATION_JSON("application", "json"),
  APPLICATION_JAVASCRIPT("application", "javascript"),
  APPLICATION_MSWORD("application", "msword"),
  APPLICATION_OCTET_STREAM("application", "octet-stream"),
  APPLICATION_PDF("application", "pdf"),
  APPLICATION_XHTML_XML("application", "xhtml+xml"),
  APPLICATION_XML("application", "xml"),
  APPLICATION_ZIP("application", "zip"),
  AUDIO_MPEG("audio", "mpeg"),
  AUDIO_OGG("audio", "ogg"),
  IMAGE_GIF("image", "gif"),
  IMAGE_JPEG("image", "jpeg"),
  IMAGE_PNG("image", "png"),
  MULTIPART_FORM_DATA_TYPE("multipart", "form-data"),
  TEXT_CSS("text", "css"),
  TEXT_CSV("text", "csv"),
  TEXT_HTML("text", "html"),
  TEXT_PLAIN("text", "plain"),
  TEXT_XML("text", "xml"),
  WILDCARD_TYPE(MediaType.MEDIA_TYPE_WILDCARD, MediaType.MEDIA_TYPE_WILDCARD)
  ;
  //@formatter:on

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private final String type;
  private final String subType;

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Returns the MediaTypeConstant object for the given type string or null if not found. 
   * 
   * @param typeDef The type definition (e.g. "image/jpg") (must not be null).
   * @return the defined MediaTypeConstant or null.
   */
  public static MediaTypeConstant get(String typeDef)
  {
    if (typeDef != null)
    {
      for (MediaTypeConstant mediaType : values())
      {
        if (mediaType.isType(typeDef))
        {
          return mediaType;
        }
      }
    }
    return null;
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Creates a new MediaTypeConstant.
   * 
   * @param type The main type (must not be null).
   * @param subType The sub type (must not be null).
   * @throws IllegalArgumentException If any of the arguments is null.
   */
  private MediaTypeConstant(String type, String subType)
  {
    this.type = type;
    this.subType = subType;
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the string representation (e.g. "text/html") of this media type.
   */
  @Override
  public String asString()
  {
    return getType() + MEDIA_TYPE_SEPARATOR + getSubType();
  }

  @Override
  public String getType()
  {
    return type;
  }

  @Override
  public String getSubType()
  {
    return subType;
  }

  /**
   * Returns true if the given string equals (case-insensitive) the receiver's
   * string representation.
   */
  public boolean isType(String typeDef)
  {
    return asString().equalsIgnoreCase(typeDef);
  }

  public boolean isWildcardType()
  {
    return getType().equals(MEDIA_TYPE_WILDCARD);
  }

  public boolean isWildcardSubType()
  {
    return getSubType().equals(MEDIA_TYPE_WILDCARD);
  }

  @Override
  public String toString()
  {
    return String.format("%s('%s')", getClass().getSimpleName(), asString());
  }
}
