// ===========================================================================
// CONTENT  : CLASS MediaTypeDefinition
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 14/05/2020
// HISTORY  :
//  14/05/2020  mdu  CREATED
//
// Copyright (c) 2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.net;

import org.pfsw.bif.text.IStringPair;

/**
 * Class that allows free definition of MediaType objects.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class MediaTypeDefinition implements MediaType
{
  private final String type;
  private final String subType;

  /**
   * Creates a new instance based on the given type and sub-type.
   */
  public static MediaTypeDefinition create(String type, String subType)
  {
    return new MediaTypeDefinition(type, subType);
  }

  /**
   * Creates a new instance based on the given type definition which
   * must comply with media type syntax (i.e. "type/subtype").
   * 
   * @param typeDef The media type definition as string (must not be null).
   * @return The new instance.
   * @throws IllegalArgumentException if the given string cannot be parsed. 
   */
  public static MediaTypeDefinition create(String typeDef)
  {
    IStringPair stringPair;

    stringPair = MediaTypes.parse(typeDef);
    return create(stringPair.getString1(), stringPair.getString2());
  }

  /**
   * Creates a new instance based on the given type and sub-type.
   */
  public MediaTypeDefinition(String type, String subType)
  {
    super();
    this.type = type;
    this.subType = subType;
  }

  @Override
  public String getType()
  {
    return this.type;
  }

  @Override
  public String getSubType()
  {
    return this.subType;
  }

  /**
   * Returns the string representation (e.g. "text/html") of this media type.
   */
  @Override
  public String asString()
  {
    return getType() + MEDIA_TYPE_SEPARATOR + getSubType();
  }

  @Override
  public boolean equals(Object obj)
  {
    if (!(obj instanceof MediaType))
    {
      return false;
    }
    String ownString = equalityString();
    String otherString = equalityString((MediaType)obj);
    return ownString.equals(otherString);
  }

  @Override
  public int hashCode()
  {
    return equalityString().hashCode();
  }

  @Override
  public String toString()
  {
    return String.format("%s('%s')", getClass().getSimpleName(), asString());
  }

  private String equalityString() 
  {
    return equalityString(this);
  }
  
  private String equalityString(MediaType mediaType) 
  {
    return mediaType.getClass().getCanonicalName() + mediaType.asString();    
  }
}
