// ===========================================================================
// CONTENT  : CLASS MediaTypes
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 14/05/2020
// HISTORY  :
//  14/05/2020  mdu  CREATED
//
// Copyright (c) 2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.net;

import static org.pfsw.text.StringUtil.*;

import org.pfsw.bif.text.IStringPair;

/**
 * Static utility methods for working with media types.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class MediaTypes
{
  /**
   * Returns the MediaType object for the given type string or null if not found. 
   * 
   * @param typeDef The type definition (e.g. "image/jpg") (must not be null).
   * @return the defined MediaType or null.
   * @throws IllegalArgumentException If the given typeDef is null or otherwise no valid media type definition.
   */
  public static MediaType getMediaType(String typeDef)
  {
    MediaType mediaType;

    if (typeDef == null)
    {
      throw new IllegalArgumentException("Media type definition must not be null");
    }

    mediaType = MediaTypeConstant.get(typeDef);
    return (mediaType == null) ? MediaTypeDefinition.create(typeDef) : mediaType;
  }

  /**
   * Splitting up a given media type string into its type and sub-type.
   * 
   * @param typeDef The media type string to be split-up in its type and sub-type (must not be null).
   * @return A string pair that contains the type and sub-type.
   * @throws IllegalArgumentException if parsing the string fails.
   */
  public static IStringPair parse(String typeDef)
  {
    IStringPair pair;

    if (SU.isNullOrBlank(typeDef))
    {
      throw new IllegalArgumentException("Argument typeDef is null or blank!");
    }
    pair = SU.splitStringPair(typeDef, MediaType.MEDIA_TYPE_SEPARATOR);
    if (SU.isNullOrBlank(pair.getString1()))
    {
      throw new IllegalArgumentException("Specified type is null or blank!");
    }
    if (SU.isNullOrBlank(pair.getString2()))
    {
      throw new IllegalArgumentException("Specified subType is null or blank!");
    }
    return pair;
  }

  // no instances supported
  private MediaTypes()
  {
    super();
  }
}
