// ===========================================================================
// CONTENT  : CLASS NetUtil
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.2 - 24/05/2020
// HISTORY  :
//  20/08/2014  mdu  CREATED
//  10/05/2015  mdu  added  --> withLeadingSlash(), withTrailingSlash(), withoutLeadingSlash(), withoutTrailingSlash()
//  24/05/2020  mdu  added  --> asURL(), asURI()
//
// Copyright (c) 2014-2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.net;

import java.io.Closeable;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.nio.charset.Charset;

import org.pfsw.bif.exceptions.InvalidUrlException;
import org.pfsw.net.internal.LoggerProvider;
import org.pfsw.text.StringUtil;

/**
 * Utility and convenience methods for network programming, provided as singleton.
 * The class has a protected default constructor that allows to extend it and
 * potentially override methods.
 *
 * @author Manfred Duchrow
 * @version 1.2
 */
public class NetUtil
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  public static final NetUtil NU = new NetUtil();

  protected static final char SLASH = '/';

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Returns the only instance this class supports (design pattern "Singleton")
   */
  public static NetUtil current()
  {
    return NetUtil.NU;
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  protected NetUtil()
  {
    super();
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Silently closes the given object. 
   * If the given object is null, nothing will be done. If an IOException occurs
   * false will be returned.
   * 
   * @param closeable The object to close (may be null).
   * @return true if executed without problem, false if an IOException occurred.
   */
  public boolean close(Closeable closeable)
  {
    if (closeable != null)
    {
      try
      {
        closeable.close();
      }
      catch (@SuppressWarnings("unused") IOException ex)
      {
        return false;
      }
    }
    return true;
  }

  /**
   * Returns the given string character and URL encoded.
   * 
   * @param string The string to be encoded (if null, null will be returned).
   * @param charset The character encoding to be used on the string before doing the URL encoding (must not be null).
   * @return The encoded string or null in case of a completely unexpected UnsupportedEncodingException.
   */
  public String urlEncode(String string, Charset charset)
  {
    if (string == null)
    {
      return null;
    }
    try
    {
      return URLEncoder.encode(string, charset.name());
    }
    catch (UnsupportedEncodingException ex)
    {
      // Should never happen because charset is already checked for supported encoding!
      LoggerProvider.getComponentLogger().logError("[" + LoggerProvider.PFNET0003 + "] org.pfsw.net.NetUtil.urlEncode(String, Charset): ", ex);
      return null;
    }
  }

  /**
   * Returns the given string character and URL decoded.
   * 
   * @param string The string to be decoded (if null, null will be returned).
   * @param charset The character encoding to be used on the string after the URL decoding (must not be null).
   * @return The decoded string or null in case of a completely unexpected UnsupportedEncodingException.
   */
  public String urlDecode(String string, Charset charset)
  {
    if (string == null)
    {
      return null;
    }
    try
    {
      return URLDecoder.decode(string, charset.name());
    }
    catch (UnsupportedEncodingException ex)
    {
      // Should never happen because charset is already checked for supported encoding!
      LoggerProvider.getComponentLogger().logError("[" + LoggerProvider.PFNET0004 + "] org.pfsw.net.NetUtil.urlDecode(String, Charset): ", ex);
      return null;
    }
  }

  /**
   * Returns the given string guaranteed with a leading slash. 
   * @param string The string that should have a leading slash (must not be null).
   */
  public String withLeadingSlash(String string)
  {
    return str().withLeadingChar(string, SLASH);
  }

  /**
   * Returns the given string guaranteed with a trailing slash. 
   * @param string The string that should have a trailing slash (must not be null).
   */
  public String withTrailingSlash(String string)
  {
    return str().withTrailingChar(string, SLASH);
  }

  /**
   * Returns the given string guaranteed with no leading slash. 
   * @param string The string that should not have a leading slash (must not be null).
   */
  public String withoutLeadingSlash(String string)
  {
    return str().withoutLeadingChar(string, SLASH);
  }

  /**
   * Returns the given string guaranteed with no trailing slash. 
   * @param string The string that should not have a trailing slash (must not be null).
   */
  public String withoutTrailingSlash(String string)
  {
    return str().withoutTrailingChar(string, SLASH);
  }

  /**
   * Returns the given string as URI if possible.
   * In contrary to the constructor of URI this method throws a runtime exception.
   * 
   * @param uriString The string to be converted to a URI (must not be null).
   * @return The URI created from the given string.
   * @throws InvalidUrlException if the given string is no valid URI.
   */
  public URI asURI(String uriString)
  {
    try
    {
      return new URI(uriString);
    }
    catch (URISyntaxException e)
    {
      throw new InvalidUrlException(e, "Not a valid URI: %s", uriString);
    }
  }

  /**
   * Returns the given string as URL if possible.
   * In contrary to the constructor of URL this method throws a runtime exception.
   * 
   * @param urlString The string to be converted to a URL (must not be null).
   * @return The URL created from the given string.
   * @throws InvalidUrlException if the given string is no valid URL.
   */
  public URL asURL(String urlString)
  {
    try
    {
      return new URL(urlString);
    }
    catch (MalformedURLException e)
    {
      throw new InvalidUrlException(e, "Not a valid URL: %s", urlString);
    }
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected StringUtil str()
  {
    return StringUtil.current();
  }
}
