// ===========================================================================
// CONTENT  : CLASS UniformResourceLocator
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.1 - 24/05/2020
// HISTORY  :
//  17/11/2010  mdu  CREATED
//	15/11/2011	mdu		added -> public void removeParameter(String name)
//	30/03/2012	mdu		added	-> clearParameters()
//  06/08/2014  mdu   added -> new constructors with the separate parts of a URL
//  29/11/2014  mdu   changed --> added Serializable
//  10/05/2015  mdu   changed --> using NetUtil for with/without slash methods
//  05/05/2017  mdu   changed --> using NamedValueList instead of NamedTextList for parameters
//  24/05/2020  mdu   added   --> create(URI), create(URL), create(String), getParametersMap()
//
// Copyright (c) 2010-2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.net;

import java.io.Serializable;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.Map;

import org.pfsw.bif.exceptions.InvalidUrlException;
import org.pfsw.bif.text.IStringPair;
import org.pfsw.bif.text.IStringRepresentation;
import org.pfsw.net.http.HttpUtil;
import org.pfsw.text.StringUtil;
import org.pfsw.util.CollectionUtil;
import org.pfsw.util.NamedValue;
import org.pfsw.util.NamedValueList;

/**
 * Helper class to parse and manipulate URLs in a convenient way.
 *
 * @author Manfred Duchrow
 * @version 2.1
 */
public class UniformResourceLocator implements Serializable
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final long serialVersionUID = 1432625831785740081L;

  private static final boolean DEBUG = "true".equals(System.getProperty("org.pfsw.net.debug", "false"));

  protected static final StringUtil SU = StringUtil.current();
  protected static final CollectionUtil CU = CollectionUtil.current();
  protected static final NetUtil NU = NetUtil.current();

  public static final String SCHEME_HTTP = HttpUtil.SCHEME_HTTP;
  public static final String SCHEME_HTTPS = HttpUtil.SCHEME_HTTPS;

  public static final int HTTP_DEFAULT_PORT = HttpUtil.HTTP_DEFAULT_PORT;
  public static final int HTTPS_DEFAULT_PORT = HttpUtil.HTTPS_DEFAULT_PORT;

  protected static final int DEFAULT_PORT = 0;
  protected static final String DEFAULT_SCHEME = SCHEME_HTTP;

  /** "://" */
  public static final String SCHEME_HOST_SEPARATOR = HttpUtil.SCHEME_HOST_SEPARATOR;

  /** "http://" */
  public static final String HTTP_SCHEME_PREFIX = HttpUtil.HTTP_PREFIX;
  /** "https://" */
  public static final String HTTPS_SCHEME_PREFIX = HttpUtil.HTTPS_PREFIX;

  /** The character '?' that is used to separate the query parameters from the URL */
  public static final String URL_QUERY_SEPARATOR = "?";
  /** The character '&' that is used to separate consecutive query parameters from each other */
  public static final String URL_PARAM_SEPARATOR = "&";
  /** The character '=' that is used to separate name and value of query parameters */
  public static final String URL_PARAM_ASSIGN = "=";
  /** The character ':' that is used to separate hostname from the port */
  public static final String URL_PORT_SEPARATOR = ":";
  /** The character '/' that is used to separate naming elements of the URL */
  public static final String SLASH = "/";

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private final NamedValueList<List<String>> parameters = new NamedValueList<List<String>>();
  private String serverName = null;
  private String requestURI = null;
  private String scheme = DEFAULT_SCHEME;
  private int port = DEFAULT_PORT;

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Returns the default port for scheme HTTP (80) or HTTPS (443).
   * For any other, unknown scheme it returns -1.
   * 
   * @param scheme The case-insensitive scheme name (must not be null).
   */
  public static int getDefaultPortForScheme(String scheme)
  {
    if (SCHEME_HTTP.equalsIgnoreCase(scheme))
    {
      return HTTP_DEFAULT_PORT;
    }
    if (SCHEME_HTTPS.equalsIgnoreCase(scheme))
    {
      return HTTPS_DEFAULT_PORT;
    }
    return -1;
  }

  /**
   * Creates a new instance.
   */
  public static UniformResourceLocator create()
  {
    return new UniformResourceLocator();
  }

  /**
   * Creates a new instance on the given URL string.
   * 
   * @param url The URL string (must not be null).
   * @return A new UniformResourceLocator based on the given URL string.
   * @throws InvalidUrlException if the given string is not a valid URL.
   */
  public static UniformResourceLocator create(String url)
  {
    try
    {
      return new UniformResourceLocator(url);
    }
    catch (MalformedURLException e)
    {
      throw new InvalidUrlException(e, "The string '%s' is no valid URL.", url);
    }
  }

  /**
   * Creates a new instance on the given URL.
   * 
   * @param url The URL (must not be null).
   * @return A new UniformResourceLocator based on the given URL.
   */
  public static UniformResourceLocator create(URL url)
  {
    return create(url.toString());
  }

  /**
   * Creates a new instance on the given URI.
   * 
   * @param uri The URI (must not be null).
   * @return A new UniformResourceLocator based on the given URI.
   */
  public static UniformResourceLocator create(URI uri)
  {
    return create(uri.toString());
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with the given URL
   */
  public UniformResourceLocator(String url) throws MalformedURLException
  {
    this();
    parseURL(url);
  }

  /**
   * Initialize the new instance with default values.
   */
  public UniformResourceLocator()
  {
    super();
  }

  /**
   * Initialize the new instance with the various parts of a URL.
   */
  public UniformResourceLocator(String scheme, String serverName, int port, String requestURI, String... queryParams)
  {
    this();
    if (SU.isNullOrBlank(scheme))
    {
      throw new IllegalArgumentException("Argument scheme is null or blank!");
    }
    if (SU.isNullOrBlank(serverName))
    {
      throw new IllegalArgumentException("Argument serverName is null or blank!");
    }
    if (port != DEFAULT_PORT)
    {
      if ((port <= 0) || (port > 65535))
      {
        throw new IllegalArgumentException("Argument port is not in the range of valid ports: 1-65535!");
      }
    }
    if (SU.isNullOrBlank(requestURI))
    {
      throw new IllegalArgumentException("Argument requestURI is null or blank!");
    }
    if (!requestURI.startsWith("/"))
    {
      throw new IllegalArgumentException("Argument requestURI must start with a slash: /");
    }
    setScheme(scheme.toLowerCase());
    setServerName(serverName);
    setPort(port);
    setRequestURI(requestURI);
    addAllQueryParameterDefinitions(queryParams);
  }

  /**
   * Initialize the new instance with the various parts of a URL.
   */
  public UniformResourceLocator(String scheme, String serverName, String requestURI, String... queryParams)
  {
    this(scheme, serverName, DEFAULT_PORT, requestURI, queryParams);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns true if this URL contains at least one query parameter
   */
  public boolean hasQueryParameter()
  {
    return getParameters().size() > 0;
  }

  /**
   * Returns true if this URL contains the query parameter with the given name.
   */
  public boolean hasQueryParameter(String name)
  {
    return getParameters().getValue(name) != null;
  }

  /**
   * Returns true if this URL contains the query parameter with the given name.
   */
  public boolean hasQueryParameter(IStringRepresentation name)
  {
    if (name == null)
    {
      return false;
    }
    return hasQueryParameter(name.asString());
  }

  /**
   * Returns the query string, that is everything following the '?'.
   */
  public String getQueryString()
  {
    StringBuffer buffer;
    NamedValueList<List<String>> params;
    NamedValue<List<String>> param;

    if (getParameters().size() == 0)
    {
      return null;
    }

    buffer = new StringBuffer();
    params = getParameters();
    for (int i = 0; i < params.size(); i++)
    {
      param = params.namedValueAt(i);
      if (param.value().isEmpty())
      {
        if (buffer.length() > 0)
        {
          buffer.append(URL_PARAM_SEPARATOR);
        }
        buffer.append(param.name());
      }
      else
      {
        for (String value : param.value())
        {
          if (buffer.length() > 0)
          {
            buffer.append(URL_PARAM_SEPARATOR);
          }
          buffer.append(param.name()).append(URL_PARAM_ASSIGN).append(value);
        }
      }
    }
    return buffer.toString();
  }

  public UniformResourceLocator setQueryString(String query)
  {
    parseQueryString(query);
    return this;
  }

  /**
   * Return the server name
   */
  public String getServerName()
  {
    return serverName();
  }

  public UniformResourceLocator setServerName(String name)
  {
    serverName(name);
    return this;
  }

  /**
   * Return the defined port. If the port is not defined (i.e zero (0))
   * it returns the default port for the current scheme.
   */
  public int getPort()
  {
    if (DEFAULT_PORT == port())
    {
      if (SCHEME_HTTP.equals(getScheme()))
      {
        return HTTP_DEFAULT_PORT;
      }
      if (SCHEME_HTTPS.equals(getScheme()))
      {
        return HTTPS_DEFAULT_PORT;
      }
    }
    return port();
  }

  /**
   * Set the port of the URL
   */
  public UniformResourceLocator setPort(int newValue)
  {
    port(newValue);
    return this;
  }

  /**
   * Returns the request URI
   */
  public String getRequestURI()
  {
    return requestURI();
  }

  public UniformResourceLocator setRequestURI(String uri)
  {
    requestURI(NU.withLeadingSlash(uri));
    return this;
  }

  /**
   * Sets the given filename to the end of the requestURI.
   * If there is already a filename it will be replaced.
   */
  public UniformResourceLocator setFilename(String filename)
  {
    setPathRelative(filename);
    return this;
  }

  /**
   * Sets the given path relative to the current requestURI.
   * Any name after the last slash in the current requestURI will be replaced.
   * Separating slashes will be handled automatically.
   * If the given path is null or blank it will be ignored.
   * 
   * @param path The path to be set relative to the current path (may be null).
   */
  public UniformResourceLocator setPathRelative(String path)
  {
    String uri;

    if (SU.notNullOrBlank(path))
    {
      uri = getRequestURI();
      if (uri.endsWith(SLASH))
      {
        uri += NU.withoutLeadingSlash(path);
      }
      else
      {
        uri = SU.cutTail(uri, SLASH);
        uri = uri + NU.withLeadingSlash(path);
      }
      setRequestURI(uri);
    }
    return this;
  }

  /**
   * Appends the given path to the end of the current requestURI.
   * Nothing will be replaced, just appended. The separating slash
   * will be added or removed automatically as needed.
   * If the given path is null or blank it will be ignored (nothing appended).
   * Any leading or trailing whitespace characters will be removed from the path.
   * 
   * @param path The path to be appended (may be null).
   */
  public UniformResourceLocator appendPath(String path)
  {
    String uri;

    if (SU.notNullOrBlank(path))
    {
      uri = getRequestURI();
      uri = NU.withTrailingSlash(uri) + NU.withoutLeadingSlash(path.trim());
      setRequestURI(uri);
    }
    return this;
  }
  public String getScheme()
  {
    return scheme();
  }

  public UniformResourceLocator setScheme(String scheme)
  {
    scheme(scheme);
    return this;
  }

  /**
   * Returns the names of all query parameters as Enumeration. 
   */
  public Enumeration<String> getParameterNames()
  {
    return CU.asEnumeration(getParameters().names());
  }

  /**
   * Returns the names of all query parameters as List. 
   */
  public List<String> getQueryParameterNames()
  {
    return getParameters().names();
  }

  /**
   * Returns the number of query parameters 
   */
  public int getNumberOfParameters()
  {
    return getParameters().size();
  }

  /**
   * Returns the (first) value of the parameter with the given name or null.
   */
  public String getParameter(IStringRepresentation name)
  {
    if (name == null)
    {
      return null;
    }
    return getParameter(name.asString());
  }

  /**
   * Returns the (first) value of the parameter with the given name or null.
   */
  public String getParameter(String name)
  {
    List<String> paramValues;

    paramValues = getParameterValues(name);
    if (paramValues == null)
    {
      return null;
    }
    if (paramValues.isEmpty())
    {
      return "";
    }
    return paramValues.get(0);
  }

  /**
   * Returns the values of the parameter with the given name or null
   * if the parameter is not present.
   */
  public List<String> getParameterValues(IStringRepresentation name)
  {
    if (name == null)
    {
      return null;
    }
    return getParameterValues(name.asString());
  }

  /**
   * Returns the values of the parameter with the given name or null
   * if the parameter is not present.
   */
  public List<String> getParameterValues(String name)
  {
    NamedValue<List<String>> param;

    param = getParameters().findNamedValue(name);
    if (param == null)
    {
      return null;
    }
    return new ArrayList<String>(param.value());
  }

  /**
   * Returns all query parameters or this URL in a map.
   * The returned map is a copy and therefore changes to it will have no impact on the 
   * parameters of this object.
   * 
   * @return An (perhaps empty) map of all query parameters.
   */
  public Map<String, List<String>> getParametersMap()
  {
    return getParameters().asHashMap();
  }

  /**
   * Sets the parameter with the given name to the specified value.
   * If the parameter already existed it will be replaced.
   */
  public UniformResourceLocator setParameter(IStringRepresentation name, String... values)
  {
    if (name == null)
    {
      return this;
    }
    return setParameter(name.asString(), values);
  }

  /**
   * Sets the parameter with the given name to the specified value.
   * If the parameter already existed it will be replaced.
   */
  public UniformResourceLocator setParameter(IStringRepresentation name, IStringRepresentation... values)
  {
    if (name != null)
    {
      for (IStringRepresentation value : values)
      {
        if (value != null)
        {
          setParameter(name, value.asString());
        }
      }
    }
    return this;
  }

  /**
   * Sets the parameter with the given name to the specified value.
   * If the parameter already existed it will be replaced.
   */
  public UniformResourceLocator setParameter(String name, String... values)
  {
    NamedValue<List<String>> param;

    param = getOrCreateParameter(name);
    param.getValue().clear();
    addValuesToParameter(param, values);
    return this;
  }

  /**
   * Adds the given value to the parameter with the specified name.
   * If the parameter did not yet exist, it will be created.
   * Null values will be skipped.
   */
  public UniformResourceLocator addToParameter(IStringRepresentation name, String... values)
  {
    if (name == null)
    {
      return this;
    }
    return addToParameter(name.asString(), values);
  }

  /**
   * Adds the given value to the parameter with the specified name.
   * If the parameter did not yet exist, it will be created.
   * Null values will be skipped.
   */
  public UniformResourceLocator addToParameter(IStringRepresentation name, IStringRepresentation... values)
  {
    if (name != null)
    {
      for (IStringRepresentation value : values)
      {
        if (value != null)
        {
          addToParameter(name, value.asString());
        }
      }
    }
    return this;
  }

  /**
   * Adds the given value to the parameter with the specified name.
   * If the parameter did not yet exist, it will be created.
   * Null values will be skipped.
   */
  public UniformResourceLocator addToParameter(String name, String... values)
  {
    NamedValue<List<String>> param;

    param = getOrCreateParameter(name);
    addValuesToParameter(param, values);
    return this;
  }

  /**
   * Remove the parameter with the given named from the list of query parameters.
   */
  public UniformResourceLocator removeParameter(IStringRepresentation name)
  {
    if (name == null)
    {
      return this;
    }
    return removeParameter(name.asString());
  }

  /**
   * Remove the parameter with the given named from the list of query parameters.
   */
  public UniformResourceLocator removeParameter(String name)
  {
    getParameters().removeKey(name);
    return this;
  }

  /**
   * Remove all query parameters from this URL.
   */
  public UniformResourceLocator clearParameters()
  {
    getParameters().clear();
    return this;
  }

  /**
   * Returns true, if the port is not explicitly set which means the default 
   * port has to be used.
   */
  public boolean isDefaultPort()
  {
    if (DEFAULT_PORT == port())
    {
      return true;
    }

    if ((SCHEME_HTTP.equals(getScheme())) && (HTTP_DEFAULT_PORT == port()))
    {
      return true;
    }

    if ((SCHEME_HTTPS.equals(getScheme())) && (HTTPS_DEFAULT_PORT == port()))
    {
      return true;
    }

    return false;
  }

  /**
   * Returns a string containing the server name and optionally the port number,
   * separated by a colon (':').
   * The port will NOT be appended to the server name if
   * <ul>
   * <li>the port is 0
   * <li>the scheme is http and the port is 80
   * <li>the scheme is https and the port is 443
   * </ul>
   * In all other cases the port will be appended to the server name.
   */
  public String getServerNameAndPort()
  {
    StringBuffer buffer;

    buffer = new StringBuffer(40);
    appendServerNameAndPort(buffer);
    return buffer.toString();
  }

  /**
   * Creates a copy of this object.
   * Changing the copy has no impact on the original object.
   */
  public UniformResourceLocator copy()
  {
    UniformResourceLocator copy;

    copy = newInstance();
    copy.setScheme(getScheme());
    copy.setServerName(getServerName());
    copy.setPort(getPort());
    copy.setRequestURI(getRequestURI());
    copy.getParameters().addAll(copyParameters());
    return copy;
  }

  /**
   * Returns true if this is a relative URL
   */
  public boolean isRelative()
  {
    return (getScheme() == null) || (getServerName() == null);
  }

  /**
   * Returns true if this is an absolute URL
   */
  public boolean isAbsolute()
  {
    return !isRelative();
  }

  /**
   * Returns the full URL including parameters as string representation
   */
  @Override
  public String toString()
  {
    return asString();
  }

  /**
   * Returns the full URL including parameters as string representation.
   */
  public String asString()
  {
    StringBuffer buffer;

    buffer = new StringBuffer(100);

    if (isAbsolute())
    {
      buffer.append(getScheme());
      buffer.append(SCHEME_HOST_SEPARATOR);
      appendServerNameAndPort(buffer);
    }
    buffer.append(getRequestURI());
    if (getQueryString() != null)
    {
      buffer.append(URL_QUERY_SEPARATOR);
      buffer.append(getQueryString());
    }
    return buffer.toString();
  }

  public URL toURL()
  {
    try
    {
      return new URL(asString());
    }
    catch (MalformedURLException e)
    {
      handleException(e);
      return null;
    }
  }

  public URI toURI()
  {
    try
    {
      return new URI(asString());
    }
    catch (URISyntaxException e)
    {
      handleException(e);
      return null;
    }
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected void parseURL(final String urlStr) throws MalformedURLException
  {
    String[] parts;
    boolean relative;

    checkIfValidURL(urlStr);

    relative = isRelativeToHost(urlStr);
    if (relative)
    {
      parts = SU.splitNameValue(urlStr, URL_QUERY_SEPARATOR);
    }
    else
    {
      parts = SU.splitNameValue(urlStr, SCHEME_HOST_SEPARATOR);
      setScheme(parts[0]);
      parts = SU.splitNameValue(parts[1], URL_QUERY_SEPARATOR);
    }
    if (parts[1].length() > 0)
    {
      setQueryString(parts[1]);
      parseQueryString(parts[1]);
    }
    if (relative)
    {
      setRequestURI(parts[0]);
    }
    else
    {
      parts = SU.splitNameValue(parts[0], SLASH);
      setRequestURI("/" + parts[1]);
      parts = SU.splitNameValue(parts[0], URL_PORT_SEPARATOR);
      setServerName(parts[0]);
      if (parts[1].length() > 0)
      {
        try
        {
          setPort(Integer.parseInt(parts[1]));
        }
        catch (NumberFormatException e)
        {
          handleException(e);
          throw new MalformedURLException("Invalid port definition: " + parts[1]);
        }
      }
    }
  }

  protected void addValuesToParameter(final NamedValue<List<String>> param, String... values)
  {
    for (String string : values)
    {
      if (string != null)
      {
        param.getValue().add(string);
      }
    }
  }

  protected void parseQueryString(String queryString)
  {
    String[] params;

    clearParameters();

    if (queryString == null)
    {
      return;
    }
    params = SU.parts(queryString, URL_PARAM_SEPARATOR);
    addAllQueryParameterDefinitions(params);
  }

  protected void appendServerNameAndPort(StringBuffer buffer)
  {
    buffer.append(getServerName());
    if (!isDefaultPort())
    {
      buffer.append(URL_PORT_SEPARATOR);
      buffer.append(getPort());
    }
  }

  protected NamedValue<List<String>>[] copyParameters()
  {
    List<NamedValue<List<String>>> copiedParameters;
    NamedValue<List<String>> paramCopy;

    copiedParameters = new ArrayList<NamedValue<List<String>>>(getParameters().size());
    for (NamedValue<List<String>> param : getParameters().namedValueArray())
    {
      paramCopy = createParameterObject(param.name());
      paramCopy.value().addAll(param.value());
      copiedParameters.add(paramCopy);
    }
    return CU.toArray(copiedParameters);
  }

  protected boolean isRelativeToHost(String urlStr)
  {
    return urlStr.startsWith(SLASH);
  }

  @SuppressWarnings("unused")
  protected void checkIfValidURL(String urlStr) throws MalformedURLException
  {
    if (isRelativeToHost(urlStr))
    {
      return;
    }
    new URL(urlStr);
    try
    {
      new URI(urlStr);
    }
    catch (URISyntaxException ex)
    {
      throw new MalformedURLException(ex.getMessage());
    }
  }

  protected void addAllQueryParameterDefinitions(final String[] queryParams)
  {
    IStringPair stringPair;
    for (String param : queryParams)
    {
      if (param.contains(URL_PARAM_ASSIGN))
      {
        stringPair = SU.splitStringPair(param, URL_PARAM_ASSIGN);
        addToParameter(stringPair.getString1(), stringPair.getString2());
      }
      else
      {
        setParameter(param);
      }
    }
  }

  protected NamedValue<List<String>> getOrCreateParameter(String name)
  {
    NamedValue<List<String>> param;

    param = getParameters().findNamedValue(name);
    if (param == null)
    {
      param = createParameter(name);
    }
    return param;
  }

  protected NamedValue<List<String>> createParameter(String name)
  {
    NamedValue<List<String>> param;

    param = createParameterObject(name);
    getParameters().add(param);
    return param;
  }

  protected NamedValue<List<String>> createParameterObject(String name)
  {
    return new NamedValue<List<String>>(name, new ArrayList<String>());
  }

  protected UniformResourceLocator newInstance()
  {
    return new UniformResourceLocator();
  }

  protected void handleException(Exception ex)
  {
    if (DEBUG)
    {
      ex.printStackTrace();
    }
  }

  protected NamedValueList<List<String>> getParameters()
  {
    return parameters;
  }

  protected String serverName()
  {
    return serverName;
  }

  protected void serverName(String newValue)
  {
    serverName = newValue;
  }

  protected String requestURI()
  {
    return requestURI;
  }

  protected void requestURI(String newValue)
  {
    requestURI = newValue;
  }

  protected String scheme()
  {
    return scheme;
  }

  protected void scheme(String newValue)
  {
    scheme = newValue;
  }

  protected int port()
  {
    return port;
  }

  protected void port(int newValue)
  {
    port = newValue;
  }
}
