// ===========================================================================
// CONTENT  : CLASS AuthorizationHeader
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 22/05/2015
// HISTORY  :
//  22/05/2015  mdu  CREATED
//
// Copyright (c) 2015, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.net.http;

/**
 * An instance of this class represents the parsed request header "Authorization".
 * So it is quite easy to determine what type (auth-scheme) of authorization is 
 * provided and what credentials are sent within.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class AuthorizationHeader
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  public static final String AUTH_SCHEME_BASIC = "Basic";
  public static final String AUTH_SCHEME_DIGEST = "Digest";
  public static final String AUTH_SCHEME_BEARER = "Bearer";

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private String authScheme;
  private String authData;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  public AuthorizationHeader()
  {
    super();
  }
  
  public AuthorizationHeader(String authScheme, String authData)
  {
    this();
    setAuthScheme(authScheme);
    setAuthData(authData);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  public String getAuthScheme()
  {
    return authScheme;
  }
  
  /**
   * Returns the authentication data that was provided with the authentication scheme.
   * This data is not yet decoded or decrypted in any way. 
   */
  public String getAuthData()
  {
    return authData;
  }
  
  public void setAuthScheme(String authScheme)
  {
    this.authScheme = authScheme;
  }
  
  public void setAuthData(String authData)
  {
    this.authData = authData;
  }

  public boolean isBasicAuthScheme() 
  {
    return isAuthScheme(AUTH_SCHEME_BASIC);
  }

  public boolean isDigestAuthScheme() 
  {
    return isAuthScheme(AUTH_SCHEME_DIGEST);
  }
  
  public boolean isBearerAuthScheme() 
  {
    return isAuthScheme(AUTH_SCHEME_BEARER);
  }
  
  /**
   * Returns true if the underlying authentication scheme equals (case-insensitive) 
   * the given name.
   * 
   * @param name The (auth scheme) name to check against.
   */
  public boolean isAuthScheme(String name) 
  {
    return name.equalsIgnoreCase(getAuthScheme());
  }
  
  /**
   * Returns the auth-scheme together with the auth-data separated by a space.
   * <p>
   * Example: "Basic emda8ghk823fvdcs99sJSst="
   */
  public String asString() 
  {
    return getAuthScheme() + " " + getAuthData();
  }

  /**
   * Returns the full request header name and value.
   * <p>
   * Example: "Authorization: Bearer e8785jAp9dnSNo/Twqpds8a7w4j="
   */
  public String asHeaderString() 
  {
    return RequestHeaderNames.AUTHORIZATION + ": " + asString();
  }
}
