// ===========================================================================
// CONTENT  : CLASS CaseInsensitiveMultiValueMap
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 07/08/2014
// HISTORY  :
//  07/08/2014  mdu  CREATED
//
// Copyright (c) 2014, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.net.http;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.Vector;

import org.pfsw.text.StringUtil;
import org.pfsw.util.CollectionUtil;

/**
 * A map that treats all keys case-insensitive and allows multiple values
 * per key.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
class CaseInsensitiveMultiValueMap
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  protected static final CollectionUtil CU = CollectionUtil.current();
  protected static final StringUtil SU = StringUtil.current();

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private final Map<String, List<String>> data = new TreeMap<String, List<String>>(String.CASE_INSENSITIVE_ORDER);

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  public CaseInsensitiveMultiValueMap()
  {
    super();
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected Map<String, List<String>> getData()
  {
    return data;
  }

  protected List<String> getDataValues(String key)
  {
    return getData().get(key);
  }

  /**
   * Add all values from the given map.
   * 
   * @param map The keys and values to be a added (if null, nothing will be done).
   */
  protected void addAllFrom(Map<String, List<String>> map)
  {
    if (CU.isNullOrEmpty(map))
    {
      return;
    }
    for (Map.Entry<String, List<String>> entry : map.entrySet())
    {
      if ((entry.getKey() != null) && (entry.getValue() != null))
      {
        getData().put(entry.getKey(), entry.getValue());
      }
    }
  }

  /**
   * Adds the given value to the values of the header with the specified name.
   * @param key Name of the header
   * @param value Value to be added
   */
  protected void addKeyValue(String key, String value)
  {
    List<String> values;

    values = getDataValues(key);
    if (values == null)
    {
      values = new Vector<String>();
    }
    values.add(value);
    getData().put(key, values);
  }

  protected void setKeyValue(String name, String value)
  {
    List<String> values;

    values = new ArrayList<String>();
    values.add(value);
    getData().put(name, values);
  }

  protected String getFirstValue(String key)
  {
    List<String> values;

    values = getDataValues(key);
    if (!CU.isNullOrEmpty(values))
    {
      return values.get(0);
    }
    return null;
  }

  protected void removeKey(String key)
  {
    getData().remove(key);
  }

  protected Set<String> getKeys()
  {
    return getData().keySet();
  }

  protected String[] getValuesArrayOf(String key)
  {
    return SU.asStrings(getDataValues(key));
  }

  protected int size()
  {
    return getData().size();
  }

  protected Map<String, String[]> asMap()
  {
    Map<String, String[]> map;

    map = new HashMap<String, String[]>(size());
    for (Map.Entry<String, List<String>> entry : getData().entrySet())
    {
      map.put(entry.getKey(), SU.asStrings(entry.getValue()));
    }
    return Collections.unmodifiableMap(map);
  }
}
