// ===========================================================================
// CONTENT  : CLASS ClientRequest
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.0 - 18/12/2020
// HISTORY  :
//  19/08/2014  mdu  CREATED
//  18/12/2020  mdu   changed -> fluent API support
//
// Copyright (c) 2014-2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.net.http;

import java.net.URL;
import java.util.Map;

import org.pfsw.bif.http.HttpMethod;
import org.pfsw.net.NetUtil;

/**
 * Definition of a client request that can be executed with a UserAgent.
 *
 * @author Manfred Duchrow
 * @version 2.0
 */
public class ClientRequest extends CommunicationData<ClientRequest>
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final NetUtil NU = NetUtil.current();

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private HttpMethod httpMethod = HttpMethod.GET;
  private URL url;

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  public static ClientRequest create()
  {
    return new ClientRequest();
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  public ClientRequest()
  {
    super();
  }
  
  public ClientRequest(URL url)
  {
    this(HttpMethod.GET, url);
  }

  public ClientRequest(HttpMethod method, URL url)
  {
    this();
    setURL(url);
    setHttpMethod(method);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  public ClientRequest copy() 
  {
    //@formatter:off
    return ClientRequest.create()
        .setHttpMethod(getHttpMethod())
        .setURL(getURL())
        .setCharset(getCharset())
        .setContentMediaType(getContentMediaType())
        .setContentDataStream(getContentDataStream())
    ;
    //@formatter:on
  }
  
  public HttpMethod getHttpMethod()
  {
    return httpMethod;
  }

  public ClientRequest setHttpMethod(HttpMethod httpMethod)
  {
    this.httpMethod = httpMethod;
    return self();
  }

  public URL getURL()
  {
    return url;
  }

  public ClientRequest setURL(URL url)
  {
    this.url = url;
    return self();
  }
  
  @Override
  public String toString()
  {
    return String.format("%s('%s')", getClass().getSimpleName(), getURL());
  }

  /**
   * Adds the given media types (e.g. "text/html") to the ACCEPT request header.
   */
  public ClientRequest accept(String... mediaTypes)
  {
    for (String mediaType : mediaTypes)
    {
      addHeader(RequestHeaderNames.ACCEPT, mediaType);
    }
    return self();
  }

  /**
   * Appends the given map as content key/value pairs. All keys and values
   * will automatically be URL encoded using the currently defined character encoding. 
   */
  public ClientRequest appendContent(Map<String, String> parameters)
  {
    checkIfHttpMethodAllowsBody();
    for (Map.Entry<String, String> param : parameters.entrySet())
    {
      if (getContentLength() > 0)
      {
        appendContent("&");
      }
      appendContent(NU.urlEncode(param.getKey(), getCharsetEncoding()));
      appendContent("=");
      appendContent(NU.urlEncode(param.getValue(), getCharsetEncoding()));
    }
    return self();
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected void checkIfHttpMethodAllowsBody()
  {
    if (!getHttpMethod().hasBody())
    {
      throw new IllegalStateException("The current request method " + getHttpMethod().getName() + " does not allow content!");
    }
  }

}
