// ===========================================================================
// CONTENT  : ABSTRACT CLASS CommunicationData
// AUTHOR   : Manfred Duchrow
// VERSION  : 3.0 - 18/12/2020
// HISTORY  :
//  20/08/2014  mdu  CREATED
//  30/12/2016  mdu   changed -> all around content type and charset encoding
//  18/12/2020  mdu   changed -> fluent API support
//
// Copyright (c) 2014-2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.net.http;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringWriter;
import java.io.Writer;
import java.nio.charset.Charset;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.pfsw.bif.IExtensibleFluentApi;
import org.pfsw.bif.http.HttpHeaderName;
import org.pfsw.bif.text.ICharsets;
import org.pfsw.bif.text.IStringPair;
import org.pfsw.logging.Logger;
import org.pfsw.net.MediaType;
import org.pfsw.net.MediaTypes;
import org.pfsw.net.internal.LoggerProvider;
import org.pfsw.text.StringUtil;

/**
 * Common fields and methods for request and response objects in HTTP communication.
 *
 * @author Manfred Duchrow
 * @version 2.0
 */
abstract class CommunicationData<T extends CommunicationData> implements IExtensibleFluentApi<T>
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final StringUtil SU = StringUtil.current();
  private static final String CONTENT_TYPE = RequestHeaderNames.CONTENT_TYPE;
  private static final String CHARSET_SEPARATOR = ";";
  private static final String CHARSET_PARAMETER = "charset";
  private static final Charset DEFAULT_CHARSET = ICharsets.UTF_8;

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private final Headers headers = new Headers();
  private ByteArrayOutputStream contentDataStream = null;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  protected CommunicationData()
  {
    super();
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Sets the "Content-Type" header.
   * If mediaType is null the header will be removed.
   * @param mediaType The media type to set. If null the header will be removed.
   * @param charset The charset to be set. If null it will be ignored.
   */
  public T setContentType(MediaType mediaType, Charset charset)
  {
    setContentTypeHeader(mediaType, charset);
    return self();
  }

  /**
   * Returns the character encoding defined in the Content-Type header
   * or null if it is not set.
   */
  public Charset getCharset()
  {
    return asCharset(getContentTypeHeader());
  }

  /**
   * Sets the character encoding in the Content-Type header
   * or removes it if the given parameter is null.
   * @throws IllegalStateException If no media type is specified yet.
   */
  public T setCharset(Charset charset)
  {
    MediaType mediaType;

    mediaType = getContentMediaType(); // Preserve media type definition
    if ((mediaType == null) && (charset != null))
    {
      throw new IllegalStateException("Setting a charset without a media type is not allowed!");
    }
    setContentTypeHeader(mediaType, charset);
    return self();
  }

  /**
   * Returns the media type specified in the "Content-Type" header
   * of null if not set. 
   */
  public MediaType getContentMediaType()
  {
    return asMediaType(getContentTypeHeader());
  }

  /**
   * Sets the media type in the Content-Type header
   * or removes the entire header if the given parameter is null.
   */
  public T setContentMediaType(MediaType mediaType)
  {
    Charset charset;

    charset = getCharset(); // Preserve charset definition
    setContentTypeHeader(mediaType, charset);
    return self();
  }

  public boolean hasContentData()
  {
    return (getContentDataStream() != null) && (getContentDataStream().size() > 0);
  }

  public int getContentLength()
  {
    if (hasContentData())
    {
      return getContentBuffer().size();
    }
    return 0;
  }

  /**
   * Appends the given bytes to the content. 
   */
  public T appendContent(byte[] data)
  {
    if ((data != null) && (data.length > 0))
    {
      getContentBuffer().write(data, 0, data.length);
    }
    return self();
  }

  /**
   * Appends the bytes from the given stream to the content. 
   * 
   * @param inStream The stream from which to read the bytes (must not be null).
   * @throws IOException If any problem with the input stream occurs.
   */
  public T appendContent(InputStream inStream) throws IOException
  {
    byte[] buffer;
    int numBytes;

    buffer = new byte[getReadBufferSize()];
    numBytes = inStream.read(buffer);
    while (numBytes != -1)
    {
      getContentBuffer().write(buffer, 0, numBytes);
      numBytes = inStream.read(buffer);
    }
    return self();
  }

  /**
   * Appends the given text to the content using the currently defined
   * character encoding.
   * 
   * @param text The text to be appended (null will be ignored).
   * @throws IllegalStateException If no encoding is set.
   */
  public void appendContent(String text)
  {
    byte[] data;

    if (text == null)
    {
      return;
    }
    if (getCharsetEncoding() == null)
    {
      throw new IllegalStateException("Content cannot be set as string because no character encoding is defined yet!");
    }
    data = text.getBytes(getCharsetEncoding());
    appendContent(data);
  }

  /**
   * Returns the content data or null if no content is available.
   */
  public byte[] getContentData()
  {
    if (hasContentData())
    {
      return getContentBuffer().toByteArray();
    }
    return null;
  }

  /**
   * Returns the content data as reader (with correct char encoding) or null 
   * if no content is available.
   */
  public Reader getContentReader()
  {
    ByteArrayInputStream inStream;

    if (!hasContentData())
    {
      return null;
    }
    inStream = new ByteArrayInputStream(getContentData());
    return new BufferedReader(new InputStreamReader(inStream, getCharsetEncoding()));
  }

  /**
   * Returns the content data as text (with correct char encoding) or null 
   * if no content is available.
   */
  public String getContentText()
  {
    Reader reader;
    char[] buffer;
    int numChars;
    Writer result;

    if (!hasContentData())
    {
      return null;
    }
    buffer = new char[getReadBufferSize()];
    result = new StringWriter(getContentLength());
    reader = getContentReader();
    try
    {
      numChars = reader.read(buffer);
      while (numChars != -1)
      {
        result.write(buffer, 0, numChars);
        numChars = reader.read(buffer);
      }
    }
    catch (IOException ex)
    {
      getLogger().logError("[" + LoggerProvider.PFNET0002 + "] Failed to convert content to string.", ex);
    }
    return result.toString();
  }

  /**
   * Add all name/values from the given map.
   * 
   * @param headerMap The header names and values to be a added (if null, nothing will be done).
   */
  public T addAllHeaders(Map<String, List<String>> headerMap)
  {
    getHeaders().addAllHeaders(headerMap);
    return self();
  }

  /**
   * Add all name/values from the given map.
   * 
   * @param headerMap The header names and values to be a added (if null, nothing will be done).
   */
  public T addHeaders(Map<HttpHeaderName, List<String>> headerMap)
  {
    if (headerMap != null)
    {
      for (Entry<HttpHeaderName, List<String>> headerDef : headerMap.entrySet())
      {
        for (String value : headerDef.getValue())
        {
          addHeader(headerDef.getKey(), value);
        }
      }
    }
    return self();
  }

  /**
   * Add a header value to the header with the given name. If the header does
   * not exist yet it will be created.
   */
  public T addHeader(String name, String value)
  {
    getHeaders().addHeader(name, value);
    return self();
  }

  /**
   * Add a header value to the header with the given name. If the header does
   * not exist yet it will be created.
   */
  public T addHeader(HttpHeaderName name, String value)
  {
    return addHeader(name.asString(), value);
  }

  /**
   * A a header with a single value. If the header does not exist yet it will 
   * be created, otherwise its value(s) will be replaced by the given value.
   */
  public T setHeader(String name, String value)
  {
    getHeaders().setHeader(name, value);
    return self();
  }

  /**
   * A a header with a single value. If the header does not exist yet it will 
   * be created, otherwise its value(s) will be replaced by the given value.
   */
  public T setHeader(HttpHeaderName name, String value)
  {
    return setHeader(name.asString(), value);
  }

  /**
   * Returns a single value (the first) of the header with the given name. 
   * If the header does not exist null will be returned.
   */
  public String getHeader(String name)
  {
    return getHeaders().getHeader(name);
  }

  /**
   * Returns a single value (the first) of the header with the given name. 
   * If the header does not exist null will be returned.
   */
  public String getHeader(HttpHeaderName name)
  {
    return getHeader(name.asString());
  }

  /**
   * Returns the header with the given name and all its associated values. 
   */
  public T removeHeader(String name)
  {
    getHeaders().removeHeader(name);
    return self();
  }

  /**
   * Returns the header with the given name and all its associated values. 
   */
  public T removeHeader(HttpHeaderName name)
  {
    return removeHeader(name.asString());
  }

  @SuppressWarnings("unchecked")
  @Override
  public T self()
  {
    return (T)this;
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected MediaType asMediaType(String contentType)
  {
    IStringPair stringPair;

    if (SU.isNullOrBlank(contentType))
    {
      return null;
    }
    stringPair = SU.splitStringPair(contentType, CHARSET_SEPARATOR);
    if (SU.isNullOrBlank(stringPair.getString1()))
    {
      return null;
    }
    return MediaTypes.getMediaType(stringPair.getString1());
  }

  protected Charset asCharset(String contentType)
  {
    IStringPair stringPair;

    if (SU.isNullOrBlank(contentType))
    {
      return null;
    }
    stringPair = SU.splitStringPair(contentType, CHARSET_SEPARATOR);
    if (SU.isNullOrBlank(stringPair.getString2()))
    {
      return null;
    }
    stringPair = SU.splitStringPair(stringPair.getString2(), "=");
    if ((!CHARSET_PARAMETER.equalsIgnoreCase(stringPair.getString1())) && SU.isNullOrBlank(stringPair.getString2()))
    {
      return null;
    }
    return Charset.forName(stringPair.getString2());
  }

  /**
   * Returns the "Content-Type" header unparsed.
   * That is with media type and charset.
   */
  protected String getContentTypeHeader()
  {
    return getHeader(CONTENT_TYPE);
  }

  /**
   * Sets the "Content-Type" header.
   * If mediaType is null the header will be removed.
   */
  protected T setContentTypeHeader(MediaType mediaType, Charset charset)
  {
    String contentType;

    if (mediaType == null)
    {
      removeHeader(CONTENT_TYPE);
    }
    else
    {
      contentType = mediaType.asString();
      if (charset != null)
      {
        contentType = contentType + CHARSET_SEPARATOR + CHARSET_PARAMETER + "=" + charset.toString().toLowerCase();
      }
      setHeader(CONTENT_TYPE, contentType);
    }
    return self();
  }

  protected Charset getCharsetEncoding()
  {
    Charset charset;

    charset = getCharset();
    return (charset == null) ? getDefaultCharset() : charset;
  }

  protected Charset getDefaultCharset()
  {
    return DEFAULT_CHARSET;
  }

  protected ByteArrayOutputStream getContentBuffer()
  {
    if (getContentDataStream() == null)
    {
      setContentDataStream(new ByteArrayOutputStream());
    }
    return getContentDataStream();
  }

  protected ByteArrayOutputStream getContentDataStream()
  {
    return this.contentDataStream;
  }

  protected T setContentDataStream(ByteArrayOutputStream bodyData)
  {
    this.contentDataStream = bodyData;
    return self();
  }

  protected Headers getHeaders()
  {
    return headers;
  }

  protected int getReadBufferSize()
  {
    return 4096;
  }

  protected Logger getLogger()
  {
    return LoggerProvider.getComponentLogger();
  }

}
