// ===========================================================================
// CONTENT  : CLASS HttpUtil
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.0 - 18/12/2020
// HISTORY  :
//  19/08/2014  mdu  CREATED
//  22/05/2015  mdu   added   -> parseAuthorizationHeaderValue()
//  10/04/2016  mdu   added   -> download() methods
//  18/12/2020  mdu   added   -> download() methods with userAgent parameter
//
// Copyright (c) 2014-2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.net.http;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.Proxy;
import java.net.URL;
import java.nio.charset.Charset;
import java.util.concurrent.atomic.AtomicLong;

import org.pfsw.bif.callback.IProgressListener;
import org.pfsw.bif.text.ICharsets;
import org.pfsw.bif.text.IStringPair;
import org.pfsw.text.StringUtil;
import org.pfsw.util.IOUtil;

/**
 * Utility and convenience methods for HTTP programming, provided as singleton.
 * The class has a protected default constructor that allows to extend it and
 * potentially override methods.
 *
 * @author Manfred Duchrow
 * @version 2.0
 */
public class HttpUtil
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  public static final HttpUtil HU = new HttpUtil();

  public static final Charset HTTP_DEFAULT_CHAR_ENCODUNG = ICharsets.ISO_8859_1;

  public static final String SCHEME_HOST_SEPARATOR = "://";

  public static final String SCHEME_HTTP = "http";
  public static final String SCHEME_HTTPS = "https";

  public static final String HTTP_PREFIX = SCHEME_HTTP + SCHEME_HOST_SEPARATOR;
  public static final String HTTPS_PREFIX = SCHEME_HTTPS + SCHEME_HOST_SEPARATOR;

  public static final int HTTP_DEFAULT_PORT = 80;
  public static final int HTTPS_DEFAULT_PORT = 443;

  public static final int DEFAULT_BUFFER_SIZE = 16364;

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Returns the only instance this class supports (design pattern "Singleton")
   */
  public static HttpUtil current()
  {
    return HttpUtil.HU;
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  protected HttpUtil()
  {
    super();
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns whether or not the given string starts with a proper HTTP protocol 
   * prefix.
   * It does not verify if the given string is a valid URL.
   */
  public boolean isHttpURL(String url)
  {
    String urlStr = url.toLowerCase(); // Actually HTTP:// is also valid!
    return urlStr.startsWith(HTTP_PREFIX);
  }

  /**
   * Returns whether or not the given string starts with a proper HTTPS protocol 
   * prefix.
   * It does not verify if the given string is a valid URL.
   */
  public boolean isHttpsURL(String url)
  {
    String urlStr = url.toLowerCase(); // Actually HTTPS:// is also valid!
    return urlStr.startsWith(HTTPS_PREFIX);
  }

  /**
   * Parses the given header value according to the rules of an Authorization header.
   * That is, an authentication scheme followed by a space and some credential data is expected. 
   * 
   * @param headerValue The header value to be parsed (may be null).
   * @return Returns the parsed header or null in any case of parsing problem.
   */
  public AuthorizationHeader parseAuthorizationHeaderValue(String headerValue)
  {
    IStringPair stringPair;

    if (stringUtil().isNullOrBlank(headerValue))
    {
      return null;
    }
    stringPair = stringUtil().splitStringPair(headerValue.trim(), " ");
    if (stringUtil().isAnyNullOrBlank(stringPair.getString1(), stringPair.getString2()))
    {
      return null;
    }
    return new AuthorizationHeader(stringPair.getString1().trim(), stringPair.getString2().trim());
  }

  /**
   * Downloads (via GET request) the content of the specified URL and stores it into the given file.
   * Uses a default buffer size of 16kB.
   * 
   * @param url A valid HTTP or HTTPS URL that can be used to execute a GET request.
   * @param file The file into which the downloaded content should be stored.
   * @return The number of downloaded bytes.
   * @throws IOException If any reading of writing fails.
   */
  public long download(URL url, File file) throws IOException
  {
    // Specifying null ensures that global system properties are respected.
    return download(url, file, (Proxy)null);
  }

  /**
   * Downloads (via GET request) the content of the specified URL and stores it into the given file.
   * Uses a default buffer size of 16kB.
   * 
   * @param userAgent The user agent to do act as client for the download.
   * @param url A valid HTTP or HTTPS URL that can be used to execute a GET request.
   * @param file The file into which the downloaded content should be stored.
   * @return The number of downloaded bytes.
   * @throws IOException If any reading of writing fails.
   */
  public long download(UserAgent userAgent, URL url, File file) throws IOException
  {
    return download(userAgent, url, file, IOUtil.IGNORING_PROGRESS_LISTENER);
  }
  
  /**
   * Downloads (via GET request) the content of the specified URL and stores it into the given file.
   * Uses a default buffer size of 16kB.
   * 
   * @param url A valid HTTP or HTTPS URL that can be used to execute a GET request.
   * @param file The file into which the downloaded content should be stored.
   * @param listener An optional progress listener to display progress or cancel the download.
   * @return The number of downloaded bytes.
   * @throws IOException If any reading of writing fails.
   */
  public long download(URL url, File file, IProgressListener listener) throws IOException
  {
    return download(url, file, (Proxy)null, DEFAULT_BUFFER_SIZE, listener);
  }

  /**
   * Downloads (via GET request) the content of the specified URL and stores it into the given file.
   * 
   * @param url A valid HTTP or HTTPS URL that can be used to execute a GET request.
   * @param file The file into which the downloaded content should be stored.
   * @return The number of downloaded bytes.
   * @throws IOException If any reading of writing fails.
   */
  public long download(URL url, File file, Proxy proxy) throws IOException
  {
    return download(url, file, proxy, DEFAULT_BUFFER_SIZE, IOUtil.IGNORING_PROGRESS_LISTENER);
  }

  /**
   * Downloads (via GET request) the content of the specified URL and stores it into the given file.
   * Uses a default buffer size of 16kB.
   * 
   * @param url A valid HTTP or HTTPS URL that can be used to execute a GET request.
   * @param file The file into which the downloaded content should be stored.
   * @param listener An optional progress listener to display progress or cancel the download.
   * @return The number of downloaded bytes.
   * @throws IOException If any reading of writing fails.
   */
  public long download(URL url, File file, Proxy proxy, IProgressListener listener) throws IOException
  {
    return download(url, file, proxy, DEFAULT_BUFFER_SIZE, listener);
  }

  /**
   * Downloads (via GET request) the content of the specified URL and stores it into the given file.
   * Uses a default buffer size of 16kB.
   * 
   * @param userAgent The user agent to do act as client for the download.
   * @param url A valid HTTP or HTTPS URL that can be used to execute a GET request.
   * @param file The file into which the downloaded content should be stored.
   * @param listener An optional progress listener to display progress or cancel the download.
   * @return The number of downloaded bytes.
   * @throws IOException If any reading of writing fails.
   */
  public long download(UserAgent userAgent, URL url, File file, IProgressListener listener) throws IOException
  {
    return download(userAgent, url, file, DEFAULT_BUFFER_SIZE, listener);
  }
  
  /**
   * Downloads (via GET request) the content of the specified URL and stores it into the given file.
   * 
   * @param url A valid HTTP or HTTPS URL that can be used to execute a GET request.
   * @param file The file into which the downloaded content should be stored.
   * @param bufferSize The size of the buffer (usually larger buffer - better performance).
   * @param listener An optional progress listener to display progress or cancel the download.
   * @return The number of downloaded bytes.
   * @throws IOException If any reading of writing fails.
   */
  public long download(URL url, File file, Proxy proxy, int bufferSize, IProgressListener listener) throws IOException
  {
    OutputStream outputStream;

    outputStream = new FileOutputStream(file);
    return download(url, outputStream, proxy, bufferSize, listener);
  }

  /**
   * Downloads (via GET request) the content of the specified URL and stores it into the given file.
   * 
   * @param userAgent The user agent to do act as client for the download.
   * @param url A valid HTTP or HTTPS URL that can be used to execute a GET request.
   * @param file The file into which the downloaded content should be stored.
   * @param bufferSize The size of the buffer (usually larger buffer - better performance).
   * @param listener An optional progress listener to display progress or cancel the download.
   * @return The number of downloaded bytes.
   * @throws IOException If any reading of writing fails.
   */
  public long download(UserAgent userAgent, URL url, File file, int bufferSize, IProgressListener listener) throws IOException
  {
    ClientRequest request;
    OutputStream outputStream;
    
    request = new ClientRequest(url);
    outputStream = new FileOutputStream(file);
    return download(userAgent, request, outputStream, bufferSize, listener);
  }
  
  /**
   * Downloads (via GET request) the content of the specified URL and stores it into the given stream.
   * 
   * @param url A valid HTTP or HTTPS URL that can be used to execute a GET request.
   * @param outputStream The stream into which the downloaded content should be stored.
   * @param bufferSize The size of the buffer (usually larger buffer - better performance).
   * @param listener An optional progress listener to display progress or cancel the download.
   * @return The number of downloaded bytes.
   * @throws IOException If any reading of writing fails.
   */
  public long download(URL url, OutputStream outputStream, Proxy proxy, int bufferSize, IProgressListener listener) throws IOException
  {
    ClientRequest request;

    request = new ClientRequest(url);
    return download(request, outputStream, proxy, bufferSize, listener);
  }

  /**
   * Downloads the content of the specified URL and stores it into the given stream.
   * 
   * @param request A valid HTTP or HTTPS request that can be used to execute the download.
   * @param outputStream The stream into which the downloaded content should be stored.
   * @param bufferSize The size of the buffer (usually larger buffer - better performance).
   * @param listener An optional progress listener to display progress or cancel the download.
   * @return The number of downloaded bytes.
   * @throws IOException If any reading of writing fails.
   */
  public long download(final ClientRequest request, final OutputStream outputStream, final Proxy proxy, final int bufferSize, final IProgressListener listener) throws IOException
  {
    UserAgent userAgent = UserAgent.create().setProxy(proxy);
    return download(userAgent, request, outputStream, bufferSize, listener);
  }

  /**
   * Downloads the content of the specified URL and stores it into the given stream.
   * 
   * @param userAgent The user agent to do act as client for the download.
   * @param request A valid HTTP or HTTPS request that can be used to execute the download.
   * @param outputStream The stream into which the downloaded content should be stored.
   * @param bufferSize The size of the buffer (usually larger buffer - better performance).
   * @param listener An optional progress listener to display progress or cancel the download.
   * @return The number of downloaded bytes.
   * @throws IOException If any reading of writing fails.
   */
  public long download(final UserAgent userAgent, final ClientRequest request, final OutputStream outputStream, final int bufferSize, final IProgressListener listener) throws IOException
  {
    IResponseStreamHandler responseStreamHandler;
    final AtomicLong copiedBytes = new AtomicLong(0L);
    
    responseStreamHandler = new IResponseStreamHandler()
    {
      @Override
      public void handleStream(InputStream inputStream) throws IOException
      {
        try
        {
          copiedBytes.set(ioUtil().copyStream(inputStream, outputStream, bufferSize, listener));
        }
        finally
        {
          ioUtil().close(inputStream);
          ioUtil().close(outputStream);
        }
      }
    };
    try
    {
      userAgent.sendRequest(request, responseStreamHandler);      
    }
    finally
    {
      ioUtil().close(outputStream);
    }
    return copiedBytes.get();
  }
  
  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected StringUtil stringUtil()
  {
    return StringUtil.current();
  }

  protected IOUtil ioUtil()
  {
    return IOUtil.current();
  }

}
