// ===========================================================================
// CONTENT  : CLASS ProxyBuilder
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 18/12/2020
// HISTORY  :
//  18/12/2020  mdu  CREATED
//
// Copyright (c) 2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.net.http;

import java.net.InetSocketAddress;
import java.net.Proxy;
import java.net.Proxy.Type;

import org.pfsw.bif.configuration.ConfigRuntimeException;
import org.pfsw.bif.exceptions.InvalidUrlException;
import org.pfsw.net.UniformResourceLocator;
import org.pfsw.text.StringUtil;

/**
 * A builder to create {@link Proxy} instances.
 * 
 * @author M.Duchrow
 */
public class ProxyBuilder
{
  private static final StringUtil SU = StringUtil.current();
  private static final HttpUtil HU = HttpUtil.current();

  public static final String DEFAULT_HTTP_PROXY_ENV_VAR = "HTTP_PROXY";

  private Type type;
  private InetSocketAddress address;

  /**
   * Returns a new ProxyBuilder instance.
   */
  public static ProxyBuilder create()
  {
    return new ProxyBuilder();
  }

  /**
   * Returns a new ProxyBuilder instance initialized with the given hostname and port.
   * Defaults to type HTTP.
   */
  public static ProxyBuilder create(String hostname, int port)
  {
    return create().address(new InetSocketAddress(hostname, port)).type(Type.HTTP);
  }

  /**
   * Returns a ProxyBuilder initialized from environment variable "HTTP_PROXY".
   * 
   * @return The newly created builder.
   * @throws ConfigRuntimeException If the specified environment variable cannot be found.
   * @throws InvalidUrlException If the value of the specified environment variable is no valid URL.
   */
  public static ProxyBuilder fromDefaultEnvVar()
  {
    return fromEnvVar(DEFAULT_HTTP_PROXY_ENV_VAR);
  }

  /**
   * Returns a ProxyBuilder initialized from the given environment variable.
   * 
   * @param envVarName The name of the environment variable that specifies the proxy URL (must not be null).
   * @return The newly created builder.
   * @throws ConfigRuntimeException If the specified environment variable cannot be found.
   * @throws InvalidUrlException If the value of the specified environment variable is no valid URL.
   */
  public static ProxyBuilder fromEnvVar(String envVarName)
  {
    String proxyUrl = System.getenv(envVarName);

    if (SU.isNullOrBlank(proxyUrl))
    {
      throw new ConfigRuntimeException("Environment variable '%s' not found!", envVarName);
    }
    return from(proxyUrl);
  }

  /**
   * Returns a ProxyBuilder initialized from the given proxy server URL.
   * 
   * @param proxyUrl The proxy URL with scheme, host and port (must not be null).
   * @return The newly created builder.
   * @throws InvalidUrlException If the value of the specified environment variable is no valid URL.
   */
  public static ProxyBuilder from(String proxyUrl)
  {
    UniformResourceLocator url;

    if (!(HU.isHttpURL(proxyUrl) || HU.isHttpsURL(proxyUrl)))
    {
      throw new InvalidUrlException("Specified protocol '%s' not supported!", proxyUrl);
    }
    url = UniformResourceLocator.create(proxyUrl);
    return create(url.getServerName(), url.getPort());
  }

  protected ProxyBuilder()
  {
    super();
  }

  public ProxyBuilder type(Type proxyType)
  {
    setType(proxyType);
    return this;
  }

  public ProxyBuilder address(InetSocketAddress inetSocketAddress)
  {
    setAddress(inetSocketAddress);
    return this;
  }

  public Proxy build()
  {
    return new Proxy(getType(), getAddress());
  }

  protected Type getType()
  {
    return this.type;
  }

  protected void setType(Type type)
  {
    this.type = type;
  }

  protected InetSocketAddress getAddress()
  {
    return this.address;
  }

  protected void setAddress(InetSocketAddress address)
  {
    this.address = address;
  }
}
