// ===========================================================================
// CONTENT  : CLASS ServerResponse
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.0 - 18/12/2020
// HISTORY  :
//  20/08/2014  mdu  CREATED
//  18/12/2020  mdu   added   --> create(), fluent API support
//
// Copyright (c) 2014-2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.net.http;

import java.net.HttpURLConnection;
import java.nio.charset.Charset;
import java.nio.charset.UnsupportedCharsetException;

import org.pfsw.bif.http.StandardHttpHeaderName;
import org.pfsw.net.internal.LoggerProvider;

/**
 * Holds all data received from a server as response to a single HTTP request.
 *
 * @author Manfred Duchrow
 * @version 2.0
 */
public class ServerResponse extends CommunicationData<ServerResponse>
{
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private final int responseCode;
  private String responseMessage;

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  public static ServerResponse create(int responseCode)
  {
    return new ServerResponse(responseCode);
  }

  public static ServerResponse create(int responseCode, String responseMessage)
  {
    return new ServerResponse(responseCode, responseMessage);
  }
  
  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  public ServerResponse(int responseCode)
  {
    super();
    this.responseCode = responseCode;
  }

  public ServerResponse(int responseCode, String responseMessage)
  {
    this(responseCode);
    setResponseMessage(responseMessage);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  public int getResponseCode()
  {
    return responseCode;
  }

  public String getResponseMessage()
  {
    return responseMessage;
  }

  public ServerResponse setResponseMessage(String responseMessage)
  {
    this.responseMessage = responseMessage;
    return self();
  }

  /**
   * Sets the character encoding to be used on the received content when it gets 
   * converted to text.
   *  
   * @param charEncoding The encoding name (e.g. "UTF-8"). If null the default HTTP encoding will be set.
   */
  public ServerResponse setCharEncoding(String charEncoding)
  {
    Charset enc;

    if (charEncoding == null)
    {
      enc = HttpUtil.HTTP_DEFAULT_CHAR_ENCODUNG;
    }
    else
    {
      try
      {
        enc = Charset.forName(charEncoding);
      }
      catch (@SuppressWarnings("unused") UnsupportedCharsetException ex)
      {
        getLogger().logInfo("[{0}] Received unsupported character encoding {1}. Using default {2} instead.", LoggerProvider.PFNET0001, charEncoding, HttpUtil.HTTP_DEFAULT_CHAR_ENCODUNG);
        enc = HttpUtil.HTTP_DEFAULT_CHAR_ENCODUNG;
      }
    }
    setCharset(enc);
    return self();
  }

  /**
   * Returns true if the response has returned with the OK status code (i.e. 200).
   */
  public boolean isOk()
  {
    return isCodeIn(HttpURLConnection.HTTP_OK);
  }
  
  /**
   * Returns true if the response code is one of the redirect codes 301, 302, 303, 307, 308. 
   */
  public boolean isRedirect() 
  {
    return isCodeIn(301, 302, 303, 307, 308);
  }

  /**
   * Returns true if the response has a status code that equals
   * one of the given codes.
   */
  public boolean isCodeIn(int... codes)
  {
    for (int code : codes)
    {
      if (getResponseCode() == code)
      {
        return true;
      }
    }
    return false;
  }

  /**
   * Returns true if the response has a status code that is in the specified
   * code range (fromCode and toCode both included).
   */
  public boolean isCodeInRange(int fromCode, int toCode)
  {
    if ((getResponseCode() >= fromCode) && (getResponseCode() <= toCode))
    {
      return true;
    }
    return false;
  }
  
  /**
   * Returns the value of the "Location" header or null if it is not present.
   */
  public String getLocationHeader() 
  {
    return getHeader(StandardHttpHeaderName.LOCATION);
  }
  
  @Override
  public String toString()
  {
    return String.format("%s(%d, '%s')", getClass().getSimpleName(), getResponseCode(), getResponseMessage());
  }
}
