// ===========================================================================
// CONTENT  : CLASS UserAgent
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.0 - 18/12/2020
// HISTORY  :
//  19/08/2014  mdu  CREATED
//  28/02/2016  mdu   changed --> Support chunked encoding
//  10/04/2016  mdu   added   --> Proxy support
//  18/12/2020  mdu   added   --> follow redirects, fluent API
//
// Copyright (c) 2014-2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.net.http;

import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.Proxy;
import java.net.URL;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.HttpsURLConnection;

import org.pfsw.bif.http.HttpMethod;
import org.pfsw.logging.Logger;
import org.pfsw.logging.Logger2;
import org.pfsw.logging.LoggerFactoryProvider;
import org.pfsw.net.internal.LoggerProvider;
import org.pfsw.text.StringUtil;

/**
 * A class that handles communication via HTTP for a client. That is, it
 * executes requests and receives the corresponding responses.
 *
 * @author Manfred Duchrow
 * @version 2.0
 */
public class UserAgent
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final Logger2 LOG = LoggerFactoryProvider.getLogger2(UserAgent.class);
  private static final StringUtil SU = StringUtil.current();

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  // TODO Implement a CookieManager and set it here to a variable

  private final Headers headers = new Headers();
  private int readTimeout = 10000; // 10 seconds
  private int connectTimeout = 5000; // 5 seconds
  private String userAgentName = "PF-Net/1.0";
  private Proxy proxy = null;
  private boolean followRedirects = false;
  private HostnameVerifier hostnameVerifier = null;

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Creates a new instance of a UserAgent with no proxy and not following redirects.
   */
  public static UserAgent create()
  {
    return new UserAgent();
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  public UserAgent()
  {
    this(null);
  }

  public UserAgent(Proxy proxy)
  {
    super();
    setProxy(proxy);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the user agent's proxy. If null is provided, it means that
   * the system property settings regarding HTTP proxy are respected.
   */
  public Proxy getProxy()
  {
    return proxy;
  }

  /**
   * Sets the user agent's proxy. If null is provided, it means that
   * the system property settings regarding HTTP proxy are respected.
   */
  public UserAgent setProxy(Proxy proxy)
  {
    this.proxy = proxy;
    return self();
  }

  /**
   * Add a header value to the header with the given name. If the header does
   * not exist yet it will be created.
   */
  public UserAgent addHeader(String name, String value)
  {
    getHeaders().addHeader(name, value);
    return self();
  }

  /**
   * A a header with a single value. If the header does not exist yet it will 
   * be created, otherwise its value(s) will be replaced by the given value.
   */
  public UserAgent setHeader(String name, String value)
  {
    getHeaders().setHeader(name, value);
    return self();
  }

  /**
   * Returns a single value (the first) of the header with the given name. 
   * If the header does not exist null will be returned.
   */
  public String getHeader(String name)
  {
    return getHeaders().getHeader(name);
  }

  /**
   * Returns the header with the given name and all its associated values. 
   */
  public UserAgent removeHeader(String name)
  {
    getHeaders().removeHeader(name);
    return self();
  }

  /**
   * Sends a request based on the given request definition and the user agent's
   * own state.
   * 
   * @param request The definition of the request to be sent (must not be null).
   * @return A response object containing all data received from the server.
   * @throws IOException Any problem during remote communication.
   */
  public ServerResponse sendRequest(ClientRequest request) throws IOException
  {
    return sendRequest(request, null);
  }

  /**
   * Sends a request based on the given request definition and the user agent's
   * own state.
   * 
   * @param request The definition of the request to be sent (must not be null).
   * @param responseStreamHandler An optional (may be null) handler for the response stream.
   * If provided the ServerResponse methods hasContentData(), getContentData(), getConentText(), 
   * getContentReader() must not be used anymore!
   * @return A response object containing the data received from the server except maybe the
   * content body if that was handled already by the given responseStreamHandler.
   * @throws IOException Any problem during remote communication.
   */
  public ServerResponse sendRequest(ClientRequest request, IResponseStreamHandler responseStreamHandler) throws IOException
  {
    ServerResponse response;

    response = executeRequest(request, responseStreamHandler);
    if (response.isRedirect() && isFollowRedirects())
    {
      response = handleRedirectResponse(request, responseStreamHandler, response);
    }
    return response;
  }

  public int getReadTimeout()
  {
    return readTimeout;
  }

  public UserAgent setReadTimeout(int readTimeout)
  {
    this.readTimeout = readTimeout;
    return self();
  }

  public int getConnectTimeout()
  {
    return connectTimeout;
  }

  public UserAgent setConnectTimeout(int connectTimeout)
  {
    this.connectTimeout = connectTimeout;
    return self();
  }

  public String getUserAgentName()
  {
    return userAgentName;
  }

  public UserAgent setUserAgentName(String userAgentName)
  {
    this.userAgentName = userAgentName;
    return self();
  }

  public boolean isFollowRedirects()
  {
    return this.followRedirects;
  }

  public UserAgent setFollowRedirects(boolean followRedirects)
  {
    this.followRedirects = followRedirects;
    return self();
  }

  public HostnameVerifier getHostnameVerifier()
  {
    return this.hostnameVerifier;
  }

  public UserAgent setHostnameVerifier(HostnameVerifier hostnameVerifier)
  {
    this.hostnameVerifier = hostnameVerifier;
    return self();
  }

  public boolean hasProxy()
  {
    return getProxy() != null;
  }

  public UserAgent self()
  {
    return this;
  }

  @Override
  public String toString()
  {
    return String.format("%s()", getClass().getSimpleName());
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected HttpURLConnection openHttpConnection(URL url) throws IOException
  {
    if (hasProxy())
    {
      return (HttpURLConnection)url.openConnection(getProxy());
    }
    return (HttpURLConnection)url.openConnection();
  }

  protected void sendRequestBody(HttpURLConnection connection, ClientRequest request) throws IOException
  {
    DataOutputStream outStream;

    outStream = new DataOutputStream(connection.getOutputStream());
    try
    {
      outStream.write(request.getContentData());
      outStream.flush();
    }
    finally
    {
      outStream.close();
    }
  }

  protected void applyUserAgent(HttpURLConnection connection)
  {
    String agentName;
    agentName = connection.getRequestProperty(RequestHeaderNames.USER_AGENT) + " " + getUserAgentName();
    connection.addRequestProperty(RequestHeaderNames.USER_AGENT, agentName);
  }

  protected void applyRequestHeaders(HttpURLConnection connection, Headers headerDef)
  {
    String[] headerValues;

    for (String headerName : headerDef.getHeaderNames())
    {
      headerValues = headerDef.getHeaderValues(headerName);
      if (SU.isNullOrEmpty(headerValues))
      {
        getLogger().logDebug("Skipped empty request header: {0}", headerName);
      }
      else
      {
        // TODO Check what separator is to be used!!!
        connection.setRequestProperty(headerName, SU.asString(headerValues, " "));
      }
    }
  }

  /**
   * Sends a request based on the given request definition and the user agent's
   * own state.
   * 
   * @param request The definition of the request to be sent (must not be null).
   * @param responseStreamHandler An optional (may be null) handler for the response stream.
   * If provided the ServerResponse methods hasContentData(), getContentData(), getConentText(), 
   * getContentReader() must not be used anymore!
   * @return A response object containing the data received from the server except maybe the
   * content body if that was handled already by the given responseStreamHandler.
   * @throws IOException Any problem during remote communication.
   */
  protected ServerResponse executeRequest(ClientRequest request, IResponseStreamHandler responseStreamHandler) throws IOException
  {
    HttpURLConnection connection;
    ServerResponse response;

    connection = openHttpConnection(request.getURL());
    try
    {
      connection.setRequestMethod(request.getHttpMethod().getName());
      connection.setReadTimeout(getReadTimeout());
      connection.setConnectTimeout(getConnectTimeout());
      connection.setDoInput(true);
      applyUserAgent(connection);
      applyRequestHeaders(connection, getHeaders());
      applyRequestHeaders(connection, request.getHeaders());
      adjustForSSL(connection);
      if (request.hasContentData())
      {
        connection.setDoOutput(true);
        connection.setRequestProperty(RequestHeaderNames.CONTENT_LENGTH, Integer.toString(request.getContentLength()));
        sendRequestBody(connection, request);
      }
      response = retrieveResponse(connection, responseStreamHandler);
      LOG.debugf("Received %d for execution of %s %s ", response.getResponseCode(), request.getHttpMethod().asString(), request.getURL());
    }
    finally
    {
      connection.disconnect();
    }
    return response;
  }

  protected ServerResponse retrieveResponse(final HttpURLConnection connection, final IResponseStreamHandler responseStreamHandler) throws IOException
  {
    InputStream inStream;
    final ServerResponse response;
    int responseCode;
    IResponseStreamHandler inStreamHandler;

    //    responseCode = HttpUtil.current().getResponseCode(connection);
    responseCode = connection.getResponseCode();
    response = new ServerResponse(responseCode, connection.getResponseMessage());
    response.addAllHeaders(connection.getHeaderFields());

    if (isSupposedToHaveContent(response))
    {
      inStream = (responseCode < 400) ? connection.getInputStream() : connection.getErrorStream();
      if (inStream != null)
      {
        inStreamHandler = responseStreamHandler;
        if (inStreamHandler == null)
        {
          inStreamHandler = new IResponseStreamHandler()
          {
            @Override
            public void handleStream(InputStream stream) throws IOException
            {
              response.appendContent(stream);
            }
          };
        }
        inStreamHandler.handleStream(inStream);
      }
    }
    return response;
  }

  protected ServerResponse handleRedirectResponse(ClientRequest request, IResponseStreamHandler responseStreamHandler, ServerResponse response)
    throws IOException
  {
    String location;
    URL redirectTarget;
    ClientRequest redirectRequest;

    if (request.getHttpMethod() == HttpMethod.HEAD)
    {
      // HEAD requests are never followed!
      return response;
    }
    location = response.getLocationHeader();
    if (SU.isNullOrBlank(location))
    {
      // Without new location no redirect possible!
      return response;
    }
    redirectTarget = new URL(location);
    redirectRequest = request.copy().setURL(redirectTarget);
    if (response.isCodeIn(301, 302, 303))
    {
      redirectRequest.setHttpMethod(HttpMethod.GET);
    }
    LOG.debugf("Follow redirect %s response: %s %s", response.getResponseCode(), redirectRequest.getHttpMethod().asString(), redirectRequest.getURL());
    return sendRequest(redirectRequest, responseStreamHandler);
  }

  protected boolean isSupposedToHaveContent(ServerResponse response)
  {
    int contentLength;
    boolean isChunked;

    contentLength = SU.asInteger(response.getHeader(ResponseHeaderNames.CONTENT_LENGTH), 0);
    isChunked = "chunked".equalsIgnoreCase(response.getHeader(ResponseHeaderNames.TRANSFER_ENCODING));

    return (canHaveContent(response.getResponseCode()) && ((contentLength > 0) || isChunked));
  }

  protected boolean canHaveContent(int responseCode)
  {
    switch (responseCode)
    {
      case 204 :
      case 205 :
      case 300 :
      case 301 :
      case 302 :
      case 303 :
      case 304 :
      case 305 :
      case 306 :
      case 307 :
      case 308 :
      case 401 :
        return false;
      default :
        return true;
    }
  }
  
  protected void adjustForSSL(HttpURLConnection connection)
  {
    HttpsURLConnection httpsConnection;

    if (!(connection instanceof HttpsURLConnection))
    {
      return;
    }
    httpsConnection = (HttpsURLConnection)connection;
    if (getHostnameVerifier() != null)
    {      
      httpsConnection.setHostnameVerifier(getHostnameVerifier());
    }
  }

  protected Headers getHeaders()
  {
    return headers;
  }

  protected Logger getLogger()
  {
    return LoggerProvider.getComponentLogger();
  }
}
