// ===========================================================================
// CONTENT  : INTERFACE XMLTagInterpreterFactory
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.2 - 04/09/2009
// HISTORY  :
//  10/07/1999 	duma  CREATED
//	28/02/2002	duma	changed	-> Support for SAX 2.0
//	29/06/2002	duma	bugfix	-> After end tag, this controller must become content handler again
//	04/09/2009	mdu		added		-> Support for LexicalHandler
//
// Copyright (c) 1999-2009, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.pax;

import java.util.Stack;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;

/**
 * This defines the API for classes that can create tag interpreter
 * in instances for specific tag names in a defined tag set.
 *
 * @author Manfred Duchrow
 * @version 2.2
 */
public abstract class BaseXMLTagInterpreterController extends DefaultHandler implements XMLTagInterpreterController
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final boolean DEBUG = "true".equals(System.getProperty("org.pfsw.pax.debug", "false"));

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private XMLReader reader = null;
  private Stack<XMLTagInterpreter> interpreterStack = null;
  private XMLTagInterpreterFactory interpreterFactory = null;
  private boolean lexicalHandlerSupported = false;
  private LexicalHandlerDispatcher lexicalHandlerHolder = new LexicalHandlerDispatcher();

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initializes the instance variables with default values.    <br>
   * Subclasses, that are overriding this constructor must call
   * super() in the first line of their constructor !
   */
  protected BaseXMLTagInterpreterController()
  {
    setInterpreterStack(new Stack<XMLTagInterpreter>());
    setInterpreterFactory(createTagInterpreterFactory());
  }

  /**
   * Initializes the instance variables with default values.    <br>
   * Subclasses, that are overriding this constructor must call
   * super(parser) in the first line of their constructor !
   *
   * @param xmlReader The SAX reader that reads the XML source.
   */
  public BaseXMLTagInterpreterController(XMLReader xmlReader)
  {
    this();
    setXmlReader(xmlReader);
    initLexicalHandler();
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  public LexicalHandlerDispatcher getLexicalHandlerHolder()
  {
    return this.lexicalHandlerHolder;
  }

  public void setLexicalHandlerHolder(LexicalHandlerDispatcher newValue)
  {
    this.lexicalHandlerHolder = newValue;
  }

  /**
   * Gives back the control of the element interpretation to the receiver.   <br>
   * The controller than takes the last tag interpreter from its stack
   * and passes over control to it.
   *
   * @param result The result of a sub tag interpretation.
   */
  @Override
  public void returnControl(Object result) throws SAXException
  {
    XMLTagInterpreter previous = null;
    XMLTagInterpreter current = null;

    current = popFromStack();
    trace("Returned from " + current.getTagName() + " - interpreter.");
    if (isStackEmpty()) // This only happens after the documents end tag
    {
      getReader().setContentHandler(this);
      handleFinalResult(result);
    }
    else
    {
      previous = peekStack();
      trace("Restarting interpreter for " + previous.getTagName());
      restartInterpreter(previous, current.getTagName(), result);
    }
  }

  /**
   * Supports SAX 2.0
   *
   * @see org.xml.sax.ContentHandler#startElement(String,String,String,Attributes)
   */
  @Override
  public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException

  {
    startElement(qName, attributes);
  }

  public void startElement(String name, Attributes attributes) throws SAXException
  {
    XMLTagInterpreter interpreter = null;

    interpreter = getInterpreterFor(name);
    if (interpreter != null)
    {
      putOnStack(interpreter);
      interpreter.setController(this);
      startInterpreter(interpreter, name, attributes);
    }
    else
    {
      throw (new SAXException("Interpreter for <" + name + "> not found !"));
    }
  }

  /**
   * Returns the XML reader that will be used to parse XML streams
   */
  public XMLReader getXmlReader()
  {
    return getReader();
  }

  /**
   * Sets the XML reader that will be used to parse XML streams
   * 
   * @param aReader The reader this controller should use to parse XML streams
   */
  public void setXmlReader(XMLReader aReader)
  {
    if (getReader() != null)
    {
      getReader().setContentHandler(null);
    }

    setReader(aReader);

    if (aReader != null)
    {
      aReader.setContentHandler(this);
    }
  }

  // =========================================================================
  // ABSTRACT PROTECTED INSTANCE METHODS
  // =========================================================================
  /**
   * This method is called, when the end tag of the document was reached.    <br>
   * It gets the result of the last tag interpreter.
   *
   * @param result The result of the all embracing document tag.
   */
  abstract protected void handleFinalResult(Object result);

  /**
   * Returns a new instance of a tag interpreter factory.   <br>
   * This method is called during initialization of the controller.
   * Its returned object is put into the variable interpreterFactory.
   */
  abstract protected XMLTagInterpreterFactory createTagInterpreterFactory();

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  /**
   * Starts the given tag interpreter for the first time.    <br>
   * The first time is, when the corresponding start tag was found.
   * The interpreter will now be installed in the SAX parser as the
   * current document handler until it returns control to this controller
   * through the method <i>returnControl()</i> or <i>startElement()</i>.<br>
   *
   * @param interpreter The tag interpreter to be started.
   * @param elementName The name of the tag.
   * @param attributes All attributes and their values found in the start tag.
   * @see #returnControl(Object)
   * @see #startElement(String, Attributes)
   */
  protected void startInterpreter(XMLTagInterpreter interpreter, String elementName, Attributes attributes) throws SAXException
  {
    trace("Start interpreter for <" + elementName + "> first time");
    interpreter.start(elementName, attributes);
    getReader().setContentHandler(interpreter);
    getLexicalHandlerHolder().setLexicalHandler(interpreter);
  }

  /**
   * Restarts the given tag interpreter after a sub tag interpretation
   * was finished.    <br>
   * If an XML tag contains inner tags, the control is passed to the
   * controller and from there to an appropriate tag interpreter.
   * When this tag interpreter has finished, it returns a business object,
   * which then is passed over to the superior tag interpreter via the
   * <i>restart()</i> method. So this superior interpreter can put this
   * result into its own business object.
   *
   * @param interpreter The tag interpreter to be restarted.
   * @param subTagName The name of the sub tag, the result corresponds to.
   * @param result The result object of the interpretation of the sub tag.
   * @see XMLTagInterpreter#restart(String, Object)
   */
  protected void restartInterpreter(XMLTagInterpreter interpreter, String subTagName, Object result) throws SAXException
  {
    interpreter.restart(subTagName, result);
    getReader().setContentHandler(interpreter);
  }

  protected XMLTagInterpreter getInterpreterFor(String tagName) throws SAXException
  {
    XMLTagInterpreter interpreter = null;

    interpreter = getInterpreterFactory().getInterpreterFor(tagName);
    return interpreter;
  }

  /**
   * Returns the last element from the stack and removes it.
   *
   * @return The interpreter on top of the stack.
   */
  protected XMLTagInterpreter popFromStack()
  {
    return (getInterpreterStack().pop());
  }

  /**
   * Returns the last element from the stack without removing it.
   *
   * @return The interpreter on top of the stack.
   */
  protected XMLTagInterpreter peekStack()
  {
    return (getInterpreterStack().peek());
  }

  /**
   * Puts the given tag interpreter on top of the stack.
   */
  protected void putOnStack(XMLTagInterpreter interpreter)
  {
    getInterpreterStack().push(interpreter);
  }

  /**
   * Returns whether the stack is empty or not.
   */
  protected boolean isStackEmpty()
  {
    return getInterpreterStack().isEmpty();
  }

  protected void initLexicalHandler()
  {
    try
    {
      if (getXmlReader().getProperty(PROP_LEXICAL_HANDLER) == null)
      {
        getXmlReader().setProperty(PROP_LEXICAL_HANDLER, getLexicalHandlerHolder());
        setLexicalHandlerSupported(true);
      }
    }
    catch (SAXException ex)
    {
      if (DEBUG)
      {
        ex.printStackTrace();
      }
      setLexicalHandlerSupported(false);
      trace("LexicalHandler not supported by parser.");
    }
  }

  /**
   * Method for debugging output.   <br>
   * Must be overridden by subclasses for producing output.
   * Is used also by all interpreters.
   */
  protected void trace(@SuppressWarnings("unused") String text)
  {
    // Nothing by default
  }

  protected XMLReader getReader()
  {
    return reader;
  }

  protected void setReader(XMLReader aValue)
  {
    this.reader = aValue;
  }

  protected Stack<XMLTagInterpreter> getInterpreterStack()
  {
    return this.interpreterStack;
  }

  protected void setInterpreterStack(Stack<XMLTagInterpreter> aValue)
  {
    this.interpreterStack = aValue;
  }

  protected XMLTagInterpreterFactory getInterpreterFactory()
  {
    return this.interpreterFactory;
  }

  protected void setInterpreterFactory(XMLTagInterpreterFactory aValue)
  {
    interpreterFactory = aValue;
  }

  /** 
   * Returns true if the LexicalHandler callback is supported by the XML parser 
   */
  protected boolean isLexicalHandlerSupported()
  {
    return lexicalHandlerSupported;
  }

  protected void setLexicalHandlerSupported(boolean newValue)
  {
    lexicalHandlerSupported = newValue;
  }
}