// ===========================================================================
// CONTENT  : CLASS BaseXMLTagInterpreterFactory
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.0 - 28/02/2002
// HISTORY  :
//  11/07/1999 	duma  CREATED
//	28/02/2002	duma	changed	-> Support for SAX 2.0
//
// Copyright (c) 1999-2002, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.pax;

import java.util.HashMap;
import java.util.Map;

import org.xml.sax.SAXException;

/**
 * This is the basic implementation of a registry of tag interpreter classes
 * an their associated tag names.
 *
 * @author Manfred Duchrow
 * @version 2.0
 */
public abstract class BaseXMLTagInterpreterFactory implements XMLTagInterpreterFactory
{
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private Map<String, String> classNameRegistry = new HashMap<String, String>();
  private Map<String, Class<XMLTagInterpreter>> classRegistry = new HashMap<String, Class<XMLTagInterpreter>>();

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns an instance of the tag interpreter class associated that
   * is associated to the specified tag name.
   *
   * @param tagName The name of the tag, the interpreter is looked up for.
   * @return The tag interpreter instance or null, if the tag isn't supported.
   */
  @Override
  public XMLTagInterpreter getInterpreterFor(String tagName) throws SAXException
  {
    Class<XMLTagInterpreter> interpreterClass = null;
    XMLTagInterpreter interpreter = null;

    interpreterClass = getInterpreterClassFor(tagName);
    if (interpreterClass != null)
    {
      try
      {
        interpreter = interpreterClass.newInstance();
      }
      catch (Exception ex)
      {
        String msg = "Could not create new instance of interpreter class '";
        msg = msg + interpreterClass.getName() + "'\n";
        msg = msg + ex.toString();
        throw (new SAXException(msg));
      }
    }
    return interpreter;
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  /**
   * Register a tag interpreter class name for a particular tag name.   <br>
   *
   * @param tagName The name of the tag, the interpreter class is responsible for.
   * @param className The full class name (with package !) for the interpreter class.
   */
  protected void registerTagInterpreter(String tagName, String className)
  {
    getClassNameRegistry().put(tagName, className);
  }

  /**
   * Returns a tag interpreter class associated with the given tag name.
   *
   * @param tagName The name of the tag, the interpreter is searched for.
   * @return The tag interpreter instance or null, if the tag isn't supported.
   */
  @SuppressWarnings("unchecked")
  protected Class<XMLTagInterpreter> getInterpreterClassFor(String tagName) throws SAXException
  {
    String className = null;
    Class<XMLTagInterpreter> interpreterClass = null;

    interpreterClass = getClassRegistry().get(tagName);
    if (interpreterClass == null)
    {
      className = getClassNameRegistry().get(tagName);
      if (className != null)
      {
        try
        {
          interpreterClass = (Class<XMLTagInterpreter>)Class.forName(className);
          getClassRegistry().put(tagName, interpreterClass);
        }
        catch (Exception ex)
        {
          throw new SAXException("Interpreter class '" + className + "' not found !", ex);
        }
      }
    }
    return interpreterClass;
  }

  /** 
   * Returns the internal mapping of tag names to their corresponding
   * interpreter classes.    <br>
   * key : tag name (String)  <br>
   * value : tag interpreter class (Class)
   */
  protected Map<String, String> getClassNameRegistry()
  {
    return this.classNameRegistry;
  }

  /** 
   * Sets the internal mapping of tag names to their corresponding
   * interpreter classes.   <br>
   * key : tag name (String)  <br>
   * value : tag interpreter class (Class)
   */
  protected void setClassNameRegistry(Map<String, String> aValue)
  {
    this.classNameRegistry = aValue;
  }

  /** 
   * Returns the internal mapping of tag names to their corresponding
   * interpreter class names.    <br>
   * key : tag name (String)  <br>
   * value : tag interpreter class name (String)
   */
  protected Map<String, Class<XMLTagInterpreter>> getClassRegistry()
  {
    return this.classRegistry;
  }

  /** 
   * Sets the internal mapping of tag names to their corresponding
   * interpreter class names.    <br>
   * key : tag name (String)  <br>
   * value : tag interpreter class name (String)
   */
  protected void setClassRegistry(Map<String, Class<XMLTagInterpreter>> aMap)
  {
    this.classRegistry = aMap;
  }
}