// ===========================================================================
// CONTENT  : CLASS MarkupWriter
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.3 - 30/07/2004
// HISTORY  :
//  28/07/2000  duma  CREATED
//	30/06/2002	duma	changed	->	Use line.separator rather than "\n"
//	30/06/2002	duma	changed	->	Use " as qoute instead of '
//	22/12/2003	duma	changed	->	Use logger() to report exceptions rather than printStackTrace()
//	30/07/2004	duma	added		->	single quotes
//
// Copyright (c) 2000-2004, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.pax;

import java.io.IOException;
import java.io.Writer;

import org.pfsw.logging.Logger;

/**
 * A class that helps to write out a markup language (e.g. XML) to a stream.
 *
 * @author Manfred Duchrow
 * @version 1.3
 */
@SuppressWarnings("resource")
public class MarkupWriter
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final String NEWLINE = System.getProperty("line.separator");

  private static final char SINGLE_QUOTE_CHAR = '\'';
  private static final char DOUBLE_QUOTE_CHAR = '"';

  // =========================================================================
  // CLASS VARIABLES
  // =========================================================================
  private static char[] SpecialCharacters = initSpecialCharacters();
  private static String[] Placeholders = initPlaceholders();

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private Writer writer = null;
  private int indentLevel = 0;
  private int indentIncrement = 2;
  private char quoteChar = DOUBLE_QUOTE_CHAR;

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  private static char[] initSpecialCharacters()
  {
    char[] characters = { '<', '>', '&' };
    return characters;
  }

  private static String[] initPlaceholders()
  {
    String[] ph = { "&lt;", "&gt;", "&amp;" };
    return ph;
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  public MarkupWriter(Writer writer)
  {
    super();
    setWriter(writer);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  public int getIndentLevel()
  {
    return indentLevel;
  }

  public int getIndentIncrement()
  {
    return indentIncrement;
  }

  public void setIndentIncrement(int newValue)
  {
    indentIncrement = newValue;
  }

  public void writeStartTag(String tagName)
  {
    beginStartTag(tagName);
    closeTag();
  }

  public void beginStartTag(String tagName)
  {
    beginTag();
    write(tagName);
  }

  public void writeEndTag(String tagName)
  {
    beginTag();
    write("/");
    write(tagName);
    closeTag();
  }

  public void writeEmptyTag(String tagName)
  {
    beginTag();
    write(tagName);
    finishEmptyTag();
  }

  public void finishEmptyTag()
  {
    writeSpace();
    write("/");
    closeTag();
  }

  public void finishTag()
  {
    closeTag();
  }

  public void writeAttribute(String attrName, String value)
  {
    writeSpace();
    write(attrName);
    writeAssignment();
    writeQuote();
    writeCharData(value, true);
    writeQuote();
  }

  public void writeText(String text)
  {
    writeCharData(text, false);
  }

  public void newIndentedLine()
  {
    write(NEWLINE);
    writeIndentation();
  }

  public void newIncIndentedLine()
  {
    incIndent();
    newIndentedLine();
  }

  public void newDecIndentedLine()
  {
    decIndent();
    newIndentedLine();
  }

  public String contents()
  {
    return getWriter().toString();
  }

  public void incIndent()
  {
    indentLevel += indentIncrement;
  }

  public void decIndent()
  {
    indentLevel -= indentIncrement;
  }

  /**
   * Configures the writer to write single quotes ( ' ) around attribute values.
   */
  public void useSingleQuotes()
  {
    setQuoteChar(SINGLE_QUOTE_CHAR);
  }

  /**
   * Configures the writer to write double quotes ( " ) around attribute values.
   */
  public void useDoubleQuotes()
  {
    setQuoteChar(DOUBLE_QUOTE_CHAR);
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected void beginTag()
  {
    write("<");
  }

  protected void closeTag()
  {
    write(">");
  }

  protected void writeSpace()
  {
    write(" ");
  }

  protected void writeAssignment()
  {
    write("=");
  }

  protected void writeQuote()
  {
    write(getQuoteChar());
  }

  protected void writeCharData(String text, boolean asAttrValue)
  {
    // Here we have to translate special characters into their
    // markup language representation. e.g. < is &lt;
    for (int index = 0; index < text.length(); index++)
      write(translateCharacter(text.charAt(index), asAttrValue));
  }

  protected void write(String text)
  {
    try
    {
      getWriter().write(text);
    }
    catch (IOException ex)
    {
      logger().logException(ex);
    }
  }

  protected void write(char ch)
  {
    try
    {
      getWriter().write(ch);
    }
    catch (IOException ex)
    {
      logger().logException(ex);
    }
  }

  protected void flush()
  {
    try
    {
      getWriter().flush();
    }
    catch (IOException ex)
    {
      logger().logException(ex);
    }
  }

  protected void writeIndentation()
  {
    for (int i = 1; i <= indentLevel; i++)
      writeSpace();
  }

  protected int findSpecialChar(char ch)
  {
    for (int i = 0; i < SpecialCharacters.length; i++)
    {
      if (SpecialCharacters[i] == ch)
        return i;
    }
    return -1;
  }

  @SuppressWarnings("incomplete-switch")
  protected String translateCharacter(char ch, boolean transQuote)
  {
    if (ch == getQuoteChar())
    {
      if (transQuote)
      {
        switch (ch)
        {
          case SINGLE_QUOTE_CHAR :
            return "&apos;";
          case DOUBLE_QUOTE_CHAR :
            return "&quot;";
        }
      }
    }
    int index = findSpecialChar(ch);
    if (index >= 0)
    {
      return Placeholders[index];
    }
    return String.valueOf(ch);
  }

  protected Logger logger()
  {
    return LoggerProvider.getLogger();
  }

  protected Writer getWriter()
  {
    return writer;
  }

  protected void setWriter(Writer newValue)
  {
    writer = newValue;
  }

  protected void setIndentLevel(int newValue)
  {
    indentLevel = newValue;
  }

  protected char getQuoteChar()
  {
    return quoteChar;
  }

  protected void setQuoteChar(char newValue)
  {
    quoteChar = newValue;
  }
}
