// ===========================================================================
// CONTENT  : CLASS NamespacePrefixMapper
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 30/01/2011
// HISTORY  :
//  30/01/2011  mdu  CREATED
//
// Copyright (c) 2011, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.pax;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.xml.namespace.NamespaceContext;

/**
* An implementation that handles namespace prefix to URI and URI to prefix
* mapping.
*
* @author M.Duchrow
* @version 1.0
*/
public class NamespacePrefixMapper implements NamespaceContext
{
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private final Map<String, String> mappings;

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Creates a new instance with the given mappings.
   */
  public static NamespacePrefixMapper create(Map<String, String> prefixToURIMappings)
  {
    return new NamespacePrefixMapper(prefixToURIMappings);
  }

  /**
   * Creates a new instance with an empty mapping.
   */
  public static NamespacePrefixMapper create()
  {
    return new NamespacePrefixMapper();
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with the given mappings.
   */
  public NamespacePrefixMapper(Map<String, String> prefixToURIMappings)
  {
    super();
    this.mappings = prefixToURIMappings;
  }

  /**
   * Initialize the new instance with an empty mapping.
   */
  public NamespacePrefixMapper()
  {
    this(new HashMap<String, String>());
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  @Override
  public String getNamespaceURI(String prefix)
  {
    return getMappings().get(prefix);
  }

  @Override
  public String getPrefix(String namespaceURI)
  {
    for (Map.Entry<String, String> mapping : getMappings().entrySet())
    {
      if (mapping.getValue().equals(namespaceURI))
      {
        return mapping.getKey();
      }
    }
    return null;
  }

  @Override
  public Iterator<String> getPrefixes(String namespaceURI)
  {
    List<String> result;

    result = new ArrayList<String>();
    for (Map.Entry<String, String> mapping : getMappings().entrySet())
    {
      if (mapping.getValue().equals(namespaceURI))
      {
        result.add(mapping.getKey());
      }
    }
    return result.iterator();
  }

  /**
   * Add the prefix and URI to the mappings.
   * 
   * @return The mapper itself to support fluent API. 
   */
  public NamespacePrefixMapper addMapping(String nsPrefix, String nsURI)
  {
    if ((nsPrefix != null) && (nsURI != null))
    {
      getMappings().put(nsPrefix, nsURI);
    }
    return this;
  }

  protected Map<String, String> getMappings()
  {
    return this.mappings;
  }
}
