// ===========================================================================
// CONTENT  : CLASS XMLWriteInterpreter
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.0 - 28/02/2002
// HISTORY  :
//  28/07/2000  duma  CREATED
//	28/02/2002	duma	changed	-> Support for SAX 2.0
//
// Copyright (c) 2002, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.pax;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

/**
 * The instances of this class are responsible for writing the tag's
 * name, attributes and contents to the given MarkupWriter.
 *
 * @author Manfred Duchrow
 * @version 2.0
 */
public class XMLWriteInterpreter extends BaseXMLTagInterpreter
{
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private String tagName = null;
  private MarkupWriter markupWriter = null;
  private boolean isEmptyTag = true;
  private boolean lastWasInnerTag = false;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  public XMLWriteInterpreter(String name, MarkupWriter writer)
  {
    super();
    setTagName(name);
    setMarkupWriter(writer);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  @Override
  public String getTagName()
  {
    return tagName;
  }

  @Override
  public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
  {
    checkEmptyTag();
    getMarkupWriter().newIncIndentedLine();
    super.startElement(uri, localName, qName, attributes);
  }

  @Override
  public void start(String elementName, Attributes attributes) throws SAXException
  {
    getMarkupWriter().beginStartTag(elementName);
    for (int index = 0; index < attributes.getLength(); index++)
    {
      getMarkupWriter().writeAttribute(attributes.getQName(index), attributes.getValue(index));
    }
  }

  @Override
  public void restart(String subTagName, Object subResult) throws SAXException
  {
    getMarkupWriter().decIndent();
    setLastWasInnerTag(true);
  }

  /**
   * Returns the result object, created by this interpreter from the XML data .   <br>
   *
   * @return The business object resulting from the parsed XML data.
   */
  @Override
  public Object getResult()
  {
    return null;
  }

  /**
   * This method is called whenever character data is received from the parser.   <br>
   * In contrary to the ContentHandler.charachters() method it gets the
   * data as string rather than a char[] array with start and length. <br>
   * Subclasses must override this method, if their tag contains PCDATA.
   * Here it's not defined as an abstract method, because the method's empty
   * implementation is convenient for tag interpreters without PCDATA.
   * <br>
   * @param data The data received from the parser.
   * @see org.xml.sax.ContentHandler#characters(char[],int,int)
   */
  @Override
  public void characterData(String data)
  {
    checkEmptyTag();
    getMarkupWriter().writeText(data);
    setLastWasInnerTag(false);
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected void checkEmptyTag()
  {
    if (isEmptyTag())
    {
      setEmptyTag(false);
      getMarkupWriter().finishTag();
    }
  }

  /**
   * This method is called just before control is passed back to the controller.   <br>
   * It gives the interpreter a chance to do any necessary completion of
   * its result object.
   */
  @Override
  protected void finalizeResult()
  {
    if (isEmptyTag())
    {
      getMarkupWriter().finishEmptyTag();
    }
    else
    {
      if (setLastWasInnerTag())
      {
        getMarkupWriter().newIndentedLine();
      }
      getMarkupWriter().writeEndTag(getTagName());
    }
  }

  protected void setTagName(String newValue)
  {
    tagName = newValue;
  }

  protected MarkupWriter getMarkupWriter()
  {
    return markupWriter;
  }

  protected void setMarkupWriter(MarkupWriter newValue)
  {
    markupWriter = newValue;
  }

  public boolean isEmptyTag()
  {
    return this.isEmptyTag;
  }

  public void setEmptyTag(boolean isEmptyTag)
  {
    this.isEmptyTag = isEmptyTag;
  }

  public boolean setLastWasInnerTag()
  {
    return this.lastWasInnerTag;
  }

  public void setLastWasInnerTag(boolean lastWasInnerTag)
  {
    this.lastWasInnerTag = lastWasInnerTag;
  }
}