// ===========================================================================
// CONTENT  : CLASS ClassRegistry
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.4 - 11/02/2017
// HISTORY  :
//  25/01/2003  mdu CREATED
//	28/02/2003	mdu	added		->	keys(), values(), classes(), classNames(), getClassInfo()
//	17/11/2010	mdu	changed	->	to generic type
//	23/11/2014	mdu	changed	->	key must be string
//  11/02/2017  mdu added   ->  getFirst(), isEmpty()
//
// Copyright (c) 2003-2017, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.plugin;

import static org.pfsw.util.CollectionUtil.*;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.pfsw.reflect.ClassInfo;

/**
 * A class registry is a container that keeps classes registered under a
 * logical key. For a query by key it can return 
 * <ul>
 * <li>the full qualified class name</li>
 * <li>the Class object</li>
 * <li>an instance of the class</li>
 * </ul>
 *
 * @author Manfred Duchrow
 * @version 1.4
 */
public class ClassRegistry<T>
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final boolean DEBUG = "true".equals(System.getProperty("org.pfsw.plugin.debug", "false"));

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private Class<T> entryType = null;
  private Map<String, ClassInfo<T>> registry = null;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  public ClassRegistry()
  {
    super();
    registry(new HashMap<String, ClassInfo<T>>());
  }

  /**
   * Initialize the new instance with a type all classes that are registered
   * must comply to. That is, they must be the identical class, a subclass of 
   * this type or an implementer of this type if it is an interface.
   * 
   * @param typeOfRegisteredClasses The type to which the classes that can be 
   * 				registered must be compatible to
   */
  public ClassRegistry(Class<T> typeOfRegisteredClasses)
  {
    this();
    setEntryType(typeOfRegisteredClasses);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the name of the class which is registered under the given key.
   * If nothing is registered under the key this method returns null.
   * 
   * @param key The key under which the class is registered
   */
  public String getClassName(Object key)
  {
    ClassInfo<T> classInfo;

    classInfo = get(key);
    return (classInfo == null) ? null : classInfo.getClassName();
  }

  /**
   * Returns the class which is registered under the given key.
   * If nothing is registered under the key this method returns null.
   * 
   * @param key The key under which the class is registered
   */
  public Class<T> getClass(String key)
  {
    ClassInfo<T> classInfo;

    classInfo = get(key);
    return (classInfo == null) ? null : classInfo.getClassObject();
  }

  /**
   * Returns a new instance of the class which is registered under the 
   * given key.
   * If nothing is registered under the key this method returns null.
   * 
   * @param key The key under which the class is registered.
   */
  public T newInstance(String key)
  {
    ClassInfo<T> classInfo;

    classInfo = get(key);
    return (classInfo == null) ? null : classInfo.createInstance();
  }

  /**
   * Registers the class with the given name under the specified key.
   * 
   * @param key The key under which the class is registered
   * @param className The fully qualified name of the class
   * @return true, if the class was found, successfully validated and registered
   */
  public boolean register(String key, String className)
  {
    try
    {
      put(key, className);
      return true;
    }
    catch (ClassRegistryException ex)
    {
      handleException(ex);
      return false;
    }
  }

  /**
   * Registers the class with the given name under the specified key.
   * 
   * @param key The key under which the class is registered.
   * @param className The fully qualified name of the class.
   * @throws ClassRegistryException If the class can't be found or if it is of 
   * 					the wrong type or can't be instantiated.
   */
  public void put(String key, String className) throws ClassRegistryException
  {
    assertArgNotNull("put", "key", key);
    assertArgNotNull("put", "className", className);

    ClassInfo<T> classInfo;

    classInfo = new ClassInfo<T>(className);
    put(key, classInfo);
  }

  /**
   * Registers the class under the specified key.
   * 
   * @param key The key under which the class is registered
   * @param aClass The class to be registered
   * @return true, if the class was successfully validated and registered
   */
  public boolean register(String key, Class<T> aClass)
  {
    try
    {
      put(key, aClass);
      return true;
    }
    catch (ClassRegistryException ex)
    {
      handleException(ex);
      return false;
    }
  }

  /**
   * Registers the given class under the specified key.
   * 
   * @param key The key under which the class is registered
   * @param aClass The class to be registered
   * @throws ClassRegistryException If the class is of the wrong type or 
   * 					can't be instantiated
   */
  public void put(String key, Class<T> aClass) throws ClassRegistryException
  {
    assertArgNotNull("put", "key", key);
    assertArgNotNull("put", "aClass", aClass);

    ClassInfo<T> classInfo;

    classInfo = new ClassInfo<T>(aClass);
    put(key, classInfo);
  }

  /**
   * Removes the entry in the registry which is defined by the given key.
   * @return true, if anything has been found and removed, otherwise false.
   */
  public boolean remove(Object key)
  {
    Object obj;

    if (key != null)
    {
      obj = registry().remove(key);
      return obj != null;
    }
    return false;
  }

  /**
   * Returns all keys known in this registry.
   */
  public String[] keys()
  {
    Set<String> keys;

    keys = registry().keySet();
    return keys.toArray(new String[keys.size()]);
  }

  /**
   * Returns all class info objects that are currently in this registry.
   */
  @SuppressWarnings("unchecked")
  public ClassInfo<T>[] values()
  {
    if (registry().isEmpty())
    {
      return new ClassInfo[0];
    }
    return CU.toArray(registry().values());
  }

  /**
   * Returns the first class info object of null if the registry is empty.
   */
  public ClassInfo<T> getFirst()
  {
    if (isEmpty())
    {
      return null;
    }
    return registry().values().iterator().next();
  }

  /**
   * Returns all class objects that are currently in this registry.
   */
  @SuppressWarnings({ "unchecked", "cast" })
  public Class<T>[] classes()
  {
    List<Class<T>> classes;

    classes = collectClasses();
    if (classes.isEmpty())
    {
      return (Class<T>[])new Class[0];
    }
    return CU.toArray(classes);
  }

  /**
   * Returns all class names of the classes that are currently in this registry.
   */
  public String[] classNames()
  {
    List<String> names;

    names = collectNames();
    return names.toArray(new String[names.size()]);
  }

  /**
   * Returns the class information object stored under the given key or null
   * if nothing is found.
   */
  public ClassInfo<T> getClassInfo(String key)
  {
    return get(key);
  }

  /**
   * Returns true if the given type is compatible to the underlying entry type.
   * If no underlying entry type exists all types are assumed valid and true 
   * will be returned.
   * 
   * @param type The type to be checked.
   */
  public boolean isValidType(Class<?> type)
  {
    if (isTyped())
    {
      if (!getEntryType().isAssignableFrom(type))
      {
        return false;
      }
    }
    return true;
  }

  public int size()
  {
    return registry().size();
  }

  public boolean isEmpty()
  {
    return registry().isEmpty();
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected ClassInfo<T> get(Object key)
  {
    if (key == null)
    {
      return null;
    }

    return registry().get(key);
  }

  protected void put(String key, ClassInfo<T> classInfo) throws ClassRegistryException
  {
    assertArgNotNull("put", "key", key);
    validate(classInfo);
    registry().put(key, classInfo);
  }

  protected void validate(ClassInfo<T> classInfo) throws ClassRegistryException
  {
    assertClassFound(classInfo);
    assertValidType(classInfo);
    assertInstanceCreation(classInfo);
  }

  protected void assertClassFound(ClassInfo<T> classInfo) throws ClassRegistryException
  {
    Class<T> aClass;

    aClass = classInfo.getClassObject();
    if (aClass == null)
    {
      throw new ClassRegistryException("Class " + classInfo.getClassName() + " not found");
    }
  }

  protected void assertValidType(ClassInfo<T> classInfo) throws ClassRegistryException
  {
    if (!isValidType(classInfo.getClassObject()))
    {
      throw new ClassRegistryException("Class " + classInfo.getClassName() + " is not compatible to type " + getEntryType().getName());
    }
  }

  protected void assertInstanceCreation(ClassInfo<T> classInfo) throws ClassRegistryException
  {
    try
    {
      classInfo.newInstance();
    }
    catch (Exception ex)
    {
      throw new ClassRegistryException("Unable to create instances of class " + classInfo.getClassName() + " (" + ex.getMessage() + ")");
    }
  }

  protected boolean isTyped()
  {
    return getEntryType() != null;
  }

  protected void assertArgNotNull(String methodName, String argName, Object arg)
  {
    if (arg == null)
    {
      throw new IllegalArgumentException("Argument '" + argName + "' not valid in method '" + getClass().getName() + "." + methodName + "'");
    }
  }

  protected void illegalArgumentException(String methodName, String argName)
  {
    throw new IllegalArgumentException("Argument '" + argName + "' not valid in method '" + getClass().getName() + "." + methodName + "'");
  }

  /**
   * Returns all class names that are currently in this registry.
   */
  protected List<String> collectNames()
  {
    Collection<ClassInfo<T>> values;
    List<String> result;

    values = registry().values();
    result = new ArrayList<String>(values.size());
    for (ClassInfo<T> classInfo : values)
    {
      result.add(classInfo.getClassName());      
    }
    return result;
  }

  /**
   * Returns all classes that are currently in this registry.
   */
  protected List<Class<T>> collectClasses()
  {
    Collection<ClassInfo<T>> values;
    List<Class<T>> result;
    
    values = registry().values();
    result = new ArrayList<Class<T>>(values.size());
    for (ClassInfo<T> classInfo : values)
    {
      result.add(classInfo.getClassObject());
    }
    return result;
  }
  
  protected void handleException(Exception ex) 
  {
    if (DEBUG)
    {
      LoggerProvider.getLogger().logException(ex);
    }
  }
  
  protected Class<T> getEntryType()
  {
    return entryType;
  }

  protected void setEntryType(Class<T> newValue)
  {
    entryType = newValue;
  }

  protected Map<String, ClassInfo<T>> registry()
  {
    return registry;
  }

  protected void registry(Map<String, ClassInfo<T>> newValue)
  {
    registry = newValue;
  }
}
