// ===========================================================================
// CONTENT  : CLASS PluginCollector
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.6 - 11/02/2017
// HISTORY  :
//  02/02/2003  mdu  CREATED
//	06/11/2003	mdu	added			->	DEBUG
//	22/12/2003	mdu	changed		->	Use logger for debug output
//	26/12/2004	mdu	changed		->	Support path definition by property "org.pf.plugin.path"
//	17/11/2010	mdu	changed		->	Support loading plugin definitions via getResourceStream()
//  23/11/2014  mdu changed   ->  gerneric type / Support ofPluginRegistration
//  29/01/2017  mdu changed   ->  extended mechanism to determine classloader.
//  11/02/2017  mdu changed   ->  use PluginRegistrationLoader 
//
// Copyright (c) 2003-2017, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.plugin;

import java.io.IOException;
import java.io.InputStream;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.pfsw.file.Classpath;
import org.pfsw.file.ClasspathElement;
import org.pfsw.file.FileUtil;

/**
 * The responsibility of this class is to detect all properties files with a
 * specific name on a given classpath and load the defined classes that fit to
 * a given class/interface into a class regitry.
 * <p>
 * Since version 1.3 it is possible to specify a system property named
 * "org.pf.plugin.path" to define the path on which the plugin collector
 * must search for plug-ins. 
 * <p>
 * Since version 1.4 it is also possible to load all plugin definitions
 * from a single file that is looked-up via the standard classloader mechanism
 * getResourceStream().
 * <p>
 * Since version 1.5 it is also possible to load plugin definitions
 * from classes that implement the {@link PluginRegistration} interface. These
 * classes are found by the Java service loader mechanism on the whole classpath
 * (based on classloader hierarchy).
 *
 * @author Manfred Duchrow
 * @version 1.6
 */
public class PluginCollector<TType>
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  protected static final String DEBUG_PREFIX = "org.pf.plugin.PluginCollector: ";

  // =========================================================================
  // CLASS VARIABLES
  // =========================================================================
  public static boolean DEBUG = "true".equals(System.getProperty("org.pf.plugin.PluginCollector.debug"));

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private ClassRegistry<TType> classRegistry = null;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  public PluginCollector()
  {
    this((ClassRegistry<TType>)null);
  }

  /**
   * Initialize the new instance the type the plugins to collect must
   * be compliant with.
   */
  public PluginCollector(Class<TType> pluginType)
  {
    this(new ClassRegistry<TType>(pluginType));
  }

  /**
   * Initialize the new instance with a class registry.
   */
  public PluginCollector(ClassRegistry<TType> registry)
  {
    super();
    setClassRegistry(registry == null ? new ClassRegistry<TType>() : registry);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the class registry that contains all loaded plugins.
   */
  public ClassRegistry<TType> getClassRegistry()
  {
    return classRegistry;
  }

  /**
   * Load the class definitions from the file with the given name.
   * The file will be searched for utilizing the classloader of this class.
   * 
   * @param filename The name of the properties file that contains the class definitions
   * @return The class registry that contains the new definitions if the file has been found.
   */
  @SuppressWarnings("resource")
  public ClassRegistry<TType> loadPluginsViaClassLoader(String filename)
  {
    InputStream stream;

    stream = getClass().getResourceAsStream(filename);
    if (stream != null)
    {
      if (DEBUG)
      {
        debug("Reading plugin file " + filename + "via classloader.");
      }
      readMapping(stream, filename);
    }
    else
    {
      if (DEBUG)
      {
        debug("Classloader could not find file " + filename);
      }
    }
    return getClassRegistry();
  }

  /**
   * Load the class definitions from all files with the given name that can
   * be found in the system classpath.
   * 
   * @param filename The name of the properties file that contains the class definitions
   * @return The class registry that contains the new definitions from all the files that have been found.
   */
  public ClassRegistry<TType> loadPlugins(String filename)
  {
    return loadPlugins(filename, getPluginClasspath());
  }

  /**
   * Load the class definitions from all files with the given name that can
   * be found in the specified classpath.
   * 
   * @param filename The name of the properties file that contains the class definitions
   * @param classpath The classpath the specified filename to be looked up in
   */
  public ClassRegistry<TType> loadPlugins(String filename, String classpath)
  {
    readPlugins(filename, classpath);
    return getClassRegistry();
  }

  /**
   * Load the class definitions from all files with the given name that can
   * be found in the specified classpath.
   * 
   * @param filename The name of the properties file that contains the class definitions
   * @param classpath The classpath on which to lookup the the specified filename
   */
  public ClassRegistry<TType> loadPlugins(String filename, Classpath classpath)
  {
    readPlugins(filename, classpath);
    return getClassRegistry();
  }

  /**
   * Load the class definitions from all PluginRegistration implementing classes
   * that can be found via Java service loader mechanism in the given classloader's 
   * hierarchy. If the given classcloader is null the first non-null classloader of 
   * this list will be used.
   * <ol>
   *   <li>The current thread's context class loader</li>
   *   <li>the class loader of this class</li>
   *   <li>the system class loader</li>
   * </ol>
   * 
   * @param classLoader The class loader from which to start looking for services
   * that implement interface {@link PluginRegistration}.
   */
  public ClassRegistry<TType> loadPluginsViaPluginRegistration(final ClassLoader classLoader)
  {
    List<PluginRegistration> plugins;

    plugins = PluginRegistrationLoader.getPlugins(classLoader);
    for (PluginRegistration registration : plugins)
    {
      addMappings(registration.getClassesToRegister());
    }
    return getClassRegistry();
  }

  /**
   * Load the class definitions from all PluginRegistration implementing classes
   * that can be found via Java service loader mechanism using the classloader of 
   * this class.
   */
  public ClassRegistry<TType> loadPluginsViaPluginRegistration()
  {
    return loadPluginsViaPluginRegistration(null);
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================  
  /**
   * Read all classes specified in the given file, if it exists in the specified
   * classpath element. Add those classes to the registry if they can be
   * instantiated and are of the correct type.
   *  
   * @param filename The name of the properties file that contains the class definitions
   * @param strClasspath The classpath the specified filename to be looked up in
   */
  protected void readPlugins(String filename, String strClasspath)
  {
    Classpath classpath;

    classpath = new Classpath(strClasspath);
    classpath.removeDuplicates();
    readPlugins(filename, classpath);
  }

  /**
   * Read all classes specified in the given file, if it exists in the specified
   * classpath element. Add those classes to the registry if they can be
   * instantiated and are of the correct type.
   *  
   * @param filename The name of the properties file that contains the class definitions
   * @param classpath The classpath the specified filename to be looked up in
   */
  protected void readPlugins(String filename, Classpath classpath)
  {
    ClasspathElement[] elements;

    if (DEBUG)
    {
      debug("Lookup file <" + filename + "> in <" + classpath + ">");
    }
    elements = classpath.elementsContaining(filename);
    for (int i = 0; i < elements.length; i++)
    {
      if (DEBUG)
      {
        debug("Found file <" + filename + "> in <" + elements[i].getName() + ">");
      }
      readMapping(elements[i], filename);
    }
  }

  @SuppressWarnings("resource")
  protected void readMapping(ClasspathElement element, String filename)
  {
    InputStream stream = null;

    try
    {
      stream = element.open(filename);
    }
    catch (IOException ex)
    {
      if (DEBUG)
      {
        debug("Error when opening file <" + filename + "> in <" + element.getName() + "> : " + ex.toString());
      }
      return;
    }
    readMapping(stream, filename);
  }

  /**
   * Reads all plugin definitions from the given stream into the underlying registry.
   * This method definitely closes the given stream (even in case of an exception).
   * 
   * @param stream The stream from which to read the definitions. 
   * @param filename The filename that corresponds to the stream (just for debugging).
   */
  protected void readMapping(InputStream stream, String filename)
  {
    Properties properties;

    try
    {
      properties = new Properties();
      properties.load(stream);
      addMappings(properties);
    }
    catch (IOException ex)
    {
      if (DEBUG)
      {
        debug("Failed to load properties from file " + filename + "    : " + ex.toString());
      }
    }
    finally
    {
      fileUtil().close(stream);
    }
  }

  protected void addMappings(Properties properties)
  {
    String value;

    for (String key : properties.stringPropertyNames())
    {      
      value = properties.getProperty(key);
      if (DEBUG)
      {
        debug("Register plugin key=" + key + " | value=" + value);
      }
      getClassRegistry().register(key, value);
    }
  }

  @SuppressWarnings("unchecked")
  protected void addMappings(Map<String, Class<?>> classesToRegister)
  {
    Class<TType> type;

    for (Map.Entry<String, Class<?>> entry : classesToRegister.entrySet())
    {
      if (getClassRegistry().isValidType(entry.getValue()))
      {
        type = (Class<TType>)entry.getValue();
        if (DEBUG)
        {
          debug("Register plugin class key=" + entry.getKey() + " | value=" + type.getCanonicalName());
        }
        getClassRegistry().register(entry.getKey(), type);
      }
    }
  }

  /**
   * Returns the classpath that is used to lookup any plug-in information files.
   */
  protected Classpath getPluginClasspath()
  {
    return PluginUtil.current().getPluginClasspath();
  }

  protected void setClassRegistry(ClassRegistry<TType> newValue)
  {
    classRegistry = newValue;
  }

  protected FileUtil fileUtil()
  {
    return FileUtil.current();
  }

  protected PluginUtil pluginUtil()
  {
    return PluginUtil.current();
  }

  protected void debug(String text)
  {
    LoggerProvider.getLogger().logDebug(DEBUG_PREFIX + text);
  }
}
