// ===========================================================================
// CONTENT  : ABSTRACT CLASS PluginLoader
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.2 - 11/02/2017
// HISTORY  :
//  25/05/2015  mdu  CREATED
//  29/01/2017  mdu   changed --> improved classloader handling
//  11/02/2017  mdu   added   --> loadFirstFoundInstance(), defaultToFirstEntry(), getAllPluginsViaPluginRegistration
//
// Copyright (c) 2015-2017, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.plugin;

import java.util.ServiceLoader;

import org.pfsw.bif.configuration.ConfigurationException;
import org.pfsw.bif.configuration.IConfigSettings;
import org.pfsw.reflect.ClassInfo;

/**
 * An abstract class that implements the generic logic to load a configured
 * implementation class for a specific interface through different variants
 * of loading mechanisms. The following list describes the order of mechanisms tried.
 * The first that successfully creates an instance will end the process and return that instance.
 * <ol>
 * <li>Check if a configuration parameter for the implementation class is set.
 * If that's the case, an instance of this class is created and returned.
 * </li>
 * <li>Check if an configuration parameter for the alias of the implementation class is set.
 * If that's the case, all plugin-registration implementations on the classpath will be checked for
 * aliases and associated classes that implement the TInterface.
 * The the class for the configured alias will be loaded and returned.
 * </li>
 * <li>Finally the Java ServiceLoader mechanism will be used to find an implementation class
 * for the TInterface.</li>
 * </ol>
 *
 * @author Manfred Duchrow
 * @version 1.2
 */
abstract public class PluginLoader<TInterface>
{
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private final String classNameConfigParamName;
  private final String aliasConfigParamName;
  private final Class<TInterface> interfaceClass;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  protected PluginLoader(Class<TInterface> interfaceClass, String classNameConfigParamName, String aliasConfigParamName)
  {
    super();
    this.interfaceClass = interfaceClass;
    this.classNameConfigParamName = classNameConfigParamName;
    this.aliasConfigParamName = aliasConfigParamName;
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the instance that is specified by class name configuration parameter 
   * or indirectly by the alias name optionally defined in the alias configuration parameter.
   * This class must be available to the given class loader.
   * <p>
   * If the class configuration parameter is not specified and the alias names is also not
   * available then the class will be looked-up using the Java service loader mechanism.
   * In that case the first found implementation class for interface TInterface will be taken.
   * Wherever a classloader is needed the current thread's context class loader will be used.
   * 
   * @param config The configuration from which the parameters for class name or alias name will be taken (must not be null).
   * @return The instance of the configured TInterface service.
   * @throws ConfigurationException If no implementation class was configured or it cannot be instantiated.
   */
  public TInterface loadPlugin(IConfigSettings config) throws ConfigurationException
  {
    return loadPlugin(null, config, true);
  }

  /**
   * Returns the instance that is specified by class name configuration parameter 
   * or indirectly by the alias name optionally defined in the alias configuration parameter.
   * This class must be available to the given class loader.
   * <p>
   * If the class configuration parameter is not specified and the alias names is also not
   * available then either a ConfigurationException
   * gets thrown (fallbackToServiceLoader=false) or the class will be looked-up
   * using the service loader mechanism (fallbackToServiceLoader=true).
   * In that case the first found implementation class for interface TInterface 
   * will be taken.
   * 
   * @param classLoader An optional class loader. If null the current threads context classloader will be used.
   * @param config The configuration from which the parameters for class name or alias name will be taken (must not be null).
   * @param fallbackToServiceLoader Definition whether or not finally the Java ServiceLoader mechanism should be tried.
   * @return The instance of the configured TInterface service.
   * @throws ConfigurationException If no implementation class was configured or it cannot be instantiated.
   */
  public TInterface loadPlugin(ClassLoader classLoader, IConfigSettings config, boolean fallbackToServiceLoader) throws ConfigurationException
  {
    String implClassName;
    String implClassAlias;
    TInterface instance = null;

    implClassName = config.getValue(getClassNameConfigParamName());
    if (implClassName == null)
    {
      implClassAlias = config.getValue(getAliasConfigParamName());
      if (implClassAlias == null)
      {
        if (defaultToFirstEntry())
        {
          instance = loadFirstFoundInstance(classLoader);
        }
      }
      else
      {
        instance = loadInstanceByAlias(classLoader, implClassAlias);
      }
      if (instance != null)
      {
        return instance;
      }
      if (fallbackToServiceLoader)
      {
        return loadInstanceFromService(classLoader);
      }
      throw new ConfigurationException("No class or alias specified in configiration parameter " + getParameterNames());
    }
    return loadInstance(classLoader, implClassName);
  }

  /**
   * Returns all plugins that can be found via the PluginRegistration mechanism
   * for the underlying interface definition.
   */
  public ClassRegistry<TInterface> getAllPluginsViaPluginRegistration()
  {
    return getAllPluginsViaPluginRegistration(null);
  }

  /**
   * Returns all plugins that can be found via the PluginRegistration mechanism
   * for the underlying interface definition using the given class loader as
   * preferred class loader for the Java ServiceLoader lookup.
   * 
   * @param loader The preferred class loader (may be null).
   */
  public ClassRegistry<TInterface> getAllPluginsViaPluginRegistration(ClassLoader loader)
  {
    return autoLoadPlugins(loader);
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the instance that is specified in the first 
   * <strong>META-INF/services/full-qualified-TInterface-name</strong>
   * file on the classpath.
   * 
   * @param classLoader An optional class loader to lookup the service implementors (may be null). 
   * @return The instance of the configured TInterface service.
   * @throws ConfigurationException If no implementation class was configured or it cannot be instantiated.
   */
  protected TInterface loadInstanceFromService(final ClassLoader classLoader) throws ConfigurationException
  {
    return loadInstance(classLoader, (Class<TInterface>)null);
  }

  /**
   * Returns the instance for the given className. It uses the given
   * classLoader to find the class.
   * 
   * @param classLoader The class loader to use for loading the class. If null the class loader
   * of this class will be used.
   * @param className The name of the class that implements interface TInterface (must not be null).
   * @return The instance of the TInterface implementing class specified by the className.
   * @throws ConfigurationException If the class cannot be loaded or instantiated.
   */
  @SuppressWarnings("unchecked")
  protected TInterface loadInstance(final ClassLoader classLoader, final String className) throws ConfigurationException
  {
    Class<TInterface> implClass;
    ClassLoader loader;

    loader = determineClassLoader(classLoader);
    try
    {
      implClass = (Class<TInterface>)loader.loadClass(className);
    }
    catch (Exception ex)
    {
      throw new ConfigurationException(ex);
    }
    return loadInstance(classLoader, implClass);
  }

  /**
   * Returns the instance for the given alias. It uses the given
   * classLoader to automatically load all alias to implementation class mappings.
   * 
   * @param classLoader The class loader to use for loading the class mapping. 
   * If null the context classloader or the class loader of this class will be used.
   * @param aliasName The name of the alias that is associated to a class that implements interface TInterface (must not be null).
   * @return The instance of the TInterface implementing class specified by the aliasName or null if not found.
   */
  protected TInterface loadInstanceByAlias(final ClassLoader classLoader, final String aliasName)
  {
    ClassLoader loader;
    ClassRegistry<TInterface> plugins;

    loader = determineClassLoader(classLoader);
    plugins = autoLoadPlugins(loader);
    return plugins.newInstance(aliasName);
  }

  /**
   * Returns the and instance of the first found class definition. It uses the given
   * classLoader to automatically load all implementation class mappings via the
   * {@link PluginRegistration} mechanism.
   * 
   * @param classLoader The class loader to use for loading the class mapping. 
   * If null the context classloader or the class loader of this class will be used.
   * @return The instance of the TInterface implementing class specified by the aliasName or null if not found.
   */
  protected TInterface loadFirstFoundInstance(final ClassLoader classLoader)
  {
    ClassLoader loader;
    ClassRegistry<TInterface> plugins;
    ClassInfo<TInterface> classInfo;

    loader = determineClassLoader(classLoader);
    plugins = autoLoadPlugins(loader);
    classInfo = plugins.getFirst();
    if (classInfo == null)
    {
      return null;
    }
    return classInfo.createInstance();
  }

  /**
   * Returns the instance for the given class.
   * 
   * @param classLoader An optional classloader to be used for looking up the interface implementations.
   * @param implClass The class to instantiate. If null the service loader 
   * mechanism will be used (looking for file META-INF/services/TInterface).
   * @return The instance of the TInterface specified by given class.
   * @throws ConfigurationException If the class cannot be instantiated.
   */
  protected TInterface loadInstance(final ClassLoader classLoader, final Class<TInterface> implClass) throws ConfigurationException
  {
    if (implClass == null)
    {
      return createInstanceFromServiceDeclaration(classLoader);
    }
    try
    {
      return implClass.newInstance();
    }
    catch (Exception ex)
    {
      throw new ConfigurationException(ex);
    }
  }

  protected TInterface createInstanceFromServiceDeclaration(final ClassLoader classLoader)
  {
    ServiceLoader<TInterface> serviceLoader;

    serviceLoader = ServiceLoader.load(getInterfaceClass(), determineClassLoader(classLoader));
    for (TInterface instance : serviceLoader)
    {
      // Just use the first that was found
      return instance;
    }
    return null;
  }

  protected ClassRegistry<TInterface> autoLoadPlugins(ClassLoader loader)
  {
    PluginCollector<TInterface> pluginCollector;

    pluginCollector = new PluginCollector<TInterface>(getInterfaceClass());
    return pluginCollector.loadPluginsViaPluginRegistration(loader);
  }

  protected ClassLoader determineClassLoader(ClassLoader classLoader)
  {
    return pluginUtil().getClassLoader(classLoader);
  }

  protected String getParameterNames()
  {
    String result = "";

    if (getClassNameConfigParamName() != null)
    {
      result += getClassNameConfigParamName();
    }
    if (getAliasConfigParamName() != null)
    {
      if (result.length() > 0)
      {
        result += " or ";
      }
      result += getAliasConfigParamName();
    }
    return result;
  }

  /**
   * The return value of this method defines if the first entry in the found
   * plugins should be returned as default if no explicit alias name has been
   * provided in the configuration settings.
   * <p>
   * Here it returns always false - so no default will be provided.
   * <br>
   * Subclasses may override this method to alter the behavior.  
   */
  protected boolean defaultToFirstEntry()
  {
    return false;
  }

  protected Class<TInterface> getInterfaceClass()
  {
    return this.interfaceClass;
  }

  protected String getClassNameConfigParamName()
  {
    return this.classNameConfigParamName;
  }

  protected String getAliasConfigParamName()
  {
    return this.aliasConfigParamName;
  }

  protected PluginUtil pluginUtil()
  {
    return PluginUtil.current();
  }
}
