// ===========================================================================
// CONTENT  : CLASS PluginRegistry
// AUTHOR   : M.Duchrow
// VERSION  : 2.1 - 21/05/2017
// HISTORY  :
//  25/02/2007  mdu  CREATED
//  23/11/2014  mdu   changed --> generic typing
//  21/05/2017  mdu   added   --> loadPluginsByServiceLoader()
//
// Copyright (c) 2007-2017, by M.Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.plugin;

import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Array;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.ServiceLoader;

import org.pfsw.file.Classpath;
import org.pfsw.file.ClasspathElement;
import org.pfsw.file.FileUtil;
import org.pfsw.file.PropertiesFileLoader;
import org.pfsw.reflect.ClassInfo;
import org.pfsw.reflect.Dynamic;
import org.pfsw.reflect.ReflectUtil;
import org.pfsw.text.StringPattern;
import org.pfsw.text.StringUtil;
import org.pfsw.util.CollectionUtil;
import org.pfsw.util.NamedValueList;
import org.pfsw.util.OrderedProperties;

/**
 * A special registry that is capable to hold any number of plug-ins of
 * any type.
 *
 * @author M.Duchrow
 * @version 2.0
 */
public class PluginRegistry
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  protected static final boolean DEBUG = "true".equals(System.getProperty("org.pf.plugin.PluginRegistry.debug"));
  protected static final String DEBUG_PREFIX = "org.pf.plugin.PluginRegistry: ";

  protected static final String PLUGIN_DEF_SEPARATOR = ";";
  protected static final String OPTIONS_START = "[";
  protected static final String OPTIONS_END = "]";
  protected static final StringPattern PLUGIN_OPTIONS_PATTERN = StringPattern.create(OPTIONS_START + "*" + OPTIONS_END + "?*");

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private NamedValueList<PluginHolder> registry;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values (i.e. an empty registry).
   */
  public PluginRegistry()
  {
    super();
    setRegistry(new NamedValueList<PluginHolder>());
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Load all plug-ins that are defined in any library in the specified file.
   * 
   * @param filename The name of the file where the plug-in definition is looked-up.
   * @return The number of plug-ins that have been loaded to the registry
   */
  public int loadPluginsFrom(String filename)
  {
    int sum;

    sum = readPlugins(filename, getPluginClasspath());
    if (DEBUG)
    {
      debug("Registered " + sum + " plug-ins");
    }
    return sum;
  }

  /**
   * Utilizes the Java {@link ServiceLoader} mechanism to load all plugins
   * implementing the given type into this registry.
   * 
   * @param pluginType The type (i.e. interface) of the plug-in (must not be null).
   * @param classLoaders An optional array of class loaders. The first non-null classloader
   * will be used for the lookup. If none is provided the thread context classloader and 
   * the class loader of this class will be tried. 
   * @return The number of plug-ins that have been loaded to the registry
   */
  public <T> int loadPluginsByServiceLoader(Class<T> pluginType, ClassLoader... classLoaders)
  {
    int count = 0;
    ClassLoader classLoader;
    ServiceLoader<T> found;

    classLoader = plugUtil().getClassLoader(classLoaders);
    found = ServiceLoader.load(pluginType, classLoader);
    for (T plugin : found)
    {
      registerPlugin(pluginType, plugin);
      count++;
    }
    return count;
  }

  /**
   * Returns the plug-in with the specified ID or null
   * if no such plug-in can be found in the registry.
   * If more than one plug-in is registered under the same ID the result
   * is the first one found.
   * 
   * @param pluginId The identifier of the plug-in (must not be null)
   * @return An instance of the found plug-in or null
   */
  public Object getPlugin(String pluginId)
  {
    PluginHolder pluginHolder;

    pluginHolder = getRegistry().valueAt(pluginId);
    if (pluginHolder != null)
    {
      return pluginHolder.getInstance();
    }
    return null;
  }

  /**
   * Returns the plug-in with the specified ID and the given type or null
   * if no such plug-in can be found in the registry.
   * 
   * @param pluginId The identifier of the plug-in (must not be null).
   * @param pluginType The type of the plug-in (must not be null).
   * @return An instance of the found plug-in or null.
   */
  public <T> T getPlugin(String pluginId, Class<T> pluginType)
  {
    PluginHolder pluginHolder;

    for (int i = 0; i < getRegistry().size(); i++)
    {
      if (pluginId.equals(getRegistry().nameAt(i)))
      {
        pluginHolder = getRegistry().valueAt(i);
        if (pluginHolder.isPluginType(pluginType))
        {
          return pluginHolder.getInstance(pluginType);
        }
      }
    }
    return null;
  }

  /**
   * Returns all plug-ins that are of the specified type. That also implies
   * plug-ins that are subclasses or implementors of the given type.
   * 
   * @param pluginType The type the plug-ins to look for must comply with.
   * @return An array of found plug-in instances (never null).
   */
  @SuppressWarnings("unchecked")
  public <T> T[] getPluginsOfType(final Class<T> pluginType)
  {
    List<T> foundPlugins;
    PluginHolder pluginHolder;
    Iterator<PluginHolder> iter;
    T plugin;

    foundPlugins = new ArrayList<T>(getRegistry().size());
    iter = getRegistry().values().iterator();
    while (iter.hasNext())
    {
      pluginHolder = iter.next();
      if (pluginHolder.isPluginType(pluginType))
      {
        plugin = pluginHolder.getInstance(pluginType);
        if (plugin != null)
        {
          foundPlugins.add(plugin);
        }
      }
    }

    if (foundPlugins.size() == 0)
    {
      return (T[])Array.newInstance(pluginType, 0);
    }
    return coll().toArray(foundPlugins, pluginType);
  }

  /**
   * Returns the IDs of all plug-ins that are of the specified type. That also implies
   * plug-ins that are subclasses or implementors of the given type.
   * 
   * @param pluginType The type the plug-ins to look for must comply with.
   * @return An array of IDs (never null).
   */
  public String[] getPluginIDsOfType(Class<?> pluginType)
  {
    List<String> foundIDs;
    PluginHolder pluginHolder;

    foundIDs = new ArrayList<String>(getRegistry().size());
    for (int i = 0; i < getRegistry().size(); i++)
    {
      pluginHolder = getRegistry().valueAt(i);
      if (pluginHolder.isPluginType(pluginType))
      {
        foundIDs.add(getRegistry().nameAt(i));
      }
    }
    if (foundIDs.isEmpty())
    {
      return StringUtil.EMPTY_STRING_ARRAY;
    }
    return str().asStrings(foundIDs);
  }

  /**
   * Remove all registered plugins.
   */
  public void clear()
  {
    getRegistry().clear();
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected boolean registerPlugin(PluginHolder pluginHolder)
  {
    String id;
    Object plugin;

    plugin = pluginHolder.getInstance();
    if (plugin == null)
    {
      return false;
    }
    id = getIdOfPlugin(plugin, pluginHolder.getExternalId());
    getRegistry().add(id, pluginHolder);
    return true;
  }

  protected <T> void registerPlugin(@SuppressWarnings("unused") Class<T> pluginType, T plugin)
  {
    String className;
    ClassInfo<T> classInfo;
    PluginHolder pluginHolder;
    String id;

    className = plugin.getClass().getName();
    id = getIdOfPlugin(plugin, className);
    classInfo = new ClassInfo<T>(className);
    pluginHolder = new PluginHolder(id, classInfo);
    pluginHolder.setSoleInstance(plugin);
    pluginHolder.isSingleton(true);
    registerPlugin(pluginHolder);
  }

  protected String getIdOfPlugin(Object plugin, String defaultId)
  {
    String id = null;
    Method method;
    IIdentifiablePlugin idPlugin;

    if (plugin instanceof IIdentifiablePlugin)
    {
      idPlugin = (IIdentifiablePlugin)plugin;
      id = idPlugin.getPluginId();
    }
    else
    {
      method = find_getPluginId_Method(plugin);
      if (method != null)
      {
        id = (String)Dynamic.invoke(plugin, method, null);
      }
    }
    if (id == null)
    {
      id = defaultId;
    }
    return id;
  }

  /**
   * Read all classes specified in the given file, if it exists in the specified
   * classpath element. Add those classes to the registry if they can be
   * instantiated and are of the correct type.
   *  
   * @param filename The name of the properties file that contains the class definitions
   * @param classpath The classpath the specified filename to be looked up in
   * 
   * @return The number of plug-ins that have been added to the registry
   */
  protected int readPlugins(String filename, Classpath classpath)
  {
    ClasspathElement[] elements;
    int sum = 0;

    if (DEBUG)
    {
      debug("Lookup file <" + filename + "> in <" + classpath + ">");
    }
    elements = classpath.elementsContaining(filename);
    for (int i = 0; i < elements.length; i++)
    {
      if (DEBUG)
      {
        debug("Found file <" + filename + "> in <" + elements[i].getName() + ">");
      }
      sum += readPluginsFrom(elements[i], filename);
    }
    return sum;
  }

  @SuppressWarnings("resource")
  protected int readPluginsFrom(ClasspathElement element, String filename)
  {
    InputStream stream = null;
    OrderedProperties<?> properties;

    try
    {
      stream = element.open(filename);
      properties = PropertiesFileLoader.loadFullPropertiesFile(stream);
    }
    catch (IOException e)
    {
      if (DEBUG)
      {
        debug("Exception reading file <" + filename + "> : " + e.toString());
      }
      return 0;
    }
    finally
    {
      fileUtil().close(stream);
    }
    return addDefinitionsToRegistry(properties);
  }

  protected int addDefinitionsToRegistry(OrderedProperties<?> definitions)
  {
    String[] keys;
    int sum = 0;

    keys = definitions.getPropertyNames();
    for (int i = 0; i < keys.length; i++)
    {
      if (parseAndAddToRegistry(keys[i], definitions.getProperty(keys[i])))
      {
        sum++;
      }
    }
    return sum;
  }

  @SuppressWarnings("rawtypes")
  protected boolean parseAndAddToRegistry(String id, String pluginDefinition)
  {
    PluginHolder holder;
    String[] parts;
    ClassInfo<?> classInfo;
    String className;
    String classOptions = null;
    Properties config;
    String[] keyValue;

    parts = str().parts(pluginDefinition, PLUGIN_DEF_SEPARATOR);
    if (str().isNullOrEmpty(parts))
    {
      return false;
    }
    className = parts[0];
    if (DEBUG)
    {
      debug("Found <" + className + "> with id <" + id + ">");
    }
    if (PLUGIN_OPTIONS_PATTERN.matches(className))
    {
      classOptions = str().getDelimitedSubstring(className, OPTIONS_START, OPTIONS_END);
      className = str().cutHead(className, OPTIONS_END);
    }
    classInfo = new ClassInfo(className);
    if (!validateClass(classInfo))
    {
      return false;
    }
    holder = new PluginHolder(id, classInfo);
    if ("1".equals(classOptions))
    {
      holder.isSingleton(true);
    }
    config = new Properties();
    for (int i = 1; i < parts.length; i++)
    {
      keyValue = str().splitNameValue(parts[i], "=");
      if (!str().isNullOrEmpty(keyValue))
      {
        config.setProperty(keyValue[0], keyValue.length > 1 ? keyValue[1] : "");
      }
    }
    holder.setConfiguration(config);
    return registerPlugin(holder);
  }

  protected boolean validateClass(ClassInfo<?> classInfo)
  {
    if (validateClassFound(classInfo))
    {
      return validateInstanceCreation(classInfo);
    }
    return false;
  }

  protected boolean validateClassFound(ClassInfo<?> classInfo)
  {
    Class<?> aClass;

    aClass = classInfo.getClassObject();
    if (aClass == null)
    {
      LoggerProvider.getLogger().logError("Plug-in class <" + classInfo.getClassName() + "> not found!");
      return false;
    }
    return true;
  }

  protected boolean validateInstanceCreation(ClassInfo<?> classInfo)
  {
    try
    {
      classInfo.newInstance();
      return true;
    }
    catch (Exception ex)
    {
      LoggerProvider.getLogger().logError("Plug-in class <" + classInfo.getClassName() + "> cannot be instantiated.", ex);
      return false;
    }
  }

  protected Method find_getPluginId_Method(Object object)
  {
    return ReflectUtil.current().findMethod(object.getClass(), IIdentifiablePlugin.ID_METHOD_NAME, null, Modifier.PUBLIC);
  }

  /**
   * Returns the classpath that is used to lookup any plug-in information files.
   */
  protected Classpath getPluginClasspath()
  {
    return PluginUtil.current().getPluginClasspath();
  }

  protected NamedValueList<PluginHolder> getRegistry()
  {
    return registry;
  }

  protected void setRegistry(NamedValueList<PluginHolder> newValue)
  {
    registry = newValue;
  }

  protected FileUtil fileUtil()
  {
    return FileUtil.current();
  }

  protected StringUtil str()
  {
    return StringUtil.current();
  }

  protected CollectionUtil coll()
  {
    return CollectionUtil.current();
  }

  protected PluginUtil plugUtil()
  {
    return PluginUtil.current();
  }

  protected void debug(String text)
  {
    LoggerProvider.getLogger().logDebug(DEBUG_PREFIX + text);
  }
}
