// ===========================================================================
// CONTENT  : CLASS PluginUtil
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.3 - 29/01/2017
// HISTORY  :
//  27/12/2004  mdu  CREATED
//	25/03/2006	mdu		changed	-->	Extracted determination of lookup path to FileUtil
//	02/03/2007	mdu		changed	-->	removed dead code
//  29/01/2017  mdu   added   --> getClassLoader()
//
// Copyright (c) 2004-2017, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.plugin;

import org.pfsw.file.Classpath;
import org.pfsw.file.FileUtil;
import org.pfsw.util.CollectionUtil;
import org.pfsw.util.ObjectNotFoundException;

/**
 * Provides helper methods around the PF plug-in mechanism.
 *
 * @author Manfred Duchrow
 * @version 1.3
 */
public class PluginUtil
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  public static final PluginUtil PU = new PluginUtil();

  private static final String SYS_PROP_PLUGIN_PATH_LEGACY = "org.pf.plugin.path";
  public static final String SYS_PROP_PLUGIN_PATH = "org.pfsw.plugin.path";

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================

  /**
   * Returns the only instance this class supports (design pattern "Singleton")
   */
  public static PluginUtil current()
  {
    return PluginUtil.PU;
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  protected PluginUtil()
  {
    super();
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the classpath that is used to lookup any plug-in information files.
   */
  public Classpath getPluginClasspath()
  {
    String sysPropName;

    sysPropName = (System.getProperty(SYS_PROP_PLUGIN_PATH) == null) ? SYS_PROP_PLUGIN_PATH_LEGACY : SYS_PROP_PLUGIN_PATH;
    return fileUtil().getLookupPath(sysPropName);
  }

  /**
   * Returns the first non-null classloader from the following order:
   * <ol>
   *   <li>suggestedClassLoaders</li>
   *   <li>The current thread's context class loader</li>
   *   <li>the class loader of this class</li>
   *   <li>the system class loader</li>
   * </ol>
   * or throws an IllegalStateException if no classloader can be found.
   */
  public ClassLoader getClassLoader(final ClassLoader... suggestedClassLoaders)
  {
    ClassLoader[] classLoaders;

    classLoaders = new ClassLoader[0];
    classLoaders = collUtil().append(classLoaders, suggestedClassLoaders);
    classLoaders = collUtil().append(classLoaders, Thread.currentThread().getContextClassLoader(), getClass().getClassLoader(), ClassLoader.getSystemClassLoader());
    try
    {
      return collUtil().getFirstNonNull(classLoaders);
    }
    catch (@SuppressWarnings("unused") ObjectNotFoundException ex)
    {
      throw new IllegalStateException("No classloader found!");
    }
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected FileUtil fileUtil()
  {
    return FileUtil.current();
  }

  protected CollectionUtil collUtil()
  {
    return CollectionUtil.current();
  }

}
