// ===========================================================================
// CONTENT  : CLASS Base64UrlEncodedRandomStringGenerator
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 25/04/2020
// HISTORY  :
//  25/04/2020  mdu  CREATED
//
// Copyright (c) 2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.security;

import org.pfsw.bif.ByteSequenceGenerator;
import org.pfsw.bif.conversion.IBase64Encoder;
import org.pfsw.text.StringGenerator;
import org.pfsw.util.Base64UrlEncoder;
import org.pfsw.util.RandomByteSequenceGenerator;

/**
 * A generator for random strings that uses a newly generated random byte sequence
 * and returns it base64url encoded (without padding). 
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class Base64UrlEncodedRandomStringGenerator implements StringGenerator
{
  /**
   * Creates a new instance that creates new strings based on a {@link ByteSequenceGenerator}
   * with the given length for the by sequences. 
   * 
   * @param byteSequenceLength The length of the random byte sequences to be used. 
   */
  public static Base64UrlEncodedRandomStringGenerator create(int byteSequenceLength)
  {
    return new Base64UrlEncodedRandomStringGenerator(byteSequenceLength);
  }

  /**
   * Creates a new instance that creates new strings based on the given {@link ByteSequenceGenerator}.
   * 
   * @param byteSequenceGenerator The generator for new byte sequences. 
   */
  public static Base64UrlEncodedRandomStringGenerator create(ByteSequenceGenerator byteSequenceGenerator)
  {
    return new Base64UrlEncodedRandomStringGenerator(byteSequenceGenerator);
  }

  private final IBase64Encoder base64Encoder = Base64UrlEncoder.withoutPadding();
  private final ByteSequenceGenerator byteSequenceGenerator;

  /**
   * Creates a new instance that creates new strings based on a {@link ByteSequenceGenerator}
   * with the given length for the by sequences. 
   * 
   * @param byteSequenceLength The length of the random byte sequences to be used. 
   */
  public Base64UrlEncodedRandomStringGenerator(int byteSequenceLength)
  {
    this(RandomByteSequenceGenerator.create(byteSequenceLength));
  }

  /**
   * Creates a new instance that creates new strings based on the given {@link ByteSequenceGenerator}.
   * 
   * @param byteSequenceGenerator The generator for new byte sequences. 
   */
  public Base64UrlEncodedRandomStringGenerator(ByteSequenceGenerator byteSequenceGenerator)
  {
    super();
    this.byteSequenceGenerator = byteSequenceGenerator;
  }

  @Override
  public String generateString()
  {
    byte[] bytes;

    bytes = getByteSequenceGenerator().generateByteSequence();
    return getBase64Encoder().convert(bytes);
  }

  @Override
  public String generateString(int length)
  {
    String string;

    string = generateString();
    while (string.length() < length)
    {
      string += generateString();
    }
    return string.substring(0, length);
  }

  @Override
  public String toString()
  {
    return String.format("%s()", getClass().getSimpleName());
  }

  protected IBase64Encoder getBase64Encoder()
  {
    return this.base64Encoder;
  }

  protected ByteSequenceGenerator getByteSequenceGenerator()
  {
    return this.byteSequenceGenerator;
  }
}
