// ===========================================================================
// CONTENT  : CLASS SecurityUtil
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 18/08/2014
// HISTORY  :
//  18/08/2014  mdu  CREATED
//
// Copyright (c) 2014, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.security;

import static org.pfsw.text.StringUtil.*;
import static org.pfsw.util.SysUtil.*;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.security.GeneralSecurityException;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.CertificateParsingException;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.pfsw.bif.exceptions.IORuntimeException;
import org.pfsw.bif.text.IStringPair;
import org.pfsw.security.crypto.CryptoConstants;
import org.pfsw.security.x509.GeneralName;
import org.pfsw.util.Base64Converter;
import org.pfsw.util.CollectionUtil;
import org.pfsw.util.IOUtil;

/**
 * This singleton supports utility methods around various security aspects.
 * It has a protected default constructor, so it can be extended!
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class SecurityUtil
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  public static final SecurityUtil SECU = new SecurityUtil();

  public static final String BASIC_CREDENTIALS_SEPARATOR = ":";

  public static final String AUTHENTICATION_SCHEME_BASIC = "Basic";

  public static final String AUTHENTICATION_SCHEME_DIGEST = "Digest";

  public static final String AUTHENTICATION_SCHEME_BEARER = "Bearer";

  private static final Pattern CERT_PATTERN = Pattern.compile("-+BEGIN\\s+.*CERTIFICATE[^-]*-+(?:\\s|\\r|\\n)+" + // Header
      "([a-z0-9+/=\\r\\n]+)" + // Base64 text
      "-+END\\s+.*CERTIFICATE[^-]*-+", // Footer
      Pattern.CASE_INSENSITIVE);

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Returns the only instance this class supports (design pattern "Singleton")
   */
  public static SecurityUtil current()
  {
    return SECU;
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  protected SecurityUtil()
  {
    super();
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Encodes the given username and password according to the BASIC authentication
   * scheme specification (RFC-2617).
   * <br>
   * That is base64(username:password). 
   * 
   * @param username The username (must not be null).
   * @param password The password (must not be null)
   * @return The base64 encoded concatenated username password.
   */
  public String encodeBASICCredentials(String username, String password)
  {
    String both;

    both = username + BASIC_CREDENTIALS_SEPARATOR + password;
    return Base64Converter.encode(both);
  }

  /**
   * Encodes the given username and password according to the BASIC authentication
   * scheme specification (RFC-2617).
   * <br>
   * That is base64(username:password). 
   * 
   * @param username The username (must not be null).
   * @param password The password (must not be null)
   * @param charset The charset to be used for the string encoding (must not be null).
   * @return The base64 encoded concatenated username and password.
   */
  public String encodeBASICCredentials(String username, String password, Charset charset)
  {
    String both;

    both = username + BASIC_CREDENTIALS_SEPARATOR + password;
    return Base64Converter.encodeToString(both.getBytes(charset));
  }

  /**
   * Decodes and splits up the value presented by a BASIC authentication 
   * header as to RFC-2617.
   * 
   * @param basicCredentials The base64 encoded concatenated username password string (must not be null). 
   * @param charset The charset to be used for the string encoding (must not be null).
   * @return A string pair where the first string contains the username and the second the password.
   * @throws IllegalArgumentException If the given string cannot be decoded or cannot be split into two sub-values.
   */
  public IStringPair decodeBASICCredentials(String basicCredentials, Charset charset)
  {
    byte[] bytes;
    String decoded;
    IStringPair pair;

    bytes = Base64Converter.decode(basicCredentials);
    decoded = new String(bytes, charset);
    pair = SU.splitStringPair(decoded, BASIC_CREDENTIALS_SEPARATOR);
    if (SU.isNullOrBlank(pair.getString1()) || SU.isNullOrBlank(pair.getString2()))
    {
      throw new IllegalArgumentException("Failed to split up the given credentials.");
    }
    return pair;
  }

  /**
   * Decodes and splits up the value presented by a BASIC authentication 
   * header as to RFC-2617.
   * 
   * @param basicCredentials The base64 encoded concatenated username password string (must not be null). 
   * @return A string pair where the first string contains the username and the second the password.
   * @throws IllegalArgumentException If the given string cannot be decoded or cannot be split into two sub-values.
   */
  public IStringPair decodeBASICCredentials(String basicCredentials)
  {
    String decoded;
    IStringPair pair;

    decoded = Base64Converter.decodeToString(basicCredentials);
    pair = SU.splitStringPair(decoded, BASIC_CREDENTIALS_SEPARATOR);
    if (SU.isNullOrBlank(pair.getString1()) || SU.isNullOrBlank(pair.getString2()))
    {
      throw new IllegalArgumentException("Failed to split up the given credentials.");
    }
    return pair;
  }

  /**
   * Creates the full authorization header value for the HTTP BASIC authentication
   * scheme as defined in RFC-2617.
   * Username and password will be concatenated and base64 encoded.
   * 
   * @param username The username (must not be null).
   * @param password The password (must not be null)
   * @return The complete header value (e.g. "Basic Q29yb2xpbmE6N1J1YmJlcmJvYXQyNQ==")
   */
  public String createBASICAuthorization(String username, String password)
  {
    String credentials;

    credentials = encodeBASICCredentials(username, password);
    return AUTHENTICATION_SCHEME_BASIC + " " + credentials;
  }

  /**
   * Creates the full authorization header value for the BEARER authentication
   * scheme as defined in RFC-6750.
   * 
   * @param token The token that will be base64 encoded by this method (must not be null).
   * @return The complete header value (e.g. "Bearer cy9aJsKnPX1+SDIjJz9ERlYoLyY9U0HDvFNEQEo34oKsesOfZGpkag==")
   */
  public String createBEARERAuthorization(byte[] token)
  {
    String encoded;

    encoded = Base64Converter.encodeToString(token);
    return AUTHENTICATION_SCHEME_BEARER + " " + encoded;
  }

  /**
   * Creates the full authorization header value for the BEARER authentication
   * scheme as defined in RFC-6750.
   * 
   * @param token The token that will be base64 encoded by this method (must not be null).
   * @param charset The charset to be used for the string encoding (must not be null).
   * @return The complete header value (e.g. "Bearer cy9aJsKnPX1+SDIjJz9ERlYoLyY9U0HDvFNEQEo34oKsesOfZGpkag==")
   */
  public String createBEARERAuthorization(String token, Charset charset)
  {
    return createBEARERAuthorization(token.getBytes(charset));
  }
  
  public List<GeneralName> getSubjectAlternativeNamesFrom(X509Certificate serverCertificate) 
  {
    Collection<List<?>> alternativeNames;
    List<GeneralName> result = new ArrayList<GeneralName>();
    Integer asn1Tag = null;
    String asn1Value;
    
    try
    {
      alternativeNames = serverCertificate.getSubjectAlternativeNames();
    }
    catch (CertificateParsingException e)
    {
      throw new SecurityRuntimeException(e);
    }
    for (List<?> nameList : alternativeNames)
    {
      for (Object token : nameList)
      {
        if (token instanceof Integer)
        {
          asn1Tag = (Integer)token;
        }
        else
        {
          asn1Value = (String)token;
          if (asn1Tag != null)
          {
            result.add(GeneralName.create(asn1Tag, asn1Value));            
          }
          asn1Tag = null;
        }
      }
    }
    return result;
  }

  /**
   * Read all certificates from the given PKCS#7 stream.
   * The given stream will be closed in any case after this call.
   * 
   * @param inStream A stream that is expected to contain X.509 certificates in PKCS#7 format.
   * @return The certificates read from the given PKCS#7 stream.
   * @throws CertificateException For any case of format or IO errors.
   */
  public Collection<X509Certificate> readX509CertificatesFromPKCS7(final InputStream inStream) throws CertificateException
  {
    CertificateFactory certFactory;
    Collection<? extends Certificate> certificates;
    Collection<X509Certificate> result;

    try
    {
      certFactory = CertificateFactory.getInstance(CryptoConstants.CERT_STANDARD_X509);
      certificates = certFactory.generateCertificates(inStream);
      result = new ArrayList<X509Certificate>(certificates.size());
      for (Certificate certificate : certificates)
      {
        result.add((X509Certificate)certificate);
      }
      return result;
    }
    finally
    {
      SYSU.close(inStream);
    }
  }

  /**
   * Reads a certificate from the given PKCS#7 stream.
   * 
   * @param inStream A stream that is expected to contain X.509 certificates in PKCS#7 format.
   * @return The (first) X.509 certificate read from the given PKCS#7 stream.
   * @throws CertificateException For any case of format or IO errors.
   */
  public X509Certificate readX509CertificateFromPKCS7(final InputStream inStream) throws CertificateException
  {
    Collection<X509Certificate> result;

    result = readX509CertificatesFromPKCS7(inStream);
    if (result.isEmpty())
    {
      return null;
    }
    return result.iterator().next();
  }

  /**
   * Reads a certificate from the given PKCS#7 file.
   * 
   * @param file A file that is expected to contain X.509 certificates in PKCS#7 format.
   * @return The (first) X.509 certificate read from the given PKCS#7 file.
   * @throws CertificateException For any case of format or IO errors.
   */
  public X509Certificate readX509CertificateFromPKCS7(final File file) throws CertificateException
  {
    FileInputStream inStream;

    try
    {
      inStream = new FileInputStream(file);
    }
    catch (FileNotFoundException ex)
    {
      throw new CertificateException("Certificate file " + file.getAbsolutePath() + " not found", ex);
    }
    return readX509CertificateFromPKCS7(inStream);
  }

  /**
   * Reads a certificate chain from the given PEM file.
   * 
   * @param pemFile The file to read from (must not be null).
   * @return The full certificate chain found in the file.
   */
  public X509Certificate[] readX509CertificateChainFromPEM(File pemFile)
  {
    String contents;
    Matcher matcher;
    CertificateFactory certificateFactory;
    List<X509Certificate> certificates;

    try
    {
      contents = IOUtil.current().readTextFrom(new FileReader(pemFile));
    }
    catch (IOException e)
    {
      throw new IORuntimeException(e, "Cannot read PEM file: %s", pemFile.getAbsolutePath());
    }
    matcher = CERT_PATTERN.matcher(contents);
    try
    {
      certificateFactory = CertificateFactory.getInstance("X.509");
      certificates = new ArrayList<X509Certificate>();

      int start = 0;
      while (matcher.find(start))
      {
        byte[] buffer = Base64Converter.decode(matcher.group(1));
        certificates.add((X509Certificate)certificateFactory.generateCertificate(new ByteArrayInputStream(buffer)));
        start = matcher.end();
      }

      return CollectionUtil.current().toArray(certificates);
    }
    catch (GeneralSecurityException e)
    {
      throw new SecurityRuntimeException(e, "Error when trying to read certificate chain from PEM file: %s", pemFile.getAbsolutePath());
    }
  }
}
