// ===========================================================================
// CONTENT  : CLASS SignatureAlgorithm
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 27/09/2014
// HISTORY  :
//  27/09/2014  mdu  CREATED
//
// Copyright (c) 2014, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.security;

import java.security.NoSuchAlgorithmException;
import java.security.Signature;
import java.util.ArrayList;
import java.util.List;

import org.pfsw.text.StringUtil;

/**
 * Each instance of this class represents a signature algorithm. 
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class SignatureAlgorithm
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final StringUtil SU = StringUtil.current();

  public static final SignatureAlgorithm MD5_WITH_RSA;
  public static final SignatureAlgorithm SHA1_WITH_DSA;
  public static final SignatureAlgorithm SHA1_WITH_RSA;
  public static final SignatureAlgorithm SHA256_WITH_RSA;
  public static final SignatureAlgorithm SHA384_WITH_RSA;
  public static final SignatureAlgorithm SHA512_WITH_RSA;
  public static final SignatureAlgorithm SHA1_WITH_ECDSA;
  public static final SignatureAlgorithm SHA256_WITH_ECDSA;
  public static final SignatureAlgorithm SHA384_WITH_ECDSA;
  public static final SignatureAlgorithm SHA512_WITH_ECDSA;

  private static final List<SignatureAlgorithm> REGISTRY;

  static
  {
    REGISTRY = new ArrayList<SignatureAlgorithm>();

    MD5_WITH_RSA = internalCreate("MD5withRSA", null);
    SHA1_WITH_DSA = internalCreate("SHA1withDSA", null);
    SHA1_WITH_RSA = internalCreate("SHA1withRSA", null);
    SHA256_WITH_RSA = internalCreate("SHA256withRSA", "RS256");
    SHA384_WITH_RSA = internalCreate("SHA384withRSA", "RS384");
    SHA512_WITH_RSA = internalCreate("SHA512withRSA", "RS512");
    SHA1_WITH_ECDSA = internalCreate("SHA1withECDSA", null);
    SHA256_WITH_ECDSA = internalCreate("SHA256withECDSA", "ES256");
    SHA384_WITH_ECDSA = internalCreate("SHA384withECDSA", "ES384");
    SHA512_WITH_ECDSA = internalCreate("SHA512withECDSA", "ES512");
  }

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private final String name;
  private final String jwaName;

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  public static SignatureAlgorithm findByName(String algorithmName)
  {
    if (algorithmName != null)
    {
      for (SignatureAlgorithm sigAlg : REGISTRY)
      {
        if (algorithmName.equals(sigAlg.getName()))
        {
          return sigAlg;
        }
      }
    }
    return null;
  }

  public static SignatureAlgorithm findByJWAName(String jwaName)
  {
    if (jwaName != null)
    {
      for (SignatureAlgorithm sigAlg : REGISTRY)
      {
        if (jwaName.equals(sigAlg.getJWAName()))
        {
          return sigAlg;
        }
      }
    }
    return null;
  }

  /**
   * Registers a SignatureAlgorithm for re-use.
   * @return true if the algorithm has been added to the registry, false if 
   * it has not been added (e.g. because it was a duplicate).
   */
  public static boolean register(SignatureAlgorithm algorithm)
  {
    if (algorithm == null)
    {
      return false;
    }
    if (REGISTRY.contains(algorithm))
    {
      return false;
    }
    REGISTRY.add(algorithm);
    return true;
  }

  private static SignatureAlgorithm internalCreate(String algorithmName, String jwaName)
  {
    SignatureAlgorithm algorithm;

    algorithm = new SignatureAlgorithm(algorithmName, jwaName);
    register(algorithm);
    return algorithm;
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Creates the SignatureAlgorithm object for the given name.
   * 
   * @param algorithmName The name of the signature algorithm (must not be null).
   * @throws SignatureAlgorithm If algorithmName is null or blank.
   */
  public SignatureAlgorithm(String algorithmName)
  {
    this(algorithmName, null);
  }

  /**
   * Creates the SignatureAlgorithm object for the given names.
   * 
   * @param algorithmName The name of the signature algorithm (must not be null).
   * @param jwaName An optional name from the JWA specification (may be null).
   * @throws SignatureAlgorithm If algorithmName is null or blank.
   */
  public SignatureAlgorithm(String algorithmName, String jwaName)
  {
    super();
    if (SU.isNullOrBlank(algorithmName))
    {
      throw new IllegalArgumentException("The signature algorithm name must not be null or blank!");
    }
    this.name = algorithmName;
    this.jwaName = jwaName;
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the signature algorithm name.
   */
  public String getName()
  {
    return name;
  }

  /**
   * Returns the JSON Web Algorithm (JWA) name.
   * 
   * @return The name form the JWA specification or null if this algorithm has no JWA name.
   */
  public String getJWAName()
  {
    return jwaName;
  }

  /**
   * Returns the signature object or throws an NoSuchAlgorithmException if
   * this signature algorithm is not supported.
   */
  public Signature asSignature() throws NoSuchAlgorithmException
  {
    return Signature.getInstance(getName());
  }

  @Override
  public int hashCode()
  {
    return getName().hashCode();
  }

  @Override
  public boolean equals(Object obj)
  {
    if (obj instanceof SignatureAlgorithm)
    {
      SignatureAlgorithm other = (SignatureAlgorithm)obj;
      getName().equals(other.getName());
    }
    return false;
  }

  @Override
  public String toString()
  {
    return String.format("SignatureAlgorithm(%s)", getName());
  }
}
