// ===========================================================================

// CONTENT  : CLASS AutoAuthenticationManager
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 12/09/2004
// HISTORY  :
//  12/09/2004  mdu  CREATED
//
// Copyright (c) 2004, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.security.authentication;

import java.net.Authenticator;
import java.net.PasswordAuthentication;
import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.pfsw.text.StringUtil;
import org.pfsw.text.Version;

/**
 * This manager class is responsible to hold a registry of LocationCredentials
 * objects. With the method aboutToAccess( URL ) it selects the appropriate 
 * Authenticator from its regestry and sets in as default into the 
 * java.net.Authenticator class. That allows to do automatic authentication for
 * different (URLs).
 * <br>
 * <b>Be aware that this code is not thread-safe!</b>
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class AutoAuthenticationManager extends Authenticator
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  protected static final Version HOST_SUPPORT_VERSION = new Version("1.4");
  protected static final Version JAVA_VERSION = new Version(System.getProperty("java.version"));

  // =========================================================================
  // CLASS VARIABLES
  // =========================================================================
  private static AutoAuthenticationManager soleInstance = new AutoAuthenticationManager();

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private Map<String, LocationCredentials> locationCredentials = new HashMap<String, LocationCredentials>();
  private LocationCredentials activeCredentials = null;
  private AbstractAuthenticator fallbackAuthenticator = null;

  // =========================================================================
  // PUBLIC CLASS METHODS
  // =========================================================================
  /**
   * Must be called once, to activate this authentication manager mechanism
   * within the java.net.Authenticator.
   * 
   * @return The single instance of this class
   */
  public static AutoAuthenticationManager install()
  {
    Authenticator.setDefault(instance());
    return instance();
  }

  /**
   * Can be called to remove this authentication manager from 
   * the java.net.Authenticator.
   */
  public static void uninstall()
  {
    Authenticator.setDefault(null);
  }

  /**
   * Returns the only instance this class supports (design pattern "Singleton")
   */
  public static AutoAuthenticationManager instance()
  {
    return soleInstance;
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  private AutoAuthenticationManager()
  {
    super();
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the fallback authenticator which will be called if no credentials
   * can be found in the registered credtentials.
   */
  public AbstractAuthenticator getFallbackAuthenticator()
  {
    return fallbackAuthenticator;
  }

  /**
   * Sets the fallback authenticator which will be called if no credentials
   * can be found in the registered credtentials.
   */
  public void setFallbackAuthenticator(AbstractAuthenticator newValue)
  {
    fallbackAuthenticator = newValue;
  }

  /**
   * This method must be called to activate the authenticator that provides
   * credentials to be able to access the given URL.
   * 
   * @param url The URL that might need authentication to be accessed
   * @return true if an authenticator for the given URL was activated
   */
  public boolean aboutToAccess(String url)
  {
    LocationCredentials credentials;

    credentials = findCredentialsFor(url);
    setActiveCredentials(credentials);
    return hasActiveCredentials();
  }

  /**
   * This method must be called to activate the authenticator that provides
   * credentials to be able to access the given URL.
   * 
   * @param url The URL that might need authentication to be accessed
   * @return true if an authenticator for the given URL was activated
   */
  public boolean aboutToAccess(URL url)
  {
    if (url == null)
    {
      setActiveCredentials(null);
      return hasActiveCredentials();
    }
    return aboutToAccess(url.toString());
  }

  /**
   * Registers the given location credentials under the specified id.
   * The id can be used to remove the credentials once again. 
   */
  public void register(String id, LocationCredentials credentials)
  {
    if ((!str().isNullOrBlank(id)) && credentials != null)
    {
      getLocationCredentials().put(id, credentials);
    }
  }

  /**
   * Removes the location credentials registered under the specified key from
   * the registry.
   * Returns the removed credentials or null if not found.
   */
  public LocationCredentials remove(String id)
  {
    if (str().isNullOrBlank(id))
    {
      return null;
    }
    return getLocationCredentials().remove(id);
  }

  /**
   * Removes all registered credentials and the fallback authenticator.
   */
  public void reset()
  {
    setLocationCredentials(new HashMap<String, LocationCredentials>());
    setFallbackAuthenticator(null);
    setActiveCredentials(null);
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  @Override
  protected PasswordAuthentication getPasswordAuthentication()
  {
    PasswordAuthentication auth;

    if (hasActiveCredentials())
    {
      auth = getActiveCredentials().getPasswordAuthentication();
    }
    else
    {
      auth = findPasswordAuthenticationForRealm();
    }
    if ((auth == null) && (getFallbackAuthenticator() != null))
    {
      fillFallbackAuthenticator();
      auth = getFallbackAuthenticator().getPasswordAuthentication();
    }
    setActiveCredentials(null);
    return auth;
  }

  protected PasswordAuthentication findPasswordAuthenticationForRealm()
  {
    LocationCredentials creds;

    creds = findCredentialsForRealm();
    if (creds == null)
    {
      return null;
    }
    return creds.getPasswordAuthentication();
  }

  /**
   * Returns the location credentials registered under the specified key or 
   * null if not found.
   */
  protected LocationCredentials getCredentials(String id)
  {
    if (str().isNullOrBlank(id))
    {
      return null;
    }
    return getLocationCredentials().get(id);
  }

  protected LocationCredentials findCredentialsFor(String url)
  {
    Iterator<LocationCredentials> iter;
    LocationCredentials creds;

    if (str().isNullOrBlank(url))
    {
      return null;
    }

    iter = getLocationCredentials().values().iterator();
    while (iter.hasNext())
    {
      creds = iter.next();
      if (creds.appliesToURL(url))
      {
        return creds;
      }
    }
    return null;
  }

  protected LocationCredentials findCredentialsForRealm()
  {
    if (isBasicHttp())
    {
      return findCredentialsForRealm(getRequestingPrompt());
    }
    return null;
  }

  protected LocationCredentials findCredentialsForRealm(String realm)
  {
    Iterator<LocationCredentials> iter;
    LocationCredentials creds;

    if (str().isNullOrBlank(realm))
    {
      return null;
    }

    iter = getLocationCredentials().values().iterator();
    while (iter.hasNext())
    {
      creds = iter.next();
      if (creds.appliesToRealm(realm))
      {
        return creds;
      }
    }
    return null;
  }

  protected void fillFallbackAuthenticator()
  {
    AbstractAuthenticator auth;

    auth = getFallbackAuthenticator();
    if (supportsHostNames())
    {
      auth.setRequestingHost(getRequestingHost());
    }
    else
    {
      auth.setRequestingHost(null);
    }
    auth.setRequestingPort(getRequestingPort());
    auth.setRequestingPrompt(getRequestingPrompt());
    auth.setRequestingProtocol(getRequestingProtocol());
    auth.setRequestingScheme(getRequestingScheme());
    auth.setRequestingSite(getRequestingSite());
  }

  protected boolean hasActiveCredentials()
  {
    return getActiveCredentials() != null;
  }

  protected boolean isBasicHttp()
  {
    return ("Basic".equalsIgnoreCase(getRequestingScheme()) && "http".equalsIgnoreCase(getRequestingProtocol()));
  }

  protected boolean supportsHostNames()
  {
    return !JAVA_VERSION.isLessThan(HOST_SUPPORT_VERSION);
  }

  protected Map<String, LocationCredentials> getLocationCredentials()
  {
    return locationCredentials;
  }

  protected void setLocationCredentials(Map<String, LocationCredentials> newValue)
  {
    locationCredentials = newValue;
  }

  protected LocationCredentials getActiveCredentials()
  {
    return activeCredentials;
  }

  protected void setActiveCredentials(LocationCredentials newValue)
  {
    activeCredentials = newValue;
  }

  protected StringUtil str()
  {
    return StringUtil.current();
  }

}
