// ===========================================================================
// CONTENT  : CLASS LocationCredentials
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 12/09/2004
// HISTORY  :
//  12/09/2004  mdu  CREATED
//
// Copyright (c) 2004, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.security.authentication;

import java.net.PasswordAuthentication;

import org.pfsw.text.StringPattern;
import org.pfsw.text.StringPatternCollection;
import org.pfsw.text.StringUtil;

/**
 * Holds the userId and a password for a collection of URL patterns.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class LocationCredentials
{
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private String realmName = null;
  private StringPatternCollection locationPatterns = null;
  private String userId = null;
  private PasswordAuthentication credentials = null;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with a userId and corresponding passwords.
   * URL patterns can be added later with method addPattern().
   * 
   * @param userId The userId to be used for authentication
   * @param password The password to be used for authentication
   */
  public LocationCredentials(String userId, String password)
  {
    super();
    setUserId(userId);
    setLocationPatterns(new StringPatternCollection());
    if (password != null)
    {
      setCredentials(new PasswordAuthentication(userId, password.toCharArray()));
    }
  }

  /**
   * Initialize the new instance with a userId and corresponding passwords.
   * URL patterns can be added later with method addPattern().
   * 
   * @param realm The realm that is accessible with the given credentials
   * @param userId The userId to be used for authentication
   * @param password The password to be used for authentication
   */
  public LocationCredentials(String realm, String userId, String password)
  {
    this(userId, password);
    setRealm(realm);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the password based authentication data.
   */
  public PasswordAuthentication getPasswordAuthentication()
  {
    if (getCredentials() == null)
    {
      fillCredentials();
    }
    return getCredentials();
  }

  /**
   * Returns the realm this credentials apply to
   */
  public String getRealm()
  {
    return getRealmName();
  }

  /**
   * Sets the realm this credentials apply to
   */
  public void setRealm(String realm)
  {
    setRealmName(realm);
  }

  /**
   * Adds the given pattern to the URL patterns that can be accessed with this 
   * object's authentication credentials.
   * 
   * @param pattern A URL pattern that might contain '*' as wildcards. 
   */
  public void addPattern(String pattern)
  {
    StringPattern strPattern;

    if (!str().isNullOrBlank(pattern))
    {
      strPattern = StringPattern.create(pattern);
      addPattern(strPattern);
    }
  }

  /**
   * Adds the given pattern to the URL patterns that can be accessed with this 
   * object's authentication credentials.
   * 
   * @param pattern A URL pattern that might contain '*' as wildcards. 
   */
  public void addPattern(StringPattern pattern)
  {
    if (pattern != null)
    {
      getLocationPatterns().add(pattern);
    }
  }

  /**
   * Returns true if the given realm and URL match the settings in this object.
   */
  public boolean appliesTo(String realm, String url)
  {
    if ((realm == null) || (url == null))
    {
      return false;
    }
    return appliesToRealm(realm) && appliesToURL(url);
  }

  /**
   * Returns true if the given URL matches any of the underlying patterns.
   */
  public boolean appliesToURL(String url)
  {
    if (url == null)
    {
      return false;
    }
    return getLocationPatterns().matchesAny(url);
  }

  /**
   * Returns true if the given realm matches the realm name of this object.
   */
  public boolean appliesToRealm(String realm)
  {
    if ((realm == null) || (getRealm() == null))
    {
      return false;
    }
    return getRealm().equalsIgnoreCase(realm);
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  /**
   * Subclasses must override this method in order to get the credentials from
   * somewhere and set a PasswordAuthentication object using setCredentials().
   * The userId might be set already and can be accessed via getUserId().
   */
  protected void fillCredentials()
  {
    // Can't do anything here
  }

  protected String getRealmName()
  {
    return this.realmName;
  }

  protected void setRealmName(String newValue)
  {
    this.realmName = newValue;
  }

  protected StringPatternCollection getLocationPatterns()
  {
    return this.locationPatterns;
  }

  protected void setLocationPatterns(StringPatternCollection newValue)
  {
    this.locationPatterns = newValue;
  }

  protected String getUserId()
  {
    return this.userId;
  }

  protected void setUserId(String newValue)
  {
    this.userId = newValue;
  }

  protected PasswordAuthentication getCredentials()
  {
    return this.credentials;
  }

  protected void setCredentials(PasswordAuthentication newValue)
  {
    this.credentials = newValue;
  }

  protected StringUtil str()
  {
    return StringUtil.current();
  }
}
