// ===========================================================================
// CONTENT  : CLASS HashedString
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.0 - 16/11/2013
// HISTORY  :
//  24/06/2008  mdu  CREATED
//	17/11/2013	mdu  added   -> SHA-2 (224, 256, 383, 512)
//
// Copyright (c) 2008-2013, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.security.crypto;

import static org.pfsw.security.crypto.CryptoConstants.*;

import java.util.Arrays;

import org.pfsw.text.CheckedCharsetName;
import org.pfsw.text.StringUtil;
import org.pfsw.util.Base64Converter;

/**
 * Container for a SHA hashed strings including salt for SSHA.
 * Supports SHA-1 (160 bit) as well as SHA-2 (256, 383, 512 bit).
 * SHA-2 224 bit is not supported before Java 8 or with a crypto provider
 * explicitly added to the classpath.
 *
 * @author Manfred Duchrow
 * @version 2.0
 */
public class HashedString
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  protected static final CryptoUtil CRYPTU = CryptoUtil.current();

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private byte[] strHash = null;
  private byte[] salt = null;
  private boolean isHashed = false;
  private HashMechanism hashMechanism = null;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with a hashed string.
   * 
   * @param string A base64 encoded hash of a string with a prefix 
   * (e.g. {SHA} or {SSHA} or {SHA256} )
   */
  public HashedString(String string)
  {
    super();
    init(string);
  }

  public HashedString(byte[] content, String hashType)
  {
    this(content, null, hashType);
  }

  public HashedString(byte[] content, HashMechanism mechanism)
  {
    this(content, null, mechanism);
  }

  public HashedString(byte[] content, byte[] salt, String hashType)
  {
    this(content, salt, HashMechanism.findHashMechanism(hashType));
  }

  public HashedString(byte[] content, byte[] salt, HashMechanism mechanism)
  {
    super();
    init(content, salt, mechanism);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the hash value (without appended salt, if any).
   */
  public byte[] getHashValue()
  {
    return getStrHash();
  }

  /**
   * Returns the salt that has been used to hash the string or null if 
   * no salt was used.
   */
  public byte[] getSaltValue()
  {
    return getSalt();
  }

  /**
   * Returns the hash value with appended salt value (if any). If no salt value
   * was used then the result is the same as of {@link #getHashValue()}.
   */
  public byte[] getHashWithSaltValue()
  {
    byte[] bytes;

    bytes = (getSalt() == null) ? getStrHash() : CRYPTU.concatArrays(getStrHash(), getSalt());
    return bytes;
  }

  /**
   * Returns the hash value (without appended salt, if any) as Base64 encoded string.
   */
  public String getHashValueBase64()
  {
    return Base64Converter.encodeToString(getHashValue());
  }

  /**
   * Returns the hash value with appended salt value (if any) as Base64 encoded string. 
   * If no salt value was used then the result is the same as of {@link #getHashValueBase64()}.
   */
  public String getHashWithSaltValueBase64()
  {
    return Base64Converter.encodeToString(getHashWithSaltValue());
  }

  /**
   * Returns true if the value is salted.
   */
  public boolean isSalted()
  {
    if (getHashMechanism() == null)
    {
      return false;
    }
    return getHashMechanism().isSalted();
  }

  /**
   * Compares the given clear text string with the hashed value (which was hashed as UTF-8).
   * Returns true if the hash of the given string equals the hash in this object.
   */
  public boolean isEqualUTF8(String string)
  {
    return isEqual(CheckedCharsetName.UTF_8.getBytes(string));
  }

  /**
   * Compares the given clear text string with the hashed value.
   * Returns true if the hash of the given string equals the hash
   * in this object.
   */
  public boolean isEqual(String string)
  {
    return isEqual(stringToBytes(string));
  }

  /**
   * Compares the given plain text string with the hashed string.
   * Returns true if the hash of the given plain text equals the hash
   * in this object.
   */
  public boolean isEqual(byte[] plainText)
  {
    byte[] otherSaltedString;
    byte[] otherHash;

    if (isHashed())
    {
      if (isSalted())
      {
        otherSaltedString = new byte[plainText.length + getSalt().length];
        System.arraycopy(plainText, 0, otherSaltedString, 0, plainText.length);
        System.arraycopy(getSalt(), 0, otherSaltedString, plainText.length, getSalt().length);
      }
      else
      {
        otherSaltedString = plainText;
      }
      otherHash = getHashMechanism().getAlgorithm().computeHash(otherSaltedString);
    }
    else
    {
      otherHash = plainText;
    }
    return Arrays.equals(getStrHash(), otherHash);
  }

  /**
   * Returns the contents base64 with a curly bracket prefix that defines
   * the used hashing mechanism.
   * <P>  
   * Example: "{SSHA256}lKv8GwCDbjB6wvYroiAKaGZXUdWYgtYAnpQWLTrWHPUSzuooRrOT5Yp70mbNwfx0IziQ+g=="
   */
  public String asString()
  {
    byte[] bytes;

    bytes = getHashWithSaltValue();
    if (!isHashed())
    {
      // char encoding does not matter here because it s not hashed - it is plain text
      return new String(bytes);
    }
    return CRYPTU.base64EncodedWithPrefix(bytes, getHashMechanism());
  }

  @Override
  public String toString()
  {
    return getClass().getSimpleName() + "(" + asString() + ")";
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  /**
   * Initialized this HashedString instance from a Base64 encoded hash that is
   * prefixed with an algorithm identifier.
   * @param string Base64 encoded hash with prefix.
   */
  protected void init(final String string)
  {
    String prefix;
    String base64Hash;
    byte[] hash;
    byte[][] hashAndSalt;
    HashMechanism mechanism;

    prefix = str().prefix(string, ENC_MECHANISM_END);
    if (prefix == null) // not hashed at all
    {
      setStrHash(stringToBytes(string));
      setIsHashed(false);
      return;
    }

    setIsHashed(true);
    prefix = prefix.toUpperCase() + ENC_MECHANISM_END;
    base64Hash = str().suffix(string, ENC_MECHANISM_END);
    hash = Base64Converter.decode(base64Hash);

    mechanism = HashMechanism.findUnsaltedHashMechanism(prefix);
    if (mechanism != null)
    {
      setHashMechanism(mechanism);
      setStrHash(hash);
    }
    else
    {
      mechanism = HashMechanism.findSaltedHashMechanism(prefix);
      if (mechanism != null)
      {
        setHashMechanism(mechanism);
        hashAndSalt = split(hash, mechanism.getByteLength());
        setStrHash(hashAndSalt[0]);
        setSalt(hashAndSalt[1]);
      }
    }
  }

  protected void init(final byte[] content, final byte[] saltData, final HashMechanism mechanism)
  {
    byte[] bytesToHash;

    setHashMechanism(mechanism);
    setSalt(saltData);
    if (mechanism == null)
    {
      setIsHashed(false);
      setStrHash(content);
    }
    else
    {
      setIsHashed(true);
      if (isSalted())
      {
        if (getSalt() == null)
        {
          setSalt(CRYPTU.generateSalt20());
        }
        bytesToHash = CRYPTU.concatArrays(content, getSalt());
      }
      else
      {
        bytesToHash = content;
      }
      setStrHash(CRYPTU.computeHash(bytesToHash, mechanism.getAlgorithmName()));
    }
  }

  protected byte[][] split(byte[] src, int n)
  {
    byte[] l, r;
    if (src.length <= n)
    {
      l = src;
      r = new byte[0];
    }
    else
    {
      l = new byte[n];
      r = new byte[src.length - n];
      System.arraycopy(src, 0, l, 0, n);
      System.arraycopy(src, n, r, 0, r.length);
    }
    byte[][] lr = { l, r };
    return lr;
  }

  /**
   * Converts the given string to a byte array.
   * This default implementation is using the platform's default character encoding.
   * Subclasses my override this method to use other encoding (e.g. UTF-8).
   */
  protected byte[] stringToBytes(final String string)
  {
    if (string == null)
    {
      return null;
    }
    return string.getBytes();
  }

  protected byte[] getStrHash()
  {
    return strHash;
  }

  protected void setStrHash(byte[] newValue)
  {
    this.strHash = newValue;
  }

  protected byte[] getSalt()
  {
    return this.salt;
  }

  protected void setSalt(byte[] newValue)
  {
    this.salt = newValue;
  }

  protected boolean isHashed()
  {
    return this.isHashed;
  }

  protected void setIsHashed(boolean newValue)
  {
    this.isHashed = newValue;
  }

  protected HashMechanism getHashMechanism()
  {
    return this.hashMechanism;
  }

  protected void setHashMechanism(HashMechanism info)
  {
    this.hashMechanism = info;
  }

  protected StringUtil str()
  {
    return StringUtil.current();
  }
}
