// ===========================================================================
// CONTENT  : CLASS AbstractPKCEGenerator
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 24/04/2020
// HISTORY  :
//  24/04/2020  mdu  CREATED
//
// Copyright (c) 2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.security.oauth.pkce;

import org.pfsw.security.Base64UrlEncodedRandomStringGenerator;
import org.pfsw.text.StringGenerator;

/**
 * A customizable generic implementation of a {@link PKCEGenerator} where subclasses 
 * determine what code challenge method to be used and how to generate that data.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public abstract class AbstractPKCEGenerator implements PKCEGenerator
{
  /**
   * Returns the generator for random strings.
   */
  private static StringGenerator createDefaultRandomStringGenerator()
  {
    // length must be a value with length x where 32 <= x <= 92 to match base64 encoded length range 43-128.
    return Base64UrlEncodedRandomStringGenerator.create(57);
  }

  private final CodeChallengeMethod codeChallengeMethod;
  private final StringGenerator randomStringGenerator;

  /**
   * Creates a new instance for the given code challenge method and a default random string generator
   * and true for createBase64EncodedCodeVerifiers.
   * 
   * @param codeChallengeMethod The mechanism to be used for deriving the code challenge from the code verifier (must not be null).
   */
  public AbstractPKCEGenerator(CodeChallengeMethod codeChallengeMethod)
  {
    this(codeChallengeMethod, createDefaultRandomStringGenerator());
  }

  /**
   * Creates a new instance with the given arguments.
   * 
   * @param codeChallengeMethod The mechanism to be used for deriving the code challenge from the code verifier (must not be null).
   * @param randomStringGenerator A string generator that produces a different string for each invocation (must not be null).
   *   It will be used to generate new code verifier strings.
   */
  public AbstractPKCEGenerator(CodeChallengeMethod codeChallengeMethod, StringGenerator randomStringGenerator)
  {
    super();
    this.codeChallengeMethod = codeChallengeMethod;
    this.randomStringGenerator = randomStringGenerator;
  }

  @Override
  public ProofKeyForCodeExchange generate()
  {
    String codeVerifier;
    String codeChallenge;

    codeVerifier = generateRandomCodeVerifier();
    codeChallenge = createCodeChallenge(codeVerifier);
    return new ProofKeyForCodeExchange(codeVerifier, codeChallenge, getCodeChallengeMethod());
  }

  /**
   * Returns the code challenge method this generator is using to
   * create the code challenge.
   */
  public CodeChallengeMethod getCodeChallengeMethod()
  {
    return this.codeChallengeMethod;
  }
  
  @Override
  public String toString()
  {
    return String.format("%s(%s)", getClass().getSimpleName(), getCodeChallengeMethod());
  }

  /**
   * Returns a new random string to be used as code verifier.
   * The generated string must conform to the specification
   * https://tools.ietf.org/html/rfc7636#section-4.1
   * <p>
   * This method provides a default string generation mechanism
   * that can be overridden by subclasses if needed, but usually the better
   * approach for customization is providing a different StringGenerator 
   * during instance creation.
   */
  protected String generateRandomCodeVerifier()
  {
    return getRandomStringGenerator().generateString();
  }

  /**
   * Returns the generator for random strings.
   */
  protected StringGenerator getRandomStringGenerator()
  {
    return this.randomStringGenerator;
  }

  /**
   * Returns the derived code challenge string for the given code verifier
   * by applying the {@link #getCodeChallengeMethod()}. 
   * 
   * @param codeVerifier The newly created random code verifier.
   * @return The calculated code challenge.
   */
  protected abstract String createCodeChallenge(String codeVerifier);
}
