// ===========================================================================
// CONTENT  : CLASS PlainPKCEGenerator
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 24/04/2020
// HISTORY  :
//  24/04/2020  mdu  CREATED
//
// Copyright (c) 2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.security.oauth.pkce;

import org.pfsw.text.StringGenerator;

/**
 * An RFC-7636 PKCE proof key generator for the "plain" code challenge method. 
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class PlainPKCEGenerator extends AbstractPKCEGenerator
{
  private static PlainPKCEGenerator INSTANCE = create();

  /**
   * Returns always the same instance of this class.
   */
  public static PlainPKCEGenerator instance()
  {
    return INSTANCE;
  }

  /**
   * Creates a new instance with a default random string generator for new code verifier strings.
   */
  public static PlainPKCEGenerator create()
  {
    return new PlainPKCEGenerator();
  }

  /**
   * Creates a new instance with the given random string generator.
   * 
   * @param randomStringGenerator A string generator that produces a different string for each invocation (must not be null).
   *   It will be used to generate new code verifier strings.
   */
  public static PlainPKCEGenerator create(StringGenerator randomStringGenerator)
  {
    return new PlainPKCEGenerator(randomStringGenerator);
  }

  /**
   * Returns a newly generated proof key.
   */
  public static ProofKeyForCodeExchange newProofKey()
  {
    return create().generate();
  }

  /**
   * Creates a new instance with a default random string generator for new code verifier strings.
   */
  public PlainPKCEGenerator()
  {
    super(CodeChallengeMethod.PLAIN);
  }

  /**
   * Creates a new instance with the given random string generator.
   * 
   * @param randomStringGenerator A string generator that produces a different string for each invocation (must not be null).
   *   It will be used to generate new code verifier strings.
   */
  public PlainPKCEGenerator(StringGenerator randomStringGenerator)
  {
    super(CodeChallengeMethod.PLAIN, randomStringGenerator);
  }

  @Override
  protected String createCodeChallenge(String codeVerifier)
  {
    return codeVerifier;
  }
}
