// ===========================================================================
// CONTENT  : CLASS ProofKeyForCodeExchange
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 24/04/2020
// HISTORY  :
//  24/04/2020  mdu  CREATED
//
// Copyright (c) 2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.security.oauth.pkce;

/**
 * Holder of a code verifier and code challenge as specified by RFC-7637 (PKCE).
 * <br>
 * See https://tools.ietf.org/html/rfc763 
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class ProofKeyForCodeExchange
{
  private final String codeVerifier; // plain text random string
  private final String codeChallenge; // plain copy or hash of code verifier
  private final CodeChallengeMethod codeChallengeMethod;

  public ProofKeyForCodeExchange(String codeVerifier, String codeChallenge, CodeChallengeMethod codeChallengeMethod)
  {
    super();
    this.codeVerifier = codeVerifier;
    this.codeChallenge = codeChallenge;
    this.codeChallengeMethod = codeChallengeMethod;
  }

  /**
   * Returns the plain text code verifier string.
   */
  public String getCodeVerifier()
  {
    return this.codeVerifier;
  }

  /**
   * Returns the plain text or hashed code challenge string.
   */
  public String getCodeChallenge()
  {
    return this.codeChallenge;
  }

  /**
   * Returns the method that was used to derive the code challenge from the code verifier.
   */
  public CodeChallengeMethod getCodeChallengeMethod()
  {
    return this.codeChallengeMethod;
  }

  @Override
  public String toString()
  {
    return String.format("%s('%s', '%s')", getClass().getSimpleName(), getCodeChallengeMethod().asString(), getCodeVerifier());
  }
}
