// ===========================================================================
// CONTENT  : CLASS S256PKCEGenerator
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 24/04/2020
// HISTORY  :
//  24/04/2020  mdu  CREATED
//
// Copyright (c) 2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.security.oauth.pkce;

import org.pfsw.bif.conversion.IBase64Encoder;
import org.pfsw.security.crypto.CheckedHashAlgorithm;
import org.pfsw.security.crypto.CryptoUtil;
import org.pfsw.text.StringGenerator;
import org.pfsw.util.Base64UrlEncoder;

/**
 * An RFC-7636 PKCE proof key generator for the "S256" code challenge method. 
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class S256PKCEGenerator extends AbstractPKCEGenerator
{
  private static final S256PKCEGenerator INSTANCE = create();
  private static final IBase64Encoder BASE64_ENCODER = Base64UrlEncoder.withoutPadding();

  /**
   * Returns always the same instance of this class.
   */
  public static S256PKCEGenerator instance()
  {
    return INSTANCE;
  }

  /**
   * Creates a new instance with a default random string generator for new code verifier strings.
   */
  public static S256PKCEGenerator create()
  {
    return new S256PKCEGenerator();
  }

  /**
   * Creates a new instance with the given random string generator.
   * 
   * @param randomStringGenerator A string generator that produces a different string for each invocation (must not be null).
   *   It will be used to generate new code verifier strings.
   */
  public static S256PKCEGenerator create(StringGenerator randomStringGenerator)
  {
    return new S256PKCEGenerator(randomStringGenerator);
  }

  /**
   * Returns a newly generated proof key.
   */
  public static ProofKeyForCodeExchange newProofKey()
  {
    return create().generate();
  }

  /**
   * Creates a new instance with a default random string generator for new code verifier strings.
   */
  public S256PKCEGenerator()
  {
    super(CodeChallengeMethod.S256);
  }

  /**
   * Creates a new instance with the given random string generator.
   * 
   * @param randomStringGenerator A string generator that produces a different string for each invocation (must not be null).
   *   It will be used to generate new code verifier strings.
   */
  public S256PKCEGenerator(StringGenerator randomStringGenerator)
  {
    super(CodeChallengeMethod.S256, randomStringGenerator);
  }

  @Override
  protected String createCodeChallenge(String codeVerifier)
  {
    byte[] hash;

    hash = CryptoUtil.current().computeHash(codeVerifier.getBytes(), CheckedHashAlgorithm.SHA_2_256);
    return encodeBase64(hash);
  }
  
  /**
   * This method is responsible for doing a proper base64 URL encoding of 
   * the given hash value.
   */
  protected String encodeBase64(byte[] hash)
  {
    return BASE64_ENCODER.convert(hash);
  }
}
