// ===========================================================================
// CONTENT  : CLASS MultiX509TrustManager
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 02/08/2021
// HISTORY  :
//  02/08/2021  mdu  CREATED
//
// Copyright (c) 2021, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.security.x509;

import static org.pfsw.util.CollectionUtil.*;

import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.List;

import javax.net.ssl.X509TrustManager;

/**
 * An {@link X509TrustManager} that is capable of holding an arbitrary number of other {@link X509TrustManager}s
 * and delegating certificate checks to them until a check is successful or finally if none is successful
 * a CertificateException gets thrown. 
 * 
 * @author M.Duchrow
 */
public class MultiX509TrustManager implements X509TrustManager
{
  private final List<X509TrustManager> trustManagers = new ArrayList<X509TrustManager>();

  public MultiX509TrustManager()
  {
    super();
  }

  @Override
  public X509Certificate[] getAcceptedIssuers()
  {
    List<X509Certificate> acceptedIssuers = new ArrayList<X509Certificate>();

    for (X509TrustManager trustManager : getTrustManagers())
    {
      CU.addAll(acceptedIssuers, trustManager.getAcceptedIssuers());
    }
    return CU.toArray(acceptedIssuers);
  }

  @Override
  public void checkClientTrusted(final X509Certificate[] chain, final String authType) throws CertificateException
  {
    X509TrustManagerCheck check;

    check = new X509TrustManagerCheck()
    {
      @Override
      public void checkWith(X509TrustManager trustManager) throws CertificateException
      {
        trustManager.checkClientTrusted(chain, authType);
      }
    };
    checkWithTrustManagers(check);
  }

  @Override
  public void checkServerTrusted(final X509Certificate[] chain, final String authType) throws CertificateException
  {
    X509TrustManagerCheck check;

    check = new X509TrustManagerCheck()
    {
      @Override
      public void checkWith(X509TrustManager trustManager) throws CertificateException
      {
        trustManager.checkServerTrusted(chain, authType);
      }
    };
    checkWithTrustManagers(check);
  }
  
  /**
   * Adds the given trust manager if it is not null and not yet contained.
   */
  public void add(X509TrustManager trustManager) 
  {
    if ((trustManager != null) && !getTrustManagers().contains(trustManager))
    {
      getTrustManagers().add(trustManager);
    }
  }

  /**
   * Returns true if no trust manager is contained.
   */
  public boolean isEmpty()
  {
    return getTrustManagers().isEmpty();
  }

  /**
   * Removes all contained trust managers.
   */
  public void clear()
  {
    getTrustManagers().clear();
  }

  @Override
  public String toString()
  {
    return String.format("%s()", getClass().getSimpleName());
  }

  protected void checkWithTrustManagers(X509TrustManagerCheck check) throws CertificateException
  {
    CertificateException exception;

    exception = new CertificateException("No trust manager found to check certificate chain!");
    if (!isEmpty())
    {
      for (X509TrustManager trustManager : getTrustManagers())
      {
        try
        {
          check.checkWith(trustManager);
          return; // check successful - we are done
        }
        catch (CertificateException e)
        {
          exception = e;
        }
      }
    }
    throw exception;
  }

  protected List<X509TrustManager> getTrustManagers()
  {
    return this.trustManagers;
  }
}
