// ===========================================================================
// CONTENT  : CLASS X509TrustManagerBuilder
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 02/08/2021
// HISTORY  :
//  02/08/2021  mdu  CREATED
//
// Copyright (c) 2021, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.security.x509;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.security.GeneralSecurityException;
import java.security.KeyStore;
import java.util.ArrayList;
import java.util.List;

import javax.net.ssl.TrustManager;
import javax.net.ssl.TrustManagerFactory;
import javax.net.ssl.X509TrustManager;

import org.pfsw.bif.exceptions.IORuntimeException;
import org.pfsw.security.KeystoreType;
import org.pfsw.security.SecurityRuntimeException;
import org.pfsw.util.IOUtil;

public class X509TrustManagerBuilder
{
  private final X509TrustManager jreTrustManager;
  private final List<X509TrustManager> trustManagers = new ArrayList<X509TrustManager>();

  /**
   * Create a new builder.
   * 
   * @throws SecurityRuntimeException if initialization with default JRE trust store fails.
   */
  public static X509TrustManagerBuilder create()
  {
    return new X509TrustManagerBuilder();
  }

  protected X509TrustManagerBuilder()
  {
    super();
    this.jreTrustManager = findDefaultTrustManager(null);
  }
  
  /**
   * Add the JRE/JDK default trust store.
   */
  public X509TrustManagerBuilder addDefaultTrustStore() 
  {
    // We are relying here on the fact that no null value and no duplicate will be added.
    add(getJreTrustManager());
    return this;
  }

  /**
   * Add the given trust store.
   * 
   * @param fileName The path/name of the keystore file (must not be null).
   * @param password The password to unlock the keystore file. 
   */
  public X509TrustManagerBuilder addTrustStore(String fileName, String password) 
  {
    FileInputStream inStream;
    
    try
    {
      inStream = new FileInputStream(fileName);
    }
    catch (FileNotFoundException e)
    {
      throw new IORuntimeException(e, "Unable to open trust store file: %s", fileName);
    }
    return addTrustStore(inStream, password);
  }
  
  /**
   * Add the given trust store.
   * 
   * @param trustStore The truststore file stream (must not be null).
   * @param password The password to unlock the truststore file. 
   */
  public X509TrustManagerBuilder addTrustStore(InputStream trustStore, String password) 
  {
    KeyStore keystore;
    X509TrustManager trustManager;
    
    try
    {
      keystore = KeyStore.getInstance(KeystoreType.JKS.asString());
      keystore.load(trustStore, password.toCharArray());
      trustManager = findDefaultTrustManager(keystore);
      add(trustManager);
    }
    catch (GeneralSecurityException e)
    {
      throw new SecurityRuntimeException(e, "Unable to create trust keystore");
    }
    catch (IOException e)
    {
      throw new IORuntimeException(e, "Unable to load truststore stream");
    }
    finally
    {
      IOUtil.current().close(trustStore);
    }
    return this;
  }
  
  public MultiX509TrustManager build() 
  {
    MultiX509TrustManager multiX509TrustManager;
    
    multiX509TrustManager = new MultiX509TrustManager();
    for (X509TrustManager trustManager : getTrustManagers())
    {
      multiX509TrustManager.add(trustManager);
    }
    return multiX509TrustManager;
  }
  
  /**
   * Adds the given trust manager if it is not null and not yet contained.
   */
  protected void add(X509TrustManager trustManager) 
  {
    if ((trustManager != null) && !getTrustManagers().contains(trustManager))
    {
      getTrustManagers().add(trustManager);
    }
  }
  
  protected X509TrustManager findDefaultTrustManager(KeyStore keyStore)
  {
    TrustManagerFactory tmf;

    try
    {
      tmf = TrustManagerFactory.getInstance(TrustManagerFactory.getDefaultAlgorithm());
      tmf.init(keyStore); // If keyStore is null, tmf will be initialized with the default JRE trust store.
    }
    catch (GeneralSecurityException e)
    {
      throw new SecurityRuntimeException(e, "Failed to initialize trust manager with keystore.");
    }

    for (TrustManager tm : tmf.getTrustManagers())
    {
      if (tm instanceof X509TrustManager)
      {
        return (X509TrustManager)tm;
      }
    }
    return null;
  }

  protected X509TrustManager getJreTrustManager()
  {
    return this.jreTrustManager;
  }

  protected List<X509TrustManager> getTrustManagers()
  {
    return this.trustManagers;
  }
}
