// ===========================================================================
// CONTENT  : CLASS Element
// AUTHOR   : Manfred Duchrow
// VERSION  : 3.1 - 23/06/2017
// HISTORY  :
//  28/02/2002  duma  CREATED
//	17/05/2002	duma	adedd		->	find() and findChildren() methods	
//	25/05/2002	duma	added		->	removeAttribute(), removeText()
//	04/07/2002	duma	added		->	Constructor with tag name
//	10/10/2003	duma	added		->	Constructor with tag name and text
//	19/03/2004	duma	changed	->	Support for namespace information
//	12/01/2008	mdu		added		->	EMPTY_ARRAY, toArray(), toList()
//	07/02/2010	mdu		added		->	Better namspace and prefix handling
//  17/01/2014  mdu   added   ->  Support of processing instructions
//  23/06/2017  mdu   added   ->  Various getChild() and getChildren() methods
//
// Copyright (c) 2002-2017, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.six;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.pfsw.text.StringUtil;
import org.pfsw.util.CollectionUtil;
import org.pfsw.util.NamedTextList;

/**
 * Each element represents one tag occurrence in a XML stream.
 * The element's name corresponds to the XML tag name.
 *
 * @author Manfred Duchrow
 * @version 3.1
 */
public class Element implements XmlItem
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  public static final Element[] EMPTY_ARRAY = new Element[0];

  protected static final String NAMESPACE_SEPARATOR = ":";
  protected static final String NAMESPACE_PREFIX = "xmlns" + NAMESPACE_SEPARATOR;
  private static final int NOT_FOUND_INDEX = -1;

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private String name = null;
  private String text = null;
  private final List<XmlItem> internalChildItems = new ArrayList<XmlItem>();
  private NamedTextList elementAttributes = new NamedTextList();
  private Map<String, String> namespaces = null;

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Returns an Element array that contains all Element objects of the given
   * collection. If the given collection is null or empty an empty Element array
   * will be returned.
   * 
   * @param elementCollection The collection to be converted to an array.
   * @throws ClassCastException If any object in the given collection is not an Element.
   */
  public static Element[] toArray(Collection<Element> elementCollection)
  {
    if (coll().isNullOrEmpty(elementCollection))
    {
      return EMPTY_ARRAY;
    }
    return coll().toArray(elementCollection);
  }

  /**
   * Returns an ArrayList that contains all elements of the given array.
   * If the given array is null or empty then an empty ArrayList will 
   * be returned.
   * 
   * @param elements The array to convert to a list
   */
  public static List<Element> toList(Element[] elements)
  {
    if (coll().isNullOrEmpty(elements))
    {
      return new ArrayList<Element>();
    }
    return coll().toList(elements);
  }

  private static CollectionUtil coll()
  {
    return CollectionUtil.current();
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  public Element()
  {
    super();
  }

  /**
   * Initialize the new instance with the given name.
   * 
   * @param tagName The name of the new element
   */
  public Element(String tagName)
  {
    this();
    setName(tagName);
  }

  /**
   * Initialize the new instance with the given name and text.
   * 
   * @param tagName The name of the new element
   * @param text The text content of the new element
   */
  public Element(String tagName, String text)
  {
    this();
    setName(tagName);
    setText(text);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns always Type.ELEMENT.
   */
  @Override
  public Type getItemType()
  {
    return Type.ELEMENT;
  }

  /**
   * Returns the name of the element. This name corresponds to the
   * tag name in the XML representation of this element.
   */
  @Override
  public String getName()
  {
    return name;
  }

  /**
   * Set the element's name which corresponds to its XML tag name.
   * 
   * @param newName Any valid XML tag name
   */
  public void setName(String newName)
  {
    name = newName;
  }

  /**
   * Returns the local name of this element. That is, without a namespace prefix.
   */
  public String getLocalName()
  {
    return getLocalNameOf(getName());
  }

  public String getNamePrefix()
  {
    return getNamePrefixOf(getName());
  }

  /**
   * Returns the elements textual content which corresponds to the 
   * XML tag's PCDATA.
   */
  public String getText()
  {
    return text;
  }

  /**
   * Set the element's text.
   * 
   * @param newText Any valid XML character data compliant to the desired encoding
   */
  public void setText(String newText)
  {
    text = newText;
  }

  /**
   * Returns the value of the attribute with the given name. If the attribute
   * is not set (not available) this method returns null.
   * <br>
   * @param attrName The name of the attribute including a namespace prefix if applicable
   */
  public String getAttribute(String attrName)
  {
    return getElementAttributes().textAt(attrName);
  }

  /**
   * Sets the value of the attribute with the given name. If the attribute
   * exists already its value will be changed to the new one.
   * 
   * @param attrName The name of the attribute including a namespace prefix if applicable
   * @param value The value to assign to this attribute
   */
  public void setAttribute(String attrName, String value)
  {
    getElementAttributes().put(attrName, value);
  }

  /**
   * Removes the attribute with the specified name from the element's 
   * attribute list.
   * 
   * @param attrName The name of the attribute to remove (including namespace prefix)
   */
  public void removeAttribute(String attrName)
  {
    getElementAttributes().remove(attrName);
  }

  /**
   * Returns all attributes of this element.
   * This returned map may be modified, since it is a copy.
   * To add or remove attributes will not change the attributes of this element.
   * by Element.
   */
  public Map<String, String> getAttributes()
  {
    HashMap<String, String> copy;

    copy = getElementAttributes().asHashMap();
    return copy;
  }

  /**
   * Returns the names of all attributes assigned to this element.
   * If the element has no attribute an empty array will be returned.
   * The returned names are qualified names, that means they may contain
   * namespace prefixes.
   */
  public String[] getAttributeNames()
  {
    return getAttributeNames(true);
  }

  /**
   * Returns the names of all attributes assigned to this element.
   * If the element has no attribute an empty array will be returned.
   * The returned names are local names, that means they don't contain
   * namespace prefixes.
   */
  public String[] getAttributeLocalNames()
  {
    return getAttributeNames(false);
  }

  /**
   * Returns the number of contained child items.
   */
  public int getChildItemCount()
  {
    return getInternalChildItems().size();
  }

  /**
   * Returns the number of child elements.
   */
  public int getChildrenCount()
  {
    int counter = 0;

    for (XmlItem item : getInternalChildItems())
    {
      if (item instanceof Element)
      {
        counter++;
      }
    }
    return counter;
  }

  /**
   * Returns all direct child elements.
   */
  public Element[] getChildren()
  {
    return Element.toArray(collectContainedElements());
  }

  /**
   * Returns all direct child elements with the given tagName.
   */
  public Element[] getChildren(String tagName)
  {
    return getChildren(new ElementNameFilter(tagName));
  }

  /**
   * Returns all direct child elements that match the given filter.
   */
  public Element[] getChildren(IElementFilter elementFilter)
  {
    return Element.toArray(collect(elementFilter, false));
  }

  /**
   * Returns the first direct child element or null if there is no child element.
   */
  public Element getFirstChild()
  {
    return getFirstChild(IElementFilter.ALL);
  }

  /**
   * Returns the first child element that matches the given filter 
   * or null if there is no matching child element found.
   */
  public Element getFirstChild(String tagName)
  {
    return getFirstChild(new ElementNameFilter(tagName));
  }

  /**
   * Returns the first child element that matches the given filter 
   * or null if there is no matching child element found.
   */
  public Element getFirstChild(IElementFilter elementFilter)
  {
    Element element;

    for (XmlItem item : getInternalChildItems())
    {
      if (item instanceof Element)
      {
        element = (Element)item;
        if (elementFilter.matches(element))
        {
          return element;
        }
      }
    }
    return null;
  }

  /**
   * Returns true if the element has any child (Element or {@link XmlProcessingInstruction}).
   * For more specific item types see {@link #hasChildren()} and {@link #hasProcessingInstructions()}.  
   */
  public boolean hasChildItems()
  {
    return !getInternalChildItems().isEmpty();
  }

  /**
   * Returns true if the element contains at least on child element.
   * That does NOT include processing instructions.
   */
  public boolean hasChildren()
  {
    for (XmlItem item : getInternalChildItems())
    {
      if (item.getItemType() == Type.ELEMENT)
      {
        return true;
      }
    }
    return false;
  }

  /**
   * Returns true if the element contains at least one processing instruction.
   */
  public boolean hasProcessingInstructions()
  {
    for (XmlItem item : getInternalChildItems())
    {
      if (item.getItemType() == Type.PI)
      {
        return true;
      }
    }
    return false;
  }

  /**
   * Removes the given child from this element. The child itself must be in
   * the children list. That is, comparison is done by == rather than equals().
   * 
   * @param childElement The child element to be removed
   * @return true if the child element was found and removed
   */
  public boolean removeChild(Element childElement)
  {
    return removeChildItem(childElement);
  }

  /**
   * Removes the given child from this element. The child itself must be in
   * the children list. That is, comparison is done by == rather than equals().
   * 
   * @param item The child element to be removed
   * @return true if the child element was found and removed
   */
  public boolean removeChildItem(XmlItem item)
  {
    int index;

    index = indexOfChild(item);
    if (index == NOT_FOUND_INDEX)
    {
      return false;
    }
    return (item == getInternalChildItems().remove(index));
  }

  /**
   * Removes all children of this element. That includes all types 
   * {@link Element} and {@link XmlProcessingInstruction}).
   */
  public void removeAllChildren()
  {
    getInternalChildItems().clear();
  }

  /**
   * Removes the element's text content.
   */
  public void removeText()
  {
    setText(null);
  }

  /**
   * Returns true if the element has any text.
   */
  public boolean hasText()
  {
    return (getText() != null);
  }

  /**
   * Returns true if the element has any attribute.
   */
  public boolean hasAttributes()
  {
    return ((getElementAttributes() != null) && (getElementAttributes().size() > 0));
  }

  /**
   * Returns true if the element has an attribute with the given name.
   */
  public boolean hasAttribute(String attrName)
  {
    return (getAttribute(attrName) != null);
  }

  /**
   * Returns the element's name as a XML tag.
   * Example: "<Person>"
   */
  @Override
  public String toString()
  {
    return "<" + getName() + ">";
  }

  /**
   * Adds the given element as a child to the receiver
   * 
   * @param childElement The element to add
   */
  public void addChild(Element childElement)
  {
    addChildItem(childElement);
  }

  /**
   * Adds the given item as a child to the receiver.
   * 
   * @param item The item to add.
   */
  public void addChildItem(XmlItem item)
  {
    if (item == null)
    {
      return;
    }
    getInternalChildItems().add(item);
  }

  /**
   * Inserts the given newChild after the specified old child.
   * Returns true if the old child was found and the new child successfully
   * inserted after it.
   */
  public boolean insertChildAfter(XmlItem oldChild, XmlItem newChild)
  {
    int index;

    if ((oldChild == null) || (newChild == null))
    {
      return false;
    }
    index = indexOfChild(oldChild);
    if (index == NOT_FOUND_INDEX)
    {
      return false;
    }
    insertAfter(index, newChild);
    return true;
  }

  /**
   * Inserts the given newChild before the specified old child.
   * Returns true if the old child was found and the new child successfully
   * inserted after it.
   */
  public boolean insertChildBefore(XmlItem oldChild, XmlItem newChild)
  {
    int index;

    if ((oldChild == null) || (newChild == null))
    {
      return false;
    }

    index = indexOfChild(oldChild);
    if (index == NOT_FOUND_INDEX)
    {
      return false;
    }
    insertBefore(index, newChild);
    return true;
  }

  /**
   * Replaces the given oldChild by the given newChild.
   * Returns true if the old child was found and replaced.
   */
  public boolean replaceChild(XmlItem oldChild, XmlItem newChild)
  {
    int index;

    if ((oldChild == null) || (newChild == null))
    {
      return false;
    }

    index = indexOfChild(oldChild);
    if (index == NOT_FOUND_INDEX)
    {
      return false;
    }
    getInternalChildItems().set(index, newChild);
    return true;
  }

  /**
   * Returns the first element in the hierarchy that matches the given filter
   * or null if none matches.
   * 
   * @param filter The filter that defines which elements match (must not be null).
   * @return The found element or null.
   */
  public Element findFirst(IElementFilter filter)
  {
    Element[] elements;

    elements = find(filter);
    if (elements.length > 0)
    {
      return elements[0];
    }
    return null;
  }

  /**
   * Returns all sub elements matching the given filter.
   * This method returns all matching elements down the whole hierarchy.
   * 
   * @param filter The filter that defines which elements match (must not be null).
   * @return An array of elements (never null).
   */
  public Element[] find(IElementFilter filter)
  {
    List<Element> result;

    result = collect(filter, true);
    if (filter.matches(this))
    {
      result.add(0, this);
    }
    return Element.toArray(result);
  }

  /**
   * Returns all sub elements with the given tag name plus this element,
   * if it's name is equal to the given tagName.
   * This method returns all matching elements down the whole hierarchy.
   * 
   * @return An array of elements (never null)
   */
  public Element[] find(String tagName)
  {
    return find(tagName, null);
  }

  /**
   * Returns all sub elements with the given tag name and the specified
   * attribute values plus this element if it matches the tagName and
   * attributes.
   * This method returns all matching elements down the whole hierarchy.
   * 
   * @param tagName The name of the elements to be found (must not be null).
   * @param attributes Definition of attributes the elements must match (may be null).
   * @return An array of elements (never null)
   */
  public Element[] find(String tagName, Map<String, String> attributes)
  {
    List<Element> result;
    IElementFilter filter;

    filter = new ElementAndFilter(new ElementNameFilter(tagName), new ElementAttributesFilter(attributes));
    result = collect(filter, true);
    if (filter.matches(this))
    {
      result.add(0, this);
    }
    return Element.toArray(result);
  }

  /**
   * Returns all sub elements with the given tag name. 
   * That is, all direct children
   * and children of those children and so on. It goes down recursively the
   * whole hierarchy.
   * 
   * @return An array of elements (never null)
   */
  public Element[] findSubElements(String tagName)
  {
    return collectElements(tagName, null, true);
  }

  /**
   * Returns all sub elements with the given tag name and the specified
   * attribute values. That is, all direct children
   * and children of those children and so on. It goes down recursively the
   * whole hierarchy.
   */
  public Element[] findSubElements(String tagName, Map<String, String> attributes)
  {
    return collectElements(tagName, attributes, true);
  }

  /**
   * Returns all children with the given tag name. 
   * This method returns only direct children of this element.
   * It does not go deeper into the hierarchy.
   * 
   * @return An array of elements (never null)
   */
  public Element[] findChildren(String tagName)
  {
    return collectElements(tagName, null, false);
  }

  /**
   * Returns all children with the given tag name and the specified 
   * attribute values. 
   * This method returns only direct children of this element.
   * It does not go deeper into the hierarchy.
   * 
   * @return An array of elements (never null)
   */
  public Element[] findChildren(String tagName, Map<String, String> attributes)
  {
    return collectElements(tagName, attributes, false);
  }

  /**
   * Returns a shallow copy of the element, that is a new element with the same
   * name, same attributes and same text but without children.
   */
  public Element shallowCopy()
  {
    Element copy;
    String[] names;

    copy = new Element(getName());
    if (hasAttributes())
    {
      names = getAttributeNames();
      for (int i = 0; i < names.length; i++)
      {
        copy.setAttribute(names[i], getAttribute(names[i]));
      }
    }
    if (hasText())
    {
      copy.setText(getText());
    }
    return copy;
  }

  /**
   * Returns a deep copy of the element. That is a new element with the same
   * name, same attributes, the same text and with a copy of all child elements.
   */
  public Element deepCopy()
  {
    Element copy;
    Element element;

    copy = shallowCopy();
    if (hasChildren())
    {
      for (XmlItem item : getInternalChildItems())
      {
        switch (item.getItemType())
        {
          case ELEMENT :
            element = (Element)item;
            copy.addChildItem(element.deepCopy());
            break;
          case PI :
            copy.addChildItem(item);
            break;
          default :
            break;
        }
      }
    }
    return copy;
  }

  /**
   * Returns a copy of the element with all child elements copied as well.
   * This is a deep copy.
   * 
   * @see #deepCopy()
   */
  public Element copy()
  {
    return deepCopy();
  }

  /**
   * Returns a map with the names of namespaces as keys and the corresponding
   * namespace URIs as values.
   * The map contains only those namespace definitions that are set by
   * attributes of this element.
   * If no namespace is declared in this element the returned map is empty. 
   */
  public Map<String, String> getNamespaceDeclarations()
  {
    if (getNamespaces() == null)
    {
      setNamespaces(detectNamespaceDeclarations());
    }
    return getNamespaces();
  }

  /**
   * Returns all children of all types. That is Elements and XmlProcessinGinstructions.
   */
  public XmlItem[] getChildItems()
  {
    return Element.coll().toArray(getInternalChildItems(), XmlItem.class);
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  /**
   * Returns all elements with the given tag name.
   */
  protected List<Element> collect(IElementFilter elementFilter, boolean recursive)
  {
    ArrayList<Element> result;
    Element element;

    if (!hasChildren())
    {
      return new ArrayList<Element>();
    }
    result = new ArrayList<Element>(getChildrenCount());

    for (XmlItem item : getInternalChildItems())
    {
      if (item instanceof Element)
      {
        element = (Element)item;
        if (elementFilter.matches(element))
        {
          result.add(element);
        }
        if (recursive)
        {
          result.addAll(element.collect(elementFilter, true));
        }
      }
    }
    return result;
  }

  protected Element[] collectElements(String tagName, Map<String, String> attributes, boolean all)
  {
    IElementFilter filter;

    filter = new ElementAndFilter(new ElementNameFilter(tagName), new ElementAttributesFilter(attributes));
    return Element.toArray(collect(filter, all));
  }

  protected void setChildItems(List<XmlItem> items)
  {
    removeAllChildren();
    getInternalChildItems().addAll(items);
  }

  protected boolean matches(Element element, String tagName, Map<String, String> attributes)
  {
    String key;
    String expectedValue;
    String value;

    if (tagName.equals(element.getName()))
    {
      if (attributes != null)
      {
        for (Map.Entry<String, String> entry : attributes.entrySet())
        {          
          key = entry.getKey();
          value = element.getAttribute(key);
          if (value == null)
          {
            return false;
          }
          expectedValue = entry.getValue();
          if (!value.equals(expectedValue))
          {
            return false;
          }
        }
      }
      return true;
    }
    return false;
  }

  protected Map<String, String> detectNamespaceDeclarations()
  {
    String[] attrNames;
    Map<String, String> foundNamespaces;
    String nsId;
    String nsUri;

    foundNamespaces = new HashMap<String, String>();
    attrNames = getAttributeNames();
    for (int i = 0; i < attrNames.length; i++)
    {
      if (attrNames[i].startsWith(NAMESPACE_PREFIX))
      {
        nsId = str().suffix(attrNames[i], NAMESPACE_SEPARATOR);
        nsUri = getAttribute(attrNames[i]);
        foundNamespaces.put(nsId, nsUri);
      }
    }
    return foundNamespaces;
  }

  protected int indexOfChild(XmlItem child)
  {
    if ((getChildItemCount() > 0) && (child != null))
    {
      int i = 0;
      for (XmlItem item : getInternalChildItems())
      {
        if (item == child)
        {
          return i;
        }
        i++;
      }
    }
    return NOT_FOUND_INDEX;
  }

  /**
   * Inserts the given child after the given index
   */
  protected void insertAfter(int index, XmlItem item)
  {
    if (item == null)
    {
      return;
    }
    getInternalChildItems().add(index + 1, item);
  }

  /**
   * Inserts the given child before the given index
   */
  protected void insertBefore(int index, XmlItem item)
  {
    if (item == null)
    {
      return;
    }
    getInternalChildItems().add(index, item);
  }

  protected String[] getAttributeNames(boolean withPrefix)
  {
    String attrName;
    String[] names = StringUtil.EMPTY_STRING_ARRAY;
    Iterator<String> nameIterator;
    int index = 0;

    if (hasAttributes())
    {
      names = new String[getElementAttributes().size()];
      nameIterator = getElementAttributes().names().iterator();
      while (nameIterator.hasNext())
      {
        attrName = nameIterator.next();
        if (!withPrefix)
        {
          attrName = getLocalNameOf(attrName);
        }
        names[index] = attrName;
        index++;
      }
    }
    return names;
  }

  protected String getNamePrefixOf(String aName)
  {
    if (aName.indexOf(NAMESPACE_SEPARATOR) > 0)
    {
      return str().cutTail(aName, NAMESPACE_SEPARATOR);
    }
    return StringUtil.EMPTY_STRING;
  }

  protected String getLocalNameOf(String aName)
  {
    if (aName.indexOf(NAMESPACE_SEPARATOR) > 0)
    {
      return str().cutHead(aName, NAMESPACE_SEPARATOR);
    }
    return aName;
  }

  protected List<XmlItem> getInternalChildItems()
  {
    return this.internalChildItems;
  }

  protected List<Element> collectContainedElements()
  {
    return collect(IElementFilter.ALL, false);
  }

  protected NamedTextList getElementAttributes()
  {
    return this.elementAttributes;
  }

  protected void setAttributes(NamedTextList newValue)
  {
    this.elementAttributes = newValue;
  }

  protected Map<String, String> getNamespaces()
  {
    return this.namespaces;
  }

  protected void setNamespaces(Map<String, String> newValue)
  {
    this.namespaces = newValue;
  }

  protected StringUtil str()
  {
    return StringUtil.current();
  }
}
