// ===========================================================================
// CONTENT  : CLASS StringTagInterpreter
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.3 - 17/01/2014
// HISTORY  :
//  28/02/2002  duma  CREATED
//	17/03/2003	duma	changed	->	Reduced visibility from public to default
//	04/08/2009	mdu		changed	->	Leave content of CDATA[] sections unchanged
//  17/01/2014  mdu   changed ->  Support processing instructions
//
// Copyright (c) 2002-2014, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.six;

import java.util.ArrayList;
import java.util.List;

import org.pfsw.pax.BaseXMLTagInterpreter;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

/**
 * This interpreter is used to read in XML data and build a simple
 * String base element hierarchy.
 *
 * @author Manfred Duchrow
 * @version 1.3
 */
class StringTagInterpreter extends BaseXMLTagInterpreter
{
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private String tagName = null;
  private Element element = null;
  private List<XmlItem> children = null;
  private boolean isTextCDATA = false;
  private final List<TaggedBuffer> bufferList = new ArrayList<TaggedBuffer>();

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with the given tag name.
   */
  public StringTagInterpreter(String aTagName)
  {
    super();
    setTagName(aTagName);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the name of the tag, the interpreter is responsible for. <br>
   *
   * @return A string containing the name of the tag the receiver is made for.
   */
  @Override
  public String getTagName()
  {
    return tagName;
  }

  @Override
  public void start(String elementName, Attributes attributes) throws SAXException
  {
    setElement(new Element());
    getElement().setName(elementName);
    for (int i = 0; i < attributes.getLength(); i++)
    {
      getElement().setAttribute(attributes.getQName(i), attributes.getValue(i));
    }
  }

  @Override
  public void restart(String subTagName, Object subResult) throws SAXException
  {
    addChild((Element)subResult);
  }

  @Override
  public Object getResult()
  {
    return getElement();
  }

  @Override
  public void characterData(String data)
  {
    TaggedBuffer currentBuffer;

    currentBuffer = getLastTextBuffer();
    if ((currentBuffer != null) && (currentBuffer.isTextCDATA() != isTextCDATA()))
    {
      currentBuffer = null;
    }
    if (currentBuffer == null)
    {
      currentBuffer = new TaggedBuffer(isTextCDATA(), data.length());
      getBufferList().add(currentBuffer);
    }
    currentBuffer.append(data);
  }

  @Override
  public void startCDATA() throws SAXException
  {
    setIsTextCDATA(true);
  }

  @Override
  public void endCDATA() throws SAXException
  {
    setIsTextCDATA(false);
  }

  @Override
  public void processingInstruction(String target, String data) throws SAXException
  {
    addChild(new XmlProcessingInstruction(target, data));
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected void addChild(XmlItem child)
  {
    if (getChildren() == null)
    {
      setChildren(new ArrayList<XmlItem>());
    }

    getChildren().add(child);
  }

  /**
   * This method is called just before control is passed back to the controller.   <br>
   * It gives the interpreter a chance to do any necessary completion of
   * its result object.
   */
  @Override
  protected void finalizeResult()
  {
    if (getChildren() != null)
    {
      getElement().setChildItems(getChildren());
    }
    getElement().setText(finishElementText());
  }

  protected String finishElementText()
  {
    StringBuffer buffer;

    if (getBufferList().isEmpty())
    {
      return null;
    }
    buffer = new StringBuffer(calcTextLength());
    for (TaggedBuffer part : getBufferList())
    {      
      if (part.isTextCDATA())
      {
        buffer.append(part.toString());
      }
      else
      {
        buffer.append(removeWhitespaces(part.toString()));
      }
    }
    if (buffer.length() == 0)
    {
      return null;
    }
    return buffer.toString();
  }

  /**
   * This method removes all whitespace characters from the given text 
   * and replaces them by just a single blank, except at the beginning and
   * the very end of the text.
   */
  protected String removeWhitespaces(String text)
  {
    char[] chars;
    StringBuffer buffer;
    char ch;
    boolean addBlank = false;

    chars = text.toCharArray();
    buffer = new StringBuffer(chars.length);
    int i = 0;
    while (i < chars.length)
    {
      ch = chars[i];
      if ((ch == ' ') || (ch == '\n') || (ch == '\r') || (ch == '\t'))
      {
        addBlank = true;
      }
      else
      {
        if (addBlank && (buffer.length() > 0))
        {
          buffer.append(' ');
          addBlank = false;
        }
        buffer.append(ch);
        addBlank = false;
      }
      i++;
    }
    return buffer.toString();
  }

  protected TaggedBuffer getLastTextBuffer()
  {
    if (getBufferList().isEmpty())
    {
      return null;
    }
    return getBufferList().get(getBufferList().size() - 1);
  }

  protected int calcTextLength()
  {
    int length = 0;

    if (getBufferList().isEmpty())
    {
      return 0;
    }
    for (TaggedBuffer part : getBufferList())
    {      
      length += part.length();
    }
    return length;
  }

  protected void setTagName(String newValue)
  {
    tagName = newValue;
  }

  protected Element getElement()
  {
    return element;
  }

  protected void setElement(Element newValue)
  {
    element = newValue;
  }

  protected List<XmlItem> getChildren()
  {
    return children;
  }

  protected void setChildren(List<XmlItem> newValue)
  {
    children = newValue;
  }

  protected boolean isTextCDATA()
  {
    return isTextCDATA;
  }

  protected void setIsTextCDATA(boolean newValue)
  {
    isTextCDATA = newValue;
  }

  protected List<TaggedBuffer> getBufferList()
  {
    return bufferList;
  }
}
