// ===========================================================================
// CONTENT  : CLASS XmlDocument
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 17/01/2014
// HISTORY  :
//  17/01/2014  mdu  CREATED
//
// Copyright (c) 2014, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.six;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * Represents the whole XML document and contains 0..n children which can
 * be of type Element or XmlProcessingInstruction.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class XmlDocument
{
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private String encoding;
  private final List<XmlProcessingInstruction> processingInstructions = new ArrayList<XmlProcessingInstruction>();
  private Element rootElement;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  public XmlDocument()
  {
    super();
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  public String getEncoding()
  {
    return this.encoding;
  }

  public void setEncoding(String encoding)
  {
    this.encoding = encoding;
  }

  public Element getRootElement()
  {
    return this.rootElement;
  }

  public void setRootElement(Element rootElement)
  {
    this.rootElement = rootElement;
  }

  public XmlProcessingInstruction[] getProcessingInstructions()
  {
    return XmlProcessingInstruction.toArray(this.processingInstructions);
  }

  public void addProcessingInstruction(XmlProcessingInstruction instruction)
  {
    if (instruction != null)
    {
      this.processingInstructions.add(instruction);
    }
  }

  public void clearProcessingInstruction()
  {
    this.processingInstructions.clear();
  }

  public boolean hasProcessingInstructions()
  {
    return !this.processingInstructions.isEmpty();
  }

  /**
   * Returns all sub elements with the given tag name and the specified
   * attribute values plus this element if it matches the tagName and
   * attributes.
   * This method returns all matching elements down the whole hierarchy.
   * 
   * @param tagName The name of the elements to be found (must not be null).
   * @param attributes Definition of attributes the elements must match (may be null).
   * @return An array of elements (never null)
   */
  public Element[] find(String tagName, Map<String, String> attributes)
  {
    if (getRootElement() == null)
    {
      return Element.EMPTY_ARRAY;
    }
    return getRootElement().find(tagName, attributes);
  }

  /**
   * Returns all sub elements matching the given filter.
   * This method returns all matching elements down the whole hierarchy.
   * 
   * @param filter The filter that defines which elements match (must not be null).
   * @return An array of elements (never null).
   */
  public Element[] find(IElementFilter filter)
  {
    if (getRootElement() == null)
    {
      return Element.EMPTY_ARRAY;
    }
    return getRootElement().find(filter);
  }

  /**
   * Returns the first element in the hierarchy that matches the given filter
   * or null if none matches.
   * 
   * @param filter The filter that defines which elements match (must not be null).
   * @return The found element or null.
   */
  public Element findFirst(IElementFilter filter)
  {
    if (getRootElement() == null)
    {
      return null;
    }
    return getRootElement().findFirst(filter);
  }

}
