package org.pfsw.text;

import org.pfsw.bif.text.IStringPair;

/**
 * Abstract standard implementation for replacing multiple placeholders identified 
 * by a start and end delimiter in a string.
 * Placeholders that cannot be resolved will be left unchanged in the result string.
 */
public abstract class AbstractPlaceholderReplacer implements PlaceholderReplacer
{
  public static final String DEFAULT_SEPARATOR = ":";

  private final String startDelimiter;
  private final String endDelimiter;

  protected AbstractPlaceholderReplacer(String startDelimiter, String endDelimiter)
  {
    super();
    this.startDelimiter = startDelimiter;
    this.endDelimiter = endDelimiter;
  }

  @Override
  public String apply(String string)
  {
    return replacePlaceholders(string);
  }

  @Override
  public String replacePlaceholders(String input)
  {
    if (input == null)
    {
      return null;
    }
    return detectAndReplacePlaceholders(input);
  }

  public String getStartDelimiter()
  {
    return startDelimiter;
  }

  public String getEndDelimiter()
  {
    return endDelimiter;
  }

  @Override
  public String toString()
  {
    return String.format("%s(%s...%s)", getClass().getSimpleName(), getStartDelimiter(), getEndDelimiter());
  }

  protected String detectAndReplacePlaceholders(String input)
  {
    String result;
    String nextSubstring = input;
    String placeholder;
    String placeholderName;
    String placeholderValue;
    IStringPair placeolderDefinition;

    result = input;
    nextSubstring = input;
    placeholderName = str().getDelimitedSubstring(nextSubstring, getStartDelimiter(), getEndDelimiter());
    while (str().notNullOrBlank(placeholderName))
    {
      placeholder = String.format("%s%s%s", getStartDelimiter(), placeholderName, getEndDelimiter());
      nextSubstring = str().suffix(nextSubstring, placeholder);
      placeolderDefinition = separateNameAndDefaultValue(placeholderName);
      placeholderValue = getValueOfPlaceholder(placeolderDefinition.getString1(), placeolderDefinition.getString2());
      if (placeholderValue != null)
      {
        result = str().replaceAll(result, placeholder, placeholderValue);
      }
      placeholderName = str().getDelimitedSubstring(nextSubstring, getStartDelimiter(), getEndDelimiter());
    }
    return result;
  }

  protected IStringPair separateNameAndDefaultValue(String placeholderName)
  {
    if (placeholderName.contains(getDefaultSeparator()))
    {
      return str().splitStringPair(placeholderName, getDefaultSeparator());      
    }
    return StringPair.create(placeholderName);
  }

  protected String getDefaultSeparator()
  {
    return DEFAULT_SEPARATOR;
  }

  protected abstract String getValueOfPlaceholder(String placeholderName, String defaultValue);

  protected StringUtil str()
  {
    return StringUtil.current();
  }
}
