// ===========================================================================
// CONTENT  : CLASS TextBuilder
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 28/04/2021
// HISTORY  :
//  28/04/2021  mdu  CREATED
//
// Copyright (c) 2021, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.text;

import java.io.StringWriter;

import org.pfsw.bif.IExtensibleFluentApi;
import org.pfsw.bif.text.IStringRepresentation;

/**
 * A builder for text with fluent API support and placeholder replacement based of {@link String#format(String, Object...)}.
 */
public class TextBuilder<T extends TextBuilder<T>> implements Appendable, IStringRepresentation, IExtensibleFluentApi<T>
{
  private final StringWriter stringWriter;
  private String newline = System.getProperty("line.separator", "\n");

  @SuppressWarnings("rawtypes")
  public static TextBuilder create()
  {
    return new TextBuilder(new StringWriter());
  }

  protected TextBuilder(StringWriter writer)
  {
    super();
    this.stringWriter = writer;
  }

  public T print(String text, Object... args)
  {
    appendText(String.format(text, args));
    return self();
  }

  public T println(String text, Object... args)
  {
    print(String.format(text, args));
    return println();
  }

  public T println()
  {
    appendNewline();
    return self();
  }

  @Override
  public T append(char c)
  {
    appendText(Character.toString(c));
    return self();
  }

  @Override
  public T append(CharSequence csq)
  {
    appendText(csq.toString());
    return self();
  }

  @Override
  public T append(CharSequence csq, int start, int end)
  {
    getStringWriter().append(csq, start, end);
    return self();
  }

  @Override
  public String asString()
  {
    return getStringWriter().toString();
  }

  public T setNewline(String newline)
  {
    this.newline = newline;
    return self();
  }

  @SuppressWarnings("unchecked")
  @Override
  public T self()
  {
    return (T)this;
  }

  @Override
  public String toString()
  {
    return String.format("%s()", getClass().getSimpleName());
  }

  protected void appendText(String text)
  {
    getStringWriter().append(text);
  }

  protected void appendNewline()
  {
    getStringWriter().append(getNewline());
  }

  protected StringWriter getStringWriter()
  {
    return this.stringWriter;
  }

  protected String getNewline()
  {
    return this.newline;
  }
}
