// ===========================================================================
// CONTENT  : CLASS JsonClaims
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 20/09/2020
// HISTORY  :
//  20/09/2020  mdu  CREATED
//
// Copyright (c) 2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.text.json;

import java.util.List;

import org.pfsw.bif.text.IStringConstant;
import org.pfsw.text.TimeUnit;
import org.pfsw.text.TimeValue;

/**
 * Provides a simplified view of JSON based claims used with OAUTH2 and OIDC.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class JsonClaims extends JsonObjectWrapper
{
  /**
   * Creates a claim object on an empty JSON object.
   */
  public static JsonClaims create()
  {
    return new JsonClaims();
  }
  
  /**
   * Creates a claim object on the given JSON object.
   */
  public static JsonClaims create(JsonObject jsonObject)
  {
    return new JsonClaims(jsonObject);
  }
  
  /**
   * Creates a claim object on the given JSON object accessor.
   */
  public static JsonClaims create(JsonObjectAccessor jsonAccessor)
  {
    return new JsonClaims(jsonAccessor);
  }
  
  /**
   * Creates a claim object on an empty JSON object.
   */
  public JsonClaims()
  {
    this(JsonObject.create());
  }

  /**
   * Creates a claim object on the given JSON object.
   */
  public JsonClaims(JsonObject jsonObject)
  {
    this(new JsonObjectAccessor(jsonObject));
  }
  
  /**
   * Creates a claim object on the given JSON object accessor.
   */
  public JsonClaims(JsonObjectAccessor jsonAccessor)
  {
    super(jsonAccessor);
  }
  
  /**
   * Returns all claim names present in the underlying JSON object (maybe empty but never null).
   */
  public List<String> getClaimNames()
  {
    return getJsonAccessor().getFieldNames();
  }

  /**
   * Returns the JSON object at the specified field name. 
   * 
   * @param name The name of the field (must not be null).
   * @return Returns either the JSON object or null if field not present.
   */
  public JsonObject getJsonObjectField(String name)
  {
    return getJsonAccessor().getOptionalJsonObjectValue(name);
  }

  /**
   * Returns the JSON object at the specified field name. 
   * 
   * @param name The name of the field (must not be null).
   * @return Returns either the JSON object or null if field not present.
   */
  public JsonObject getJsonObjectField(IStringConstant name)
  {
    return getJsonAccessor().getOptionalJsonObjectValue(name);
  }

  /**
   * Returns the value at the specified field name. 
   * 
   * @param name The name of the field (must not be null).
   * @param type The expected return type (must not be null).
   * @return Returns either the value or null if field not present.
   */
  public <T> T getCustomField(String name, Class<T> type)
  {
    return getJsonAccessor().getOptionalTypedValue(name, type);
  }

  /**
   * Returns the value at the specified field name. 
   * 
   * @param name The name of the field (must not be null).
   * @param type The expected return type (must not be null).
   * @return Returns either the value or null if field not present.
   */
  public <T> T getCustomField(IStringConstant name, Class<T> type)
  {
    return getJsonAccessor().getOptionalTypedValue(name, type);
  }

  /**
   * Returns the value at the specified field name as a {@link TimeValue}
   * with the specified {@link TimeUnit}. 
   * 
   * @param name The name of the field (must not be null).
   * @param timeUnit The time unit of the value found in the object (must not be null).
   * @return Returns either the TimeUnit or null if field not present.
   */
  public TimeValue getTimeValue(String name, TimeUnit timeUnit)
  {
    return getJsonAccessor().getOptionalTimeValue(name, timeUnit);
  }

  /**
   * Returns the value at the specified field name as a {@link TimeValue}
   * with the specified {@link TimeUnit}. 
   * 
   * @param name The name of the field (must not be null).
   * @param timeUnit The time unit of the value found in the object (must not be null).
   * @return Returns either the TimeUnit or null if field not present.
   */
  public TimeValue getTimeValue(IStringConstant name, TimeUnit timeUnit)
  {
    return getJsonAccessor().getOptionalTimeValue(name, timeUnit);
  }

  /**
   * Returns true if this object contains no claim at all.
   */
  public boolean isEmpty()
  {
    return getJsonAccessor().isEmpty();
  }

  /**
   * Returns the token part as {@link JsonObject}.
   */
  @Override
  public JsonObject getJsonObject()
  {
    return super.getJsonObject();
  }
  
  /**
   * Returns the underlying {@link JsonObjectAccessor};
   */
  @Override
  public JsonObjectAccessor getJsonAccessor()
  {
    return super.getJsonAccessor();
  }
}
