// ===========================================================================
// CONTENT  : CLASS JsonObjectAccessor
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.1 - 20/09/2020
// HISTORY  :
//  08/01/2017  mdu  CREATED
//  16/02/2017  mdu   added --> getRequiredStringArrayValue(), getOptionalStringArrayValue()
//  08/03/2020  mdu   added --> support IStringConstant for field names
//  02/05/2020  mdu   added --> setFieldValue()
//  20/09/2020  mdu   added --> hasField(), getFieldNames(), create(), isEmpty()
//
// Copyright (c) 2017-2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.text.json;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.List;

import org.pfsw.bif.text.IStringConstant;
import org.pfsw.text.NaturalNumber;
import org.pfsw.text.TimeUnit;
import org.pfsw.text.TimeValue;

/**
 * This is a wrapper around a single JsonObject. It supports read and write access to the 
 * objects's field values in a convenient way and supports optional and required
 * field access as well as implicit type conversion.
 * All access problems are signaled by {@link JsonAccessException} which is a RuntimeException.
 *
 * @author Manfred Duchrow
 * @version 2.1
 */
public class JsonObjectAccessor
{
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private final JsonObject jsonObject;

  // =========================================================================
  // STATIC METHODS
  // =========================================================================
  /**
   * Creates an accessor on a new empty {@link JsonObject}.
   */
  public static JsonObjectAccessor create()
  {
    return create(JsonObject.create());
  }

  /**
   * Creates an accessor based on the given JSON object.
   * 
   * @param jsonObject The JSON object to be accessed (must not be null).
   * throws IllegalArgumentException if the given jsonObject is null.
   */
  public static JsonObjectAccessor create(final JsonObject jsonObject)
  {
    return new JsonObjectAccessor(jsonObject);
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Creates a new accessor based on the given JSON object.
   * 
   * @param jsonObject The JSON object to be accessed (must not be null).
   * throws IllegalArgumentException if the given jsonObject is null.
   */
  public JsonObjectAccessor(final JsonObject jsonObject)
  {
    super();
    if (jsonObject == null)
    {
      throw new IllegalArgumentException("JSON object must not be null");
    }
    this.jsonObject = jsonObject;
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the underlying JSON object.
   */
  public JsonObject getJsonObject()
  {
    return this.jsonObject;
  }

  /**
   * Returns all field names present in the underlying JSON object (maybe empty but never null).
   */
  public List<String> getFieldNames()
  {
    return new ArrayList<String>(getJsonObject().getNames());
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present or of another type that expected (according to the return type of this method).
   */
  public JsonObjectAccessor getRequiredJsonObjectAccessorValue(final IStringConstant name)
  {
    return getRequiredJsonObjectAccessorValue(name.asString());
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present or of another type that expected (according to the return type of this method).
   */
  public JsonObjectAccessor getRequiredJsonObjectAccessorValue(final String name)
  {
    JsonObject object;

    object = getRequiredJsonObjectValue(name);
    return new JsonObjectAccessor(object);
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present or of another type that expected (according to the return type of this method).
   */
  public JsonObject getRequiredJsonObjectValue(final IStringConstant name)
  {
    return getRequiredJsonObjectValue(name.asString());
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present or of another type that expected (according to the return type of this method).
   */
  public JsonObject getRequiredJsonObjectValue(final String name)
  {
    return getRequiredTypedValue(name, JsonObject.class);
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present or of another type that expected (according to the return type of this method).
   */
  public JsonArray getRequiredJsonArrayValue(final IStringConstant name)
  {
    return getRequiredJsonArrayValue(name.asString());
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present or of another type that expected (according to the return type of this method).
   */
  public JsonArray getRequiredJsonArrayValue(final String name)
  {
    return getRequiredTypedValue(name, JsonArray.class);
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present or of another type that expected (according to the return type of this method).
   */
  public String getRequiredStringValue(final IStringConstant name)
  {
    return getRequiredStringValue(name.asString());
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present or of another type that expected (according to the return type of this method).
   */
  public String getRequiredStringValue(final String name)
  {
    return getRequiredTypedValue(name, String.class);
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present or of another type that expected (according to the return type of this method).
   */
  public String[] getRequiredStringArrayValue(final IStringConstant name)
  {
    return getRequiredStringArrayValue(name.asString());
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present or of another type that expected (according to the return type of this method).
   */
  public String[] getRequiredStringArrayValue(final String name)
  {
    return getStringArray(name, true);
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present or of another type that expected (according to the return type of this method).
   */
  public Boolean getRequiredBooleanValue(final IStringConstant name)
  {
    return getRequiredBooleanValue(name.asString());
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present or of another type that expected (according to the return type of this method).
   */
  public Boolean getRequiredBooleanValue(final String name)
  {
    return getRequiredTypedValue(name, Boolean.class);
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present or of another type that expected (according to the return type of this method).
   */
  public Integer getRequiredIntegerValue(final IStringConstant name)
  {
    return getRequiredIntegerValue(name.asString());
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present or of another type that expected (according to the return type of this method).
   */
  public Integer getRequiredIntegerValue(final String name)
  {
    return getRequiredTypedValue(name, Integer.class);
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present or of another type that expected (according to the return type of this method).
   */
  public Long getRequiredLongValue(final IStringConstant name)
  {
    return getRequiredLongValue(name.asString());
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present or of another type that expected (according to the return type of this method).
   */
  public Long getRequiredLongValue(final String name)
  {
    return getRequiredTypedValue(name, Long.class);
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present or of another type that expected (according to the return type of this method).
   */
  public BigDecimal getRequiredBigDecimalValue(final IStringConstant name)
  {
    return getRequiredBigDecimalValue(name.asString());
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present or of another type that expected (according to the return type of this method).
   */
  public BigDecimal getRequiredBigDecimalValue(final String name)
  {
    return getRequiredTypedValue(name, BigDecimal.class);
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present or of another type that expected (according to the return type of this method).
   */
  public NaturalNumber getRequiredNaturalNumberValue(final IStringConstant name)
  {
    return getRequiredNaturalNumberValue(name.asString());
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present or of another type that expected (according to the return type of this method).
   */
  public NaturalNumber getRequiredNaturalNumberValue(final String name)
  {
    return getRequiredTypedValue(name, NaturalNumber.class);
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present or of another type that expected (according to the return type of this method).
   */
  public TimeValue getRequiredTimeValue(IStringConstant name, TimeUnit timeUnit)
  {
    return getRequiredTimeValue(name.asString(), timeUnit);
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present or of another type that expected (according to the return type of this method).
   */
  public TimeValue getRequiredTimeValue(String name, TimeUnit timeUnit)
  {
    return asTimeValue(getRequiredLongValue(name), timeUnit);
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present.
   */
  public Object getRequiredRawValue(final IStringConstant name)
  {
    return getRequiredRawValue(name.asString());
  }

  /**
   * Returns either the value associated to the given name or throws a {@link JsonAccessException}
   * if the value is not present.
   */
  public Object getRequiredRawValue(final String name)
  {
    return getRawValueOf(name, true);
  }

  /**
   * Returns either the value associated to the given name or null.
   */
  public JsonObjectAccessor getOptionalJsonObjectAccessorValue(final IStringConstant name)
  {
    return getOptionalJsonObjectAccessorValue(name.asString());
  }

  /**
   * Returns either the value associated to the given name or null.
   */
  public JsonObjectAccessor getOptionalJsonObjectAccessorValue(final String name)
  {
    JsonObject object;

    object = getOptionalJsonObjectValue(name);
    if (object == null)
    {
      return null;
    }
    return JsonObjectAccessor.create(object);
  }

  /**
   * Returns either the value associated to the given name or null.
   */
  public JsonObject getOptionalJsonObjectValue(final IStringConstant name)
  {
    return getOptionalJsonObjectValue(name.asString());
  }

  /**
   * Returns either the value associated to the given name or null.
   */
  public JsonObject getOptionalJsonObjectValue(final String name)
  {
    return getOptionalTypedValue(name, JsonObject.class);
  }

  /**
   * Returns either the value associated to the given name or null.
   */
  public JsonArray getOptionalJsonArrayValue(final IStringConstant name)
  {
    return getOptionalJsonArrayValue(name.asString());
  }

  /**
   * Returns either the value associated to the given name or null.
   */
  public JsonArray getOptionalJsonArrayValue(final String name)
  {
    return getOptionalTypedValue(name, JsonArray.class);
  }

  /**
   * Returns either the value associated to the given name or null.
   */
  public String getOptionalStringValue(final IStringConstant name)
  {
    return getOptionalStringValue(name.asString());
  }

  /**
   * Returns either the value associated to the given name or null.
   */
  public String getOptionalStringValue(final String name)
  {
    return getOptionalTypedValue(name, String.class);
  }

  /**
   * Returns either the value associated to the given name or null.
   */
  public String[] getOptionalStringArrayValue(final IStringConstant name)
  {
    return getOptionalStringArrayValue(name.asString());
  }

  /**
   * Returns either the value associated to the given name or null.
   */
  public String[] getOptionalStringArrayValue(final String name)
  {
    return getStringArray(name, false);
  }

  /**
   * Returns either the value associated to the given name or null.
   */
  public Boolean getOptionalBooleanValue(final IStringConstant name)
  {
    return getOptionalBooleanValue(name.asString());
  }

  /**
   * Returns either the value associated to the given name or null.
   */
  public Boolean getOptionalBooleanValue(final String name)
  {
    return getOptionalTypedValue(name, Boolean.class);
  }

  /**
   * Returns either the value associated to the given name or null.
   */
  public Integer getOptionalIntegerValue(final IStringConstant name)
  {
    return getOptionalIntegerValue(name.asString());
  }

  /**
   * Returns either the value associated to the given name or null.
   */
  public Integer getOptionalIntegerValue(final String name)
  {
    return getOptionalTypedValue(name, Integer.class);
  }

  /**
   * Returns either the value associated to the given name or null.
   */
  public Long getOptionalLongValue(final IStringConstant name)
  {
    return getOptionalLongValue(name.asString());
  }

  /**
   * Returns either the value associated to the given name or null.
   */
  public Long getOptionalLongValue(final String name)
  {
    return getOptionalTypedValue(name, Long.class);
  }

  /**
   * Returns either the value associated to the given name or null.
   */
  public BigDecimal getOptionalBigDecimalValue(final IStringConstant name)
  {
    return getOptionalBigDecimalValue(name.asString());
  }

  /**
   * Returns either the value associated to the given name or null.
   */
  public BigDecimal getOptionalBigDecimalValue(final String name)
  {
    return getOptionalTypedValue(name, BigDecimal.class);
  }

  /**
   * Returns either the value associated to the given name or null.
   */
  public NaturalNumber getOptionalNaturalNumberValue(final IStringConstant name)
  {
    return getOptionalNaturalNumberValue(name.asString());
  }

  /**
   * Returns either the value associated to the given name or null.
   */
  public NaturalNumber getOptionalNaturalNumberValue(final String name)
  {
    return getOptionalTypedValue(name, NaturalNumber.class);
  }

  /**
   * Returns the value at the specified field name as a {@link TimeValue}
   * with the specified {@link TimeUnit}. 
   * 
   * @param name The name of the field (must not be null).
   * @param timeUnit The time unit of the value found in the object (must not be null).
   * @return Returns either the TimeUnit or null if field not present.
   */
  public TimeValue getOptionalTimeValue(IStringConstant name, TimeUnit timeUnit)
  {
    return getOptionalTimeValue(name.asString(), timeUnit);
  }

  /**
   * Returns the value at the specified field name as a {@link TimeValue}
   * with the specified {@link TimeUnit}. 
   * 
   * @param name The name of the field (must not be null).
   * @param timeUnit The time unit of the value found in the object (must not be null).
   * @return Returns either the TimeUnit or null if field not present.
   */
  public TimeValue getOptionalTimeValue(String name, TimeUnit timeUnit)
  {
    return asTimeValue(getOptionalLongValue(name), timeUnit);
  }

  /**
   * Returns either the value associated to the given name or null.
   */
  public Object getOptionalRawValue(final IStringConstant name)
  {
    return getOptionalRawValue(name.asString());
  }

  /**
   * Returns either the value associated to the given name or null.
   */
  public Object getOptionalRawValue(final String name)
  {
    return getRawValueOf(name, false);
  }

  /**
   * Returns either the value associated to the given name or null if not preset.
   * @throws JsonAccessException if the value is not of the specified type. 
   */
  public <T> T getOptionalTypedValue(final IStringConstant name, Class<T> type)
  {
    return getOptionalTypedValue(name.asString(), type);
  }

  /**
   * Returns either the value associated to the given name or null if not preset.
   * @throws JsonAccessException if the value is not of the specified type. 
   */
  public <T> T getOptionalTypedValue(final String name, Class<T> type)
  {
    return getTypedValueOf(name, type, false);
  }

  /**
   * Returns either the value associated to the given name or @throws JsonAccessException 
   * if the value is not of the specified type or not present. 
   */
  public <T> T getRequiredTypedValue(final IStringConstant name, Class<T> type)
  {
    return getRequiredTypedValue(name.asString(), type);
  }

  /**
   * Returns either the value associated to the given name or @throws JsonAccessException 
   * if the value is not of the specified type or not present. 
   */
  public <T> T getRequiredTypedValue(final String name, Class<T> type)
  {
    return getTypedValueOf(name, type, true);
  }

  public <T> T getTypedValueOf(final IStringConstant name, final Class<T> type, final boolean isRequired)
  {
    return getTypedValueOf(name.asString(), type, isRequired);
  }

  public <T> T getTypedValueOf(final String name, final Class<T> type, final boolean isRequired)
  {
    T value;

    try
    {
      if (type == Long.class)
      {
        value = type.cast(getLongValueOf(name));
      }
      else if (type == NaturalNumber.class)
      {
        value = type.cast(getNaturalNumberValueOf(name));
      }
      else
      {
        value = getJsonObject().getValueOfType(name, type);
      }
    }
    catch (Exception ex)
    {
      throw new JsonAccessException(ex, "Type conversion problem for value of \"%s\". The value is not %s", name, type.getSimpleName());
    }
    if ((value == null) && isRequired)
    {
      signalRequiredFieldMissing(name);
    }
    return value;
  }

  public Object getRawValueOf(final IStringConstant name, final boolean isRequired)
  {
    return getRawValueOf(name.asString(), isRequired);
  }

  public Object getRawValueOf(final String name, final boolean isRequired)
  {
    Object value;

    value = getJsonObject().get(name);
    if ((value == null) && isRequired)
    {
      signalRequiredFieldMissing(name);
    }
    return value;
  }

  /**
   * Sets the field with the given name to the specified value.
   * 
   * @param name The name of the field.
   * @param value The value to be set.
   * @return this object instance.
   */
  public JsonObjectAccessor setFieldValue(String name, Object value)
  {
    getJsonObject().put(name, value);
    return this;
  }

  /**
   * Sets the field with the given name to the specified value.
   * 
   * @param name The name of the field.
   * @param value The value to be set.
   * @return this object instance.
   */
  public JsonObjectAccessor setFieldValue(IStringConstant name, Object value)
  {
    getJsonObject().setField(name, value);
    return this;
  }

  /**
   * Returns whether or not this JSON object contains a field with the given name.
   */
  public boolean hasField(String name)
  {
    return getJsonObject().hasField(name);
  }

  /**
   * Returns whether or not this JSON object contains a field with the given name.
   */
  public boolean hasField(IStringConstant name)
  {
    return hasField(name.asString());
  }

  /**
   * Returns true if the underlying JSON object contains no field at all.
   */
  public boolean isEmpty()
  {
    return getJsonObject().isEmpty();
  }

  public void signalRequiredFieldMissing(String fieldName) throws JsonAccessException
  {
    throw new JsonAccessException("Required field missing: \"%s\"", fieldName);
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected String[] getStringArray(final String name, boolean isRequired)
  {
    JsonArray jsonArray;
    String[] strings;

    jsonArray = getTypedValueOf(name, JsonArray.class, isRequired);
    if (jsonArray == null)
    {
      return null;
    }
    strings = new String[jsonArray.size()];
    for (int i = 0; i < strings.length; i++)
    {
      strings[i] = jsonArray.getString(i);
    }
    return strings;
  }

  protected NaturalNumber getNaturalNumberValueOf(final String name)
  {
    Long longValue;

    longValue = getLongValueOf(name, "NaturalNumber");
    if (longValue == null)
    {
      return null;
    }
    return NaturalNumber.valueOf(longValue.longValue());
  }

  protected Long getLongValueOf(final String name)
  {
    return getLongValueOf(name, "Long");
  }

  protected Long getLongValueOf(final String name, String targetTypeName)
  {
    Object value;
    Integer intValue;

    value = getJsonObject().get(name);
    if (value == null)
    {
      return null;
    }
    if (value instanceof Long)
    {
      return (Long)value;
    }
    if (value instanceof Integer)
    {
      intValue = (Integer)value;
      return Long.valueOf(intValue.longValue());
    }
    throw new ClassCastException("Field named <" + name + "> is not of type " + targetTypeName + ". It is a " + value.getClass().getName());
  }

  protected TimeValue asTimeValue(Long value, TimeUnit timeUnit)
  {
    if (value == null)
    {
      return null;
    }
    return new TimeValue(value, timeUnit);
  }
}
