// ===========================================================================
// CONTENT  : ABSTRACT CLASS JsonObjectWrapper
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.1 - 20/09/2020
// HISTORY  :
//  02/05/2020  mdu  CREATED
//  20/09/2020  mdu   added toString()
//
// Copyright (c) 2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.text.json;

/**
 * This class is a wrapper around a single {@link JsonObject} that holds it in
 * a {@link JsonObjectAccessor} for easy access.
 * <br>
 * Subclasses are supposed to implement getters and setters for the relevant attributes
 * utilizing enum values that implement {@link org.pfsw.bif.enums.IFieldName} or {@link org.pfsw.bif.text.IStringConstant}.
 * <p>
 * Example: <br>
 * <code>
 * public String getName()
 * {
 *   return getJsonAccessor().getOptionalStringValue(AttrNameConstants.NAME);
 * }
 * </code>
 *
 * @author Manfred Duchrow
 * @version 1.1
 */
public abstract class JsonObjectWrapper
{
  private final JsonObjectAccessor jsonAccessor;

  public JsonObjectWrapper()
  {
    this(new JsonObject());
  }

  public JsonObjectWrapper(JsonObject jsonObject)
  {
    this(JsonObjectAccessor.create(jsonObject));
  }

  public JsonObjectWrapper(JsonObjectAccessor jsonAccessor)
  {
    super();
    this.jsonAccessor = jsonAccessor;
  }

  @Override
  public String toString()
  {
    return String.format("%s(%s)", getClass().getSimpleName(), getJsonAccessor().getJsonObject());
  }

  /**
   * Returns the underlying {@link JsonObject};
   */
  protected JsonObject getJsonObject() 
  {
    return getJsonAccessor().getJsonObject();
  }
  
  /**
   * Returns the underlying {@link JsonObjectAccessor};
   */
  protected JsonObjectAccessor getJsonAccessor()
  {
    return this.jsonAccessor;
  }
}
