// ===========================================================================
// CONTENT  : CLASS JsonWriter
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 10/05/2020
// HISTORY  :
//  10/05/2020  mdu  CREATED
//
// Copyright (c) 2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.text.json;

import static org.pfsw.bif.text.JsonSyntaxElement.*;
import static org.pfsw.text.json.JsonUtil.*;

import java.util.Collection;

import org.pfsw.bif.text.JsonSyntaxElement;
import org.pfsw.text.FormattedTextWriter;

/**
 * Writes a given {@link JsonObject} into an {@link Appendable} 
 * using JSON string representation format.
 * <p>
 * This class is thread-safe.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class JsonWriter
{
  public static final boolean DEFAULT_OPEN_BRACES_ON_NEW_LINE = false;
  public static final int DEFAULT_INDENTATION_SPACES = 2;

  private final int indentationSpaces;
  private final boolean openingBracesOnNewLine;

  /**
   * Creates a new instance with opening braces '{' and '[' on the same line.
   * 
   * @return The created instance.
   */
  public static JsonWriter create()
  {
    return new JsonWriter();
  }

  /**
   * Creates a new instance with opening braces '{' and '[' on a new line.
   * 
   * @return The created instance.
   */
  public static JsonWriter createWithOpeningBracesOnNewLine()
  {
    return new JsonWriter(true);
  }

  /**
   * Creates a new instance with opening braces '{' and '[' on the same line
   * and the specified indentation spaces per indentation level.
   * 
   * @param indentationSpaces The number of spaces per indentation level.
   * @return The created instance.
   * @throws IllegalArgumentException if the given indentation size is less than 0 or greater than 20.
   */
  public static JsonWriter create(int indentationSpaces)
  {
    return new JsonWriter(indentationSpaces);
  }

  /**
   * Creates a new instance with opening braces '{' and '[' on a new line.
   * 
   * @param indentationSpaces The number of spaces per indentation level.
   * @return The created instance.
   * @throws IllegalArgumentException if the given indentation size is less than 0 or greater than 20.
   */
  public static JsonWriter createWithOpeningBracesOnNewLine(int indentationSpaces)
  {
    return new JsonWriter(indentationSpaces, true);
  }

  /**
   * Convenience method to convert the given {@link JsonObject} to a pretty printed 
   * JSON string representation (with opening braces on same line and 2 spaces indentation).
   * 
   * @param jsonObject The JSON object to convert to a pretty string.
   * @return The pretty string JSON representation. 
   */
  public static String prettyPrint(JsonObject jsonObject)
  {
    return create().asString(jsonObject);
  }

  /**
   * Creates a new instance with opening braces '{' and '[' on the same line
   * and 2 indentation spaces per indentation level.
   */
  public JsonWriter()
  {
    this(DEFAULT_OPEN_BRACES_ON_NEW_LINE);
  }

  /**
   * Creates a new instance with a default of 2 indentation spaces per indentation level.
   * 
   * @param openingBracesOnNewLine Defines whether or not opening braces '{' and '[' must be put on new line.
   */
  public JsonWriter(boolean openingBracesOnNewLine)
  {
    this(DEFAULT_INDENTATION_SPACES, openingBracesOnNewLine);
  }

  /**
   * Creates a new instance with opening braces '{' and '[' on the same line
   * and the specified indentation spaces per indentation level.
   * 
   * @param indentationSpaces The number of spaces per indentation level.
   * @throws IllegalArgumentException if the given indentation size is less than 0 or greater than 20.
   */
  public JsonWriter(int indentationSpaces)
  {
    this(indentationSpaces, DEFAULT_OPEN_BRACES_ON_NEW_LINE);
  }

  /**
   * Creates a new instance.
   * 
   * @param indentationSpaces The number of spaces per indentation level.
   * @param openingBracesOnNewLine Defines whether or not opening braces '{' and '[' must be put on new line.
   * @throws IllegalArgumentException if the given indentation size is less than 0 or greater than 20.
   */
  public JsonWriter(int indentationSpaces, boolean openingBracesOnNewLine)
  {
    super();
    this.indentationSpaces = indentationSpaces;
    this.openingBracesOnNewLine = openingBracesOnNewLine;
  }

  /**
   * Convenience method to convert the given {@link JsonObject} to a pretty printed
   * JSON string representation.
   * 
   * @param jsonObject The JSON object to be converted to a string.
   * @return The JSON string representation. 
   */
  public String asString(JsonObject jsonObject)
  {
    StringBuffer buffer;

    buffer = new StringBuffer();
    write(buffer, jsonObject);
    return buffer.toString();
  }

  /**
   * Writes the given {@link JsonObject} to the specified {@link Appendable} with line breaks
   * and indentation.
   * 
   * @param target The target to which write the output.
   * @param jsonObject The object to be written as JSON string representation.
   */
  @SuppressWarnings("resource")
  public void write(Appendable target, JsonObject jsonObject)
  {
    FormattedTextWriter writer;

    writer = createFormattedTextWriterFor(target);
    try
    {
      writer.writeln(OBJECT_START).indent();
      writeJsonObjectFields(writer, jsonObject);
      writeObjectEnd(writer);
    }
    catch (RuntimeException e)
    {
      throw new JsonOutputException(e, "Error when trying to write object to JSON string representation");
    }
  }

  protected void writeField(FormattedTextWriter writer, FieldData field)
  {
    JU.appendJSONString(writer, field.getName());
    writer.write(PAIR_SEPARATOR);
    writeValue(writer, field.getValue(), true);
  }

  protected void writeValue(FormattedTextWriter writer, Object value, boolean isSingleValue)
  {
    if (value instanceof Collection)
    {
      writeMultipleValues(writer, (Collection<?>)value);
    }
    else if (value instanceof JsonObject)
    {
      writeJsonObject(writer, (JsonObject)value);
    }
    else
    {
      if (isSingleValue)
      {
        writer.write(" ");
      }
      JU.appendJSONObject(writer, value);
    }
  }

  protected void writeJsonObject(FormattedTextWriter writer, JsonObject jsonObject)
  {
    writeObjectStart(writer);
    writeJsonObjectFields(writer, jsonObject);
    writeObjectEnd(writer);
  }

  protected void writeJsonObjectFields(FormattedTextWriter writer, JsonObject jsonObject)
  {
    int counter = 0;

    for (String fieldName : jsonObject.getNames())
    {
      counter++;
      if (counter > 1)
      {
        writer.write(ELEMENT_SEPARATOR);
        writer.newLine();
      }
      FieldData fieldData = new FieldData(fieldName, jsonObject.getValue(fieldName));
      writeField(writer, fieldData);
    }
  }

  protected void writeMultipleValues(FormattedTextWriter writer, Collection<?> valueCollection)
  {
    int elementCounter = 0;

    if (valueCollection.isEmpty())
    {
      writer.write(" ").write(ARRAY_START).write(ARRAY_END);
      return;
    }
    writeArrayStart(writer);
    for (Object value : valueCollection)
    {
      elementCounter++;
      if (elementCounter > 1)
      {
        writer.writeln(ELEMENT_SEPARATOR);
      }
      writeValue(writer, value, false);
    }
    writeArrayEnd(writer);
  }

  protected void writeObjectStart(FormattedTextWriter writer)
  {
    writeStart(writer, OBJECT_START);
  }

  protected void writeObjectEnd(FormattedTextWriter writer)
  {
    writer.newLine().outdent();
    writer.write(OBJECT_END);
  }

  protected void writeArrayStart(FormattedTextWriter writer)
  {
    writeStart(writer, ARRAY_START);
  }

  protected void writeArrayEnd(FormattedTextWriter writer)
  {
    writer.newLine().outdent();
    writer.write(ARRAY_END);
  }

  protected void writeStart(FormattedTextWriter writer, JsonSyntaxElement startElement)
  {
    if (isOpeningBracesOnNewLine())
    {
      writer.newLine();
      writer.write(startElement);
    }
    else
    {
      writer.write(" ").write(startElement);
    }
    writer.newLine();
    writer.indent();
  }

  /**
   * Returns how many spaces per indentation level will be used for formatting. 
   */
  public int getIndentationSpaces()
  {
    return this.indentationSpaces;
  }

  /**
   * Returns true if opening braces like '{' and '[' are put on a new line. 
   */
  public boolean isOpeningBracesOnNewLine()
  {
    return this.openingBracesOnNewLine;
  }


  protected FormattedTextWriter createFormattedTextWriterFor(Appendable target)
  {
    return FormattedTextWriter.create(target, getIndentationSpaces());
  }

  // ============== Inner classes ======================================

  protected static class FieldData
  {
    private final String name; // never null
    private final Object value; // may be null

    FieldData(String name, Object value)
    {
      super();
      this.name = name;
      this.value = value;
    }

    public String getName()
    {
      return this.name;
    }

    public Object getValue()
    {
      return this.value;
    }
  }
}
