// ===========================================================================
// CONTENT  : CLASS BasicTextFunctionResolver
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.2 - 15/11/2014
// HISTORY  :
//  26/06/1999 	duma  CREATED
//	25/01/2000	duma	moved		-> from package 'com.mdcs.text'
//	06/03/2004	duma	changed	-> chack parater == null in funcdate()
//  15/11/2014  mdu   changed -> clean code, using List instead of Vector
//
// Copyright (c) 1999-2044, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.textx;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;

/**
 * This class provides support for some basic functions,
 * that are useful in text replacement done by a TextEngine.
 *
 * @author Manfred Duchrow
 * @version 1.2
 * @see TextEngine
 */
public class BasicTextFunctionResolver implements FunctionResolver
{
  // ========================================================================
  // PRIVATE CONSTANTS
  // ========================================================================
  /**
   * Name for a function that gets one string parameter and returns it
   * with a capitalized first letter.
   * Example: UC1("random") ==> "Random"
   */
  public static final String UPPERCASE_FIRST_LETTER = "UC1";
  /**
   * Name for a function that gets one string parameter and returns it
   * with a lower case first letter.
   * Example: LC1("FirstEntry") ==> "firstEntry"
   */
  public static final String LOWERCASE_FIRST_LETTER = "LC1";
  /**
   * Name for a function that gets one string parameter and returns it
   * with all letters changed to upper case.
   * Example: UC1("random") ==> "RANDOM"
   */
  public static final String UPPERCASE = "UC";
  /**
   * Name for a function that gets one string parameter and returns it
   * with all letters changed to lower case.
   * Example: LC("FirstEntry") ==> "firstentry"
   */
  public static final String LOWERCASE = "LC";
  /**
   * Name for a function that gets one string parameter and returns it
   * with a lower case first letter.
   * Example: Date() ==> "January 12, 1999"
   */
  public static final String DATE = "Date";

  // ========================================================================
  // IMPLEMENTATION OF INTERFACE FunctionResolver
  // ========================================================================

  /**
   * Returns the value for the function with the given name.
   *
   * @param functionName The case sensitive name of the function.
   * @param parameter A collection of parameters for the function.
   * @return The evaluation of the function with the given parameters (null is a valid return value)
   * @throws UnknownFunctionException The receiver is not knowing the function.
   */
  @Override
  public Object executeFunction(String functionName, List<String> parameter) throws UnknownFunctionException, InvalidParameterException
  {
    if (functionName.equals(UPPERCASE_FIRST_LETTER))
    {
      return funcUppercaseFirstLetter(parameter);
    }
    if (functionName.equals(LOWERCASE_FIRST_LETTER))
    {
      return funcLowercaseFirstLetter(parameter);
    }
    if (functionName.equals(UPPERCASE))
    {
      return funcUppercase(parameter);
    }
    if (functionName.equals(LOWERCASE))
    {
      return funcLowercase(parameter);
    }
    if (functionName.equals(DATE))
    {
      return funcDate(parameter);
    }
    throw (new UnknownFunctionException(functionName));
  }

  /**
   * Returns if the function with the given name can be resolved by the receiver.
   *
   * @param functionName The case sensitive name of the function.
   * @return Whether the function with the given name is known or not.
   */
  @Override
  public boolean isKnownFunction(String functionName)
  {
    if (functionName.equals(UPPERCASE_FIRST_LETTER))
    {
      return true;
    }
    if (functionName.equals(LOWERCASE_FIRST_LETTER))
    {
      return true;
    }
    if (functionName.equals(UPPERCASE))
    {
      return true;
    }
    if (functionName.equals(LOWERCASE))
    {
      return true;
    }
    if (functionName.equals(DATE))
    {
      return true;
    }
    return false;
  }

  // ========================================================================
  // SUPPORTED FUNCTIONS : PROTECTED INSTANCE METHODS
  // ========================================================================

  protected String upperLowerFirstLetter(List<String> parameter, boolean upper) throws InvalidParameterException
  {
    String param1 = null;
    String result = null;

    checkParameterCount(UPPERCASE_FIRST_LETTER, parameter, 1);

    param1 = parameter.get(0);
    if (param1.length() > 0)
    {
      result = param1.substring(0, 1);
      if (upper)
      {
        result = result.toUpperCase() + param1.substring(1);
      }
      else
      {
        result = result.toLowerCase() + param1.substring(1);
      }
    }
    else
    {
      result = param1;
    }
    return result;
  }

  protected String funcUppercaseFirstLetter(List<String> parameter) throws InvalidParameterException
  {
    return upperLowerFirstLetter(parameter, true);
  }

  protected String funcLowercaseFirstLetter(List<String> parameter) throws InvalidParameterException
  {
    return upperLowerFirstLetter(parameter, false);
  }

  protected String funcUppercase(List<String> parameter) throws InvalidParameterException
  {
    checkParameterCount(UPPERCASE, parameter, 1);
    String param = parameter.get(0);
    return param.toUpperCase();
  }

  // 

  protected String funcLowercase(List<String> parameter) throws InvalidParameterException
  {
    checkParameterCount(LOWERCASE, parameter, 1);
    String param = parameter.get(0);
    return param.toLowerCase();
  }

  protected String funcDate(List<String> parameter) throws InvalidParameterException
  {
    SimpleDateFormat df = null;

    if (parameter != null)
    {
      if (parameter.size() == 0)
      {
        df = new SimpleDateFormat();
      }
      else if (parameter.size() == 1)
      {
        df = new SimpleDateFormat(parameter.get(0));
      }
      else
      {
        checkParameterCount(DATE, parameter, 1);
      }
    }
    if (df == null)
      df = new SimpleDateFormat();

    return df.format(new Date());
  }

  // ========================================================================
  // PROTECTED INSTANCE METHODS
  // ========================================================================
  protected void checkParameterCount(String funcName, List<String> parameter, int count) throws InvalidParameterException
  {
    String message = null;

    if (parameter.size() != count)
    {
      message = "Wrong number of parameters ! ";
      message = message + "Required: " + Integer.toString(count) + " ; ";
      message = message + "Given: " + Integer.toString(parameter.size()) + ".";
      throw (new InvalidParameterException(funcName, message));
    }
  }
}