// ===========================================================================
// CONTENT  : CLASS BasicVariableContainer
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.0 - 16/11/2014
// HISTORY  :
//  31/07/1999 	duma  CREATED
//	25/01/2000	duma	moved		-> 	from package 'com.mdcs.text'
//	25/01/2000	duma	added		->	setValueFor( String, boolean )
//  07/02/2000  duma  chnaged ->  text for UnknownVariableException
//	06/02/2004	duma	added		->	addFrom()
//  16/11/2014  mdu   changed ->  new constructor for Properties
//
// Copyright (c) 1999-2014, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.textx;

import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import org.pfsw.util.PropertiesMap;

/**
 * This is a basic implementation of the interface <i>VariableResolver</i>.
 * It provides a container for a simple mapping of variable name to their
 * corresponding values.
 *
 * @author Manfred Duchrow
 * @version 2.0
 */
public class BasicVariableContainer implements VariableContainer
{
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private Map<String, ?> variableMapping = null;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.    <br>
   * Subclasses that override this constructor must call <i>super()</i> first, 
   * before doing anything else.
   */
  public BasicVariableContainer()
  {
    super();
    setVariableMapping(createNewMapping());
  }

  /**
   * Initialize the new instance with the given mapping.    <br>
   * 
   * @param mapping The instance of a <i>Map</i> interface compliant class.
   */
  public BasicVariableContainer(Map<String, Object> mapping)
  {
    super();
    setVariableMapping(mapping);
  }

  /**
   * Initialize the new instance with the given properties.<br>
   * 
   * @param properties The properties to initialize the new instance with (must not be null).
   */
  public BasicVariableContainer(Properties properties)
  {
    super();

    Map<String, String> mapping = new PropertiesMap(properties);
    setVariableMapping(mapping);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================

  /**
   * Returns all variable name the resolver currently knows
   */
  @Override
  public Set<String> knownVariableNames()
  {
    return getVariableMapping().keySet();
  }

  /**
   * Returns the value for the variable with the given name.
   *
   * @param varName The case sensitive name of the variable.
   * @return The value associated to the given variable (even null is a valid value)
   * @throws UnknownVariableException The receiver is not knowing the variable.
   */
  @Override
  public Object getValue(String varName) throws UnknownVariableException
  {
    Object value = null;

    value = getVariableMapping().get(varName);
    if (value == null)
    {
      throw (new UnknownVariableException(varName));
    }
    return value;
  }

  /**
   * Returns if the variable with the given name can be resolved by the receiver.
   *
   * @param varName The case sensitive name of the variable.
   * @return Whether the variable with the given name is known or not.
   */
  @Override
  public boolean isKnownVariable(String varName)
  {
    Object value = null;

    value = getVariableMapping().get(varName);
    return (value != null);
  }

  /**
   * Sets the value of the variable with the given name.
   *
   * @param varName The case sensitive name of the variable. Must not be null !
   * @param value The new value of the variable. Must not be null !
   */
  @Override
  public void setValueFor(String varName, Object value)
  {
    if ((varName != null) && (value != null))
    {
      getVariableMapping().put(varName, value);
    }
  }

  /**
   * Sets the boolean value of the variable with the given name.
   *
   * @param varName The case sensitive name of the variable. Must not be null !
   * @param value The boolean value
   */
  public void setValueFor(String varName, boolean value)
  {
    if (varName != null)
    {
      getVariableMapping().put(varName, value ? Boolean.TRUE : Boolean.FALSE);
    }
  }

  /**
   * Sets the value of the variable with the given name.
   *
   * @param varName The case sensitive name of the variable. Must not be null !
   * @param value The new value of the variable. Must not be null !
   */
  @Override
  public void setValue(String varName, String value)
  {
    if ((varName != null) && (value != null))
    {
      getVariableMapping().put(varName, value);
    }
  }

  /**
   * Remove the variable with the specified name.
   * 
   * @param varName The name of the variable to be removed
   */
  @Override
  public void removeVariable(String varName)
  {
    if (varName != null)
    {
      getVariableMapping().remove(varName);
    }
  }

  /**
   * Adds all key value pairs from the given map to the variables, if the
   * key is a String.
   */
  public void addFrom(Map<String, Object> vars)
  {
    if (vars == null)
    {
      return;
    }
    for (Map.Entry<String, Object> entry : vars.entrySet())
    {
      setValueFor(entry.getKey(), entry.getValue());
    }
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  /**
   * Returns a new instance for the mapping.    <br>
   * Subclasses may override this method to support a different mapping class.
   *  
   * @return A new instance of a <i>Map</i> compliant class.
   */
  protected Map<String, Object> createNewMapping()
  {
    return new HashMap<String, Object>();
  }

  @SuppressWarnings("unchecked")
  protected Map<String, Object> getVariableMapping()
  {
    return (Map<String, Object>)variableMapping;
  }

  protected void setVariableMapping(Map<String, ? extends Object> aValue)
  {
    variableMapping = aValue;
  }
}
