// ===========================================================================
// CONTENT  : CLASS TextEngine
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.7.2 - 16/08/2007
// HISTORY  :
//  25/06/1999 	duma  CREATED
//	03/10/1999	duma	renamed	-> 	from TextCompletioner
//	09/12/1999	duma	added		-> 	indentation support
//	25/01/2000	duma	moved		-> 	from package 'com.mdcs.text'
//	25/01/2000	duma	changed	->	refactored
//	25/01/2000	duma	added		->	control elements #IF(), #ELSE, #ENDIF
//  01/02/2000  duma  changed ->  Getter/Setter for resolver are public now
//  07/02/2000  duma  added   ->  control element #IFDEF
//	24/03/2002	duma	added		->	support of special name characters and replacable delimiters
//	24/10/2002	duma	added		->	allowMissingPlaceholders(), useDollarCurlyBrackets()
//	06/03/2004	duma	changed	->	all constants to be protected static final and not private final
//	16/08/2007	mdu		changed	->	cleanup style, use StringBuffer
//
// Copyright (c) 1999-2007, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.textx;

import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.EmptyStackException;
import java.util.List;
import java.util.Stack;

import org.pfsw.text.StringUtil;
import org.pfsw.util.Bool;

/**
 * Instances of this class are able to replace 
 * all variables by their values in a given text
 *
 * @author Manfred Duchrow
 * @version 1.7.2
 */
public class TextEngine
{
  // =======================================================================
  // CONSTANTS
  // =======================================================================
  protected static final Character DEFAULT_VAR_START_DELIMITER = new Character('%');
  protected static final Character DEFAULT_VAR_END_DELIMITER = new Character('%');
  protected static final Character DEFAULT_TEXT_DELIMITER = new Character('\'');
  protected static final Character DEFAULT_PARAM_SEPARATOR = new Character(',');
  protected static final Character DEFAULT_FUNCPARAM_START = new Character('(');
  protected static final Character DEFAULT_FUNCPARAM_END = new Character(')');
  protected static final Character CONTROL_INDICATOR = new Character('#');

  protected static final String CONTROL_IF = "IF";
  protected static final String CONTROL_ELSE = "ELSE";
  protected static final String CONTROL_ENDIF = "ENDIF";
  protected static final String CONTROL_IFDEF = "IFDEF";

  protected static final int EMPTY_CACHE = -99;

  // =======================================================================
  // INSTANCE VARIABLES
  // =======================================================================
  private VariableResolver variableResolver = null;
  private FunctionResolver functionResolver = null;
  private StringReader sourceBuffer = null;
  private StringWriter resultBuffer = null;
  private Character nextChar = null;
  private boolean nextCharFilled = false;
  private int charCache = EMPTY_CACHE;
  private Stack<Bool> executionControl = new Stack<Bool>();
  private boolean supressMissingVariableException = false;
  private Character varStartPrefix = null;
  private Character varStartDelimiter = DEFAULT_VAR_START_DELIMITER;
  private Character varEndDelimiter = DEFAULT_VAR_END_DELIMITER;
  private String indentationFiller = null;
  private String specialNameCharacters = "";

  // =======================================================================
  // CONSTRUCTOR
  // =======================================================================
  /**
   * Initializes the new created instance with the given objects
   * for resolving variable and function names to values.
   *
   * @param varResolver The object that provides values for variables
   * @see #setVariableResolver( VariableResolver )
   */
  public TextEngine(VariableResolver varResolver)
  {
    setVariableResolver(varResolver);
  }

  /**
   * Initializes the new created instance with the given objects
   * for resolving variable and function names to values.
   *
   * @param varResolver The object that provides values for variables
   * @param funcResolver The object that provides values for function calls
   * @see #setVariableResolver( VariableResolver )
   * @see #setFunctionResolver( FunctionResolver )
   */
  public TextEngine(VariableResolver varResolver, FunctionResolver funcResolver)
  {
    this(varResolver);
    setFunctionResolver(funcResolver);
  }

  // =======================================================================
  // PUBLIC INSTANCE METHODS
  // =======================================================================
  /** 
   * Returns the variable resolver 
   */
  public VariableResolver getVariableResolver()
  {
    return variableResolver;
  }

  /** 
   * Sets the variable resolver 
   */
  public void setVariableResolver(VariableResolver vr)
  {
    variableResolver = vr;
  }

  /** 
   * Returns the function resolver 
   */
  public FunctionResolver getFunctionResolver()
  {
    return functionResolver;
  }

  /** 
   * Sets the function resolver 
   */
  public void setFunctionResolver(FunctionResolver fr)
  {
    functionResolver = fr;
  }

  /**
   * Returns the character that is specified as prefix of the placeholder
   * delimiter or null, if none is used.
   * <br>
   * Example: 
   *  %varName% -> Has no prefix, this method returns null  <br>
   *  ${varName} -> Uses '$' as prefix, this method returns Character('$')
   */
  public Character getVarStartPrefix()
  {
    return varStartPrefix;
  }

  /**
   * Sets the character that is is used as prefix of the placeholder
   * start delimiter.  <br>
   * Set to null, if no prefix is desired.
   * 
   * @param prefix The prefix character of null to use none.
   */
  public void setVarStartPrefix(Character prefix)
  {
    varStartPrefix = prefix;
  }

  /**
   * Returns the delimiter that marks the start of a variable
   */
  public Character getVarStartDelimiter()
  {
    return varStartDelimiter;
  }

  /**
   * Sets the delimiter that marks the start of a variable
   */
  public void setVarStartDelimiter(Character newValue)
  {
    varStartDelimiter = newValue;
  }

  /**
   * Returns the delimiter that marks the end of a variable
   */
  public Character getVarEndDelimiter()
  {
    return varEndDelimiter;
  }

  /**
   * Sets the delimiter that marks the end of a variable
   */
  public void setVarEndDelimiter(Character newValue)
  {
    varEndDelimiter = newValue;
  }

  /**
   * Returns all extra charcters, that are allowed in placeholder names.
   * <br>
   * By default only [A-Z][a-z][0-9] and '_' are allowed.
   */
  public String getSpecialNameCharacters()
  {
    return specialNameCharacters;
  }
  /**
   * Sets all extra charcters, that are allowed in placeholder names.
   * <br>
   * By default only [A-Z][a-z][0-9] and '_' are allowed.
   */
  public void setSpecialNameCharacters(String newValue)
  {
    specialNameCharacters = newValue;
  }

  /**
   * Returns the given text, after replaceing all placeholders with according text.
   *
   * @param text The input text containing placeholders.
   * @return the given input text with all placeholders replaced
   */
  public String completeText(String text) throws TextReplacementException
  {
    initializeBuffers(text);

    copyText();

    return resultString();
  }

  /**
   * Sets the number of spaces a text must be indented.    <br>
   *
   * @param indentSize the number of indentation spaces ( 0 means no indentation )
   */
  public void setIndentation(int indentSize)
  {
    if (indentSize <= 0)
    {
      setIndentationFiller(null);
    }
    else
    {
      setIndentationFiller(strUtil().repeat(StringUtil.CH_SPACE, indentSize));
    }
  }

  /**
   * Sets the character that indicates the start of a placeholder or
   * directive to the specified value.
   */
  public void setVarStartDelimiter(char newValue)
  {
    setVarStartDelimiter(new Character(newValue));
  }

  /**
   * Sets the character that indicates the end of a placeholder or
   * directive to the specified value.
   */
  public void setVarEndDelimiter(char newValue)
  {
    setVarEndDelimiter(new Character(newValue));
  }

  /**
   * Sets the delimiters to ${}
   * <p>
   * Placeholders then must look like ${varName}.
   */
  public void useDollarCurlyBrackets()
  {
    setVarStartPrefix(new Character('$'));
    setVarStartDelimiter(new Character('{'));
    setVarEndDelimiter(new Character('}'));
  }

  /**
   * Allows placeholders that cannot be resolved by the variable resolver.
   * That means, that the placeholder's name itself will be placed in
   * the text.
   * <br>
   * The default behaviour of a TextEngine is to throw an exception, if 
   * a placeholder can't be resolved.
   */
  public void allowMissingPlaceholders()
  {
    supressMissingVariableException(true);
  }

  /**
   * Forbids placeholders that cannot be resolved by the variable resolver.
   * That means, that an exception will be thrown, if a placeholder can't
   * be resolved.
   * <br>
   * The default behaviour of a TextEngine is to throw an exception, if 
   * a placeholder can't be resolved.
   */
  public void forbidMissingPlaceholders()
  {
    supressMissingVariableException(false);
  }

  // =======================================================================
  // PROTECTED INSTANCE METHODS
  // =======================================================================
  protected void initializeResultBuffer(int initialCapacity)
  {
    setResultBuffer(new StringWriter(initialCapacity));
    if (isIndentationOn())
    {
      getResultBuffer().write(getIndentationFiller());
    }
  }

  protected void initializeSourceBuffer(String source)
  {
    setSourceBuffer(new StringReader(source));
  }

  protected void initializeBuffers(String text)
  {
    initializeResultBuffer(text.length());
    initializeSourceBuffer(text);
  }

  protected void copyText() throws TextReplacementException
  {
    Character ch = null;

    while (hasMoreChars())
    {
      ch = nextCharacter();
      if (isInsertionStart(ch))
      {
        writeToResultBuffer(getPlaceholderValue());
      }
      else
      {
        writeToResultBuffer(ch.charValue());
        if ((isIndentationOn()) && (ch.charValue() == '\n'))
        {
          writeToResultBuffer(getIndentationFiller());
        }
      }
    }
  }

  protected boolean isInsertionStart(final Character character)
  {
    Character ch = character;

    if (hasDelimiterPrefix())
    {
      if (ch.equals(getVarStartPrefix()))
      {
        ch = nextCharacter();
        if (ch.equals(getVarStartDelimiter()))
        {
          return true;
        }
        putBackNextChar();
      }
      return false;
    }
    return ch.equals(getVarStartDelimiter());
  }

  protected boolean hasDelimiterPrefix()
  {
    return getVarStartPrefix() != null;
  }

  protected void writeToResultBuffer(String text)
  {
    if (isWritingOn())
    {
      getResultBuffer().write(text);
    }
  }

  protected void writeToResultBuffer(char ch)
  {
    if (isWritingOn())
    {
      getResultBuffer().write(ch);
    }
  }

  protected String getPlaceholderName()
  {
    StringBuffer buffer;
    Character ch;

    buffer = new StringBuffer(40);
    while (hasMoreChars())
    {
      ch = nextCharacter();
      if (isValidPlaceholderNameCharacter(ch.charValue()))
      {
        buffer.append(ch);
      }
      else
      {
        putBackNextChar();
        break;
      }
    }

    return buffer.toString();
  }

  protected String getPlaceholderValue() throws TextReplacementException
  {
    Character ch = null;
    String placeholderName = null;

    if (!hasMoreChars())
    {
      throw new TextReplacementException("Unexpected end of string reached !");
    }

    ch = nextCharacter();

    // Duplication of start delimiter will produce the delimiter itself
    if (ch.equals(getVarStartDelimiter()))
    {
      return getVarStartDelimiter().toString();
    }

    if (ch.equals(getControlIndicator()))
    {
      return evaluateControl();
    }

    putBackNextChar();

    placeholderName = getPlaceholderName();

    ch = nextCharacter();

    if (ch.equals(getVarEndDelimiter()))
    {
      return evaluatePlaceholder(placeholderName);
    }

    if (ch.equals(getFunctionParameterStart()))
    {
      return evaluateFunction(placeholderName);
    }

    throw (new TextReplacementException("Invalid termination of placeholder: " + placeholderName));
  }

  protected void checkEndAfterFunction(String funcName) throws TextReplacementException
  {
    Character ch = null;
    String msg = null;

    ch = skipSpaces();
    if (!ch.equals(getVarEndDelimiter()))
    {
      msg = "Invalid character '" + ch + "' after '" + funcName + "' !";
      throw new TextReplacementException(msg);
    }
  }

  protected void checkFunctionStart(String funcName) throws TextReplacementException
  {
    Character ch = null;
    String msg = null;

    ch = skipSpaces();
    if (!ch.equals(getFunctionParameterStart()))
    {
      msg = "Missing opening bracket in '" + funcName + "' ! ";
      throw new TextReplacementException(msg);
    }
  }

  protected void checkFunctionEnd(String funcName) throws TextReplacementException
  {
    Character ch = null;
    String msg = null;

    ch = skipSpaces();
    if (!ch.equals(getFunctionParameterEnd()))
    {
      msg = "Invalid end '" + ch + "' of " + funcName;
      throw new TextReplacementException(msg);
    }
  }

  protected void checkParameterCount(String funcName, int expected, int actual) throws TextReplacementException
  {
    String msg = null;

    if (actual != expected)
    {
      msg = "The number of parameters in " + funcName + " must be ";
      msg = msg + Integer.toString(expected) + " !";
      throw new TextReplacementException(msg);
    }
  }

  protected String evaluateControlIF() throws TextReplacementException
  {
    final String IF_CTRL = CONTROL_IF + getFunctionParameterStart() + getFunctionParameterEnd();
    List<String> parameters;

    checkFunctionStart(IF_CTRL);
    parameters = getFunctionParameters();
    checkFunctionEnd(IF_CTRL);
    checkEndAfterFunction(IF_CTRL);
    checkParameterCount(IF_CTRL, 1, parameters.size());

    if (Bool.get(parameters.get(0)).isTrue())
    {
      switchWritingOn();
    }
    else
    {
      switchWritingOff();
    }

    return "";
  }

  protected String evaluateControlIFDEF() throws TextReplacementException
  {
    final String IFDEF_CTRL = CONTROL_IFDEF + getFunctionParameterStart() + getFunctionParameterEnd();
    String varName = null;

    checkFunctionStart(IFDEF_CTRL);
    varName = getFunctionParameterName();
    checkFunctionEnd(IFDEF_CTRL);
    checkEndAfterFunction(IFDEF_CTRL);

    if ((isDefinedPlaceholder(varName)) && (evaluatePlaceholder(varName).length() > 0))
    {
      switchWritingOn();
    }
    else
    {
      switchWritingOff();
    }
    return "";
  }

  protected String evaluateControlELSE() throws TextReplacementException
  {
    checkEndAfterFunction(CONTROL_ELSE);

    try
    {
      toggleWritingSwitch();
    }
    catch (@SuppressWarnings("unused") EmptyStackException ex)
    {
      throw new TextReplacementException("Unmatched " + CONTROL_ELSE + " reached !");
    }

    return "";
  }

  protected String evaluateControlENDIF() throws TextReplacementException
  {
    checkEndAfterFunction(CONTROL_ENDIF);

    try
    {
      removeLastWritingSwitch();
    }
    catch (@SuppressWarnings("unused") EmptyStackException ex)
    {
      throw new TextReplacementException("Unmatched " + CONTROL_ENDIF + " reached !");
    }
    return "";
  }

  protected String evaluateControl() throws TextReplacementException
  {
    String controlName = null;
    String msg = null;

    controlName = getPlaceholderName();

    if (controlName.equals(CONTROL_IF))
    {
      return evaluateControlIF();
    }

    if (controlName.equals(CONTROL_ENDIF))
    {
      return evaluateControlENDIF();
    }

    if (controlName.equals(CONTROL_ELSE))
    {
      return evaluateControlELSE();
    }

    if (controlName.equals(CONTROL_IFDEF))
    {
      return evaluateControlIFDEF();
    }

    msg = "Unknown control element: " + CONTROL_INDICATOR + controlName;
    throw new TextReplacementException(msg);
  }

  protected String getTextConstant()
  {
    Character ch = null;
    boolean ready = false;
    StringBuffer buffer;

    buffer = new StringBuffer(100);
    while (hasMoreChars() && !ready)
    {
      ch = nextCharacter();
      if (ch.equals(getTextDelimiter()))
      {
        if (hasMoreChars())
        {
          ch = nextCharacter();
          if (ch.equals(getTextDelimiter()))
          {
            buffer.append(ch.toString());
          }
          else
          {
            ready = true;
            putBackNextChar();
          }
        }
        else
        {
          ready = true;
        }
      }
      else
      {
        buffer.append(ch.toString());
      }
    }
    return buffer.toString();
  }

  protected String asString(Object anObject)
  {
    String resultString = null;

    if (anObject instanceof TextRepresentation)
    {
      resultString = ((TextRepresentation)anObject).asText();
    }
    else
    {
      resultString = anObject.toString();
    }

    return resultString;
  }

  protected String evaluatePlaceholder(String varName) throws UnknownVariableException
  {
    Object value;

    // Writing could be switched off by IFDEF and therefore there must not
    // be an exception, if a varName is not known.
    // Actually the evaluation can be skipped, because writing switched off !
    if (isWritingOn())
    {
      if (supressMissingVariableException())
      {
        try
        {
          value = getVariableResolver().getValue(varName);
        }
        catch (@SuppressWarnings("unused") UnknownVariableException e)
        {
          // Ignore the exception and use the variable name instead
          value = varName;
        }
      }
      else
      {
        value = getVariableResolver().getValue(varName);
      }
      return asString(value);
    }

    return "";
  }

  protected boolean isDefinedPlaceholder(String varName)
  {
    return getVariableResolver().isKnownVariable(varName);
  }

  protected List<String> getFunctionParameters() throws TextReplacementException
  {
    Character ch = null;
    List<String> parameters;
    String arg = null;
    String varName = null;

    parameters = new ArrayList<String>();
    ch = skipSpaces();

    do
    {
      if (ch.equals(getTextDelimiter()))
      {
        arg = getTextConstant();
      }
      else
      {
        putBackNextChar();
        varName = getPlaceholderName();
        arg = evaluatePlaceholder(varName);
      }

      parameters.add(arg);

      ch = skipSpaces();
    }
    while (ch.equals(getFunctionParameterSeparator()));

    putBackNextChar();

    return parameters;
  }

  /**
   * Returns the next name for a function parameter.
   */
  protected String getFunctionParameterName() throws TextReplacementException
  {
    Character ch = null;
    String varName = null;

    ch = skipSpaces();

    if (ch.equals(getTextDelimiter()))
    {
      throw (new TextReplacementException("No text constant allowed here !"));
    }
    putBackNextChar();
    varName = getPlaceholderName();

    return varName;
  }

  protected String evaluateFunction(String funcName, List<String> parameter) throws UnknownFunctionException, InvalidParameterException
  {
    if (getFunctionResolver() == null)
    {
      throw new UnknownFunctionException(funcName);
    }
    return asString(getFunctionResolver().executeFunction(funcName, parameter));
  }

  protected String evaluateFunction(String funcName) throws TextReplacementException
  {
    List<String> parameters;

    parameters = getFunctionParameters();

    checkFunctionEnd(funcName);
    checkEndAfterFunction(funcName);

    return evaluateFunction(funcName, parameters);
  }

  protected void readAhead() throws IOException
  {
    int readResult = -1;

    if (getCharCache() == EMPTY_CACHE) // Reads only, if cache is empty !
    {
      readResult = getSourceBuffer().read();

      if (readResult < 0)
      {
        throw (new IOException("End of String"));
      }
      setCharCache(readResult);
    }
  }

  protected void readNext() throws IOException
  {
    char ch = '@';

    readAhead(); // Reads only, if cache is empty !

    ch = (char)getCharCache();
    setCharCache(EMPTY_CACHE);
    setNextChar(new Character(ch));
    setNextCharFilled(true);
  }

  protected boolean hasMoreChars()
  {
    boolean hasChar = true;

    if (isNextCharBufferEmpty())
    {
      try
      {
        readAhead();
      }
      catch (@SuppressWarnings("unused") IOException ex)
      {
        hasChar = false;
      }
    }
    return hasChar;
  }

  protected Character nextChar() throws IOException
  {
    if (isNextCharBufferEmpty())
    {
      readNext();
    }
    setNextCharFilled(false);
    return getNextChar();
  }

  /**
   * This is a convinient method for retreiving the
   * next character, without always catching IOException.
   * It should only be used in combination with hasMoreChars().
   */
  protected Character nextCharacter()
  {
    Character ch = null;
    try
    {
      ch = nextChar();
    }
    catch (@SuppressWarnings("unused") IOException ex)
    { /* deliberately ignored */
    }
    return ch;
  }

  protected void putBackNextChar()
  {
    if (isNextCharBufferEmpty())
    {
      setNextCharFilled(true);
    }
  }

  protected boolean isNextCharBufferEmpty()
  {
    return (!getNextCharFilled());
  }

  protected Character skipSpaces()
  {
    Character ch = new Character(StringUtil.CH_SPACE);

    while (hasMoreChars() && Character.isWhitespace(ch.charValue()))
      ch = nextCharacter();

    return ch;
  }

  protected String resultString()
  {
    return getResultBuffer().toString();
  }

  protected Character getFunctionParameterStart()
  {
    return DEFAULT_FUNCPARAM_START;
  }

  protected Character getFunctionParameterEnd()
  {
    return DEFAULT_FUNCPARAM_END;
  }

  protected Character getFunctionParameterSeparator()
  {
    return DEFAULT_PARAM_SEPARATOR;
  }

  protected Character getTextDelimiter()
  {
    return DEFAULT_TEXT_DELIMITER;
  }

  protected boolean isValidPlaceholderNameCharacter(char ch)
  {
    if ((Character.isJavaIdentifierStart(ch)) || (Character.isDigit(ch)) || (getSpecialNameCharacters().indexOf(ch) >= 0))
    {
      return true;
    }
    return false;
  }

  protected Character getControlIndicator()
  {
    return CONTROL_INDICATOR;
  }

  protected boolean isIndentationOn()
  {
    return (getIndentationFiller() != null);
  }

  protected boolean isWritingOn()
  {
    boolean bool = true;

    if (getExecutionControl().empty())
    {
      return true;
    }

    try
    {
      bool = getExecutionControl().peek().isTrue();
    }
    catch (@SuppressWarnings("unused") EmptyStackException ex)
    {
      return true;
    }
    return bool;
  }

  protected boolean isConditionalControlOn()
  {
    return (!getExecutionControl().empty());
  }

  protected void removeLastWritingSwitch() throws EmptyStackException
  {
    getExecutionControl().pop();
  }

  protected void switchWritingTo(Bool bool)
  {
    getExecutionControl().push(bool);
  }

  protected void toggleWritingSwitch() throws EmptyStackException
  {
    Bool old = getExecutionControl().pop();
    getExecutionControl().push(old.not());
  }

  protected void switchWritingOn()
  {
    switchWritingTo(Bool.getTrue());
  }

  protected void switchWritingOff()
  {
    switchWritingTo(Bool.getFalse());
  }

  protected StringUtil strUtil()
  {
    return StringUtil.current();
  }

  protected StringReader getSourceBuffer()
  {
    return sourceBuffer;
  }

  protected void setSourceBuffer(StringReader sb)
  {
    sourceBuffer = sb;
  }

  protected StringWriter getResultBuffer()
  {
    return resultBuffer;
  }

  protected void setResultBuffer(StringWriter rb)
  {
    resultBuffer = rb;
  }

  protected Character getNextChar()
  {
    return nextChar;
  }

  protected void setNextChar(Character ch)
  {
    nextChar = ch;
  }

  protected boolean getNextCharFilled()
  {
    return nextCharFilled;
  }

  protected void setNextCharFilled(boolean flag)
  {
    nextCharFilled = flag;
  }

  protected int getCharCache()
  {
    return charCache;
  }

  protected void setCharCache(int newValue)
  {
    charCache = newValue;
  }

  protected Stack<Bool> getExecutionControl()
  {
    return executionControl;
  }

  protected void setExecutionControl(Stack<Bool> newValue)
  {
    executionControl = newValue;
  }

  protected boolean supressMissingVariableException()
  {
    return supressMissingVariableException;
  }

  protected void supressMissingVariableException(boolean newValue)
  {
    supressMissingVariableException = newValue;
  }

  protected String getIndentationFiller()
  {
    return indentationFiller;
  }

  protected void setIndentationFiller(String newValue)
  {
    indentationFiller = newValue;
  }
}