// ===========================================================================
// CONTENT  : CLASS CollectorDictionary
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.2 - 17/02/2006
// HISTORY  :
//  27/01/2001  duma  CREATED
//	20/12/2003	duma	implemented	->	containsValue(), putAll()
//	17/02/2006	mdu		bugfix			->	putAll()
//
// Copyright (c) 2001-2006, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.util;

// ===========================================================================
// IMPORTS
// ===========================================================================
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * This class implements the java.util.Map interface.
 * Its main difference to a java.util.Hashtable is that for duplicate
 * key additions the corresponding values are kept in a list.
 * All values of this class are lists (java.util.List)
 *
 * @author Manfred Duchrow
 * @version 1.2
 */
public class CollectorDictionary<K, V> implements Map<K, List<V>>
{
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private Map<K, List<V>> objectDictionary = null;
  private boolean acceptDuplicates = false;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  public CollectorDictionary()
  {
    this.setObjectDictionary(new HashMap<K, List<V>>());
  }

  /**
   * Initialize the new instance with default values.
   */
  public CollectorDictionary(int initialCapacity)
  {
    this.setObjectDictionary(new HashMap<K, List<V>>(initialCapacity));
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  public boolean getAcceptDuplicates()
  {
    return acceptDuplicates;
  }

  public void setAcceptDuplicates(boolean newValue)
  {
    acceptDuplicates = newValue;
  }

  /**
   * Removes all mappings from this map (optional operation).
   */
  @Override
  public void clear()
  {
    this.getObjectDictionary().clear();
  }

  /**
   * Returns true if this map contains a mapping for the specified key.
   */
  @Override
  public boolean containsKey(Object key)
  {
    return this.getObjectDictionary().containsKey(key);
  }

  /**
   * Returns true if this map maps one or more keys to the specified value.
   * More formally, returns true if and only if this map contains at least
   * one mapping to a value v such that (value==null ? v==null : value.equals(v)).
   *
   * @param value - value whose presence in this map is to be tested.
   * @return true if this map maps one or more keys to the specified value.
   */
  @Override
  public boolean containsValue(Object value)
  {
    List<V> list;

    for (K key : this.keySet())
    {
      list = this.getValues(key);
      if (list.contains(value))
      {
        return true;
      }
    }
    return false;
  }

  /**
   * Returns a set view of the mappings contained in this map.
   */
  @Override
  public Set<Map.Entry<K, List<V>>> entrySet()
  {
    return this.getObjectDictionary().entrySet();
  }

  /**
   * Compares the specified object with this map for equality.
   * Returns true if the given object is also a map and the two Maps
   * represent the same mappings.
   */
  @Override
  public boolean equals(Object obj)
  {
    return this.getObjectDictionary().equals(obj);
  }

  /**
   * Returns the hash code value for this map.
   */
  @Override
  public int hashCode()
  {
    return this.getObjectDictionary().hashCode();
  }

  /**
   * Returns true if this map contains no key-value mappings.
   */
  @Override
  public boolean isEmpty()
  {
    return this.getObjectDictionary().isEmpty();
  }

  /**
   * Returns a set view of the keys contained in this map.
   */
  @Override
  public Set<K> keySet()
  {
    return this.getObjectDictionary().keySet();
  }

  /**
   * Adds a single value to the key.
   */
  public void add(K key, V value)
  {
    List<V> list;

    list = this.getValues(key);
    if (list == null)
    {
      list = this.newValue();
      this.getObjectDictionary().put(key, list);
    }
    if ((this.getAcceptDuplicates()) || (!list.contains(value)))
    {
      list.add(value);
    }
  }

  /**
   * Stores the given values under the specified key object.
   */
  public void addAll(K key, V... values)
  {
    for (V value : values)
    {
      this.add(key, value);
    }
  }

  /**
   * Stores the given value list under the specified key object.
   */
  @Override
  public List<V> put(K key, List<V> list)
  {
    return this.getObjectDictionary().put(key, list);
  }

  /**
   * Adds all values in the given collection to the key.
   */
  public void addAll(K key, Collection<V> values)
  {
    if (values != null)
    {
      for (V value : values)
      {
        this.add(key, value);
      }
    }
  }

  /**
   * Adds all entries of the given map.
   */
  public void addAll(Map<K, V> map)
  {
    if (map != null)
    {
      for (Map.Entry<K, V> entry : map.entrySet())
      {
        this.add(entry.getKey(), entry.getValue());
      }
    }
  }

  /**
   * This method is actually not supported. Use {@link #addAll(Map)} instead.
   * @throws UnsupportedOperationException
   */
  @Override
  public void putAll(Map<? extends K, ? extends List<V>> m)
  {
    throw new UnsupportedOperationException("org.pf.util.CollectorDictionary.putAll(Map<? extends K, ? extends V>)");
  }

  /**
   * Removes the mapping for this key from this map if present.
   */
  @Override
  public List<V> remove(Object key)
  {
    return this.getObjectDictionary().remove(key);
  }

  /**
   * Returns the number of key-value mappings in this map.
   */
  @Override
  public int size()
  {
    return this.getObjectDictionary().size();
  }

  /**
   * Returns a collection view of the values contained in this map.
   * In this case all elements of the returned collection are java.util.List
   * objects which contain the real values !
   */
  @Override
  public Collection<List<V>> values()
  {
    return this.getObjectDictionary().values();
  }

  /**
   * Initialize the new instance with default values.
   */
  public CollectorDictionary(boolean allowDuplicates)
  {
    this();
    this.setAcceptDuplicates(allowDuplicates);
  }

  @SuppressWarnings("unchecked")
  @Override
  public List<V> get(Object key)
  {
    return this.getValues((K)key);
  }

  /**
   * Returns an iterator over all key objects in this map.
   */
  public Iterator<K> keys()
  {
    return this.keySet().iterator();
  }

  /**
   * Returns the number of all values in all entries.
   */
  public int valuesSize()
  {
    int sum = 0;

    for (List<V> list : this.getObjectDictionary().values())
    {
      sum = sum + list.size();
    }
    return sum;
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected List<V> getValues(K key)
  {
    return this.getObjectDictionary().get(key);
  }

  protected List<V> newValue()
  {
    return new ArrayList<V>();
  }

  protected List<V> newValueWith(V obj)
  {
    List<V> value;

    value = this.newValue();
    value.add(obj);
    return value;
  }

  /**
   * Not supported - use 
   * @throws UnsupportedOperationException
   */
  @SuppressWarnings("unused") 
  protected void putAll(Object key, Collection<List<V>> values)
  {
    throw new UnsupportedOperationException("org.pf.util.CollectorDictionary.putAll(Object, Collection<List<V>>)");
  }

  protected Map<K, List<V>> getObjectDictionary()
  {
    return objectDictionary;
  }

  protected void setObjectDictionary(Map<K, List<V>> newValue)
  {
    objectDictionary = newValue;
  }
}