// ===========================================================================
// CONTENT  : CLASS Countdown
// AUTHOR   : M.Duchrow
// VERSION  : 3.0 - 26/04/2021
// HISTORY  :
//  08/08/2006  mdu  CREATED
//  03/05/2020  mdu   added   -> methods with Duration
//  26/04/2021  mdu   changed -> moved implementation to instance level
//
// Copyright (c) 2006-2021, by M.Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.util;

import org.pfsw.bif.logic.ICondition;

/**
 * Supports loops with a countdown of a defined time.
 *
 * @author M.Duchrow
 * @version 3.0
 */
public class Countdown
{
  private static final long DEFAULT_SLEEP_MS = 1;

  private long sleepTime = DEFAULT_SLEEP_MS;

  // =========================================================================
  // PUBLIC CLASS METHODS
  // =========================================================================
  /**
   * Creates a new instance.
   */
  public static Countdown create()
  {
    return new Countdown();
  }

  /**
   * Runs an empty loop until either the condition is false or the specified
   * time in milliseconds is over (more or less).
   * That particularly makes sense if the condition gets influenced by a
   * different thread. Therefore this thread sleeps one millisecond per 
   * loop iteration to allow other threads to do something.
   * 
   * @param condition A condition that is used to continue (true) or finish (false) the loop.
   * @param millis The time in milliseconds after which the loop gets terminated independently of the condition.
   */
  public static void whileTrue(ICondition condition, long millis)
  {
    create().waitWhile(condition, true, millis);
  }

  /**
   * Runs an empty loop until either the condition is false or the specified
   * time in milliseconds is over (more or less).
   * That particularly makes sense if the condition gets influenced by a
   * different thread. Therefore this thread sleeps one millisecond per 
   * loop iteration to allow other threads to do something.
   * 
   * @param condition A condition that is used to continue (true) or finish (false) the loop
   * @param duration The duration after which the loop gets terminated independently of the condition.
   */
  public static void whileTrue(ICondition condition, Duration duration)
  {
    whileTrue(condition, duration.asMilliseconds());
  }

  /**
   * Runs an empty loop until either the condition is true or the specified
   * time in milliseconds is over.
   * That particularly makes sense if the condition gets influenced by a
   * different thread. Therefore this thread sleeps one millisecond per 
   * loop iteration to allow other threads to do something.
   * 
   * @param condition A condition that is used to continue (false) or finish (true) the loop
   * @param millis The time in milliseconds after which the loop gets terminated independently of the condition.
   */
  public static void whileFalse(ICondition condition, long millis)
  {
    create().waitUntilTrue(condition, millis);
  }

  /**
   * Runs an empty loop until either the condition is true or the specified
   * time in milliseconds is over.
   * That particularly makes sense if the condition gets influenced by a
   * different thread. Therefore this thread sleeps one millisecond per 
   * loop iteration to allow other threads to do something.
   * 
   * @param condition A condition that is used to continue (false) or finish (true) the loop
   * @param duration The duration after which the loop gets terminated independently of the condition.
   */
  public static void whileFalse(ICondition condition, Duration duration)
  {
    whileFalse(condition, duration.asMilliseconds());
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  protected Countdown()
  {
    super();
  }

  // =========================================================================
  // PUBLIC METHODS
  // =========================================================================
  /**
   * Waits in a loop until either the value of the condition becomes true 
   * or the maxTimeToWait is exceeded.
   * 
   * @param condition The condition to check (might be null).
   * @param maxTimeToWait Maximum time in milliseconds to wait.
   * @return The value of the condition after waiting ended.
   */
  public boolean waitUntilTrue(ICondition condition, long maxTimeToWait)
  {
    return waitWhile(condition, false, maxTimeToWait);
  }

  /**
   * Waits in a loop until either the value of the condition becomes true 
   * or the maxTimeToWait is exceeded.
   * 
   * @param condition The condition to check (might be null).
   * @param maxTimeToWait Maximum time in milliseconds to wait.
   * @return The value of the condition after waiting ended.
   */
  public boolean waitUntilTrue(ICondition condition, Duration maxTimeToWait)
  {
    return waitUntilTrue(condition, maxTimeToWait.asMilliseconds());
  }

  /**
   * Waits in a loop until either the value of the condition becomes false 
   * or the maxTimeToWait is exceeded.
   * 
   * @param condition The condition to check (might be null).
   * @param maxTimeToWait Maximum time in milliseconds to wait.
   * @return The value of the condition after waiting ended.
   */
  public boolean waitUntilFalse(ICondition condition, long maxTimeToWait)
  {
    return waitWhile(condition, true, maxTimeToWait);
  }

  /**
   * Waits in a loop until either the value of the condition becomes false 
   * or the maxTimeToWait is exceeded.
   * 
   * @param condition The condition to check (might be null).
   * @param maxTimeToWait Maximum time in milliseconds to wait.
   * @return The value of the condition after waiting ended.
   */
  public boolean waitUntilFalse(ICondition condition, Duration maxTimeToWait)
  {
    return waitUntilFalse(condition, maxTimeToWait.asMilliseconds());
  }

  public long getSleepTime()
  {
    return this.sleepTime;
  }

  public Countdown setSleepTime(long sleepTime)
  {
    this.sleepTime = sleepTime;
    return this;
  }

  @Override
  public String toString()
  {
    return String.format("%s(sleepTime=%dms)", getClass().getSimpleName(), getSleepTime());
  }

  protected boolean waitWhile(ICondition condition, boolean continueLoopValue, long maxTimeToWait)
  {
    long done;
    boolean conditionValue = false;

    if (condition != null)
    {
      conditionValue = condition.isTrue();
      done = System.currentTimeMillis() + maxTimeToWait;
      while ((conditionValue == continueLoopValue) && (System.currentTimeMillis() <= done))
      {
        try
        {
          Thread.currentThread().sleep(getSleepTime());
          conditionValue = condition.isTrue();
        }
        catch (@SuppressWarnings("unused") InterruptedException e)
        {
          break;
        }
      }
    }
    return conditionValue;
  }
}
