// ===========================================================================
// CONTENT  : CLASS Duration
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 03/05/2020
// HISTORY  :
//  03/05/2020  mdu  CREATED
//
// Copyright (c) 2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.util;

import java.util.concurrent.atomic.AtomicLong;

import org.pfsw.bif.text.IStringRepresentation;

/**
 * Time duration split-up in days, hours, minutes, seconds and milliseconds.
 * Instances of this class are immutable.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class Duration implements Comparable<Duration>, IStringRepresentation
{
  public static final long MILLIS_PER_SECOND = 1000L;
  public static final long MILLIS_PER_MINUTE = 60L * MILLIS_PER_SECOND;
  public static final long MILLIS_PER_HOUR = 60L * MILLIS_PER_MINUTE;
  public static final long MILLIS_PER_DAY = 24L * MILLIS_PER_HOUR;
  
  public static final String DEFAULT_FORMAT = "%d[ days, ]%h[ hours, ]%m[ minutes, ]%s[ seconds, ]%t milliseconds";

  private static final long MAX_DAYS = 106751991166L;
  private static final int MAX_HOURS = 24;
  private static final int MAX_MINUTES = 59;
  private static final int MAX_SECONDS = 59;
  private static final int MAX_MILLISECONDS = 999;

  private final long fullDurationInMillis;
  private final long days;
  private final int hours;
  private final int minutes;
  private final int seconds;
  private final int milliseconds;

  /**
   * Creates a new instance based on the given milliseconds.
   * 
   * @param fullDurationInMillis The duration in milliseconds.
   * @return A new Duration instance that is equivalent to the given duration in milliseconds. 
   */
  public static Duration valueOf(long fullDurationInMillis)
  {
    return new Duration(fullDurationInMillis);
  }

  /**
   * Creates a new instance based on the given milliseconds.
   * If the duration exceeds 999 milliseconds use {@link #valueOf(long)} instead!
   * 
   * @param milliseconds The number of milliseconds - must be a value in the range of 0..999.
   * @return A new Duration instance initialized with the given values.
   * @throws IllegalArgumentException if any of the given values is outside the allowed range.
   */
  public static Duration create(int milliseconds)
  {
    return create(0, 0, 0, 0, milliseconds);
  }

  /**
   * Creates a new instance based on the given values.
   * 
   * @param seconds The number of seconds - must be a value in the range of 0..59.
   * @param milliseconds The number of milliseconds - must be a value in the range of 0..999.
   * @return A new Duration instance initialized with the given values.
   * @throws IllegalArgumentException if any of the given values is outside the allowed range.
   */
  public static Duration create(int seconds, int milliseconds)
  {
    return create(0, 0, 0, seconds, milliseconds);
  }

  /**
   * Creates a new instance based on the given values.
   * 
   * @param minutes The number of minutes - must be a value in the range of 0..59.
   * @param seconds The number of seconds - must be a value in the range of 0..59.
   * @param milliseconds The number of milliseconds - must be a value in the range of 0..999.
   * @return A new Duration instance initialized with the given values.
   * @throws IllegalArgumentException if any of the given values is outside the allowed range.
   */
  public static Duration create(int minutes, int seconds, int milliseconds)
  {
    return create(0, 0, minutes, seconds, milliseconds);
  }

  /**
   * Creates a new instance based on the given values.
   * 
   * @param hours The number of hours - must be a value in the range of 0..24.
   * @param minutes The number of minutes - must be a value in the range of 0..59.
   * @param seconds The number of seconds - must be a value in the range of 0..59.
   * @param milliseconds The number of milliseconds - must be a value in the range of 0..999.
   * @return A new Duration instance initialized with the given values.
   * @throws IllegalArgumentException if any of the given values is outside the allowed range.
   */
  public static Duration create(int hours, int minutes, int seconds, int milliseconds)
  {
    return create(0, hours, minutes, seconds, milliseconds);
  }

  /**
   * Creates a new instance based on the given values.
   * 
   * @param days The number of days - must be a value in the range of 0..106751991166.
   * @param hours The number of hours - must be a value in the range of 0..24.
   * @param minutes The number of minutes - must be a value in the range of 0..59.
   * @param seconds The number of seconds - must be a value in the range of 0..59.
   * @param milliseconds The number of milliseconds - must be a value in the range of 0..999.
   * @return A new Duration instance initialized with the given values.
   * @throws IllegalArgumentException if any of the given values is outside the allowed range.
   */
  public static Duration create(long days, int hours, int minutes, int seconds, int milliseconds)
  {
    long millis;

    assertValueRange("days", days, MAX_DAYS);
    assertValueRange("hours", hours, MAX_HOURS);
    assertValueRange("minutes", minutes, MAX_MINUTES);
    assertValueRange("seconds", seconds, MAX_SECONDS);
    assertValueRange("milliseconds", milliseconds, MAX_MILLISECONDS);
    millis = (MILLIS_PER_DAY * days) + (MILLIS_PER_HOUR * hours) + (MILLIS_PER_MINUTE * minutes) + (MILLIS_PER_SECOND * seconds) + milliseconds;
    return valueOf(millis);
  }

  private static void assertValueRange(String valueName, long value, long maxValue)
  {
    if (value < 0)
    {
      throw new IllegalArgumentException(String.format("Parameter '%s' must not be less than 0, but actually is %d", valueName, value));
    }
    if (value > maxValue)
    {
      throw new IllegalArgumentException(String.format("Parameter '%s' must not be greater than %d, but actually is %d", valueName, maxValue, value));
    }
  }

  private static int calculateValue(AtomicLong millis, long millisPerUnit)
  {
    int returnValue = 0;

    if (millis.get() >= millisPerUnit)
    {
      returnValue = (int)(millis.get() / millisPerUnit);
      millis.set(millis.get() % millisPerUnit);
    }
    return returnValue;
  }

  protected Duration(long fullDurationInMillis)
  {
    super();
    this.fullDurationInMillis = fullDurationInMillis;

    AtomicLong millis = new AtomicLong(fullDurationInMillis);
    this.days = calculateValue(millis, MILLIS_PER_DAY);
    this.hours = calculateValue(millis, MILLIS_PER_HOUR);
    this.minutes = calculateValue(millis, MILLIS_PER_MINUTE);
    this.seconds = calculateValue(millis, MILLIS_PER_SECOND);
    this.milliseconds = (int)millis.get();
  }

  /**
   * Returns the number of days.
   */
  public long getDays()
  {
    return this.days;
  }

  /**
   * Returns the number of hours.
   */
  public int getHours()
  {
    return this.hours;
  }

  /**
   * Returns the number of minutes.
   */
  public int getMinutes()
  {
    return this.minutes;
  }

  /**
   * Returns the number of seconds.
   */
  public int getSeconds()
  {
    return this.seconds;
  }

  /**
   * Returns the number of milliseconds.
   */
  public int getMilliseconds()
  {
    return this.milliseconds;
  }

  /**
   * Returns the full duration as milliseconds.
   */
  public long asMilliseconds()
  {
    return getFullDurationInMillis();
  }

  /**
   * Returns true if the duration is not under a day.
   */
  public boolean isInDaysRange()
  {
    return getDays() > 0;
  }

  /**
   * Returns true if the duration is not under an hour.
   */
  public boolean isInHoursRange()
  {
    return isInDaysRange() || (getHours() > 0);
  }

  /**
   * Returns true if the duration is not under a minute.
   */
  public boolean isInMinutesRange()
  {
    return isInHoursRange() || (getMinutes() > 0);
  }

  /**
   * Returns true if the duration is not under a second.
   */
  public boolean isInSecondsRange()
  {
    return isInMinutesRange() || (getSeconds() > 0);
  }

  /**
   * Returns the duration as a string using english text labels as suffix after each value.
   * The string will start at the largest set unit.  
   */
  @Override
  public String asString()
  {
    return asString("days", "hours", "minutes", "seconds", "milliseconds");
  }

  /**
   * Returns the duration as a string using the given text labels as suffix after each value.
   * The string will start at the largest set unit.  
   */
  public String asString(String daysText, String hoursText, String minutesText, String secondsText, String millisText)
  {
    StringBuffer buffer;

    buffer = new StringBuffer();
    if (isInDaysRange())
    {
      buffer.append(getDays()).append(" ").append(daysText).append(" ");
    }
    if (isInHoursRange())
    {
      buffer.append(getHours()).append(" ").append(hoursText).append(" ");
    }
    if (isInMinutesRange())
    {
      buffer.append(getMinutes()).append(" ").append(minutesText).append(" ");
    }
    if (isInSecondsRange())
    {
      buffer.append(getSeconds()).append(" ").append(secondsText).append(" ");
    }
    buffer.append(getMilliseconds()).append(" ").append(millisText);
    return buffer.toString();
  }
  
  /**
   * Formats the duration into a string utilizing the given template.
   * The template can have the following placeholders that will be replaced accordingly
   * by the corresponding values.
   * <ul>
   * <li>%d - days</li>
   * <li>%h - hours</li>
   * <li>%m - minutes</li>
   * <li>%s - seconds</li>
   * <li>%t - milliseconds</li>
   * </ul>
   *  
   * @param template The template with placeholders to be used for formatting (must not be null).
   * @return The formatted string.
   */
  public String format(String template)
  {
    String buffer;

    buffer = template;
    buffer = replacePlaceholder(buffer, "%d", isInDaysRange(), getDays());
    buffer = replacePlaceholder(buffer, "%h", isInHoursRange(), getHours());
    buffer = replacePlaceholder(buffer, "%m", isInMinutesRange(), getMinutes());
    buffer = replacePlaceholder(buffer, "%s", isInSecondsRange(), getSeconds());
    buffer = replacePlaceholder(buffer, "%t", true, getMilliseconds());
    return buffer.toString();    
  }

  @Override
  public int hashCode()
  {
    return Long.valueOf(getFullDurationInMillis()).hashCode();
  }

  @Override
  public boolean equals(Object obj)
  {
    if (obj instanceof Duration)
    {
      Duration other = (Duration)obj;
      return getFullDurationInMillis() == other.getFullDurationInMillis();
    }
    return false;
  }

  @Override
  public int compareTo(Duration other)
  {
    return Long.valueOf(getFullDurationInMillis()).compareTo(Long.valueOf(other.getFullDurationInMillis()));
  }

  @Override
  public String toString()
  {
    return String.format("%s(%s)", getClass().getSimpleName(), asString());
  }

  protected String replacePlaceholder(String string, String placeholder, boolean isInRange, long value) 
  {
    String buffer;
    String textToReplace;
    String optionalText;
    String replacement;
    int startIndex;
    int endIndex;
    int pos;
    
    buffer = string;
    textToReplace = placeholder;
    optionalText = "";
    
    startIndex = buffer.indexOf(placeholder);
    if (startIndex < 0)
    {
      return buffer;
    }
    pos = startIndex + placeholder.length();
    if ((pos < buffer.length()) && (buffer.charAt(pos) == '['))
    {
      endIndex = buffer.indexOf(']', pos);
      if (endIndex > pos)
      {
        textToReplace = buffer.substring(startIndex, endIndex + 1);        
        optionalText = buffer.substring(startIndex + placeholder.length() + 1, endIndex);        
      }
    }
    replacement = isInRange ? Long.toString(value) + optionalText : "";
    buffer = buffer.replace(textToReplace, replacement);
    return replacePlaceholder(buffer, placeholder, isInRange, value);
  }
  
  protected long getFullDurationInMillis()
  {
    return this.fullDurationInMillis;
  }
}
