// ===========================================================================
// CONTENT  : CLASS IOUtil
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 20/03/2016
// HISTORY  :
//  20/03/2016  mdu  CREATED
//
// Copyright (c) 2016, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.util;

//===========================================================================
//IMPORTS
//===========================================================================
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.Closeable;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.Reader;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicReference;

import org.pfsw.bif.callback.IProgressListener;
import org.pfsw.bif.text.ICharsets;
import org.pfsw.bif.text.ICheckedCharset;
import org.pfsw.bif.text.ILineProcessor;
import org.pfsw.bif.text.IStringFilter;

/**
 * Various well-tested and reusable utility methods that help to reduce large
 * amount of boilerplate code in the area of I/O (stream and reader/writer processing).
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class IOUtil
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  public static final String LINE_SEPARATOR = System.getProperty("line.separator");
  public static final int DEFAULT_BUFFER_SIZE = 16000;

  public static final IProgressListener IGNORING_PROGRESS_LISTENER = new IProgressListener()
  {
    @Override
    public boolean processed(long totalSize, long processedSize)
    {
      return IProgressListener.CONTINUE;
    }
  };

  protected static final Long NO_LIMIT = Long.valueOf(-1L);

  // =========================================================================
  // CLASS VARIABLES
  // =========================================================================
  private static IOUtil current = new IOUtil();

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Returns the only instance this class supports (design pattern "Singleton")
   */
  public static IOUtil current()
  {
    return current;
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Protected constructor allows subclassing!
   */
  protected IOUtil()
  {
    super();
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Close the given closeable ignoring any exception and checking for null before closing.
   * Returns true, if the object was closed successfully, false otherwise.
   */
  public boolean close(Closeable closeable)
  {
    if (closeable == null)
    {
      return false;
    }
    try
    {
      closeable.close();
      return true;
    }
    catch (@SuppressWarnings("unused") IOException e)
    {
      return false;
    }
  }

  /**
   * Reads all text lines from the specified input stream and passes them
   * one by one to the given line processor.
   * The processing will be terminated, if the end of the text is reached or
   * if the processor returns <b>false</b>.<br>
   * The given input stream will be closed after the execution of this method.
   * Even if an exception occurred.
   * The character encoding is the JVM default encoding.
   *
   * @param inStream The input stream that contains the text
   * @param processor The processor that receives the lines from the text
   */
  public void processTextLines(InputStream inStream, ILineProcessor processor) throws IOException
  {
    processTextLines(inStream, Charset.defaultCharset(), processor);
  }

  /**
   * Reads all text lines from the specified input stream and passes them
   * one by one to the given line processor.
   * The processing will be terminated, if the end of the text is reached or
   * if the processor returns <b>false</b>.<br>
   * The given input stream will be closed after the execution of this method.
   * Even if an exception occurred.
   *
   * @param inStream The input stream that contains the text (must not be null).
   * @param charsetName A charset name already checked to be supported (must not be null).
   * @param processor The processor that receives the lines from the text (must not be null).
   */
  public void processTextLines(InputStream inStream, ICheckedCharset charsetName, ILineProcessor processor) throws IOException
  {
    processTextLines(inStream, charsetName.getCharset(), processor);
  }

  /**
   * Reads all text lines from the specified input stream and passes them
   * one by one to the given line processor.
   * The processing will be terminated, if the end of the text is reached or
   * if the processor returns <b>false</b>.<br>
   * The given input stream will be closed after the execution of this method.
   * Even if an exception occurred.
   *
   * @param inStream The input stream that contains the text (must not be null).
   * @param charset A character set defining the encoding of the stream (must not be null).
   * @param processor The processor that receives the lines from the text (must not be null).
   */
  @SuppressWarnings("resource")
  public void processTextLines(InputStream inStream, Charset charset, ILineProcessor processor) throws IOException
  {
    InputStreamReader reader;

    assertArgumentNotNull("inStream", inStream);
    assertArgumentNotNull("charset", charset);
    assertArgumentNotNull("processor", processor);
    try
    {
      reader = new InputStreamReader(inStream, charset);
      processTextLines(reader, processor);
    }
    finally
    {
      close(inStream);
    }
  }

  /**
   * Reads all text lines from the specified reader and passes them one by one
   * to the given line processor.
   * The processing will be terminated, if the end of the text is reached or
   * if the processor returns <b>false</b>.
   *
   * @param reader The reader that contains a text stream
   * @param processor The processor that receives the lines from the text
   */
  @SuppressWarnings("resource")
  public void processTextLines(Reader reader, ILineProcessor processor) throws IOException
  {
    BufferedReader bufReader;
    String line;
    int counter = 0;
    boolean continue_reading = true;

    assertArgumentNotNull("reader", reader);
    assertArgumentNotNull("processor", processor);

    bufReader = new BufferedReader(reader);
    try
    {
      while (continue_reading && bufReader.ready())
      {
        line = bufReader.readLine();
        if (line == null)
          break;

        counter++;
        continue_reading = processor.processLine(line, counter);
      }
    }
    finally
    {
      close(bufReader);
    }
  }

  /**
   * Reads all text lines from the specified reader into a String array.
   * After that the reader will be closed. Even if an exception occurs.
   *
   * @param reader The reader which provides the text
   */
  public String[] readTextLinesFrom(Reader reader) throws IOException
  {
    return this.basicReadTextLinesFrom(reader, null);
  }

  /**
   * Reads all text lines from the specified stream into a String array.
   * After that the stream will be closed. Even if an exception occurs.
   * The character encoding is the JVM default encoding.
   *
   * @param inStream The stream from which to read the text lines
   */
  public String[] readTextLinesFrom(InputStream inStream) throws IOException
  {
    return this.readTextLinesFrom(new InputStreamReader(inStream));
  }

  /**
   * Reads all text lines matching the filter from the specified stream into a
   * String array.
   * After that the stream will be closed. Even if an exception occurs.
   * The character encoding is the JVM default encoding.
   *
   * @param inStream The stream from which to read the text lines
   * @param filter A filter that defines which lines to return (may be null -> all lines)
   */
  public String[] readTextLinesFrom(InputStream inStream, IStringFilter filter) throws IOException
  {
    return this.readTextLinesFrom(inStream, Charset.defaultCharset(), filter);
  }

  /**
   * Reads all text lines matching the filter from the specified stream into a
   * String array using the specified character encoding.
   * After that the stream will be closed. Even if an exception occurs.
   *
   * @param inStream The stream from which to read the text lines.
   * @param charsetName A charset name already checked to be supported (must not be null).
   * @param filter A filter that defines which lines to return (may be null -> all lines).
   */
  public String[] readTextLinesFrom(InputStream inStream, ICheckedCharset charsetName, IStringFilter filter) throws IOException
  {
    this.assertArgumentNotNull("inStream", inStream);
    this.assertArgumentNotNull("charsetName", charsetName);
    return this.readTextLinesFrom(inStream, charsetName.getCharset(), filter);
  }

  /**
   * Reads all text lines from the specified stream into a String array.
   * After that the stream will be closed. Even if an exception occurs.
   *
   * @param inStream The stream from which to read the text lines (must not be null).
   * @param charset A character set defining the encoding of the stream (must not be null).
   */
  @SuppressWarnings("resource")
  public String[] readTextLinesFrom(InputStream inStream, Charset charset) throws IOException
  {
    InputStreamReader reader;

    assertArgumentNotNull("inStream", inStream);
    assertArgumentNotNull("charset", charset);
    try
    {
      reader = new InputStreamReader(inStream, charset);
      return readTextLinesFrom(reader);
    }
    finally
    {
      this.close(inStream);
    }
  }

  /**
   * Reads all text lines matching the filter from the specified stream into a
   * String array using the specified character encoding.
   * After that the stream will be closed. Even if an exception occurs.
   *
   * @param inStream The stream from which to read the text lines (must not be null).
   * @param charset A character set defining the encoding of the input stream (must not be null).
   * @param filter A filter that defines which lines to return (may be null -> all lines).
   */
  @SuppressWarnings("resource")
  public String[] readTextLinesFrom(InputStream inStream, Charset charset, IStringFilter filter) throws IOException
  {
    InputStreamReader reader;

    assertArgumentNotNull("inStream", inStream);
    assertArgumentNotNull("charset", charset);

    try
    {
      reader = new InputStreamReader(inStream, charset);
      return readTextLinesFrom(reader, filter);
    }
    finally
    {
      this.close(inStream);
    }
  }

  /**
   * Reads all text lines that match the filter from the specified reader into
   * a String array. If the filter is null all lines will be returned.
   * After that the reader will be closed. Even if an exception occurs.
   *
   * @param reader The reader which provides the text
   * @param filter A filter that defines which lines to return (may be null -> all lines)
   */
  public String[] readTextLinesFrom(Reader reader, IStringFilter filter) throws IOException
  {
    return basicReadTextLinesFrom(reader, filter);
  }

  /**
   * Reads all text from the given reader and returns it as a single string.
   *  
   * @param reader The reader from which to get the content (must not be null).
   * @return The whole content of the provided reader as a string.
   * @throws IOException Any open/read problem.
   */
  public String readTextFrom(Reader reader) throws IOException
  {
    StringWriter writer;

    writer = new StringWriter(DEFAULT_BUFFER_SIZE);
    this.copyText(reader, writer);
    return writer.toString();
  }

  /**
   * Copies all text lines from the specified reader to the given writer.
   * After that the reader will be closed. Even if an exception occurs.
   * <br>
   * The writer stays open!
   *
   * @param reader The reader which provides the text to copy (must not be null).
   * @param writer The writer to which the text will be copied (must not be null).
   * @throws IOException If reading or writing fails.
   */
  public void copyText(final Reader reader, final Writer writer) throws IOException
  {
    this.copyText(reader, writer, null);
  }

  /**
   * Copies all text lines from the specified reader to the given writer.
   * After that the reader will be closed. Even if an exception occurs.
   * <br>
   * The writer stays open!
   *
   * @param reader The reader which provides the text to copy (must not be null).
   * @param writer The writer to which the text will be copied (must not be null).
   * @param progressListener A listener that will be called during the copy process before each line
   * gets actually written and once at the very end. The listener gets the current line number as argument and
   * the {@link IProgressListener#UNDEFINED_SIZE} for the total lines.
   * The listener can be used to monitor the progress as well as stopping copying by returning false. 
   * A null listener will be ignored.
   * @throws IOException If reading or writing fails.
   */
  @SuppressWarnings("resource")
  public void copyText(final Reader reader, final Writer writer, final IProgressListener progressListener) throws IOException
  {
    BufferedReader bufReader;
    ILineProcessor processor;
    final AtomicReference<IOException> exceptionHolder = new AtomicReference<IOException>(null);
    final AtomicInteger copiedLines = new AtomicInteger(0);

    bufReader = new BufferedReader(reader);
    try
    {
      processor = new ILineProcessor()
      {
        @Override
        public boolean processLine(String line, int lineNo)
        {
          boolean goOn = true;

          if (progressListener != null)
          {
            goOn = progressListener.processed(IProgressListener.UNDEFINED_SIZE, lineNo);
          }
          if (goOn)
          {
            try
            {
              if (lineNo > 1)
              {
                writer.write(LINE_SEPARATOR);
              }
              writer.write(line);
              copiedLines.incrementAndGet();
            }
            catch (IOException ex)
            {
              exceptionHolder.set(ex);
              return false;
            }
          }
          return goOn;
        }
      };
      this.processTextLines(bufReader, processor);
      if (exceptionHolder.get() == null)
      {
        if (progressListener != null)
        {
          progressListener.processed(IProgressListener.UNDEFINED_SIZE, copiedLines.get());
        }
      }
      else
      {
        throw exceptionHolder.get();
      }
    }
    finally
    {
      close(bufReader);
    }
  }

  /**
   * Reads the whole content of the given input stream and returns
   * it as a string.
   * The stream will be closed after calling this method. Even if an exception
   * occurred!
   * The stream character encoding is the JVM default encoding.
   *
   * @param inStream The input stream to read.
   * @return The text content of the given stream.
   */
  public String readTextFrom(InputStream inStream) throws IOException
  {
    return this.readTextFrom(inStream, Charset.defaultCharset());
  }

  /**
   * Reads the whole content of the given input stream and returns
   * it as a string.
   * The stream will be closed after calling this method. Even if an exception
   * occurred!
   *
   * @param inStream The input stream to read (must not be null).
   * @param charsetName A charset name already checked to be supported (must not be null).
   * @return The text content of the given stream
   */
  public String readTextFrom(InputStream inStream, ICheckedCharset charsetName) throws IOException
  {
    return this.readTextFrom(inStream, charsetName.getCharset());
  }

  /**
   * Reads the whole content of the given input stream and returns
   * it as a string.
   * The stream will be closed after calling this method. Even if an exception
   * occurred!
   *
   * @param inStream The input stream to read (must not be null).
   * @param charset A character set defining the encoding of the stream (must not be null).
   * @return The text content of the given stream.
   */
  public String readTextFrom(InputStream inStream, Charset charset) throws IOException
  {
    StringWriter writer;

    writer = new StringWriter(DEFAULT_BUFFER_SIZE);
    this.copyText(inStream, charset, writer);
    return writer.toString();
  }

  /**
   * Copies all data from the input stream to the output stream using
   * a buffer with the default size (4096 bytes).
   * After all data is copied both streams will be closed !
   * 
   * @param inStream The source stream.
   * @param outStream The target stream.
   * @return The number of bytes copied.
   */
  public long copyStream(InputStream inStream, OutputStream outStream) throws IOException
  {
    return this.copyStream(inStream, outStream, DEFAULT_BUFFER_SIZE);
  }

  /**
   * Copies all data from the input stream to the output stream using
   * a buffer with the default size (4096 bytes).
   * After all data is copied both streams will be closed !
   * 
   * @param inStream The source stream.
   * @param outStream The target stream.
   * @param listener A listener that will be called during the copy process before each chunk of data (see bufferSize)
   * gets actually processed and once at the very end. The listener can be used to monitor the progress as well as
   * stopping copying by returning false. A null listener will be ignored.
   * @return The number of bytes copied.
   */
  public long copyStream(InputStream inStream, OutputStream outStream, IProgressListener listener) throws IOException
  {
    return this.copyStream(inStream, outStream, DEFAULT_BUFFER_SIZE, listener);
  }

  /**
   * Copies all data from the input stream to the output stream using
   * a buffer of the given size in bytes.
   * After all data has been copied both streams will be closed!
   * 
   * @param inStream The source stream.
   * @param outStream The target stream.
   * @param bufSize The size of the buffer that is used for the transfer of byte chunks (must be positive value).
   * @return The number of bytes copied.
   */
  public long copyStream(InputStream inStream, OutputStream outStream, int bufSize) throws IOException
  {
    return this.copyStream(inStream, outStream, bufSize, NO_LIMIT);
  }

  /**
   * Copies all data from the input stream to the output stream using
   * a buffer of the given size in bytes.
   * After all data has been copied both streams will be closed!
   * 
   * @param inStream The source stream.
   * @param outStream The target stream.
   * @param bufSize The size of the buffer that is used for the transfer of byte chunks (must be positive value).
   * @param listener A listener that will be called during the copy process before each chunk of data (see bufferSize)
   * gets actually processed and once at the very end. The listener can be used to monitor the progress as well as
   * stopping copying by returning false. A null listener will be ignored.
   * @return The number of bytes copied.
   */
  public long copyStream(InputStream inStream, OutputStream outStream, int bufSize, IProgressListener listener) throws IOException
  {
    return this.copyStream(inStream, outStream, bufSize, NO_LIMIT, listener);
  }

  /**
   * Copies all data from the input stream to the output stream using
   * a buffer of the given size in bytes.
   * After all data has been copied both streams will be closed!
   * 
   * @param inStream The source stream.
   * @param outStream The target stream.
   * @param bufSize The size of the buffer that is used for the transfer of byte chunks (must be positive value).
   * @param limit The maximum number of bytes that will be copied. If exceeded an IOException will be thrown.
   *   Setting this parameter to null, zero or a negative value will specify no limit.
   * @return The number of bytes copied.
   */
  public long copyStream(InputStream inStream, OutputStream outStream, int bufSize, Long limit) throws IOException
  {
    return this.copyStream(inStream, outStream, bufSize, limit, IGNORING_PROGRESS_LISTENER);
  }

  /**
   * Copies all data from the input stream to the output stream using
   * a buffer of the given size in bytes.
   * After all data has been copied both streams will be closed!
   * 
   * @param inStream The source stream.
   * @param outStream The target stream.
   * @param bufSize The size of the buffer that is used for the transfer of byte chunks (must be positive value).
   * @param limit The maximum number of bytes that will be copied. If exceeded an IOException will be thrown.
   *   Setting this parameter to null, zero or a negative value will specify no limit.
   * @param listener A listener that will be called during the copy process before each chunk of data (see bufferSize)
   * gets actually processed and once at the very end. The listener can be used to monitor the progress as well as
   * stopping copying by returning false. A null listener will be ignored.
   * @return The number of bytes copied.
   */
  public long copyStream(InputStream inStream, OutputStream outStream, int bufSize, Long limit, IProgressListener listener) throws IOException
  {
    StreamCopy streamCopy;

    streamCopy = new StreamCopy(bufSize, listener);
    return streamCopy.copy(inStream, outStream, limit);
  }

  /**
   * Returns an input stream containing the given string using the provided charset
   * for encoding.
   * 
   * @param string The string to convert to a stream (must not be null).
   * @param charset The character encoding to convert the string to bytes. If null,
   * the default platform encoding will be used.
   * @return An input stream containing the provided string data.
   */
  public InputStream asStream(final String string, final Charset charset)
  {
    InputStream inStream;
    Charset encoding;

    encoding = (charset == null) ? Charset.defaultCharset() : charset;
    inStream = new ByteArrayInputStream(string.getBytes(encoding));
    return inStream;
  }

  /**
   * Returns the given string with all %xx placeholders decoded using the
   * specified charset encoding.
   *
   * @param str The string to decode.
   * @param charset The charset to be used for decoding. 
   * @return A string with all %xx placeholders decoded according to the specified charset.
   */
  public String urlDecode(String str, Charset charset)
  {
    if (str == null)
    {
      return null;
    }
    try
    {
      return URLDecoder.decode(str, charset.name());
    }
    catch (UnsupportedEncodingException e)
    {
      // Should never happen because it is always a supported charset.
      e.printStackTrace();
    }
    return str;
  }

  /**
   * Returns the given string with all %xx placeholders decoded.
   *
   * @param str The string to decode
   * @return A string with all %xx placeholders decoded (using UTF-8)
   */
  public String urlDecode(String str)
  {
    return urlDecode(str, ICharsets.UTF_8);
  }

  /**
   * Returns the given string with all characters that are not allowed directly
   * in URLs encoded as %xx placeholders. In particular it replaces spaces by "
   * %20" rather than '+'! That is reasonable in cases where files in the
   * local file system contain spaces. Such URLs (staring with "file:") cannot
   * open a stream if the URL contains a '+' because the '+' would not be
   * decoded back to a space. That is a bug (or at least an inconsistency) in
   * the URL class implementation.
   *
   * @param str The string to encode.
   * @param charset The charset to be used for encoding. 
   * @return A string with %xx encodings for all invalid URL characters.
   */
  public String urlEncode(String str, Charset charset)
  {
    String newStr;

    if (str == null)
    {
      return null;
    }
    try
    {
      newStr = URLEncoder.encode(str, charset.name());
      return newStr.replace("+", "%20");
    }
    catch (UnsupportedEncodingException e)
    {
      // Cannot happen because a Charset was used.
      e.printStackTrace();
    }
    return str;
  }

  /**
   * Returns the given string with all characters that are not allowed directly
   * in URLs encoded as %xx placeholders. In particular it replaces spaces by "
   * %20" rather than '+'! That is reasonable in cases where files in the
   * local file system contain spaces. Such URLs (staring with "file:") cannot
   * open a stream if the URL contains a '+' because the '+' would not be
   * decoded back to a space. That is a bug (or at least an inconsistency) in
   * the URL class implementation.
   *
   * @param str The string to encode.
   * @return A string with %xx (UTF-8) encodings for all invalid URL characters.
   */
  public String urlEncode(String str)
  {
    return this.urlEncode(str, ICharsets.UTF_8);
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  /**
   * Reads all text lines that match the filter from the specified reader into
   * a String array. If the filter is null all lines will be returned.
   * After that the reader will be closed. Even if an exception occurs.
   *
   * @param reader The reader which provides the text
   * @param filter A filter that defines which lines to return (may be null -> all lines)
   */
  protected String[] basicReadTextLinesFrom(final Reader reader, final IStringFilter filter) throws IOException
  {
    BufferedReader bufReader;
    ILineProcessor processor;
    final List<String> lines = new ArrayList<String>(100);

    bufReader = new BufferedReader(reader);
    try
    {
      processor = new ILineProcessor()
      {
        @Override
        public boolean processLine(String line, int lineNo)
        {
          if ((filter == null) || (filter.matches(line)))
          {
            lines.add(line);
          }
          return true;
        }
      };
      this.processTextLines(bufReader, processor);
    }
    finally
    {
      bufReader.close();
    }
    return lines.toArray(new String[lines.size()]);
  }

  protected void copyText(InputStream inStream, Charset charset, StringWriter writer) throws IOException
  {
    InputStreamReader reader = null;

    this.assertArgumentNotNull("inStream", inStream);
    this.assertArgumentNotNull("charset", charset);

    try
    {
      reader = new InputStreamReader(inStream, charset);
      this.copyText(reader, writer);
    }
    finally
    {
      this.close(reader);
      this.close(inStream);
    }
  }

  protected void assertArgumentNotNull(String argName, Object arg)
  {
    if (arg == null)
    {
      throw new IllegalArgumentException("Parameter " + argName + " must not be null!");
    }
  }
}
