// ===========================================================================
// CONTENT  : CLASS MultiValueProperty
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 11/03/2012
// HISTORY  :
//  11/03/2012  mdu  CREATED
//
// Copyright (c) 2012, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.util;

import java.util.ArrayList;
import java.util.List;

/**
 * This represents a single property with a name and a list of multiple values.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class MultiValueProperty<ValueType>
{
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private NamedValue<List<ValueType>> namedValues;
  private boolean allowDuplicates = false;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with a name.
   */
  public MultiValueProperty(String name, ValueType... values)
  {
    super();
    if (name == null)
    {
      throw new IllegalArgumentException("The name must not be null for a MultiValueProperty.");
    }
    setNamedValues(new NamedValue<List<ValueType>>(name, new ArrayList<ValueType>()));
    addValues(values);
  }

  /**
   * Initialize the new instance with a name and define whether or not 
   * duplicate values are allowed.
   */
  public MultiValueProperty(String name, boolean allowDuplicateValues, ValueType... values)
  {
    this(name, values);
    setAllowDuplicates(allowDuplicateValues);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the name of this property.
   */
  public String getName()
  {
    return getNamedValues().name();
  }

  public boolean getAllowDuplicates()
  {
    return allowDuplicates;
  }

  public void setAllowDuplicates(boolean newValue)
  {
    allowDuplicates = newValue;
  }

  /**
   * Add the given values to the list of values. 
   * 
   * @param values The values to add.
   * @return <tt>true</tt> if at least one value was added.
   */
  public boolean addValues(ValueType... values)
  {
    boolean addedSomething = false;

    for (ValueType value : values)
    {
      addedSomething = addValue(value) || addedSomething;
    }
    return addedSomething;
  }

  /**
   * Remove the given values from the list of values.
   * @param values The values to remove.
   * @return <tt>true</tt> if at least one value was in the list and has been removed.
   */
  public boolean removeValues(ValueType... values)
  {
    boolean removedSomething = false;

    for (ValueType value : values)
    {
      removedSomething = removeValue(value) || removedSomething;
    }
    return removedSomething;
  }

  /**
   * Returns true if the given value is already in the list of values.
   */
  public boolean containsValue(ValueType value)
  {
    if (value == null)
    {
      return false;
    }
    return getNamedValues().value().contains(value);
  }

  /**
   * Returns a list of the values.
   * This is a copy of the internal list, so it might be modified without
   * side-effects to this property.
   */
  public List<ValueType> getValues()
  {
    List<ValueType> copy;

    copy = new ArrayList<ValueType>(getNamedValues().value());
    return copy;
  }

  /**
   * Returns the first value of the internal list of values or null if no values
   * are available.
   */
  public ValueType getFirstValue()
  {
    if (hasValues())
    {
      return getNamedValues().value().get(0);
    }
    return null;
  }

  /**
   * Remove all values;
   */
  public void clearValues()
  {
    getNamedValues().value().clear();
  }

  /**
   * Returns true if no value is in the list of values.
   */
  public boolean isEmpty()
  {
    return getNamedValues().value().isEmpty();
  }

  /**
   * Returns true if at least one value is in the list of values.
   */
  public boolean hasValues()
  {
    return !isEmpty();
  }

  /**
   * Returns the number of values in the list of values,
   */
  public int numberOfValues()
  {
    return getNamedValues().value().size();
  }

  /**
   * Returns true if the given object is a MultiValueProperty with the same
   * name (case-sensitive) as this MultiValueProperty instance.
   */
  @SuppressWarnings("rawtypes")
  @Override
  public boolean equals(Object obj)
  {
    if (obj instanceof MultiValueProperty)
    {
      MultiValueProperty prop = (MultiValueProperty)obj;
      return getName().equals(prop.getName());
    }
    return false;
  }

  @Override
  public int hashCode()
  {
    return getClass().getName().hashCode() ^ getName().hashCode();
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  /**
   * Add the given value to the list of values. 
   * 
   * @param value The value to add.
   * @return <tt>true</tt> if the value was added.
   */
  protected boolean addValue(ValueType value)
  {
    if (value == null)
    {
      return false;
    }
    if (!getAllowDuplicates())
    {
      if (containsValue(value))
      {
        return false;
      }
    }
    return getNamedValues().value().add(value);
  }

  /**
   * Remove the given value from the list of values.
   * @param value The value to remove.
   * @return <tt>true</tt> if the value was in the list and has been removed.
   */
  protected boolean removeValue(ValueType value)
  {
    List<ValueType> values;

    if (value == null)
    {
      return false;
    }
    values = getNamedValues().value();
    return values.remove(value);
  }

  protected NamedValue<List<ValueType>> getNamedValues()
  {
    return namedValues;
  }

  protected void setNamedValues(NamedValue<List<ValueType>> newValue)
  {
    namedValues = newValue;
  }
}
