// ===========================================================================
// CONTENT  : CLASS NamedValue
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.1 - 08/09/2019
// HISTORY  :
//  14/07/2002  duma  CREATED
//	16/07/2004	duma	added		-->	copyNamedValue()
//	27/03/2010	mdu		changed	-->	to support generic types
//  08/09/2019  mdu   added   --> of()
//
// Copyright (c) 2002-2019, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.util;

/**
 * This specialized Association allows only strings as key.
 * <br>
 * Access to key and value is provided by
 * <ul>
 *   <li>name()
 *   <li>name( newName )
 *   <li>value()
 *   <li>value( newValue )
 * </ul>
 *
 * @author Manfred Duchrow
 * @version 2.1
 */
@SuppressWarnings("rawtypes")
public class NamedValue<V> extends Association<String, V> implements Comparable
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final long serialVersionUID = -3244424290618041481L;

  // =========================================================================
  // PUBLIC CLASS METHODS
  // =========================================================================
  /**
   * Creates a new instance with a name and an associated value.
   * If name is null it will be transformed to an empty String ("").
   * This is just a convenience factory method.
   */
  public static <T> NamedValue<T> of(String name, T value)
  {
    return new NamedValue<T>(name, value);
  }
  
  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with a name and a null value.
   * If name is null it will be transformed to an empty String ("").
   */
  public NamedValue(String name)
  {
    this(name, null);
  }

  /**
   * Initialize the new instance with a name and an associated value.
   * If name is null it will be transformed to an empty String ("").
   */
  public NamedValue(String name, V value)
  {
    super((name == null) ? "" : name, value);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Sets the new key. A null key will be ignored.
   */
  @Override
  public void key(String newKey)
  {
    if (newKey != null)
    {
      super.key(newKey);
    }
  }

  /**
   * Sets the name (key) of the association.
   * 
   * @param newName The new key (must not be null !)
   */
  public void name(String newName)
  {
    key(newName);
  }

  /**
   * Returns the name (key) of the association.
   */
  public String name()
  {
    return key();
  }

  /**
   * Compares this object with the specified object for order. 
   * Returns a negative integer, zero, or a positive integer as this object 
   * is less than, equal to, or greater than the specified object.
   * <br>
   * Here the comparison will be done on the names of the NamedValue objects.
   */
  @Override
  public int compareTo(Object obj)
  {
    if (obj instanceof NamedValue)
    {
      return name().compareTo(((NamedValue)obj).name());
    }
    return 1;
  }

  /**
   * Returns a copy of this object
   */
  public NamedValue<V> copyNamedValue()
  {
    NamedValue<V> copy;

    copy = new NamedValue<V>(name(), value());
    return copy;
  }
}
