// ===========================================================================
// CONTENT  : CLASS OrderedMap
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.0 - 15/11/2014
// HISTORY  :
//  12/03/2004  mdu  CREATED
//	05/08/2006	mdu		bugfix	-->	changed remove() to return the value and not the Association
//  15/11/214   mdu   changed --> generic type
//
// Copyright (c) 2004-2014, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.util;

import java.util.Collection;
import java.util.Map;
import java.util.Set;

/**
 * A Map implementation that pertains the order of its elements according to
 * when they have been added.
 *
 * @author Manfred Duchrow
 * @version 2.0
 */
public class OrderedMap<K, V> implements Map<K, V>
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private AssociationList<K, V> mapping = null;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  public OrderedMap()
  {
    super();
    setMapping(createEmptyMapping());
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Removes all entries from this map.
   * 
   * @see java.util.Map#clear()
   */
  @Override
  public void clear()
  {
    getMapping().clear();
  }

  /**
   * This method returns true, if the given key can be found in this map.
   * 
   * @see java.util.Map#containsKey(Object)
   */
  @SuppressWarnings("unchecked")
  @Override
  public boolean containsKey(Object key)
  {
    return (getMapping().findAssociation((K)key) != null);
  }

  /**
   * This method returns true, if the given value is a value of any entry
   * in this map.
   * 
   * @see java.util.Map#containsValue(Object)
   */
  @Override
  public boolean containsValue(Object value)
  {
    return getMapping().asHashMap().containsValue(value);
  }

  /**
   * Returns all entries of the map
   * 
   * @see java.util.Map#entrySet()
   */
  @Override
  public Set<Map.Entry<K, V>> entrySet()
  {
    Set<Map.Entry<K, V>> entrySet;
    Map.Entry<K, V> assoc;

    entrySet = new OrderedSet<Map.Entry<K, V>>(size());
    for (int i = 0; i < size(); i++)
    {
      assoc = getMapping().associationAt(i);
      entrySet.add(assoc);
    }
    return entrySet;
  }

  @SuppressWarnings("unchecked")
  /**
   * Returns the object that was registered under the specified key.
   * 
   * @see java.util.Map#get(Object)
   */
  @Override
  public V get(Object key)
  {
    return getMapping().valueAt((K)key);
  }

  /**
   * Returns true, if there is no entry in this map.
   * 
   * @see java.util.Map#isEmpty()
   */
  @Override
  public boolean isEmpty()
  {
    return (getMapping().size() == 0);
  }

  /**
   * Returns all keys in this map.
   * 
   * @see java.util.Map#keySet()
   */
  @Override
  public Set<K> keySet()
  {
    Set<K> keys;

    keys = new OrderedSet<K>(size());
    keys.addAll(getMapping().keys());
    return keys;
  }

  /**
   * Puts the given value under the specified key into the map.
   * 
   * @see java.util.Map#put(Object, Object)
   */
  @Override
  public V put(K key, V value)
  {
    getMapping().put(key, value);
    return value;
  }

  /**
   * Each entry of the given map will be added. 
   * 
   * @see java.util.Map#putAll(Map)
   */
  @SuppressWarnings("unchecked")
  @Override
  public void putAll(Map<? extends K, ? extends V> map)
  {
    getMapping().putAll((Map<K, V>)map);
  }

  /**
   * Removes the entry with the specified key.
   * 
   * @see java.util.Map#remove(Object)
   */
  @SuppressWarnings("unchecked")
  @Override
  public V remove(Object key)
  {
    Association<K, V> assoc;

    assoc = getMapping().removeKey((K)key);
    if (assoc == null)
    {
      return null;
    }
    return assoc.value();
  }

  /**
   * Returns how many entries currently are in this map.
   * 
   * @see java.util.Map#size()
   */
  @Override
  public int size()
  {
    return getMapping().size();
  }

  /**
   * Returns all values of the map.
   * 
   * @see java.util.Map#values()
   */
  @Override
  public Collection<V> values()
  {
    return getMapping().values();
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected AssociationList<K, V> createEmptyMapping()
  {
    return new AssociationList<K, V>();
  }

  protected AssociationList<K, V> getMapping()
  {
    return mapping;
  }

  protected void setMapping(AssociationList<K, V> newValue)
  {
    mapping = newValue;
  }

}
