// ===========================================================================
// CONTENT  : CLASS OrderedProperties
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.2 - 29/12/2020
// HISTORY  :
//  17/09/2004  mdu  CREATED
//  16/11/2014  mdu  refactored --> generic types
//  29/12/2020  mdu  added -> methods with IName
//
// Copyright (c) 2004-2020, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.Vector;

import org.pfsw.bif.text.IMutableStringPair;
import org.pfsw.bif.text.IName;

/**
 * Represents a properties collection with empty lines and comments.
 * The order will be preserved.
 * <b>DO NOT USE THE NORMAL METHODS INHERITED FROM java.util.Hashtable
 * except those that are explicitly overridden in this class!</b>
 *
 * @author Manfred Duchrow
 * @version 2.2
 */
public class OrderedProperties<T extends IMutableStringPair> extends Properties
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final long serialVersionUID = 4168763680301505984L;

  protected static final int INITIAL_CAPACITY = 50;

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private List<T> elements = null;
  private Map<String, Integer> propertyIndex = null;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  public OrderedProperties()
  {
    this(INITIAL_CAPACITY);
  }

  /**
   * Initialize the new instance with an initial capacity.
   */
  public OrderedProperties(int initialCapacity)
  {
    super();
    setElements(newElementsList(initialCapacity));
    setPropertyIndex(new HashMap<String, Integer>(initialCapacity));
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns a list containing all properties names in the order they 
   * have been added or read from a stream.
   */
  public List<String> nameList()
  {
    List<String> nameList;

    nameList = new ArrayList<String>(size());
    for (T property : getElements())
    {
      if (isValidProperty(property))
      {
        nameList.add(property.getString1());
      }
    }
    return nameList;
  }

  /**
   * Returns an array containing the properties names in the order they 
   * have been added or read from a stream.
   */
  public String[] getPropertyNames()
  {
    List<String> names;

    names = nameList();
    return names.toArray(new String[names.size()]);
  }

  /**
   * Returns how many properties are contained. 
   * That does NOT include empty and comment lines!
   * @see #numberOfElements()
   */
  @Override
  public synchronized int size()
  {
    int count = 0;

    for (T property : getElements())
    {
      if (isValidProperty(property))
      {
        count++;
      }
    }
    return count;
  }

  /**
   * Returns how many elements are contained. That includes empty and comment lines!
   * @see #size()
   */
  public int numberOfElements()
  {
    return getElements().size();
  }

  /**
   * Returns whether or not this contains any properties.
   * Empty and comment lines are ignored here.
   * @see #hasContent()
   */
  @Override
  public synchronized boolean isEmpty()
  {
    return size() < 1;
  }

  /**
   * Returns whether or not any data is contained.
   * That includes empty and comment lines are ignored here.
   * So this method can return true even if no property is defined.
   * @see #isEmpty()
   */
  public boolean hasContent()
  {
    return numberOfElements() > 0;
  }

  /**
   * Returns the value of the property with the given name or null if it is
   * not found.
   * 
   * @param name The name of the property to look for.
   */
  @Override
  public String getProperty(String name)
  {
    T property;

    property = findProperty(name);
    if (isNullOrDeleted(property))
    {
      return null;
    }
    return property.getString2();
  }

  /**
   * Returns the value of the property with the given name or null if it is
   * not found.
   * 
   * @param name The name of the property to look for.
   */
  public String getProperty(IName name)
  {
    return getProperty(name.asString());
  }
  
  /**
   * Modifies the value or adds the property specified by the given name.
   * Returns true the previous value if the property was found and modified 
   * or null if the property was added.
   * 
   * @param name The name of the property to set
   * @param value The new value to be set
   */
  @Override
  public synchronized Object setProperty(String name, String value)
  {
    T property;
    String oldValue;

    property = findProperty(name);
    if (property == null)
    {
      appendProperty(name, value);
      return null;
    }
    oldValue = property.getString2();
    property.setString2(value);
    return oldValue;
  }

  /**
   * Modifies the value or adds the property specified by the given name.
   * Returns true the previous value if the property was found and modified 
   * or null if the property was added.
   * 
   * @param name The name of the property to set.
   * @param value The new value to be set.
   */
  public Object setProperty(IName name, String value)
  {
    return setProperty(name.asString(), value);
  }
  
  /**
   * Deletes the property with the specified name.
   * Returns true if the property was deleted, otherwise false.
   * 
   * @param name The name of the property to delete
   */
  public boolean deleteProperty(String name)
  {
    return remove(name) != null;
  }

  /**
   * Deletes the property with the specified name.
   * Returns true if the property was deleted, otherwise false.
   * 
   * @param name The name of the property to delete.
   */
  public boolean deleteProperty(IName name)
  {
    return deleteProperty(name.asString());
  }
  
  /**
   * Removes the key (and its corresponding value) from this hashtable. 
   * This method does nothing if the key is not found.
   * 
   * @return the value to which the key had been mapped, or null if the key did not have a mapping.
   */
  @Override
  public synchronized Object remove(Object key)
  {
    T property;

    if (!(key instanceof String))
    {
      return null;
    }

    property = findProperty((String)key);
    if (property == null)
    {
      return null;
    }
    getPropertyIndex().remove(property.getString1());
    getElements().remove(property);
    return property.getString2();
  }

  /**
   * Searches for the property with the specified name in this property list. 
   * If the name is not found in this property list the method returns the 
   * default value argument.
   */
  @Override
  public String getProperty(String name, String defaultValue)
  {
    T property;

    property = findProperty(name);
    if (property == null)
    {
      return defaultValue;
    }
    return property.getString2();
  }

  /**
   * Searches for the property with the specified key in this property list. 
   * If the key is not found in this property list the method returns the 
   * default value argument.
   */
  public String getProperty(IName name, String defaultValue)
  {
    return getProperty(name.asString(), defaultValue);
  }
  
  /**
   * NOT SUPPORTED METHOD!
   * @throws UnsupportedOperationException
   */
  @Override
  public void list(PrintStream out)
  {
    throw new UnsupportedOperationException(getClass().getName() + ".list()");
  }

  /**
   * NOT SUPPORTED METHOD!
   * @throws UnsupportedOperationException
   */
  @Override
  public void list(PrintWriter out)
  {
    throw new UnsupportedOperationException(getClass().getName() + ".list()");
  }

  /**
   * NOT SUPPORTED METHOD!
   * @throws UnsupportedOperationException
   */
  @Override
  public synchronized void load(InputStream inStream) throws IOException
  {
    throw new UnsupportedOperationException(getClass().getName() + ".load()");
  }

  /**
   * Returns an enumeration of the properties names in the order they have been
   * added or read from a stream.
   * <p>
   * <b>Be aware that the returned enumeration is NOT backed by this 
   * OrderedProperties object!</b>  
   */
  @Override
  public Enumeration<String> propertyNames()
  {
    Vector<String> names;

    names = new Vector<String>(nameList());
    return names.elements();
  }

  @Override
  public Set<String> stringPropertyNames()
  {
    Set<String> nameSet;

    nameSet = new OrderedSet<String>(size());
    nameSet.addAll(nameList());
    return nameSet;
  }

  /**
   * NOT SUPPORTED METHOD!
   * @throws UnsupportedOperationException
   * @deprecated
   */
  @SuppressWarnings("deprecation")
  @Deprecated
  @Override
  public synchronized void save(OutputStream out, String header)
  {
    throw new UnsupportedOperationException(getClass().getName() + ".save()");
  }

  /**
   * NOT SUPPORTED METHOD!
   * @throws UnsupportedOperationException
   */
  @Override
  public synchronized void store(OutputStream out, String header) throws IOException
  {
    throw new UnsupportedOperationException(getClass().getName() + ".store()");
  }

  /**
   * Removes all properties 
   */
  @Override
  public synchronized void clear()
  {
    getElements().clear();
    getPropertyIndex().clear();
  }

  /**
   * Returns all keys of this properties list in the order they have been added.
   */
  @Override
  public synchronized Enumeration<Object> keys()
  {
    Vector<Object> names;

    names = new Vector<Object>(nameList());
    return names.elements();
  }

  /**
   * Returns all keys of this properties list in the order they have been added.
   */
  @Override
  public Set<Object> keySet()
  {

    Set<Object> nameSet;

    nameSet = new OrderedSet<Object>(size());
    nameSet.addAll(nameList());
    return nameSet;
    }

  /**
   * Returns all entries in the order they have been added
   */
  @Override
  public Set<Map.Entry<Object,Object>> entrySet()
  {
    OrderedMap<Object, Object> entries;
    Iterator<T> iter;
    T element;

    entries = new OrderedMap<Object, Object>();
    iter = getElements().iterator();
    while (iter.hasNext())
    {
      element = iter.next();
      if (isValidProperty(element))
      {
        entries.put(element.getString1(), element.getString2());
      }
    }
    return entries.entrySet();
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  /**
   * Appends the property specified by the given name with the given value..
   * Returns true if the property was successfully added at the end, 
   * otherwise false.
   * 
   * @param name The name of the property to append
   * @param value The new value to be append
   */
  protected boolean appendProperty(String name, String value)
  {
    T property;

    property = newProperty(name, value);
    addElement(property);
    return true;
  }

  /**
   * Create a new property object with the given name and value. Subclasses
   * may override this to implement their own property class.
   */
  @SuppressWarnings("unchecked")
  protected T newProperty(String name, String value)
  {
    return (T)new NamedText(name, value);
  }

  protected void addElement(T element)
  {
    Integer index;

    index = new Integer(getElements().size());
    getElements().add(element);
    registerProperty(element, index);
  }

  protected void registerProperty(T property, Integer index)
  {
    getPropertyIndex().put(property.getString1(), index);
  }

  protected T findProperty(String key)
  {
    Integer index;

    index = indexOfProperty(key);
    if (index == null)
    {
      return null;
    }
    return propertyAt(index);
  }

  protected Integer indexOfProperty(String key)
  {
    return getPropertyIndex().get(key);
  }

  protected T propertyAt(Integer index)
  {
    return propertyAt(index.intValue());
  }

  protected T propertyAt(int index)
  {
    return getElements().get(index);
  }

  /**
   * Returns true if the given property is null or marked as deleted.
   * Subclasses may override this method to enhance behavior.
   * 
   * @param property The property to check
   * @return true if the given property is null or marked as deleted
   */
  protected boolean isNullOrDeleted(T property)
  {
    return (property == null);
  }

  protected boolean isValidProperty(@SuppressWarnings("unused") T property)
  {
    return true;
  }

  protected List<T> newElementsList(int initialCapacity)
  {
    return new ArrayList<T>(initialCapacity);
  }

  protected List<T> getElements()
  {
    return this.elements;
  }

  protected void setElements(List<T> newValue)
  {
    this.elements = newValue;
  }

  protected Map<String, Integer> getPropertyIndex()
  {
    return this.propertyIndex;
  }

  protected void setPropertyIndex(Map<String, Integer> newValue)
  {
    this.propertyIndex = newValue;
  }
}
