// ===========================================================================
// CONTENT  : CLASS OrderedSet
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.3 - 18/10/2014
// HISTORY  :
//  12/03/2004  mdu  CREATED
//	19/02/2006	mdu		bugfix	->	removeAll()
//	22/02/2006	mdu		added		-> 	implements List
//  18/10/2014  mdu   changed ->  generic element type
//
// Copyright (c) 2004-2014, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.util;

import static org.pfsw.util.CollectionUtil.CU;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Set;

/**
 * Implements a Set that keeps the order of its elements according to when
 * they were added. Additionally it implements the List interface.
 * So instances of this class can either be used as Set or as List.
 *
 * @author Manfred Duchrow
 * @version 1.3
 */
public class OrderedSet<E> implements Set<E>, List<E>
{
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private List<E> elements = null;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  public OrderedSet()
  {
    this(20);
  }

  /**
   * Initialize the new set with an initial capacity.
   */
  public OrderedSet(int initialCapacity)
  {
    super();
    setElements(createInnerList(initialCapacity));
  }

  /**
   * Initialize the new set with an inner list.
   */
  protected OrderedSet(List<E> innerList)
  {
    super();
    setElements(innerList);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * @see java.util.Collection#add(java.lang.Object)
   */
  @Override
  public boolean add(E obj)
  {
    if (!getElements().contains(obj))
    {
      return getElements().add(obj);
    }
    return false;
  }

  /**
   * @see java.util.Collection#addAll(java.util.Collection)
   */
  @Override
  public boolean addAll(Collection<? extends E> coll)
  {
    boolean changed = false;
    Iterator<? extends E> iter;

    iter = coll.iterator();
    while (iter.hasNext())
    {
      if (add(iter.next()))
      {
        changed = true;
      }
    }
    return changed;
  }

  /**
   * @see java.util.Collection#clear()
   */
  @Override
  public void clear()
  {
    getElements().clear();
  }

  /**
   * @see java.util.Collection#contains(java.lang.Object)
   */
  @Override
  public boolean contains(Object obj)
  {
    return getElements().contains(obj);
  }

  /**
   * @see java.util.Collection#containsAll(java.util.Collection)
   */
  @Override
  public boolean containsAll(Collection<?> coll)
  {
    return getElements().containsAll(coll);
  }

  /**
   * @see java.util.Collection#isEmpty()
   */
  @Override
  public boolean isEmpty()
  {
    return getElements().isEmpty();
  }

  /**
   * @see java.util.Collection#iterator()
   */
  @Override
  public Iterator<E> iterator()
  {
    return getElements().iterator();
  }

  /**
   * @see java.util.Collection#remove(java.lang.Object)
   */
  @Override
  public boolean remove(Object obj)
  {
    return getElements().remove(obj);
  }

  /**
   * @see java.util.Collection#removeAll(java.util.Collection)
   */
  @Override
  public boolean removeAll(Collection<?> coll)
  {
    return getElements().removeAll(coll);
  }

  /**
   * @see java.util.Collection#retainAll(java.util.Collection)
   */
  @Override
  public boolean retainAll(Collection<?> coll)
  {
    return getElements().retainAll(coll);
  }

  /**
   * @see java.util.Collection#size()
   */
  @Override
  public int size()
  {
    return getElements().size();
  }

  /**
   * @see java.util.Collection#toArray()
   */
  @Override
  public Object[] toArray()
  {
    return CU.toArray(getElements());
  }

  /**
   * @see java.util.Collection#toArray(java.lang.Object[])
   */
  @Override
  public <T> T[] toArray(T[] array)
  {
    return getElements().toArray(array);
  }

  /**
   * Inserts the specified element at the specified position in this list  
   */
  @Override
  public void add(int index, E element)
  {
    if (!contains(element))
    {
      getElements().add(index, element);
    }
  }

  /**
   *  Inserts all of the elements in the specified collection into this list 
   *  at the specified position if they are not yet in this list.
   *  
   *  @return true if this list changed as a result of the call.
   */
  @Override
  public boolean addAll(int index, Collection<? extends E> c)
  {
    Iterator<? extends E> iter;
    boolean changed = false;
    int insertIndex = index;
    E object;

    iter = c.iterator();
    while (iter.hasNext())
    {
      object = iter.next();
      if (!contains(object))
      {
        add(insertIndex, object);
        insertIndex++;
        changed = true;
      }
    }
    return changed;
  }

  /**
   * Returns the element at the specified position in this list.
   */
  @Override
  public E get(int index)
  {
    return getElements().get(index);
  }

  /**
   * Returns the index in this list of the first occurrence of the specified 
   * element, or -1 if this list does not contain this element.
   */
  @Override
  public int indexOf(Object object)
  {
    return getElements().indexOf(object);
  }

  /**
   * Returns the index in this list of the last occurrence of the specified 
   * element, or -1 if this list does not contain this element.
   */
  @Override
  public int lastIndexOf(Object object)
  {
    return getElements().lastIndexOf(object);
  }

  /**
   * Returns a list iterator of the elements in this list (in proper sequence).
   */
  @Override
  public ListIterator<E> listIterator()
  {
    return getElements().listIterator();
  }

  /**
   * Returns a list iterator of the elements in this list (in proper sequence), 
   * starting at the specified position in this list.
   */
  @Override
  public ListIterator<E> listIterator(int index)
  {
    return getElements().listIterator(index);
  }

  /**
   * Returns the object at the specified index
   */
  @Override
  public E remove(int index)
  {
    return getElements().remove(index);
  }

  /**
   * Sets the given element at the specified index if the element is not
   * yet already in the collection!
   */
  @Override
  public E set(int index, E element)
  {
    if (!contains(element))
    {
      return getElements().set(index, element);
    }
    return get(index);
  }

  /**
   * Returns a sub list containing the elements in the range of the defined
   * indices. The result is also an OrderedSet.
   * Contrary to the definition of this method in the List interface here
   * the result list is NOT backed by this list.
   * That means, any changes to the result list do not have any impact on
   * this list!
   */
  @Override
  public List<E> subList(int fromIndex, int toIndex)
  {
    List<E> result;

    result = createInnerList(toIndex - fromIndex);
    for (int i = fromIndex; i < toIndex; i++)
    {
      result.add(get(i));
    }
    return new OrderedSet<E>(result);
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected List<E> createInnerList(int initialCapacity)
  {
    return new ArrayList<E>(initialCapacity);
  }

  protected List<E> getElements()
  {
    return elements;
  }

  protected void setElements(List<E> newValue)
  {
    elements = newValue;
  }
}
