// ===========================================================================
// CONTENT  : CLASS PointInTime
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 30/01/2021
// HISTORY  :
//  30/01/2021  mdu  CREATED
//
// Copyright (c) 2021, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.util;

import java.util.Date;

/**
 * Represents a point in time in milliseconds since midnight, January 1, 1970 UTC.
 */
public class PointInTime
{
  // timestamp in milliseconds
  private final long epochTimestamp;

  public static PointInTime create(long epochTimestampInMilliseconds)
  {
    return new PointInTime(epochTimestampInMilliseconds);
  }

  public static PointInTime create(Date date)
  {
    return create(date.getTime());
  }

  public PointInTime(long epochTimestampInMilliseconds)
  {
    super();
    this.epochTimestamp = epochTimestampInMilliseconds;
  }

  /**
   * Returns the point in time in milliseconds since midnight, January 1, 1970 UTC.
   */
  public long getEpochTimestamp()
  {
    return this.epochTimestamp;
  }

  /**
   * Returns the remaining time in milliseconds until this point in time is reached.
   * A value less or equal zero indicates that the point in time has been reached.  
   */
  public long getRemainingTime()
  {
    return getEpochTimestamp() - System.currentTimeMillis();
  }

  /**
   * Returns the remaining time until the point in time has been reached as duration.
   * Be aware that the lowest value will be a duration of 0.
   */
  public Duration getDurationUntilReached()
  {
    long remaining = getRemainingTime();
    return (remaining < 0) ? Duration.valueOf(0) : Duration.valueOf(remaining);
  }

  /**
   * Returns whether or not this point in time has been reached.
   */
  public boolean isReached()
  {
    return getRemainingTime() <= 0;
  }

  /**
   * Returns this point in time as date.
   */
  public Date asDate()
  {
    return new Date(getEpochTimestamp());
  }

  @Override
  public String toString()
  {
    return String.format("%s(%d, reached:%s)", getClass().getSimpleName(), getEpochTimestamp(), isReached());
  }
}
