// ===========================================================================
// CONTENT  : CLASS PrioritySorter
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 26/01/2020
// HISTORY  :
//  26/01/2020  mdu  CREATED
//
// Copyright (c) 2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

/**
 * This class is supposed to support sorting of objects (of type T) by an external priority.
 * That is, the objects themselves do not contain the sort criteria.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class PrioritySorter<T>
{
  private final List<PrioritizedObject<T>> prioritizedObjects = new ArrayList<PrioritySorter.PrioritizedObject<T>>();
  private int defaultPriority = 100;

  /**
   * Creates a new instance of this class.
   */
  public static <E> PrioritySorter<E> create()
  {
    return new PrioritySorter<E>();
  }
  
  /**
   * Creates a new instance of this class.
   */
  public PrioritySorter()
  {
    super();
  }

  /**
   * Sorts the given list of objects according to their associated priorities.
   * in an ascending order.
   * That is, each object that has previously been registered with a priority 
   * by calling the {@link #add(Object, int)} method will be detected by its
   * <code>equals()</code> method and then compared for sorting by its associated 
   * priority.
   * <br>
   * All those objects of the list that have not been registered or are not detected
   * will be handled with the {@link #getDefaultPriority()} value.
   * <p>
   * <strong>The order of the input list will not be changed!</strong> 
   * 
   * @param collection The collection of objects to be sorted (must not be null).
   * @return The sorted list.
   */
  public List<T> sortAscending(Collection<T> collection)
  {
    List<T> sortedList;

    sortedList = new ArrayList<T>(collection);
    Collections.sort(sortedList, createComparator());
    return sortedList;
  }

  /**
   * Sorts the given list of objects according to their associated priorities
   * in a descending order.
   * That is, each object that has previously been registered with a priority 
   * by calling the {@link #add(Object, int)} method will be detected by its
   * <code>equals()</code> method and then compared for sorting by its associated 
   * priority.
   * <br>
   * All those objects of the list that have not been registered or are not detected
   * will be handled with the {@link #getDefaultPriority()} value. 
   * <p>
   * <strong>The order of the input list will not be changed!</strong> 
   * 
   * @param collection The collection of objects to be sorted (must not be null).
   * @return The sorted list.
   */
  public List<T> sortDescending(Collection<T> collection)
  {
    List<T> sortedList;

    sortedList = sortAscending(collection);
    Collections.reverse(sortedList);
    return sortedList;
  }

  /**
   * Adds an object with its associated priority.
   * The default priority for objects that have not been added is provided
   * bay {@link #getDefaultPriority()} and is initialized to 100.
   * 
   * @param object The object to register along with the given priority (must not be null).
   * @param priority The priority value to associate to the given object.
   * @return This instance, to support fluent API.
   */
  public PrioritySorter<T> add(T object, int priority)
  {
    getPrioritizedObjects().add(new PrioritizedObject<T>(object, priority));
    return this;
  }

  /**
   * Returns the default priority for objects not registered with this sorter.
   * If not explicitly changed the value is 100.
   * 
   * @return The current default priority.
   */
  public int getDefaultPriority()
  {
    return defaultPriority;
  }

  /**
   * Sets the default priority for objects not registered with this sorter
   * to the given value.
   * 
   * @param defaultPriority The new default priority.
   * @return This instance, to support fluent API.
   */
  public PrioritySorter<T> setDefaultPriority(int defaultPriority)
  {
    this.defaultPriority = defaultPriority;
    return this;
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected List<PrioritizedObject<T>> getPrioritizedObjects()
  {
    return prioritizedObjects;
  }

  protected PrioritizedObject<T> getPrioritizedObject(T object)
  {
    for (PrioritizedObject<T> prioritizedObject : getPrioritizedObjects())
    {
      if (prioritizedObject.getObject().equals(object))
      {
        return prioritizedObject;
      }
    }
    return null;
  }

  protected Comparator<? super T> createComparator()
  {
    return new PrioritizingComparator();
  }

  // =========================================================================
  // INNER CLASSES
  // =========================================================================
  protected static class PrioritizedObject<T>
  {
    private final T object;
    private final int priority;

    protected PrioritizedObject(T object, int priority)
    {
      this.object = object;
      this.priority = priority;
    }

    public T getObject()
    {
      return object;
    }

    public int getPriority()
    {
      return priority;
    }
  }

  protected class PrioritizingComparator implements Comparator<T>
  {
    @Override
    public int compare(T object1, T object2)
    {
      Integer prio1;
      Integer prio2;

      prio1 = getPriority(getPrioritizedObject(object1));
      prio2 = getPriority(getPrioritizedObject(object2));

      return prio1.compareTo(prio2);
    }

    private Integer getPriority(PrioritizedObject<T> prioritizedObject)
    {
      return Integer.valueOf((prioritizedObject == null) ? getDefaultPriority() : prioritizedObject.getPriority());
    }
  }
}
