// ===========================================================================
// CONTENT  : CLASS ProcessStdOutRedirector
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 04/01/2019
// HISTORY  :
//  04/01/2019  mdu  CREATED
//
// Copyright (c) 2019, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.util;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;

import org.pfsw.bif.text.ILineProcessor;

/**
 * This class can be used to redirect the standard output of an external process
 * execution to an internal line processor.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class ProcessStdOutRedirector implements Runnable
{
  private final BufferedReader stdOutReader;
  private final ILineProcessor targetLineProcessor;
  
  public ProcessStdOutRedirector(Process process, ILineProcessor lineProcessor)
  {
    super();
    this.stdOutReader = new BufferedReader(new InputStreamReader(process.getInputStream()));
    this.targetLineProcessor = lineProcessor;
  }
  
  @SuppressWarnings("resource")
  public ProcessStdOutRedirector(Process process, PrintWriter targetWriter)
  {
    this(process, new PrintWriterLineProcessor(targetWriter));
  }
  
  public ProcessStdOutRedirector(Process process)
  {
    this(process, new PrintWriter(new OutputStreamWriter(System.out)));
  }
  
  /**
   * Reads all available lines from the external stdout stream and forwards them
   * to the underlying {@link ILineProcessor}.
   */
  @Override
  public void run()
  {
    String line;
    int lineCounter = 0;

    try
    {
      line = getStdOutReader().readLine();
      while (line != null)
      {
        lineCounter++;
        getTargetLineProcessor().processLine(line, lineCounter);
        line = getStdOutReader().readLine();
      }
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  
  protected BufferedReader getStdOutReader()
  {
    return this.stdOutReader;
  }
  
  protected ILineProcessor getTargetLineProcessor()
  {
    return this.targetLineProcessor;
  }
}
