// ===========================================================================
// CONTENT  : CLASS PropertiesMap
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.1 - 04/06/2015
// HISTORY  :
//  19/01/2014  mdu  CREATED
//  04/06/2015  mdu   added   --> setProperty(), getProperty(), stringPropertyNames()
//
// Copyright (c) 2014-2015, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import org.pfsw.bif.configuration.IConfigSettings;

/**
 * A Map wrapper around a properties object that allows to treat the properties
 * like a Map<String, String> which is not possible with the original Properties
 * because it is based on Hashtable<Object, Object>.
 *
 * @author Manfred Duchrow
 * @version 1.1
 */
public class PropertiesMap implements Map<String, String>, IConfigSettings
{
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private final Properties internalProperties;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  public PropertiesMap()
  {
    this(null);
  }

  public PropertiesMap(Properties properties)
  {
    super();
    if (properties == null)
    {
      internalProperties = new Properties();
    }
    else
    {
      internalProperties = properties;
    }
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the data as Properties object (a copy).
   * Modifying the returned properties does not have any impact on this object.
   */
  public Properties getProperties()
  {
    return (Properties)getInternalProperties().clone();
  }

  @Override
  public int size()
  {
    return getInternalProperties().size();
  }

  @Override
  public boolean isEmpty()
  {
    return getInternalProperties().isEmpty();
  }

  @Override
  public boolean containsKey(Object key)
  {
    if (key instanceof String)
    {
      return getInternalProperties().containsKey(key);
    }
    return false;
  }

  @Override
  public boolean containsValue(Object value)
  {
    if (value instanceof String)
    {
      return getInternalProperties().containsValue(value);
    }
    return false;
  }

  @Override
  public String get(Object key)
  {
    if (key instanceof String)
    {
      return getInternalProperties().getProperty((String)key);
    }
    return null;
  }

  @Override
  public String put(String key, String value)
  {
    String oldValue;

    oldValue = getInternalProperties().getProperty(key);
    getInternalProperties().setProperty(key, value);
    return oldValue;
  }

  @Override
  public String remove(Object key)
  {
    if (key instanceof String)
    {
      return (String)getInternalProperties().remove(key);
    }
    return null;
  }

  @Override
  public void putAll(Map<? extends String, ? extends String> map)
  {
    getInternalProperties().putAll(map);
  }

  @Override
  public void clear()
  {
    getInternalProperties().clear();
  }

  @Override
  public Set<String> keySet()
  {
    return getInternalProperties().stringPropertyNames();
  }

  @Override
  public Collection<String> values()
  {
    Collection<String> result;

    result = new ArrayList<String>(getInternalProperties().size());
    for (Object object : getInternalProperties().values())
    {
      if (object instanceof String)
      {
        result.add((String)object);
      }
    }
    return result;
  }

  @Override
  public Set<Map.Entry<String, String>> entrySet()
  {
    Set<Map.Entry<String, String>> result;

    result = new HashSet<Map.Entry<String, String>>(getInternalProperties().size());
    for (Map.Entry<Object, Object> entry : getInternalProperties().entrySet())
    {
      if ((entry.getKey() instanceof String) && (entry.getValue() instanceof String))
      {
        result.add(new MapEntry<String, String>((String)entry.getKey(), (String)entry.getValue()));
      }
    }
    return result;
  }

  /**
   * Searches for the property with the specified key in this property list. 
   * If the key is not found in this property list, the default property list, 
   * and its defaults, recursively, are then checked. The method returns null 
   * if the property is not found.
   */
  public String getProperty(String key)
  {
    return get(key);
  }

  /**
   * Searches for the property with the specified key in this property list.
   * If the key is not found in this property list, the default property list,
   * and its defaults, recursively, are then checked. The method returns the
   * default value argument if the property is not found.
   */
  public String getProperty(String key, String defaultValue)
  {
    String value;

    value = getProperty(key);
    if (value == null)
    {
      return defaultValue;
    }
    return value;
  }

  /**
   * Sets the given value as ne property value for the given key.
   */
  public void setProperty(String key, String value)
  {
    put(key, value);
  }

  /**
   * Returns a set of keys in this property list where the key and its corresponding 
   * value are strings, including distinct keys in the default property list if a key 
   * of the same name has not already been found from the main properties list. 
   * Properties whose key or value is not of type String are omitted. 
   * The returned set is not backed by the Properties object. Changes to this Properties 
   * are not reflected in the set, or vice versa.
   */
  public Set<String> stringPropertyNames()
  {
    return keySet();
  }

  @Override
  public Collection<String> getNames()
  {
    return keySet();
  }

  @Override
  public String getValue(String name)
  {
    return get(name);
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected Properties getInternalProperties()
  {
    return internalProperties;
  }
}
