// ===========================================================================
// CONTENT  : CLASS RandomRange
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 24/04/2020
// HISTORY  :
//  24/04/2020  mdu  CREATED
//
// Copyright (c) 2020, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.util;

import java.security.SecureRandom;
import java.util.Random;

/**
 * A generator of random integer values in a specified range.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class RandomRange
{
  /**
   * Creates a new instance with a {@link SecureRandom} for the generation
   * of values between minValue and maxValue (both inclusive).
   * 
   * @param minValue The minimum integer value that could be produced by this instance. 
   * @param maxValue The maximum integer value that could be produced by this instance.
   */
  public static RandomRange create(int minValue, int maxValue)
  {
    return new RandomRange(minValue, maxValue);
  }

  /**
   * Creates a new instance with the given {@link Random} implementation for the generation
   * of values between minValue and maxValue (both inclusive).
   * 
   * @param minValue The minimum integer value that could be produced by this instance. 
   * @param maxValue The maximum integer value that could be produced by this instance.
   */
  public static RandomRange create(Random random, int minValue, int maxValue)
  {
    return new RandomRange(random, minValue, maxValue);
  }

  private final Random random;
  private final int minValue;
  private final int upperBound;

  /**
   * Creates a new instance with a {@link SecureRandom} for the generation
   * of values between minValue and maxValue (both inclusive).
   * 
   * @param minValue The minimum integer value that could be produced by this instance. 
   * @param maxValue The maximum integer value that could be produced by this instance.
   */
  public RandomRange(int minValue, int maxValue)
  {
    this(new SecureRandom(), minValue, maxValue);
  }

  /**
   * Creates a new instance with the given {@link Random} implementation for the generation
   * of values between minValue and maxValue (both inclusive).
   * 
   * @param minValue The minimum integer value that could be produced by this instance. 
   * @param maxValue The maximum integer value that could be produced by this instance.
   */
  public RandomRange(Random random, int minValue, int maxValue)
  {
    super();
    this.random = random;
    this.minValue = minValue;
    this.upperBound = maxValue - minValue + 1;
  }

  /**
   * Returns a new random integer value in the specified underlying range.
   */
  public int nextInt()
  {
    return getMinValue() + getRandom().nextInt(getUpperBound());
  }

  /**
   * Returns the minimum value that can be generated by this object.
   */
  public int getMinValue()
  {
    return this.minValue;
  }
  
  /**
   * Returns the maximum value that can be generated by this object.
   */
  public int getMaxValue()
  {
    return getMinValue() + getUpperBound() - 1;
  }
  
  @Override
  public String toString()
  {
    return String.format("%s(%d..%d)", getClass().getSimpleName(), getMinValue(), getMaxValue());
  }

  protected Random getRandom()
  {
    return this.random;
  }

  protected int getUpperBound()
  {
    return this.upperBound;
  }
}
