// ===========================================================================
// CONTENT  : CLASS SimpleDate
// AUTHOR   : M.Duchrow
// VERSION  : 1.0 - 19/12/2008
// HISTORY  :
//  19/12/2008  mdu  CREATED
//
// Copyright (c) 2008, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.util;

import java.io.Serializable;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

/**
 * This class represents a simple date. That means, it is not a timestamp like
 * java.util.Date and it also starts with month 1 for January rather than with
 * zero.
 * <p>
 * For convenience it also provides some commonly string representations.
 *
 * @author M.Duchrow
 * @version 1.0
 */
public class SimpleDate implements Serializable, Comparable<SimpleDate>, Cloneable
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final long serialVersionUID = 90001L;

  protected static final SimpleDateFormat SORT_FORMAT = new SimpleDateFormat("yyyyMMdd");
  protected static final SimpleDateFormat ISO8601_FORMAT = new SimpleDateFormat("yyyy-MM-dd");
  protected static final SimpleDateFormat GERMAN_FORMAT = new SimpleDateFormat("dd.MM.yyyy");
  protected static final SimpleDateFormat BRITISH_FORMAT = new SimpleDateFormat("dd/MM/yyyy");
  protected static final SimpleDateFormat US_FORMAT = new SimpleDateFormat("MM/dd/yyyy");

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private Calendar calendar = new GregorianCalendar();

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Creates a new SimpleDate from the given dateString which must conform to 
   * the format "yyyyMMdd". 
   */
  public static SimpleDate parseSortDate(String dateString) throws ParseException
  {
    Date date;

    date = SORT_FORMAT.parse(dateString);
    return new SimpleDate(date);
  }

  /**
   * Creates a new SimpleDate from the given dateString which must conform to 
   * the format "dd.MM.yyyy". 
   */
  public static SimpleDate parseGermanDate(String dateString) throws ParseException
  {
    Date date;

    date = GERMAN_FORMAT.parse(dateString);
    return new SimpleDate(date);
  }

  /**
   * Creates a new SimpleDate from the given dateString which must conform to 
   * the format "dd/MM/yyyy". 
   */
  public static SimpleDate parseBritishDate(String dateString) throws ParseException
  {
    Date date;

    date = BRITISH_FORMAT.parse(dateString);
    return new SimpleDate(date);
  }

  /**
   * Creates a new SimpleDate from the given dateString which must conform to 
   * the format "MM/dd/yyyy". 
   */
  public static SimpleDate parseUSDate(String dateString) throws ParseException
  {
    Date date;

    date = US_FORMAT.parse(dateString);
    return new SimpleDate(date);
  }

  /**
   * Creates a new SimpleDate from the given dateString which must conform to 
   * the format "yyyy-MM-dd". 
   */
  public static SimpleDate parseISO8601Date(String dateString) throws ParseException
  {
    Date date;

    date = ISO8601_FORMAT.parse(dateString);
    return new SimpleDate(date);
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values (i.e. today's date).
   */
  public SimpleDate()
  {
    this(new Date());
  }

  /**
   * Initialize the new instance with a standard calendar.
   */
  public SimpleDate(GregorianCalendar gregorianCalendar)
  {
    super();
    setCalendar(gregorianCalendar);
  }

  /**
   * Initialize the new instance with a standard date object.
   */
  public SimpleDate(Date date)
  {
    super();
    setInternalDate(date);
  }

  /**
   * Returns a Date object representing the given parameters.
   * 
   * @param year The year of the date to create 
   * @param month The month of the date to create (1-based, 1=January)
   * @param day The day of the date to create 
   */
  public SimpleDate(int year, int month, int day)
  {
    this(DateUtil.current().newDate(year, month, day));
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the year as integer (1900..3000).
   */
  public int getYear()
  {
    return getCalendar().get(Calendar.YEAR);
  }

  /**
   * Returns the month as integer (1 = January, ..., 12 = December).
   */
  public int getMonth()
  {
    return getCalendar().get(Calendar.MONTH) + 1;
  }

  /**
   * Returns the day in month as integer (1..31).
   */
  public int getDay()
  {
    return getCalendar().get(Calendar.DAY_OF_MONTH);
  }

  /**
   * Returns the date in the format "yyyyMMdd" ("20021123").
   */
  public String asSortString()
  {
    return SORT_FORMAT.format(getInternalDate());
  }

  /**
   * Returns the date in the format "yyyy-MM-dd" ("2002-11-23").
   */
  public String asISO8601String()
  {
    return ISO8601_FORMAT.format(getInternalDate());
  }

  /**
   * Returns the date in the format "dd.MM.yyyy" ("23.11.2002").
   */
  public String asGermanString()
  {
    return GERMAN_FORMAT.format(getInternalDate());
  }

  /**
   * Returns the date in the format "dd/MM/yyyy" ("23/11/2002").
   */
  public String asBritishString()
  {
    return BRITISH_FORMAT.format(getInternalDate());
  }

  /**
   * Returns the date in the format "MM/dd/yyyy" ("11/23/2002").
   */
  public String asUSString()
  {
    return US_FORMAT.format(getInternalDate());
  }

  /**
   * Returns this date as Java Date object.
   */
  public Date asDate()
  {
    return getInternalDate();
  }

  /**
   * Returns the date in the format "yyyy-MM-dd" ("2002-11-23").
   */
  @Override
  public String toString()
  {
    return asISO8601String();
  }

  @Override
  public boolean equals(Object object)
  {
    if (object instanceof SimpleDate)
    {
      SimpleDate simpleDate = (SimpleDate)object;
      return compareTo(simpleDate) == 0;
    }
    return false;
  }

  @Override
  public int hashCode()
  {
    return asSortString().hashCode();
  }

  @Override
  public int compareTo(SimpleDate simpleDate)
  {
    return asSortString().compareTo(simpleDate.asSortString());
  }

  @Override
  public Object clone() throws CloneNotSupportedException
  {
    SimpleDate clone;

    clone = new SimpleDate();
    clone.setCalendar((Calendar)getCalendar().clone());
    return clone;
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected Date getInternalDate()
  {
    return getCalendar().getTime();
  }

  protected void setInternalDate(Date date)
  {
    getCalendar().setTime(date);
  }

  protected Calendar getCalendar()
  {
    return this.calendar;
  }

  protected void setCalendar(Calendar newValue)
  {
    this.calendar = newValue;
  }
}
