// ===========================================================================
// CONTENT  : CLASS StackedMap
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 19/03/2004
// HISTORY  :
//  19/03/2004  mdu  CREATED
//
// Copyright (c) 2004, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.util;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 * Provides a map that allows to push() and pop() internal settings to/from
 * a (LIFO) stack.
 * Accessing an element via get(key) always looks in the last map and then
 * continues looking for it (if not found) in the next map on the stack.  
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
@SuppressWarnings({ "rawtypes", "unchecked" })
public class StackedMap extends MapWrapper
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final long serialVersionUID = -43795860372737126L;

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private StackedMap parentMap = null;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  public StackedMap()
  {
    super();
  }

  /**
   * Initialize the new instance with an internal map.
   */
  public StackedMap(Map map)
  {
    super(map);
  }

  /**
   * Initialize the new instance with a parent and a value map.
   */
  protected StackedMap(StackedMap parent, Map valueMap)
  {
    super();
    parentMap(parent);
    internalMap(valueMap);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Removes all entries from this map, but not from the maps on the stack!
   * 
   * @see java.util.Map#clear()
   */
  @Override
  public void clear()
  {
    super.clear();
  }

  /**
   * This method returns true, if the given key can be found in this map or
   * any map on the stack.
   * 
   * @see java.util.Map#containsKey(Object)
   */
  @Override
  public boolean containsKey(Object key)
  {
    boolean found;

    found = super.containsKey(key);
    if ((!found) && (hasParent()))
    {
      found = parentMap().containsKey(key);
    }
    return found;
  }

  /**
   * This method returns true, if the given value is a value of any entry
   * in this map or any map on the stack.
   * 
   * @see java.util.Map#containsValue(Object)
   */
  @Override
  public boolean containsValue(Object value)
  {
    boolean found;

    found = super.containsValue(value);
    if ((!found) && (hasParent()))
    {
      found = parentMap().containsValue(value);
    }
    return found;
  }

  /**
   * Returns all entries of the map and all stacked maps.
   * 
   * @see java.util.Map#entrySet()
   */
  @Override
  public Set entrySet()
  {
    Set set;

    set = super.entrySet();
    if (hasParent())
    {
      set.addAll(parentMap().entrySet());
    }
    return set;
  }

  /**
   * Returns the object that was registered under the specified key in this
   * map or in any of the maps on the stack.
   * 
   * @see java.util.Map#get(Object)
   */
  @Override
  public Object get(Object key)
  {
    Object value;

    value = super.get(key);
    if ((value == null) && (hasParent()))
    {
      value = parentMap().get(key);
    }
    return value;
  }

  /**
   * Returns true, if there is no entry in this map and in all maps on the stack.
   * 
   * @see java.util.Map#isEmpty()
   */
  @Override
  public boolean isEmpty()
  {
    boolean empty;

    empty = super.isEmpty();
    if (empty && hasParent())
    {
      empty = parentMap().isEmpty();
    }
    return empty;
  }

  /**
   * Returns all keys in this map and the maps on the stack
   * 
   * @see java.util.Map#keySet()
   */
  @Override
  public Set keySet()
  {
    Set keys;

    keys = super.keySet();
    if (hasParent())
    {
      keys.addAll(parentMap().keySet());
    }
    return keys;
  }

  /**
   * Returns how many entries currently are in this map and the maps on the 
   * stack.
   * 
   * @see java.util.Map#size()
   */
  @Override
  public int size()
  {
    int size;

    size = super.size();
    if (hasParent())
    {
      size += parentMap().size();
    }
    return size;
  }

  /**
   * Returns all values of this map and the maps on the stack.
   * 
   * @see java.util.Map#values()
   */
  @Override
  public Collection values()
  {
    Collection values;

    values = super.values();
    if (hasParent())
    {
      values.addAll(parentMap().values());
    }
    return values;
  }

  /**
   * Pushes the given map onto the top of this stack.
   */
  public void push(Map map)
  {
    StackedMap newParent;

    newParent = new StackedMap(parentMap(), internalMap());
    parentMap(newParent);
    internalMap((map == null) ? createEmptyMap() : map);
  }

  /**
   * Pushes the an newly created map onto the top of this stack.
   */
  public void push()
  {
    push(null);
  }

  /**
   * Removes the map at the top of this stack and returns that map 
   * as the value of this function.
   */
  public Map pop()
  {
    Map valueMap;

    valueMap = internalMap();
    if (hasParent())
    {
      internalMap(parentMap().internalMap());
      parentMap(parentMap().parentMap());
    }
    return valueMap;
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================

  /**
   * Subclasses may override this method to provide an instance of a different 
   * implementation class than HashMap.
   */
  @Override
  protected Map createEmptyMap()
  {
    return new HashMap();
  }

  protected boolean hasParent()
  {
    return (parentMap() != null);
  }

  protected StackedMap parentMap()
  {
    return this.parentMap;
  }

  protected void parentMap(StackedMap newValue)
  {
    this.parentMap = newValue;
  }
}
