// ===========================================================================
// CONTENT  : CLASS StopWatch
// AUTHOR   : M.Duchrow
// VERSION  : 1.2 - 03/05/2020
// HISTORY  :
//  10/02/2007  mdu  CREATED
//  10/04/2016  mdu   added   --> createAndStart()
//  03/05/2020  mdu   added   --> duration()
//
// Copyright (c) 2007-2020, by M.Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.util;

/**
 * A simple stop watch to measure the time between a start and a stop point.
 *
 * @author M.Duchrow
 * @version 1.2
 */
public class StopWatch
{
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private long startTime = 0;
  private long stopTime = 0;

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Returns a newly created and immediately started StopWatch.
   */
  public static StopWatch create()
  {
    return new StopWatch();
  }

  /**
   * Returns a newly created and immediately started StopWatch.
   */
  public static StopWatch createAndStart()
  {
    StopWatch stopWatch;
    
    stopWatch = create();
    stopWatch.start();
    return stopWatch;
  }
  
  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  public StopWatch()
  {
    super();
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Starts the stop watch. That is, remember current time as start time.
   * @return The start time
   */
  public long start()
  {
    stopTime = 0;
    startTime = System.currentTimeMillis();
    return startTime;
  }

  /**
   * Stops the stop watch. That is, remember current time as stop time.
   * @return The stop time
   */
  public long stop()
  {
    stopTime = System.currentTimeMillis();
    return stopTime;
  }

  /**
   * Returns the duration between start and stop in milliseconds.
   * If {@link #stop()} has not been called yet it returns the duration from
   * start to now.
   */
  public long getDuration()
  {
    if (startTime <= 0)
    {
      return 0;
    }
    if (stopTime <= 0)
    {
      return System.currentTimeMillis() - startTime;
    }
    return stopTime - startTime;
  }
  
  /**
   * Returns the duration between start and stop in milliseconds.
   * If {@link #stop()} has not been called yet it returns the duration from
   * start to now.
   */
  public Duration duration() 
  {
    return Duration.valueOf(getDuration());
  }
  
  public long getStartTime()
  {
    return this.startTime;
  }
  
  public long getStopTime()
  {
    return this.stopTime;
  }

  /**
   * Resets the watch to a state before it gets started
   */
  public void reset()
  {
    startTime = 0;
    stopTime = 0;
  }
  
  @Override
  public String toString()
  {
    return String.format("%s(start=%d, stop=%d, duration=%d)", getClass().getSimpleName(), getStartTime(), getStopTime(), getDuration());
  }
}
