// ===========================================================================
// CONTENT  : CLASS SysUtil
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.9 - 24/12/2019
// HISTORY  :
//  25/12/2004  mdu  CREATED
//	03/06/2006	mdu		added		-->	get OS family name
//	28/07/2006	mdu		changed	-->	determine if isEclipse only once
//	02/03/2008	mdu		added		-->	support for system exit listener
//	17/11/2010	mdu		added		-->	isAppletEnvironment()
//	23/06/2011	mdu		changed	--> Added other names for Unix OS detection
//  27/09/2014  mdu   added   --> close()
//  20/03/2016  mdu   changed --> close() to call IOUtil
//  26/12/2018  mdu   added   --> getJavaRuntimeVersion()
//  04/01/2019  mdu   added   --> getJdkBinaryDir()
//  24/12/2019  mdu   added   --> getConfigValue()
//
// Copyright (c) 2004-2019, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.util;

import java.io.Closeable;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.Executors;

import org.pfsw.bif.lifecycle.ISystemExitListener;
import org.pfsw.bif.text.ILineProcessor;

/**
 * Provides generally useful methods for system inquiries and other system related
 * functionality.
 * <p>
 * For example it is strongly recommended to uset {@link SysUtil#exit(int)} rather
 * than {@link System#exit(int)} to terminate a JVM. It allows registered
 * listeners to be notified just before the exit happens.
 *
 * @author Manfred Duchrow
 * @version 1.9
 */
public class SysUtil
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  /**
   * An instance of this class which can be easily included
   * as static import and then used like <code>SYSU.getOsFamily()</code>.
   */
  public static final SysUtil SYSU = new SysUtil();

  public static final boolean DEBUG = "true".equals(System.getProperty("org.pfsw.util.debug"));
  
  /** "unknown" */
  public static final String OS_FAMILY_UNKNOWN = "unknown";
  /** "windows" */
  public static final String OS_FAMILY_WINDOWS = "windows";
  /** "unix" */
  public static final String OS_FAMILY_UNIX = "unix";
  /** "macintosh" */
  public static final String OS_FAMILY_MACINTOSH = "macintosh";
  
  /**
   * Contains the value of system property "java.home" as set at startup.
   */
  public static final String JAVA_HOME_DIR = System.getProperty("java.home");

  protected final static String OS_PROPERTY_NAME = "os.name";
  protected final static String PF_OS_PROPERTY_NAME = "org.pf.util.os.name";

  protected static final String SOLARIS = "SOLARIS";
  protected static final String SUN_OS = "SUNOS";
  protected static final String LINUX = "LINUX";
  protected static final String AIX = "AIX";
  protected static final String HP_UX = "HP-UX";
  protected static final String FREE_BSD = "FREEBSD";
  protected static final String DIGITAL_UNIX = "DIGITAL";
  protected static final String OSF1 = "OSF1";
  protected static final String IRIX = "IRIX";

  protected static final String[] UNIX_SYSTEMS = { SOLARIS, SUN_OS, AIX, HP_UX, LINUX, FREE_BSD, DIGITAL_UNIX, OSF1, IRIX };

  private static final String OS_FAMILY = determineOsFamilyName();
  private static final boolean IS_WINDOWS = checkIfWindows();

  private static final String SYS_PROP_SPEC_VERSION = "java.specification.version";

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private Boolean isEclipse = null;

  private List<ISystemExitListener> exitListeners = new ArrayList<ISystemExitListener>();

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Returns the only instance this class supports (design pattern "Singleton")
   */
  public static SysUtil current()
  {
    return SysUtil.SYSU;
  }

  protected static boolean checkIfWindows()
  {
    return OS_FAMILY_WINDOWS.equals(determineOsFamilyName());
  }

  protected static String determineOsFamilyName()
  {
    String osName;

    osName = System.getProperty(PF_OS_PROPERTY_NAME);
    if (osName == null)
    {
      osName = System.getProperty(OS_PROPERTY_NAME, "");
    }
    return determineOsFamilyName(osName);
  }

  protected static String determineOsFamilyName(String operationSystemName)
  {
    String osName;

    osName = operationSystemName.toUpperCase();
    if (osName.startsWith("WINDOWS"))
    {
      return OS_FAMILY_WINDOWS;
    }
    if (osName.startsWith("MAC"))
    {
      return OS_FAMILY_MACINTOSH;
    }
    if (CollectionUtil.current().indexOf(UNIX_SYSTEMS, osName) >= 0)
    {
      return OS_FAMILY_UNIX;
    }
    return OS_FAMILY_UNKNOWN;
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  protected SysUtil()
  {
    super();
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns true if the current operating-system is Windows. 
   */
  public boolean isWindows()
  {
    return IS_WINDOWS;
  }

  /**
   * Returns true if the environment is based on eclipse.
   * That is, this class is loaded within an eclipse plug-in.
   * Can also be defined by setting system property "org.pf.util.isEclipse"
   * to true or false.
   */
  public boolean isEclipse()
  {
    if (isEclipse == null)
    {
      isEclipse = Bool.toBoolean(determineIfEclipse());
    }
    return isEclipse.booleanValue();
  }

  /**
   * Returns true if the code is running as part of an applet in a browser
   * plugin JVM. 
   */
  public boolean isAppletEnvironment()
  {
    return Bool.isTrue(System.getProperty("java.version.applet"));
  }

  /**
   * Returns true if the current Java execution runtime is a JDK
   * and false if it is a JRE.
   */
  public boolean isExecutingWithJDK() 
  {
    return getJdkBinaryDir() != null;
  }
  
  /**
   * Returns the path of the currently executing JDK's binary sub directory
   * or null if it is not found (e.g. in case the runtime is a JRE).
   */
  public String getJdkBinaryDir() 
  {
    String dir;
    File folder;
    String javac;
    File compiler;
    
    dir = JAVA_HOME_DIR.replace("\\", "/");
    if (getJavaRuntimeVersion() < 9)
    {
      if (!dir.endsWith("/jre"))
      {
        return null; // No JDK
      }
      dir = dir.replace("/jre", "/bin");
    }
    else
    {
      dir = dir + "/bin";
    }
    folder = new File(dir);
    if (!folder.exists())
    {
      return null; // No JDK      
    }
    javac = isWindows() ? "javac.exe" : "javac";
    compiler = new File(folder, javac);
    if (!compiler.exists())
    {
      return null; // No JDK            
    }
    return dir;
  }
  
  /**
   * Returns the family name of the current operating system according to the 
   * name specified in system property "os.name".
   * 
   * @return "windows or "unix" or "macintosh" or "unknown" for all others
   * @see #OS_FAMILY_UNKNOWN
   * @see #OS_FAMILY_UNIX
   * @see #OS_FAMILY_MACINTOSH
   * @see #OS_FAMILY_WINDOWS
   */
  public String getOsFamily()
  {
    return OS_FAMILY;
  }
  
  /**
   * Returns the (normalized) version of the current Java Runtime.
   * That is for 4 for Java "1.4" or 8 for Java "1.8" and 10 for Java "10".  
   */
  public int getJavaRuntimeVersion() 
  {
    String versionString = System.getProperty(SYS_PROP_SPEC_VERSION);
    if (versionString.startsWith("1."))
    {
      versionString = versionString.substring(2);
    }
    return Integer.parseInt(versionString);
  }

  /**
   * Returns true if the current JRE used to execute this code is
   * supporting the Java Module System. That is, every JRE from version 9 and above.
   */
  public boolean isJavaModuleSystemSupportingRuntime() 
  {
    return getJavaRuntimeVersion() >= 9;
  }
  
  /**
   * Sets the current thread to sleep for the specified time in milliseconds
   * or until it gets interrupted.
   * Avoids throwing an InterruptedException.
   * 
   * @param milliseconds The time to sleep
   */
  public void sleep(long milliseconds)
  {
    try
    {
      Thread.currentThread().sleep(milliseconds);
    }
    catch (@SuppressWarnings("unused") InterruptedException e)
    {
      // Don't care, just wake up.
    }
  }

  /**
   * Calling this method terminates the JVM.
   * <br>
   * It should be used instead of System.exit(int) because it will notify
   * all registered ISystemExitListener before it actually calls System.exit(rc).
   * 
   *  @param rc The return code to the program that started the JVM. 
   */
  public void exit(int rc)
  {
    notifyExitListeners(rc);
    System.exit(rc);
  }

  /**
   * Calling this method terminates the JVM after the specified sleep time.
   * <br>
   * It should be used instead of System.exit(int) because it will notify
   * all registered ISystemExitListener before it actually calls System.exit(rc).
   * <br>
   * This method waits the specified sleep time before it actually exits. 
   * 
   * @param rc The return code to the program that started the JVM. 
   * @param sleepBeforeExit The time to sleep (in milliseconds) before exiting.
   */
  public void exit(int rc, long sleepBeforeExit)
  {
    sleep(sleepBeforeExit);
    exit(rc);
  }

  /**
   * Add the given listener to an internal list so that it would be called right
   * before system exit gets executed via the {@link #exit(int)} method.
   * <br>
   * If the listener is already in the internal listener list it will not be
   * added again.
   * 
   * @param listener The listener to add. Will be ignored if null.
   */
  public void addSystemExitListener(ISystemExitListener listener)
  {
    if ((listener != null) && (!getExitListeners().contains(listener)))
    {
      getExitListeners().add(listener);
    }
  }

  /**
   * Removes the given listener from the internal list so that it would not 
   * be notified anymore before system exit gets executed via the {@link #exit(int)} 
   * method.
   * <br>
   * If the listener is not in the internal listener list, nothing happens.
   * 
   * @param listener The listener to remove. Will be ignored if null.
   */
  public void removeSystemExitListener(ISystemExitListener listener)
  {
    if (listener != null)
    {
      getExitListeners().remove(listener);
    }
  }

  /**
   * Close the given closeable ignoring any exception and checking for null before closing.
   * Returns true, if the object was closed successfully, false otherwise.
   */
  public boolean close(Closeable closeable)
  {
    return IOUtil.current().close(closeable);
  }

  /**
   * Executes the specified external command with the given arguments in the specified 
   * execution directory and returns the combined result data in an object. 
   * 
   * @param execDir The directory in which the command execution should take place (may be null).
   * @param command The name of the command or program to execute (must not be null).
   * @param args An optional list of arguments to be passed to the external command. 
   * @return An result object containing the exit code of the external command and the lines written to stdout.
   * @throws IOException In any case of problems with the file system. 
   */
  public ProcessExecutionResult executeExternal(File execDir, String command, String...args) throws IOException 
  {
    ProcessBuilder builder;
    List<String> commandData = new ArrayList<String>();
    final List<String> outputLines = new ArrayList<String>();
    int exitCode;
    ILineProcessor lineProcessor;
    ProcessStdOutRedirector stdOutRedirector;
    Process process;
    
    builder = new ProcessBuilder();
    if (execDir != null)
    {      
      builder.directory(execDir);
    }
    commandData.add(command);
    CollectionUtil.current().addAll(commandData, args);
    builder.command(commandData);

    lineProcessor = new ILineProcessor()
    {
      @Override
      public boolean processLine(String line, int lineNo)
      {
        outputLines.add(line);
        return true;
      }
    };
    
    process = builder.start();
    stdOutRedirector = new ProcessStdOutRedirector(process, lineProcessor);
    Executors.newSingleThreadExecutor().submit(stdOutRedirector);
    try
    {
      exitCode = process.waitFor();
    }
    catch (InterruptedException e)
    {
      throw new IOException(e);
    }
    return new ProcessExecutionResult(exitCode, outputLines);
  }
  
  /**
   * Returns the value of an external configuration setting by checking first
   * a system property and if that is not set an environment variable.
   * If that's also not set null will be returned.
   * 
   * @param sysPropName The name of the system property to be checked (will be ignored if null).
   * @param envVarName The name of the environment variable to be checked (will be ignored if null).
   * @return The first value detected or null.
   */
  public String getConfigValue(String sysPropName, String envVarName) 
  {
    return getConfigValue(sysPropName, envVarName, null);
  }
  
  /**
   * Returns the value of an external configuration setting by checking first
   * a system property and if that is not set an environment variable and finally
   * if that's also not set returns the given default value.
   * 
   * @param sysPropName The name of the system property to be checked (will be ignored if null).
   * @param envVarName The name of the environment variable to be checked (will be ignored if null).
   * @param defaultValue The value to be returned if no value was found by the other two settings (may be null).
   * @return The first value detected or the default value.
   */
  public String getConfigValue(String sysPropName, String envVarName, String defaultValue) 
  {
    String value = null;
    
    if (sysPropName != null)
    {      
      value = System.getProperty(sysPropName);
    }
    if ((value == null) && (envVarName != null))
    {
      value = System.getenv(envVarName);
    }
    if (value == null)
    {
      value = defaultValue;
    }
    return value;
  }
  
  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  /**
   * Inform all exit listeners that now the system will be exited.
   * 
   * @param rc The return code for the exit.
   */
  protected void notifyExitListeners(final int rc)
  {
    Iterator<ISystemExitListener> iter;
    ISystemExitListener listener;

    iter = getExitListeners().iterator();
    while (iter.hasNext())
    {
      listener = iter.next();
      listener.systemAboutToExit(rc);
    }
  }

  /**
   * Returns true if the environment is based on eclipse.
   * That is, this class is loaded within an eclipse plug-in.
   * Can also be defined by setting system property "org.pf.util.isEclipse"
   * to true or false.
   */
  protected boolean determineIfEclipse()
  {
    String option;

    option = System.getProperty("org.pf.util.isEclipse");
    if (option != null)
    {
      return Bool.isTrue(option);
    }
    try
    {
      Class<?> clazz = getClass().forName("org.eclipse.core.runtime.Platform");
      return clazz != null;
    }
    catch (Exception e)
    {
      if (DEBUG)
      {
        e.printStackTrace();
      }
      return false;
    }
  }

  protected void reset()
  {
    this.isEclipse = null;
  }

  protected List<ISystemExitListener> getExitListeners()
  {
    return exitListeners;
  }

  protected void setExitListeners(List<ISystemExitListener> newValue)
  {
    exitListeners = newValue;
  }

}
