// ===========================================================================
// CONTENT  : CLASS TemporaryMap
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.4 - 05/08/2006
// HISTORY  :
//  12/10/2001  duma  CREATED
//	26/07/2002	duma	changed	-> Completely reimplemented
//	04/09/2002	duma	changed	-> Superclass from Object to MapWrapper
//	05/12/2002	duma	changed -> use HashMap and synchronize in removeExpiredEntries()
//	05/08/2006	mdu		bugfix	-> overridden remove() to avois returning Association
//
// Copyright (c) 2001-2006, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.util;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

/**
 * This implementation of a Map behaves like a HashMap except that it
 * associates a timestamp with each entry. Periodically it checks all
 * its entries' timestamp against an expire period.
 * If an entry is expired it will be automatically removed.
 * Therefore all elements just live temporarily in this map.
 *
 * @author Manfred Duchrow
 * @version 1.4
 */
public class TemporaryMap<K, V> implements Map<K, V>, TriggerClient, Serializable
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final long serialVersionUID = 3270539855831584494L;

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private Map<K, Association<Long, V>> internalMap = null;
  private long expirationTime = 60000;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  protected TemporaryMap()
  {
    super();
    internalMap(createEmptyMap());
  }

  /**
   * Initialize the new instance with values that define the expiration
   * of any object in the map and the interval for checking if expired
   * objects are in the map.
   * 
   * @param expiresAfterMs Defines after what time (milliseconds) an unused 
   * 	object must be removed (must be >= 100)
   * @param checkAfterMs Defines the interval for lookup and remove expired 
   * 	objects (must be >= 10)
   */
  public TemporaryMap(long expiresAfterMs, long checkAfterMs)
  {
    this();
    if (expiresAfterMs < 100)
    {
      throw new IllegalArgumentException("expiresAfterMs must be greater than 99 ms");
    }
    if (checkAfterMs < 10)
    {
      throw new IllegalArgumentException("checkAfterMs must be greater than 9 ms");
    }
    expirationTime(expiresAfterMs);

    Trigger.launch("TemporaryMap.Trigger", this, checkAfterMs);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  @Override
  public void clear()
  {
    internalMap().clear();
  }

  @Override
  public boolean isEmpty()
  {
    return internalMap().isEmpty();
  }

  @Override
  public int size()
  {
    return internalMap().size();
  }

  /**
   * This method returns true, if the given key can be found in this map.
   * If an entry with this key is found, its associated timestamp will be 
   * set anew, to ensure that any immediate call to get() still returns 
   * the object.
   * Otherwise it could happen, that the object is already removed by the 
   * background process that checks the timestamps and expiration times.
   * 
   * @see java.util.Map#containsKey(Object)
   */
  @Override
  public boolean containsKey(Object key)
  {
    return (get(key) != null);
  }

  /**
   * This method returns true, if the given value is a value of any entry
   * in this map.
   * If such an entry is found, its associated timestamp will be set anew,
   * to ensure that any immediate call to get() still returns the object.
   * Otherwise it could happen, that the object is already removed by the 
   * background process that checks the timestamps and expiration times.
   * 
   * @see java.util.Map#containsValue(Object)
   */
  @Override
  public boolean containsValue(Object value)
  {
    Map.Entry<K, Association<Long, V>> entry;
    Association<Long, V> assoc;

    entry = findByValue(value);
    if (entry == null)
    {
      return false;
    }
    assoc = entry.getValue();
    touch(assoc);
    return true;
  }

  /**
   * This method automatically <i>touches</i> every entry, so that its
   * expire period starts anew.
   * 
   * @see java.util.Map#entrySet()
   */
  @Override
  public Set<Map.Entry<K, V>> entrySet()
  {
    Set<Map.Entry<K, V>> entries;
    Map.Entry<K, V> newEntry;
    Association<Long, V> assoc;

    entries = new HashSet<Map.Entry<K, V>>(size());
    for (Map.Entry<K, Association<Long, V>> entry : internalMap().entrySet())
    {
      assoc = entry.getValue();
      touch(assoc);
      newEntry = new MapEntry<K, V>(entry.getKey(), assoc.value());
      entries.add(newEntry);
    }
    return entries;
  }

  /**
   * Returns the object that was registered under the specified key.
   * Here this method also renews the timestamp associated with the key's
   * entry. That is, the expiration time starts from 0 again.
   * 
   * @see java.util.Map#get(Object)
   */
  @Override
  public V get(Object key)
  {
    Association<Long, V> assoc;

    assoc = internalMap().get(key);
    if (assoc != null)
    {
      touch(assoc); // Accessing entry refreshes timestamp!
      return assoc.value();
    }
    return null;
  }

  /**
   * Returns all keys in this map.
   * <br>
   * <b>Resets the timestamp of all entries to ensure that the immediate usage
   * of the returned keys with the <i>get()</i> method produce the associated
   * values!</b>
   * 
   * @see java.util.Map#keySet()
   */
  @Override
  public Set<K> keySet()
  {
    touchAll();
    return internalMap().keySet();
  }

  /**
   * Puts the given value under the specified key into the map.
   * It also associates a timestamp with this entry. Therefore it can be
   * automatically removed after the defined expiration time.
   * 
   * @see java.util.Map#put(Object, Object)
   */
  @Override
  public V put(K key, V value)
  {
    Association<Long, V> oldAssoc;
    Association<Long, V> assoc;

    assoc = new Association<Long, V>(newTimestamp(), value);
    oldAssoc = internalMap().put(key, assoc);
    if (oldAssoc == null)
    {
      return null;
    }
    return oldAssoc.getValue();
  }

  /**
   * Each entry of the given map will be added. 
   * Each added entry gets an associated timestamp.
   * 
   * @see java.util.Map#putAll(Map)
   */
  @Override
  public void putAll(Map<? extends K, ? extends V> map)
  {
    if (map != null)
    {
      for (Map.Entry<? extends K, ? extends V> entry : map.entrySet())
      {
        put(entry.getKey(), entry.getValue());
      }
    }
  }

  /**
   * Returns all values of the map.
   * With this method the timestamps of all entries in this map will be set
   * anew. That means, their expiration time starts again at 0.
   * 
   * @see java.util.Map#values()
   */
  @Override
  public Collection<V> values()
  {
    Iterator<Association<Long, V>> iterator;
    Association<Long, V> assoc;
    Collection<V> values;

    values = new ArrayList<V>(size());
    iterator = internalMap().values().iterator();
    while (iterator.hasNext())
    {
      assoc = iterator.next();
      touch(assoc);
      values.add(assoc.value());
    }
    return values;
  }

  /**
   * Remove the entry with the given key and return the associated object.
   * 
   * @param key The key of the entry to be removed
   * @return The associated object or null if the key could not be found
   */
  @Override
  public V remove(Object key)
  {
    Association<Long, V> assoc;

    assoc = internalMap().remove(key);
    if (assoc == null)
    {
      return null;
    }
    return assoc.value();
  }

  // ---- Implementation of Interface TriggerClient --------------------------

  /**
   * <b>FOR INTERNAL USE ONLY!</b>
   * <br>
   * Will be called by a Trigger to check if object clean-up can be executed.
   * 
   * @see org.pfsw.util.TriggerClient#canBeTriggeredBy(Trigger)
   */
  @Override
  public boolean canBeTriggeredBy(Trigger trigger)
  {
    return true;
  }

  /**
   * <b>FOR INTERNAL USE ONLY!</b>
   * <br>
   * Will be called by a Trigger to execute the clean-up of expired objects.
   * 
   * @see org.pfsw.util.TriggerClient#triggeredBy(Trigger)
   */
  @Override
  public boolean triggeredBy(Trigger trigger)
  {
    removeExpiredEntries();
    return true;
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected void touch(Association<Long, V> assoc)
  {
    assoc.key(newTimestamp());
  }

  protected void touchAll()
  {
    for (Association<Long, V> assoc : internalMap().values())
    {
      touch(assoc);
    }
  }

  protected Long newTimestamp()
  {
    Long timestamp;

    timestamp = new Long(System.currentTimeMillis());
    return timestamp;
  }

  protected void removeExpiredEntries()
  {
    Iterator<Map.Entry<K, Association<Long, V>>> iterator;
    Map.Entry<K, Association<Long, V>> entry;

    synchronized (internalMap())
    {
      iterator = internalMap().entrySet().iterator();
      while (iterator.hasNext())
      {
        entry = iterator.next();
        if (isExpired(entry.getValue()))
        {
          iterator.remove();
        }
      }
    }
  }

  protected boolean isExpired(Association<Long, V> assoc)
  {
    long ts = assoc.key().longValue();
    long diff = System.currentTimeMillis() - ts;
    return (diff > expirationTime());
  }

  protected Map.Entry<K, Association<Long, V>> findByValue(Object value)
  {
    Association<Long, V> assoc;

    if (value == null)
    {
      return null;
    }
    for (Map.Entry<K, Association<Long, V>> entry : internalMap().entrySet())
    {
      assoc = entry.getValue();
      if (value.equals(assoc.value()))
      {
        return entry;
      }
    }
    return null;
  }

  /**
   * Provide a synchronized HashMap.
   */
  protected Map<K, Association<Long, V>> createEmptyMap()
  {
    return Collections.synchronizedMap(new HashMap<K, Association<Long, V>>());
  }

  protected Map<K, Association<Long, V>> internalMap()
  {
    return this.internalMap;
  }

  protected void internalMap(Map<K, Association<Long, V>> newValue)
  {
    this.internalMap = newValue;
  }

  protected long expirationTime()
  {
    return this.expirationTime;
  }

  protected void expirationTime(long newValue)
  {
    this.expirationTime = newValue;
  }
}
