// ===========================================================================
// CONTENT  : CLASS ToggleSwitch
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 26/04/2021
// HISTORY  :
//  26/04/2021  mdu  CREATED
//
// Copyright (c) 2021, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.util;

import java.util.concurrent.atomic.AtomicBoolean;

import org.pfsw.bif.execution.IExecutable;
import org.pfsw.bif.logic.ICondition;

/**
 * A condition with a given initial boolean value that changes (toggles) 
 * whenever its run() or execute() method is called.
 */
public class ToggleSwitch implements ICondition, IExecutable, Runnable
{
  private final AtomicBoolean value = new AtomicBoolean(false);

  /**
   * Creates a new instance with the given initial value.
   */
  public static ToggleSwitch create(boolean initialValue)
  {
    return new ToggleSwitch(initialValue);
  }

  /**
   * Creates a new instance with initial value <code>true</code>.
   */
  public static ToggleSwitch on()
  {
    return create(true);
  }

  /**
   * Creates a new instance with initial value <code>false</code>.
   */
  public static ToggleSwitch off()
  {
    return create(false);
  }

  protected ToggleSwitch(boolean initialValue)
  {
    super();
    value.set(initialValue);
  }

  @Override
  public synchronized boolean isTrue()
  {
    return value.get();
  }

  public boolean isFalse()
  {
    return !isTrue();
  }

  public boolean isOn()
  {
    return isTrue();
  }

  public boolean isOff()
  {
    return isFalse();
  }

  /**
   * Changes the underlying value from <code>true</code> to <code>false</code> or from <code>false</code> to <code>true</code>. 
   */
  @Override
  public void run()
  {
    toggle();
  }

  /**
   * Changes the underlying value from <code>true</code> to <code>false</code> or from <code>false</code> to <code>true</code>. 
   */
  @Override
  public void execute()
  {
    toggle();
  }

  /**
   * Changes the underlying value from <code>true</code> to <code>false</code> or from <code>false</code> to <code>true</code>. 
   */
  public synchronized void toggle()
  {
    value.set(!value.get());
  }

  @Override
  public int hashCode()
  {
    return isTrue() ? 1 : 0;
  }

  @Override
  public boolean equals(Object obj)
  {
    if ((obj != null) && (obj.getClass() == getClass()))
    {
      ToggleSwitch other = (ToggleSwitch)obj;
      return isTrue() == other.isTrue();
    }
    return false;
  }

  @Override
  public String toString()
  {
    return String.format("%s(%s)", getClass().getSimpleName(), isTrue());
  }
}
