// ===========================================================================
// CONTENT  : CLASS Trigger
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.1 - 06/03/2004
// HISTORY  :
//  26/05/2002  duma  CREATED
//	06/03/2004	duma	added		-->	terminate()
//
// Copyright (c) 2002-2004, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.util;

/**
 * A trigger is an object that runs in a separate thread and waits for 
 * a specified time until it calls a specific method of a trigger client
 * (see {@link TriggerClient}).
 * Before it actually calls the client's <i>triggerdBy()</i> method it
 * checks the client's <i>canBeTriggered()</i> method that allows it
 * or not.
 * After the client's <i>triggerdBy()</i> method was called the trigger
 * waits again until the end of its waiting period.
 *
 * @author Manfred Duchrow
 * @version 1.1
 */
public class Trigger implements Runnable
{
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private String name = null;
  private TriggerClient client = null;
  private long interval = 300000; // milliseconds
  private boolean suspended = false;
  private boolean terminated = false;

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Launch a new trigger with the given name for the given client, that
   * calls the <i>triggeredBy</i> method of the client every <i>intervalInMs</i>
   * milliseconds.<br>
   * The ne trigger will be created and immediately started in a separate 
   * thread.
   * 
   * @param name The name of the trigger (must not be null)
   * @param client The client that gets triggered (must not be null)
   * @param intervalInMs The interval after that the client regularily gets triggered (must be greater than 0)
   * @return The newly created trigger
   * @throws IllegalArgumentException if any of the given arguments is null
   * @see TriggerClient
   */
  public static Trigger launch(String name, TriggerClient client, long intervalInMs)
  {
    String invalidArg = null;
    Trigger trigger = null;
    Thread thread = null;

    if (name == null)
    {
      invalidArg = "name";
    }
    else if (client == null)
    {
      invalidArg = "client";
    }
    else if (intervalInMs <= 0)
    {
      invalidArg = "intervalInMs";
    }

    if (invalidArg != null)
    {
      throw new IllegalArgumentException(Trigger.class.getName() + ".launch() -> " + invalidArg);
    }

    trigger = new Trigger(name, client, intervalInMs);
    thread = new Thread(trigger);
    thread.start();

    return trigger;
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  protected Trigger(String name, TriggerClient client, long intervalInMs)
  {
    super();
    setName(name);
    setClient(client);
    setInterval(intervalInMs);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the name of the trigger
   */
  public String getName()
  {
    return this.name;
  }

  /**
   * Terminates the trigger. That means the thread the trigger used internally
   * will be stopped when it awakes the next time (depends on interval)
   */
  public void terminate()
  {
    setTerminated(true);
  }

  /**
   * Implementation of Runnable. <p>
   * Must not be called directly!
   */
  @Override
  public void run()
  {
    while (!isTerminated())
    {
      try
      {
        Thread.sleep(getInterval());
        if ((!isTerminated()) && (!isSuspended()))
        {
          if (canBeTriggered())
          {
            trigger();
          }
        }
      }
      catch (@SuppressWarnings("unused") InterruptedException ex)
      {
        setTerminated(true);
      }
    }
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================

  protected boolean canBeTriggered()
  {
    return getClient().canBeTriggeredBy(this);
  }

  protected void trigger()
  {
    boolean cont;

    cont = getClient().triggeredBy(this);
    if (!cont)
    {
      setSuspended(true);
    }
  }

  protected boolean isTerminated()
  {
    return getTerminated();
  }

  protected boolean isSuspended()
  {
    return getSuspended();
  }

  protected void setName(String newValue)
  {
    this.name = newValue;
  }

  protected TriggerClient getClient()
  {
    return this.client;
  }

  protected void setClient(TriggerClient newValue)
  {
    this.client = newValue;
  }

  protected long getInterval()
  {
    return this.interval;
  }

  protected void setInterval(long newValue)
  {
    this.interval = newValue;
  }

  protected boolean getSuspended()
  {
    return this.suspended;
  }

  protected void setSuspended(boolean newValue)
  {
    this.suspended = newValue;
  }

  protected boolean getTerminated()
  {
    return this.terminated;
  }

  protected void setTerminated(boolean newValue)
  {
    this.terminated = newValue;
  }
}